#ifndef _coordc_h_
#define _coordc_h_


class LinCC{
  // the coordinate transformation:
  //    y=m*(x - origin_shift)
  // the inverse is given by x=mInv*y + origin_shift
 public:
   int dim;

   double *origin_shift;
   doublematrix1 *m, *mInv;
   interval *intv_origin_shift;
   intervalmatrix1 *intv_m, *intv_mInv;


         // the copy/assignment operator
   LinCC & operator=(const LinCC & s);
   ~LinCC(); // destructor

   void reset_to_null(); // sets dim=0 and deletes arrays describing
                   // the transformation


   void doubletointerval(int orthogonal);
   void intervaltodouble();



   LinCC(int ldim,const interval IP[],const interval IM[][MAX_DIM+1],
          const interval InvIM[][MAX_DIM+1]);
          // IP- origin shift

   LinCC(int ldim,const interval IM[][MAX_DIM+1],
          const interval InvIM[][MAX_DIM+1]);
   // initiates structure with interval data
   // P(=origin_shift)  is assumed to be 0

   LinCC(int ldim,const interval IP[]);
   //  M is assumed to be Identity


   LinCC(int ldim,const double P[]);
   // m=Identity, origin_shift=P

   LinCC(int ldim);
   // identity transformation


   LinCC();  // default constructor

   LinCC(const LinCC &lincc);

   void shiftorigin(const interval IP[]);
// function updates the origin intv_origin_shift=IP


   double coord(const double x[],int i) const;
   interval coord(const interval x[],int i) const;

   void coordtran(const double x[],double y[]) const;
   void coordtran(const interval x[],interval y[]) const;
   // y=m*(x-origin_shift)

   void coordinvtran(const double y[],double x[]) const;
   void coordinvtran(const interval y[],interval x[])  const;
   // x=mInv*y + origin_shift
   // both y and x can point to the same location


   void coordtran(const interval x[],const interval Cx[][MAX_DIM+1],const interval rx[],
       interval y[],interval Cy[][MAX_DIM+1],interval ry[]) const;
     // the transformation of the set x+Cx*rx in old coordinates
     //  on output the set in new coordinates is given by
     //  y+Cy*ry = m(x - origin_shift) + m*Cx*rx
     //   y=m(x-origin_shift),  Cy=m*Cx ,   ry=rx




   void coordinvtran(const interval y[],const interval Cy[][MAX_DIM+1],const interval ry[],
       interval x[],interval Cx[][MAX_DIM+1],interval rx[]) const;
     // the transformation of the set y+Cy*ry in new coordinates
     //  on output the set in old coordinates is given by
     //   x+Cx*rx=mInv*y + origin_shift + mInv*Cy*ry
     //    x= mInv*y + origin_shift,   Cx=mInv*Cy,  rx=ry



   // transformation of tangent vectors: from old to new
   void vectcoordtran(const double x[],double y[]) const;
   void vectcoordtran(const interval x[],interval y[]) const;
   // y=m*x


   // transformation of tangent vectors: from new to old
   void vectcoordinvtran(const double y[],double x[]) const;
   void vectcoordinvtran(const interval y[],interval x[])  const;
   // x=mInv*y



   LinCC inverse() const;

  int testinverse() const;
// function returns 1 - if the identify matrix \subset intv_m*intv_mInv
  //                               and  (reversed order) the identify matrix \subset intv_mInv*intv_m
  //                 0 - otherwise


  int savebin(const char *filename,int whatsave) const;

  int savebin(FILE *fbin,int whatsave) const;
  // saves the coordinate change to a binary file
  // whatsave=0; only double's are saved
  //         =1; only intervals
  //         =2; everything
  //// structure of the file:
  //  whatsave,dim,(doubles:origin_shift,m,mInv),(intervals:intv_origin_shift,intv_m,intv_mInv)

  int readbin(const char *filename,bool recomp_inverse =false);

  int readbin(FILE *fbin,bool recomp_inverse =false);
  // reads (really reinitiates) coordinate change from the file <fbin>
    // if  recomp_inverse, then the inverse in *plincc is recomputed
   //  to make sure that computation is rigorous, can be of use, when
   // computing things first on the computer on which we don't have correct rounding



  int savetext(ostream &ftext,int whatsave) const;
  // saves the coordinate change to a text file
  // whatsave=0; only double's are saved
  //         =1; only intervals
  //         =2; everything
  //// structure of the file:
  //  whatsave,dim,(doubles:origin_shift,m,mInv),(intervals:intv_origin_shift,intv_m,intv_mInv)


  int readtext(istream &ftext);
  // reads (really reinitiates) coordinate change from the file <ftext>


   void printCC(const char name[]) const;

protected:
  void create_arrays();


};

typedef LinCC* pLinCC;


void permutecoordc(LinCC *lincc,int per[]);
// the procedure permutes the variables in the <lincc>
// the permutation is given as a new ordering  of variables(per[1],per[2],...)
//  for example: per={2,3,....}
//      means 2-nd coordinate become 1
//            3-rd become second and so on
// on output : the coordinate change works as follows:
// y_i (new) = y_{per[i]}(old)



void transLinearPart(const double lin[][MAX_DIM+1],const LinCC &linccIn,
     const LinCC &linccOut, double translin[][MAX_DIM+1]);

void transLinearPart(const interval lin[][MAX_DIM+1],const LinCC &linccIn,
     const LinCC &linccOut, interval translin[][MAX_DIM+1]);
// this function does coordinate transformation of linear map
//  <lin>, in new coordinates on input - given by linccIn
//  and output - linccOut
//  hence it returns the linear matrix reprsenting the map
//       linccOut \circ lin \circ linccIn^{-1}
//
//  This function makes sense for transforming derivative matrix of Poincare
//  map



void transLinearPart(const double lin[][MAX_DIM+1],const LinCC &lincc,
     double translin[][MAX_DIM+1]);
void transLinearPart(const interval lin[][MAX_DIM+1],const LinCC &lincc,
     interval translin[][MAX_DIM+1]);
// this function does a coordinate transformation of linear map
//  <lin>, On output <translin> contains <lin> in new coordinates
//  the fact that <lin> in new coordinates might be not a linear
//  map (this is the case when the origin s shifted) is ignored
//  This function makes sense for transforming the derivative matrix
//  of Poincare map  or the derivative of the vector field around fixed point

LinCC* insertcoord(const LinCC *coordch,int sel_coord,interval sel_coord_value);
// this function creates a coordinate transformation in the dimension
//  coordch->dim + 1, by inserting sel_coord, and setting the origin for
//  this coordinate to <sel_coord_value>

LinCC* LinCCincreasedim(LinCC *coordch,int newdim);
// this function creates a coordinate transformation in dimension
//   newdim - by adding identity on coordinates i=coordch->dim+1,\dots,newdim


LinCC* composeLinCC(const LinCC *c1,const LinCC *c2);
// the function creates a new change of coordinates c1 \circ c2

LinCC *genLinCCfd(int pdim,double M[][MAX_DIM+1],int thisIsInvM);
// creates a new LinCC - with P=0; and rigorously computed M^{-1}
//   if thisIsInvM, then the parameter M - is the inverse
//   of the transition from old to new coordinates


int LinCC_thesame(const LinCC *pcc1,const LinCC *pcc2,int  whattocheck);
// 1 - is returned when the coordinate changes are the same
// 0 - otherwise
//   if whattocheck==0 - we check doubles
//                1 - we check intervals
//                2 - we check both doubles and intervals

#endif
