#include<iostream>
#include<fstream>
using namespace std;

#include"defo.h"
#include"fnames.h"
#include"vfkso.h"
#include"section.h"
#include"inclpoin.h"
#include"galks.h"
#include"oneshot.h"
#include"multiint.h"
#include"zgfiles.h"
#include"ksperiso.h"
#include"kspoinc.h"
#include"perfixp.h"
#include"ksinclpo.h"
#include"ksunst.h"


// #define USE_LOGNORM  // 20.V.2008 - for tests, in order to see whether I can get proofs using log norm
      // instead of componentwise estimates -- THIS SHOULD BE COMMENTED OUT FOR OPTIMAL PERFORMANCE


void KSinfo_unstbsymper(interval imu)
// displays info about reasonable parameter values
// for the proof of the unstable symmetric periodic orbit
{
  #if defined(WIN95) && defined(KRAK)
    rootFrame.Clear();
  #endif
  if(abs(imu - 0.1215) < 1e-2){
    cout << "proof parameters for mu=0.1215: " << endl;
    cout << "suggested parameters: dim=11, h=2e-3,delta=1e-4,order=8 " << endl;
    cout << "iter =2,  dimdiag=10 " << endl;
    #if !defined(KSPERPAP)
    cout <<  "section = 2 ,  perpendicular section " << endl;
    #endif
    cout << "one iterate takes around 18 sec on Kepler (Linux 3GHz)" <<  endl;
    cout << "------------------------------" << endl;
    cout << "for mu=0.1212 a symmetric orbit NOT on the attractor " << endl;
    cout << "suggested parameters: dim=14, h=5e-4, delta=5e-5, order=7 " << endl;
    cout << "iter =1,   dimdiag=9 " << endl;
    #if !defined(KSPERPAP)
    cout <<  "section = 2 ,  perpendicular section " << endl;
    #endif
    cout << "computation time/per iterate on 3GHz (Kepler) - around 190 sec " << endl;
    cout << "-------------------------------------" << endl;
  }
  if(imu < 0.1){
    cout << "proof parameters for mu=0.02991 " << endl;
    cout << "suggested parameters: dim=25, h=1e-4, delta=5e-5, order=5 " << endl;
    cout << "iter =3 (2 is enough for stable dir),   dimdiag=12 " << endl;
    #if !defined(KSPERPAP)
    cout <<  "section = 2 ,  perpendicular section " << endl;
    #endif
    cout << "comp. time/per iterate (on kepler 3 GHz ) - around 600 sec = 10 min " << endl;
    cout << "-------------------------------" << endl;
  }
}


void KSinfo_unstbper(interval /* imu*/)
// displays info about reasonable parameter values
// for the proof of the unstable non-symmetric periodic orbit
{
  cout << "for mu=0.1212 a non-symmetric orbit on the attractor " << endl;
  cout << "suggested parameters: dim=13, h=4e-4, delta=2e-5, order=5 " ;
  cout << "iter =1 " << endl;
  #if !defined(KSPERPAP)
  cout <<  "section = 2 ,  perpendicular section " << endl;
  #endif
  cout << "computation time per iterate - around 185  sec (on Linux kepler 3GHz)" << endl;
  cout << "-------------------------------------" << endl;
}


void  ksunstb_repproof_status(ostream &report,int locproved,
            int full_galerkinproof,int incl_proven)
{
  if(locproved){
    char inclinfo[80];
    strcpy(inclinfo,"inclusions for i > 1 proved ");
    report << inclinfo << endl;
    cout << inclinfo << endl;
    if(full_galerkinproof){
    } else{
      if(incl_proven){
        cout << "but still a full check of apriori bounds is needed " << endl;
        report << "but still a full check of apriori bounds is needed " << endl;
      }
    }
  } else
    report << "P_{i>1}(x) is not contained in x_{i>1} " << endl;
}




// the proof of the existence of unstable periodic orbit in the situation
// of only one unstable direction, with the leading eigenvalue less than -1

int ksunstb_verifybdcond(ostream &report, InclPoincareMap &inclPMap,
       interval ixb[],interval r0[],int numLohner,SetWithTail &y,int sign1face,
       int useothercoord,LinCC *coordch)
// we check whether the boundary condition is satisfied for
//  right face - when sign1face==1
//  left face - when sign1face==-1
{
  int proved=0;


  estimates estm;

  int divtable[MAX_DIM+1];
  int dimdiv=0;

  intervalmatrix1 *C;
  C=new intervalmatrix1[inclPMap.dim+1];
  identity(C,inclPMap.dim);  // this must be the identity


  interval *xs,*r,*fx;
  xs=new interval[inclPMap.dim+1];
  r=new interval[inclPMap.dim+1];
  fx=new interval[inclPMap.dim+1];

  SetWithTail fy(y.dim);

  copyvector(ixb,xs,inclPMap.dim);
  if(sign1face==1){
    xs[1]+=Right(r0[1]);
  }else{
    xs[1]+=Left(r0[1]);
  }
  copyvector(r0,r,inclPMap.dim);
  r[1]=0;


  report << endl << endl;
  report << "Checking the boundary conditions" << endl;
 // int iterno=1;
  gal_saveinitialcond(report,xs,r,inclPMap.dim,y);

  time_t before,after;
  before=time(NULL);

  int ok;
  ok=inclPoinMap_map_divide(inclPMap,divtable,dimdiv,xs,C,r,numLohner,y,fx,fy,
                      useothercoord,coordch,useothercoord,coordch,&estm);
  delete [] C;
  C=NULL;

  if(ok){
    after=time(NULL);
    double compt_time=difftime(after,before);
    if(compt_time > 1){
      report << "inclusion" << "  computation time="  << compt_time << "  sec " << endl;
      if(compt_time > 2){
        cout << "  computation time="  << compt_time << "  sec " << endl;
      }
    }
  }else{
    report << "not computed " << endl;
    // continue;
  }
  if(ok){ // we check now the condition
    if(sign1face==1){
      proved=(fx[1] < Left(ixb[1])+Right(r0[1]));
    } else{
      proved=(fx[1] > Right(ixb[1])+Left(r0[1]));
    }
    cout << "boundary condition  for face=" << sign1face ;
    report << "boundary condition  for face=" << sign1face ;
    if(!proved){
      cout << "  NOT " ;
      report << " NOT " ;
    }
    cout << "  satisfied " << endl;
    report <<   "  satisfied " << endl;
  }

  gal_save_res(report,fx,NULL,inclPMap.dim,fy,0);

  delete [] fx;
  delete [] xs;
  delete [] r;

  return proved;
}


int galunstb_rev_orientation(FPoincareMap &PMap,InclPoincareMap &inclPMap,
       Galerkin_Proj_KuraShi &galode,  double xb[],double delta[], int numLohner,
        SetWithTail &y0,ofstream &report,int coord,int dimdiag,
        int iternum,int do_divisions,int testasymmetry,
         InclPoincareMap* symPMap)
// xb - a good approximation of the periodic orbit, the fixed point for Poincare map
// numLohner - one of the constants LOHNERxx
// iternum - the number of iterates
// coord - the type of coordinates
//  diagdiam - the dimension used in some diagonalization process used
//             to construct 'good' coordinates
//  do_divisions - if ==1, then starting from the second iterate the domain
//              is divided into parts for the computations
// this function takes the set N=xb + \\Pi_i[-delta[i],delta[i]]
// it computes P(N) ; for the next iterate N=P(N) \\cap N and so on
//// returns 1 - when a set N is found such that P_k(N) \subset N_k, for k >1
//  for k=1 (the first coordinate we want P(R(N)) to be mapped to the left
// of R(N) and P(L(N)) mapped to the right of L(N), where R(N) is the right face
//   x \in R(N) iff x \in N and x_1 = \delta
//  and L(N) the left face :  x \in L(N) iff x \in N_1=-\delta
///// it must be that <inclPMap> - is pointing onto <galode> as an the inclusion
//   probably one can do it better from the formal point of view
//  I need this to  have an access to some functions in <galode>
{
  cout << " ------------------------ " << endl;
  #if defined(WIN95) && defined(KRAK)
    rootFrame.Clear();
  #endif

  int ok=1;


  double Dp[MAX_DIM+1][MAX_DIM+1];
  if(coord!=standardcoord) {

    cout << "Real computation of Dp " << endl;
    double trash[MAX_DIM+1];

    ok=PMap.map(xb,trash,Dp);
    if(ok){
      const int buflen=80;
      char buffer[buflen];
      report << "Dp (nonrigorous) computed using h=" <<   PMap.dgettimestep() <<
                  "  order=" << PMap.dgetorder() << "  method=" <<
                   num_method_name(buffer,buflen,PMap.dgetmethod()) << endl;
      report <<  "Diagonalization dimension=" <<  dimdiag << endl;
    }else{
      cout << "unable to compute DP " << endl;
      return 0;
    }
  }


  interval ixp[MAX_DIM+1];
  tointerval(xb,ixp,PMap.dim);
  LinCC* coordch=NULL;
  if(ok)
    ok=buildcoordchange(coord,Dp,ixp,PMap.dim,&coordch,dimdiag);

  if(!ok){
    delete coordch;
    coordch=NULL;
    return  0;
  }

  report << "the starting point in standard section coordinates: " << endl;
  saveAllCoord(report,xb,PMap.dim);

  const int buflen=80;
  char buffer[buflen];
  report << osa_coordsname(buffer,buflen,coord) << endl;

  Set xs;
  interval ixb[MAX_DIM+1],fx[MAX_DIM+1];
  zero(ixb,PMap.dim); // in our coordinate frame xb is the origin

  intervalmatrix1 *C;
  C=new intervalmatrix1[PMap.dim+1];
  identity(C,PMap.dim); // C - must be an identity for the intersection
                       // r0 \\cap Pmap.map(x0+C*r0)
                       // to work correctly
  interval r0[MAX_DIM+1];
  for(int i=1;i<=PMap.dim;i++)
    r0[i]=interval(-delta[i],delta[i]);

  SetWithTail  y(y0),fy(y0.dim); // used to handle the tail



  double d=delta[1];
  cout << "the size of the set " << d << "   ";
  cout << "h=" << inclPMap.igettimestep() << "  ";
  cout << "order=" << inclPMap.igetorder() << endl;

  int strategy=gen_new_shift;
  if(d==0)
    strategy=gen_from_point;

  inclPMap.param.write(report);
  report << "strategy for gen. of new set = " << namehsetstrategy(strategy,buffer,buflen) << endl;
  report.flush();


  estimates estm;


  int incl_proven=0; // tells that we have inclusion on main variables
  int full_galerkinproof=0; // full check
  int gennew_ic;


  int dimdiv=0;
  int divtable[MAX_DIM+1];

  int useothercoord=1;

  int iterno=1;

  while(ok && (iterno<=iternum)){

    gennew_ic=1;
    report << endl << "iterate=" << iterno << endl;
    gal_saveinitialcond(report,ixb,r0,PMap.dim,y);

    xs.newvalues(PMap.dim,numLohner,ixb,C,r0); // this is our starting set

    if(r0[1].left_b()==r0[1].right_b()) d=0;
    else d=1.0;


    cout << "iterate=" << iterno <<  "/" << iternum <<  endl;

    writedivtable(cout,divtable,dimdiv);
    if(dimdiv>0) writedivtable(report,divtable,dimdiv);


    time_t before,after;
    before=time(NULL);

    int withdivisions=0;
    if(do_divisions){
      if(iterno> 1){
        withdivisions=(double(iterno)/iternum > 0.5);
      }else withdivisions=do_divisions;
    }
    if(withdivisions){
      ok=inclPoinMap_map_divide(inclPMap,divtable,dimdiv,ixb,C,r0,numLohner,y,fx,fy,
                      useothercoord,coordch,useothercoord,coordch,&estm);
    }else{
      ok=inclPMap.imapCCE(xs,y,fx,fy,useothercoord,coordch,useothercoord,coordch,&estm);
    }

    if(ok){
      after=time(NULL);
      double compt_time=difftime(after,before);
      if(compt_time > 1){
        report << "inclusion" << "  computation time="  << compt_time << "  sec " << endl;
        if(compt_time > 2){
          cout << "  computation time="  << compt_time << "  sec " << endl;
        }
      }
    }else{
      report << "not computed " << endl;
      continue;
    }
    gal_save_res(report,fx,r0,PMap.dim,fy,(d!=0));


    int numbigratios=0;
    double largestratio=0;
    if(d!=0){ // d - is an initial 1/2-diameter of r0 at the beginning
      compute_ratios(r0,fx,PMap.dim,&numbigratios,&largestratio);
      cout << "Number of ratios > 1 is = " << numbigratios << endl;
      cout << "largest ratio = " << largestratio << endl;
    }else{
      numbigratios=PMap.dim;
    }

    if(d!=0){
      if(do_divisions && (iterno==1) && (numbigratios > 0)){
        dimdiv=numbigratios+1;
        for(int i=1;i<=dimdiv;i++)
           divtable[i]=2;
      }
    }



    interval x[MAX_DIM+1],cp[MAX_DIM+1];
    for(int i=1;i<=PMap.dim;i++)
      x[i]=ixb[i]+r0[i];


    int mvar_incl_proved=1;
    for(int i=2;((i<=PMap.dim) && mvar_incl_proved);i++){
      mvar_incl_proved=fx[i].subset(x[i]);
    }
    int tail_incl_proved=(*(fy.ptail)).subset(*(y.ptail));
    int locproved=mvar_incl_proved && tail_incl_proved;

    if(!mvar_incl_proved && (numbigratios==1)){
      write_no_inclusion(report,fx,x,PMap.dim,"fx","x");
    }

    if(!tail_incl_proved)
      write_no_tail_inclusion(report,fy.ptail,y.ptail);

    if(locproved){
      if(!galode.fullcheckstatus()){
        incl_proven=1;
        gennew_ic=0;
        galode.fullcheckstatus(1); // new iteration will be with the full validation
      } else{
        incl_proven=1;
        full_galerkinproof=1;
        gennew_ic=(iterno!=iternum);
      }
    }


    ksunstb_repproof_status(report,locproved,full_galerkinproof,incl_proven);


    // the generation of a new set for the next iteration
    if(gennew_ic){

      SetWithTail ynew(y.dim);
      interval *xnew;
      xnew=new interval[PMap.dim+1];
      int withtail=1;
      if(strategy==gen_from_point){
        ok=genhset_from_point(report,iterno,PMap.dim,withtail,
         galode.get_tailevolution(),x,fx,y,fy,xnew,ynew);

      }
      if(strategy==gen_new_intersect){
        ok=gen_hset_with_tail_cap_infl(report,0,iterno,PMap.dim,withtail,
                 galode.get_tailevolution(),x,fx,y,fy,xnew,ynew);
      }
      if(strategy==gen_new_shift){
        ok=corr_based_on_diameters(incl_proven,PMap.dim,withtail,
                  galode.get_tailevolution(),x,fx,y,fy,xnew,ynew);
      }

      if(ok){
        copyvector(xnew,ixb,PMap.dim);
        DecomposeIMulti(ixb,r0,PMap.dim);
        y=ynew;
      }
      delete [] xnew;
    }

    if(!locproved && incl_proven){
      incl_proven=0;
    }

    iterno++;
  }
  //  if the proof went through, then the estimate for  fixed point is given
  // by ixb + r0 - in coordinates - <coordch> (new coordinates in <coordch>)


  if(ok && incl_proven){ // now we do final verification
    fx[1]=ixb[1]+r0[1];
    copyvector(fx,ixb,PMap.dim);
    DecomposeIMulti(ixb,r0,PMap.dim);
    y=fy;

    // our candidate set is ixb + r0 + y.ptail
    // we already have checked that  we have an entry in all directions but the first
    // one, in this direction we check the image of faces
    int signface=1;
    cout << "computation of the image of the right face " << endl;
    int Rfaceok=ksunstb_verifybdcond(report,inclPMap,ixb,r0,numLohner,y,signface,
        useothercoord,coordch);


    cout << "computation of the image of the  left face " << endl;
    signface=-1;
    int Lfaceok= ksunstb_verifybdcond(report,inclPMap,ixb,r0,numLohner,y,signface,
        useothercoord,coordch);

    incl_proven=Rfaceok && Lfaceok;

    if(incl_proven){
      cout << "PROOF COMPLETED SUCCESSFULLY" << endl;
      report << "PROOF COMPLETED SUCCESSFULLY" << endl;
    }

    if(testasymmetry){
      asymmetry_test(report,symPMap,ixb,r0,numLohner,y,divtable,dimdiv,coordch);
    }

  }


  if(ok){

    report << endl << "Estimates: " << endl;
    writeGlEstm(report,&estm,PMap.igetorder(),fabs(PMap.igettimestep()));


    if(incl_proven){
      // the computation of the 'error' of the periodic orbit found
      // 1. transform ixb+r to the ambient space coordinates
      // 2. compute an 'error' in various norms

      // Ad.1
      interval xfull[MAX_DIM+1],rfull[MAX_DIM+1];
      from_sec_to_ambient_space(inclPMap.pInsection,ixb,r0,xfull,
                    useothercoord,coordch);
      DecomposeIMulti(xfull,rfull,inclPMap.pode->dim);



      // Ad.2
      double dl2=galode.l2norm(rfull,y).right_b();
      double dh1=galode.h1norm(rfull,y).right_b();
      double dc0=galode.c0norm(rfull,y).right_b();
      double dc1=galode.c1norm(rfull,y).right_b();

      report << endl;
      report << "Center point for our bounds for periodic orbit: " << endl;
      saveAllCoord(report,xfull,inclPMap.pode->dim,0);

      y.ptail->save(report,2,"Tail");


      report << " Error estimates :" << endl;
      report  << "  diff. estm  L2=" << dl2 << "   H1=" << dh1 << "  C0="<<
      dc0 <<  "  C1=" << dc1 << endl;

      // the bounds listed above are obtained, by ignoring really the fact
      // that in the near-tail we have info on the a_k, namely the mid point
      // below we want to compute 'error' taking into account this info.
      SetWithTail ymid(y);
      for(int i=y.ptail->m+1;i<=y.ptail->big_m;i++){
        ymid.ptail->a[i]-=mid(ymid.ptail->a[i]); // we subtract the mid point
                                                        // in the near tail
      }
      dl2=galode.l2norm(rfull,ymid).right_b();
      dh1=galode.h1norm(rfull,ymid).right_b();
      dc0=galode.c0norm(rfull,ymid).right_b();
      dc1=galode.c1norm(rfull,ymid).right_b();
      report << " Error estimates with respect to the mid point in near tail:"
                    << endl;
      report  << "  diff. estm  L2=" << dl2 << "   H1=" << dh1 << "  C0="<<
      dc0 <<  "  C1=" << dc1 << endl;


    }

    // saving the section and coordinates:
    gal_sectionDataSave(report,inclPMap,useothercoord,coordch);

  }
  delete [] C;
  delete coordch;
  return incl_proven;
}

/// END - proof of an unstable periodic orbit with one unstable direction and
//              eigenvalue < -1



int kscandidate_usnstb(KuraShi &kurashi,FPoincareMap *PMap,int iter_num,
          double xb[],double xs[],double err)
// In xb and xs we return a candidate for the fixed point for <iter_num>-iterate
  //  xb - is on section and is expressed in section coordinates
  //  xs - is xb, but it is expressed in coordinates of the ambient space
// the candidate is obtained after the newton algorithm applied to a selected
// point
{

  KuramotoPoint(xs,kurashi.dim,chooseksp(kurashi.mu));

  cout << "moving point to section " << endl;
  int ok=PMap->transporttosection(xs,xb); //  <xb> is now  on section
                       // in section coordinates
  PMap->pInsection->sectiontofull(xb,xs);



  if(!ok){
    cerr << "unsuccessfull - probably a bigger tmax is  needed " << endl;
     #if defined(WIN95) && defined(KRAK)
       cerr << "press any key " << endl;
       waitBt();
    #endif
  }


  if(ok){
    cout << "nonrigorous search for a  fixed point " << endl;

    int max_iter=10;

    int fixp=1;
    double maxstep=0.1;
    int iter;
    ok=newtonsearch(*PMap,iter_num,fixp,xb,max_iter,maxstep,err,&iter);

    PMap->pInsection->sectiontofull(xb,xs);

    if(!ok){
      cerr << "Fixed point not found" << endl;
       #if defined(WIN95) && defined(KRAK)
         cerr << "press any key " << endl;
         waitBt();
      #endif
    }
  }
  return (ok);
}








void KSinclunstable(KuraShi &kurashi)
{

  int iterfixp=1;  // the fixed point (nonsymmetric) is for <iterfixp>-iterate
  int symperorbit=1; // symmetric orbit
  int testasymmetry=0;
  symorbit_question(&symperorbit,&testasymmetry);

  int manyattempts=0;    // tells whether we try various proof parameters - currently it is
                    //     the size only, later we will add the dimension

  int tailevolution=1; // tells if we allow for tail to evolve
  int uselocgalerr=0;

  double delta[MAX_DIM+1];
  double maxdelta,mindelta;

  InclPoincareParType param;
  param.h=1e-3;
  param.h=kurashi.timestep(param.h);
  param.tmax=4.0;
  param.onesteppar.order=3;
  param.onesteppar.method=TAYLORM;
  param.onesteppar.doC0intersection=1;
  param.onesteppar.onestepmethod=dfincl_C0simpleLo;
  param.onesteppar.orderdf=param.onesteppar.order;


  int emptytail=0;
  int startsec=2;
  int sec_perpendicular=0;
  int partition=50;
  double expcoeff; //  it should be  slightly bigger than 1.0

  int big_m=3*kurashi.dim;
  int set_partition=1;
  int coord=eigenvectors;
  int iternum=3;

  int numLohner;

  int dimdiag=kurashi.dim-1;

   #if defined(WIN95) && defined(KRAK)
    rootFrame.Clear();
  #endif

    cout << "-------------------" << endl;
  if(symperorbit)
    KSinfo_unstbsymper(kurashi.imu);
  else
    KSinfo_unstbper(kurashi.imu);

  param_ghpp_comp(kurashi,manyattempts,tailevolution,&startsec,
    &sec_perpendicular,&partition,&expcoeff,&numLohner,delta,&emptytail,
       &mindelta,&maxdelta,&iternum,&set_partition,param,&coord,&dimdiag);

  #if defined(KRAK) && defined(WIN95)
  rootFrame.Clear();
  #endif


  FPoincareMap *PMap;

  ksinclpreparePMap(&PMap,kurashi,startsec,symperorbit,param);


  double xs[MAX_DIM+1],xb[MAX_DIM+1];
     // xs and xb represent the same starting point, xb is on section
     //  (and expressed in section coordinates) xs contains the ambient space
     //   coordinates for xb


  double err;
  if(manyattempts)
    err=mindelta/10.0;
  else err=delta[1]/10.0;
  if(err < 1e-10)
    err=1e-6;


  // we prepare a candidate for our orbit
   if(!kscandidate_usnstb(kurashi,PMap,iterfixp,xb,xs,err)){
    delete  PMap;
    PMap=NULL;
    return;
  }

  if(sec_perpendicular){
    ksredefine_perpend_PMap(&PMap,kurashi,xs,xb,symperorbit);
  }
  // until this moment we have defined a deterministic Poincare map
  // and we know the candidate for the fixed point <xb>- this will be center
  // of our initial condition


  // NOW WE INITIALIZE DIFFERENTIAL INCLUSIONS

  interval *galerr;
  int boxnum=0;
  intervalmatrix1 *boxes;
  KuraShi_Tail *ks_tail;



  // computation of bounds and Galerkin errors

  int num_coord_iso=0;
  double bias=0;


  Galerkin_Proj_KuraShi *inclode;
  InclPoincareMap *inclPMap;
  InclPoincareMap *asymtestPMap;



  ksincl_createinclpmaps(&inclode,&inclPMap,&asymtestPMap,PMap,param,kurashi,&galerr,
     &boxes,&boxnum,&ks_tail,big_m,partition,bias,expcoeff,num_coord_iso,
             uselocgalerr,tailevolution,symperorbit,testasymmetry);

  #if defined(USE_LOGNORM)
  inclode->Ldelta_method= Ldelta_lognorm; //      the method used in LDelta computation
  inclode->Ldelta_norm=NORM_MAX;  // if the logarithmic norm is used in LDelta
                 // computation, then this parameter specifies the norm
  #endif


  char repname[80];
  strcpy(repname,catwork);
  strcat(repname,"ksunstb");
  strcat(repname,orbext);
  ofstream  report(repname,ios::app);
  ksinclcomp_description(report,kurashi,symperorbit,startsec,partition,
          expcoeff,num_coord_iso,tailevolution,0,sec_perpendicular,xs,numLohner);

#if defined(USE_LOGNORM)
  report << "USING LOG_NORM - MAX_NORM " << endl;
#endif

  report <<"An attempt to prove the existence of an unstable periodic orbit " << endl;

  int withtail=1;
  SetWithTail y0(kurashi.dim,galerr,withtail,ks_tail); // in fact it should'nt matter
             // what we plug-in as the second parameter (galerr),
             // because in y0 is the tail
             // is given by ks_tail, and y0.y=galerr - are ignored, and recomputed
             // when necessary


  if(emptytail){
    KuraShi_Tail ksty0(kurashi.dim,y0.ptail->big_m); // this defines ksty0=0
    y0.assigntail(ksty0);
  }

  if(!manyattempts){
    galunstb_rev_orientation(*PMap,*inclPMap,*inclode,xb,delta,numLohner,y0,report,coord,
             dimdiag,iternum,set_partition,testasymmetry,asymtestPMap);
  }else{
    incl_poinc_manysizes(galunstb_rev_orientation,*PMap,*inclPMap,
           *inclode,numLohner,xb,maxdelta,mindelta,y0,report,coord, dimdiag,
                        iternum,set_partition,testasymmetry,asymtestPMap);
  }
  delete  asymtestPMap;
  asymtestPMap=NULL;

  delete inclPMap;
  inclPMap=NULL;
  delete inclode;
  inclode=NULL;
  delete PMap;
  PMap=NULL;
  delete [] boxes;
  boxes=NULL;
  delete [] galerr;
  galerr=NULL;

  delete ks_tail;
  ks_tail=NULL;

  cout << "the report is in [" << repname << "]" << endl;
   #if defined(WIN95) && defined(KRAK)
  //  if(!manyattempts)
  //    waitBt();
  #endif

}





