
#include<iostream>
#include<fstream>
using namespace std;

#include<cstdlib>

#include"defo.h"
#include"coordc.h"
#include"zdynsys.h"
#include"section.h"
#include"perfixp.h"
#include"func.h"
#include"zgfiles.h"
#include"multiint.h"
#include"qrdecmp.h"
#include"oneshot.h"
#include"fnames.h"






void inflate_mix_intervals(interval x,interval y,interval *xret,double cmix)
// we assume that y \subset x
// we inflate a bit <y> and return it in <xret> , on output y \subset xret \subset x
//  cmix \in [0,1)
{
  if(!y.subset(x)){
    cerr << "inflate_mix_intervals - y is not contained in x " << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  double le,re;
  if(x.left_b() < y.left_b()){
    le=cmix*x.left_b() + (1.0 - cmix)*y.left_b();
    if(le < x.left_b()) le=x.left_b();  // this could eventually happen
                     // due to roundoff errors or cmix not in [0,1]
  } else{
    le=x.left_b();
  }
  if(x.right_b() > y.right_b()){
    re=cmix*x.right_b() + (1.0 - cmix)*y.right_b();
    if(re > x.right_b()) re=x.right_b();     // this could eventually happen
                     // due to roundoff errors or cmix not in [0,1]
  } else{
    re=x.right_b();
  }
  *xret=interval(le,re);
}



void shift_mod_interval(int mustbeinx1, interval x1,interval x2,interval *xnew,double cmix)
// on output xnew =center(x2) + [-1,1]*d/2  (eventually intersected with x1)
//          where d <= diam(x1) and
//      if diam(x2) < diam(x1) then d=(1-cmix)diam(x2) + cmix diam(x1)
//   if mustbeinx1, then we additionally take care to make sure that xnew is contained in x1
/// In particular from the construction it follows that on output the following
//  condition is true:
//           if x2 \subset x1, then x2 \subset xnew
//
{

  if((cmix >=1) || (cmix <= 0)){
    cerr << "shift_mod_interval - cmix=" << cmix << " is out of range (0,1) " << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  double d=diameter(x1).right_b();
  double diamx2=diameter(x2).left_b();
  double c=x2.mid().left_b();
  if(diamx2 < d){
    d=(1.0-cmix)*diamx2 + cmix*d;
  }
  d=d/2;
  *xnew=interval(c-d,c+d);
  if(x2.subset(x1)){
    // in this case we should already have x2 \subset *xnew
    // by the construction of *xnew, but this could be destroyed by round-off
    // errors, hence :
    *xnew=Ihull(*xnew,x2);
  }
  if(mustbeinx1){
    if(!xnew->subset(x1)){
      double le,re;
      le=xnew->left_b();
      if(le < x1.left_b()) le=x1.left_b();

      re=xnew->right_b();
      if(re > x1.right_b()) re=x1.right_b();
      *xnew=interval(le,re);
    }
  }
}




void  write_no_inclusion(ostream &report,const interval fx[],const interval x[],
        int dim,const char *nfx,const char *nx)
{
  report << "no inclusion for: " << endl;
  for(int i=1;i<=dim;i++){
    if(!in(fx[i],x[i],0)){
      report << "i=" << i << "  " << nfx << "[i]=" << fx[i] << "  " <<
        nx << "[i]=" << x[i] ;

      if(diameter(x[i])>0){
        interval r=diameter(fx[i])/diameter(x[i]);
        report <<  "  diam-ratio=" << r.right_b() << endl;
      }
      report << endl;
    }
  }
}

void compute_ratios(const interval r0[],const interval fx[],int dim,int *numbigratios,
        double *largestratio)
{
  *largestratio=0.0;
  *numbigratios=0;
  for(int i=1;i<=dim;i++){
    if(diameter(r0[i]) >0){
      double r=diameter(fx[i]).right_b()/diameter(r0[i]).right_b();
      if(r > 1) (*numbigratios)++;
      if(r > *largestratio) *largestratio=r;
    } else{
      (*numbigratios)++;
    }
  }
}





char* osa_shiftmethodname(char* name,int maxlen,int method)
{
  strcpy(name,"");
  if(method==middleofPN)
    strncpy(name,"new guess=center of PN",maxlen-1);

  if(method==NPN_midlept)
    strncpy(name,"new guess=center of the intersection of N  and PN",maxlen-1);

  return name;
}

char *osa_dptypename(char *name,int maxlen,int dptype)
{
  strcpy(name,"");
  if(dptype==dpuppertriangular)
    strncpy(name,"upper triangular matrix choosen as the model for Dp",maxlen-1);

  if(dptype==dpdiagonal)
    strncpy(name,"diagonal matrix choosen as the model for Dp",maxlen-1);

  return name;
}

int osa_coord(int sugested_coord)
// choice of the type of coordinates
{
  const int buflen=80;
  char name[buflen];
  if(strlen(osa_coordsname(name,buflen,sugested_coord))>0)
    cout << "suggested coordinates: " << name << endl;
  cout << "choose coordinates: " << endl;
  for(int i=0;i<10;i++){
    if(strlen(osa_coordsname(name,buflen,i))>0)
      cout << i<< " - " << name << endl;
  }
  int selcoord;
  cin >> selcoord;
  #if defined(WIN95) && defined(KRAK)
  cout << endl;
  #endif
  return selcoord;
}


int newguess(double d[],interval PN[],int dim,interval newfixp[],
   int newfpmethod)
// N=(-d[i],d[i])
// PN - is computed image of N
// On output:
//   newfixedp - will contain a new guess for fixed point, constrcuted according
//      to <newfpmethod>
{
  int ok=1;
  interval set[MAX_DIM+1];
  for(int i=1;i<=dim;i++)
    set[i]=PN[i];
  if(newfpmethod==NPN_midlept){

    interval di[MAX_DIM+1];
    for(int i=1;i<=dim;i++)
      di[i]=interval(-d[i],d[i]);

    ok=commonpart(set,di,set,dim);
  }
  if(ok){
    for(int i=1;i<=dim;i++)
      newfixp[i]=mid(set[i]);
  }
  return(ok);
}










void computemodeldp(int dptype,double dp[][MAX_DIM+1],LinCC &lincc,
      double dpmodel[][MAX_DIM+1])
// It may happen that later I will need to include the possibility of
// 2x2 blocks on the diagonal
{
  int pdim=lincc.dim;
  transLinearPart(dp,lincc,dpmodel);
  if(dptype==dpdiagonal){
    for(int i=1;i<=pdim;i++){
      for(int j=1;j<=pdim;j++){
        if(i!=j) dpmodel[i][j]=0.0;
      }
    }
  }
  if(dptype==dpuppertriangular){
    for(int i=1;i<=pdim;i++){
      for(int j=1;j<i;j++)
        dpmodel[i][j]=0.0;
    }
  }
}



double lipschitzpartoferror(interval intvDp[][MAX_DIM+1],
   double dpmodel[][MAX_DIM+1],int dim)
// the function returns a rigorous bound in the max norm (L^\infty) for
//  the difference intvDp - dpmodel
{
  interval diff[MAX_DIM+1][MAX_DIM+1];
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++)
      diff[i][j]=intvDp[i][j]-dpmodel[i][j];
  return(normMax(diff,dim).right_b());
}


int dpdiag_findproofpar(ostream &report,int dim ,double dpmodel[][MAX_DIM+1],
   interval eps[],interval dp[][MAX_DIM+1],double delta[],double h,int order,
    double d[],double *tstep)
// Function finds the shape of the set and time step, which appear to
//  be good for checking brouwer thm. in one shot
//  dpmodel - must be diagonal 2-dim matrix
// dim - dimension - now it must be 2
//  eps = x0 - <P>x0 is the estimate of numerical method error
//  dp - interval estimate for DP
//  delta[] - the inital set on which <l> and <eps> where computed
//       basically the return set <d> shall be contained in <delta>
//       ib fact the initial set is given as: \Pi_{i=1}^{dim}(-delta[i],delta[i])
//  h - time step for which <l> and <eps> where computed,
//       used here for control purpose: *tstep (on output) shall be less
//        than <h>
//  order - order of numerical method
//  ----------
//  Output parameters
//    d- the proposed set, again the set is  \Pi_{i=1}^{dim}(-d[i],d[i])
//   *tstep - proposed time-step
{

  int retval=1;
  double l[MAX_DIM+1][MAX_DIM+1];
  for(int i=1;i<=dim;i++)
    for(int j=1;j<=dim;j++)
      l[i][j]=abs(dp[i][j]-dpmodel[i][j]).right_b();


  for(int i=1;i<=dim;i++){
    if(fabs(dpmodel[i][i]) > 1){
      retval=0;
      report << i << "-th eignevalue not contracting :" << dpmodel[i][i] << endl;
      cout << i << "-th eignevalue not contracting :" << dpmodel[i][i] << endl;
    }
  }
  if(!retval)
    return(retval);

  double E[MAX_DIM+1];
  for(int i=1;i<=dim;i++)
    E[i]=0.5*diameter(eps[i]).right_b()/pow(h,order);


  // this is purely 2-dim
  double A,B,C,D;
  A=1.0 - fabs(dpmodel[1][1]) - l[1][1];
  B=l[1][2];
  C=l[2][1];
  D=1.0 - fabs(dpmodel[2][2]) - l[2][2];

  if((A<0) || (D<0) || (A*D - B*C < 0)){
    report << "Lipschitz error part too large - ignoring it " << endl;
    cout << "Lipschitz error part too large - ignoring it " << endl;
    A=1.0 - fabs(dpmodel[1][1]);
    B=0.0;
    C=0.0;
    D=1.0 - fabs(dpmodel[2][2]);
  }
  double r=(abs(eps[1]).right_b()*C + abs(eps[2]).right_b()*A)/(abs(eps[1]).right_b()*D +
         abs(eps[2]).right_b()*B);
  report << "ratio=" << r << endl;
  cout  << "ratio=" << r << endl;

  // now we now ratio, hence d[2]=r*d[1]
  //          but we have bounds d[i] <= delta[i]
  //  we investigate the intersection of the line d[2]=r*d[1] with these
  //    bound and we choose the point closer to zero
  //  these intersections are given by (delta[1],r*delta[1])
  //                                    and (delta[2]/r,delta[2])
  if(delta[1] < delta[2]/r){
    d[1]=delta[1];
    d[2]=r*delta[1];
  } else{
    d[1]=delta[2]/r;
    d[2]=delta[2];
  }


  *tstep=(A*d[1] - B*d[2])/E[1]; // this tstep^order
  double temp=(D*d[2]-C*d[1])/E[2];
  if(temp < *tstep) *tstep=temp;

  //  *tstep=power(*tstep,1.0/order);
  *tstep=exp((1.0/order)*log(*tstep));
  if(*tstep > h) *tstep=h;

  report << "proposed tstep=" << *tstep << endl;
  cout << "proposed tstep=" << *tstep << endl;

  return(retval);

}


double brouwer_desired_L(double dpmodel[MAX_DIM+1][MAX_DIM+1],int dim,
   double gapfactor)
// dpmodel - upper triangular
// dim - must be 2 now
//  computes L, so that r_1 - r_2 = gapfactor*(1-lambda[1])/A
{
  double lambda[3];
  for(int i=1;i<=dim;i++)
    lambda[i]=fabs(dpmodel[i][i]);
  double A=fabs(dpmodel[1][2]);

  double r1=(1.0 - lambda[1])/A;
  return(r1*(1.0-gapfactor)/(1.0/A + 1.0/(1.0-lambda[2])));
}

double brouwer_desired_2_ratio(double dpmodel[][MAX_DIM+1],double r1[],
    double theta[],int dim,double gapfactor)
// dpmodel - upper triangular
// dim - must be 2 now
//  r1 - first ratios
//  theta - estimate for x0 - P(x0)
//// function returns Delta_1 so that for second ratios we have
//     r2[1]-r2[2]=gapfactor*(r1[1]-r1[2])

{
  double lambda[3];
  for(int i=1;i<=dim;i++)
    lambda[i]=fabs(dpmodel[i][i]);
  double A=fabs(dpmodel[1][2]);

  double nom=fabs(theta[1])/A + fabs(theta[2])/(1.0-lambda[2]);
  double denom=(r1[1]-r1[2])*(1.0 - gapfactor);
  double d1=nom/denom;


  // we still have to make sure that r2[2] will be less than 1
  double d2=fabs(theta[2])/((1.0-r1[2])*(1.0 - lambda[2]));

  return max(d1,d2);
}




int dpuptrian_findproofpar(ostream &report,int dim ,double dpmodel[][MAX_DIM+1],
   interval eps[],double L,double delta[],double h,int order,
    double d[],double *tstep,int decomposeeps)
// Function finds the shape of the set and time step, which appear to
//  be good for checking brouwer thm. in one shot
//  dpmodel - must be upper traingular 2-dim matrix
// dim - dimension - now it must be 2
//  eps = x0 - <P>x0 is the estimate of numerical method error
// L - estimate for Lipschitz constant for P - dPmodel
//  delta[] - the inital set on which <l> and <eps> where computed
//       basically the return set <d> shall be contained in <delta>
//       ib fact the initial set is given as: \Pi_{i=1}^{dim}(-delta[i],delta[i])
//  h - time step for which <l> and <eps> where computed,
//       used here for control purpose: *tstep (on output) shall be less
//        than <h>
//  order - order of numerical method
//  ----------
//  Output parameters
//    d- the proposed set, again the set is  \Pi_{i=1}^{dim}(-d[i],d[i])
//   *tstep - proposed time-step
{

  int ok=1;

  double lambda[3];
  for(int i=1;i<=dim;i++)
    lambda[i]=fabs(dpmodel[i][i]);
  double A=fabs(dpmodel[1][2]);

  // testing if all eigenvalues are contracting
  int i=1;
  while((i<=dim) && ok){
    ok=(lambda[i] < 1);
    i++;
  }
  if(!ok){
    report << (i-1)<< "-eigenvalue is bigger than 1" << endl;
    cout << (i-1)<< "-eigenvalue is bigger than 1" << endl;
    return ok;
  }
  // end of testing eigenvalues

  if(1 <= (lambda[1] + L)){
    report << "L - too large " << endl;
    L=brouwer_desired_L(dpmodel,dim,0.1);
    report << "new L=" << L << endl;
  }

  double r1[3]; // first ratios
  r1[2]=L/(1.0-lambda[2]); // lower bound
  r1[1]=(1.0 - lambda[1] - L)/A; // upper bound

  report << "Interesting ratios : " << endl;
  report << "L/(1-l_2) = " << r1[2] << " < d_2/d_1 < " <<
    "(1 - l_1 - L)/A =" << r1[1] << endl;

  ok=((r1[2] < r1[1]) && (r1[2] <1));

  if(ok){
    cout << "1-st ratios ok. " << endl;
    report << "1-st ratios ok " << endl;
  } else{
    report << "1-st ratios incorrect " << endl;
    cout <<  "1-st ratios incorrect " << endl;
    L=brouwer_desired_L(dpmodel,dim,0.1);
    report << "new L=" << L << endl;
    ok=1;

    r1[2]=L/(1.0-lambda[2]); // lower bound
    r1[1]=(1.0 - lambda[1] - L)/A; // upper bound

    report << "Recomputed  ratios : " << endl;
    report << "L/(1-l_2) = " << r1[2] << " < d_2/d_1 < " <<
      "(1 - l_1 - L)/A =" << r1[1] << endl;
  }

  double ratio;
  d[1]=delta[1];

  double r2[3]; // second ratio's - including the error: x0 - Px0 (when
              // decomposeeps)
  int r2_ok=0;

  if(decomposeeps){
    double theta[3]; // estimates for x0 -P(x0) - without numerical errors
    for(int i=1;i<=dim;i++)
      theta[i]=abs(mid(eps[i])).right_b();

    r2[1]=r1[1]-theta[1]/(A*d[1]);
    r2[2]=r1[2]+theta[2]/(d[1]*(1-lambda[2]));


    ok=((r2[2] < r2[1]) && (r2[2] <1));

    if(ok){
      cout << "2-nd ratios ok. " << endl;
      report << "2-nd ratios ok " << endl;
      r2_ok=1;
    } else{
      report << "2-nd ratios bad " << endl;
      double d1=brouwer_desired_2_ratio(dpmodel,r1,theta,dim,0.1);
      report << "delta_1=" << d1 << "   will eventually fix it " << endl;
      r2_ok=0;
      ok=1;
    }
  }
  if(!r2_ok){
    if(decomposeeps)
      report << "using 1-st ratios " << endl;

    for(int i=1;i<=dim;i++)
      r2[i]=r1[i];
  }

  double c[MAX_DIM+1];
  if(decomposeeps)
    for(int i=1;i<=dim;i++)
      c[i]=0.5*diameter(eps[i]).right_b()/pow(h,order);
  else
    for(int i=1;i<=dim;i++)
      c[i]=0.5*abs(eps[i]).right_b()/pow(h,order);

  c[1]=c[1]/(d[1]*A);  // c_1/(A*\Delta_1)
  c[2]=c[2]/(d[1]*(1.0 - lambda[2])); // c_2/(\Delta_1*(1 - \lambda_2))
  double t=(r2[1] - r2[2])/(c[1]+c[2]);
  ratio=r2[2] + t*c[2];
  d[2]=ratio*d[1];


  *tstep=pow(t,1.0/order);
 // (*tstep)*=0.5;  // just to have more space - and this is essential
                  // if I don't have this factor things don't work even for
                  //  5e-4 and order=5

  if(*tstep > h) *tstep=h;

  report << "ratio = " << ratio << endl;
  report << "tstep = " << *tstep << endl << endl ;

  return ok;
}




int rigorous_dp(FPoincareMap &Pmap,interval x[],interval Px[],interval intvDp[][MAX_DIM+1],
    estimates *estm, int useothercoord/*=0*/,LinCC *pcoordch/*=NULL*/)
// the function  estimates DP and P(x) rigorously on the set x for the map <Pmap>
//  is <useothercoord> - then x on input is expressed in <*pcoordch> coordinates
//  and on output <Px> and <intvDp> is also expressed in these coordinates
{
 /* interval xc[MAX_DIM+1],r0[MAX_DIM+1];

  interval C[MAX_DIM+1][MAX_DIM+1];
  identity(C,Pmap.dim);
  copyvector(x,xc,Pmap.dim);
  DecomposeIMulti(xc,r0,Pmap.dim);
  int ok=Pmap.imapDE(xc,C,r0,Px,intvDp,useothercoord,pcoordch,useothercoord,
              pcoordch,estm); */

  Set xic(Pmap.dim,x);



  int ok=Pmap.imapDE(xic,Px,intvDp,useothercoord,pcoordch,useothercoord,
              pcoordch,estm);
  return(ok);
}


int  estimate_dp(FPoincareMap &Pmap,double xb[],double delta[],
    interval intvDp[][MAX_DIM+1],int realonly)
// the function tries to estimate DP on the set xb + [-delta[i],delta[i]]
//  for the map <Pmap>
//  if <realonly> - double (nonrigorous) computations
//         otherwise we do first C1-computations, only if they look bad we
//       eventually perform nonrigorous ones
{

  double toldp=0.2;  // this is a factor used to estimate

  estimates estm;

  interval d[MAX_DIM+1];
  for(int i=1;i<=Pmap.dim;i++)
    d[i]=xb[i]+interval(-delta[i],delta[i]);

  int ok=1;
  if(!realonly){
    interval P_d_trash[MAX_DIM+1];
    rigorous_dp(Pmap,d,P_d_trash,intvDp,&estm);
    if(ok){
      // now we will test if the spread of intvDp is reasonable
      interval middp[MAX_DIM+1][MAX_DIM+1];
      mid(intvDp,middp,Pmap.dim);
      double diamby2=diameter(intvDp,Pmap.dim).right_b()/2.0;
      double maxtermmiddp=abs(middp[1][1]).right_b();
      for(int i=1;i<=Pmap.dim;i++)
        for(int j=1; j<=Pmap.dim;j++)
          if(abs(middp[i][j])>maxtermmiddp)
            maxtermmiddp=abs(middp[i][j]).right_b();

      if(diamby2/maxtermmiddp > toldp){
        // the diameter of intDp is too large compared to maximal mid tern
        ok=0;
        cout << "rigorous C1 estimates for Dp unreasonable for eigenvectors " << endl;
      }
    }
  }


  // now we will look for something nonrigorous

  if((!ok) || realonly){
    cout << "computing empirical estimates for Dp " << endl;
    double x[MAX_DIM+1];
    double Dp[MAX_DIM+1][MAX_DIM+1];
    ok=Pmap.map(xb,x,Dp);

    if(ok)
      tointerval(Dp,intvDp,Pmap.dim);


    int maxnumpts=4;
    int ptno=1;
    while((ptno<=maxnumpts) && ok){

      double xs[MAX_DIM+1];
      // a generation of a new point
      for(int i=1;i<=Pmap.dim;i++)
        xs[i]=xb[i]+zglrandom(-delta[i],delta[i]);


      ok=Pmap.map(xs,x,Dp);
      for(int i=1;i<=Pmap.dim;i++)
        for(int j=1;j<=Pmap.dim;j++)
          intvDp[i][j]=Ihull(intvDp[i][j],interval(Dp[i][j]));

      ptno++;
    }
  }
  return(ok);
}


int estimate_eps(FMap &Pmap,double xb[],interval eps[])
// On output :
//   <eps> is an enclosure of x0 - P(x0)
//  it contains basically two ingredients
//    the real difference x0 - P(x0)
//    P(x0) - computed here has some nonzero diameter which is due to
//           to numerics: round-off errors, discretization errors,
//            section errors
{

  interval x0[MAX_DIM+1]; // starting point
  for(int i=1;i<=Pmap.dim;i++)
    x0[i]=xb[i];
  int ok;


  cout << "C0 computation to get estimates for error term " << endl;
  ok=Pmap.imap(x0,eps);

  if(ok){
    for(int i=1;i<=Pmap.dim;i++)
      eps[i]=x0[i]- eps[i];
  }

  return ok;
}



int checkBrouwer(ostream &report,FPoincareMap &Pmap,double tstep,double d[],
      LinCC *coordch,interval PN[])
// function performs a check of the Brower theorem P(N) \subset N
///// the proposed set N=d - in new coordinates
////  *coordch - is the transition from standard to new coordinates,
//           origin in new coordintes is a candidate for fixed point

// On output: 1 - if assumptions of Brouwer theorem are satisfied
//              0 - otherwise
//      PN - contains enclosure for computed P(N)
//           in new coordinates

{

  FPoincareMap_param param=Pmap.get_param();
  // we have to build the set
  interval trash[MAX_DIM+1][MAX_DIM+1];
  interval x0[MAX_DIM+1];
  for(int i=1;i<=Pmap.dim;i++) x0[i]=0.0; // in new coordinates x0 is a origin


  interval newfixp[MAX_DIM+1]; // this will be center of the image set

  interval C0[MAX_DIM+1][MAX_DIM+1];
  identity(C0,Pmap.dim);
  interval r0[MAX_DIM+1];
  for(int i=1;i<=Pmap.dim;i++)
    r0[i]=interval(-d[i],d[i]);


  int proved;

  Pmap.isettimestep(tstep);
  cout << "Computing for  h=" << tstep << endl;

  Pmap.param.par_interval.write(report);
  if(Pmap.dim==2) report << "ratio=";
  else report << "ratios Delta_i/Delta_1:" << endl;
  for(int i=2;i<=Pmap.dim;i++)
    report << (d[i]/d[1]) << "   " ;
  report << endl;

  int useothercoord=1;
  estimates estm;

  //int ok=Pmap.imapDE(x0,C0,r0,PN,useothercoord,coordch,useothercoord,coordch,&estm);
  Set xstart(Pmap.dim,x0,C0,r0);
  int ok=Pmap.imapDE(xstart,PN,useothercoord,coordch,useothercoord,coordch,&estm);

  if(!ok){
    cout << "checkBrouwer - unable to compute Pmap " << endl;
    Pmap.set_param(param);
    return 0;
  }

  report << "PN (new coordinates) " << endl;
  saveAllCoord(report,PN,Pmap.dim,1);

  report << "The set N: " << endl;
  saveAllCoord(report,d,Pmap.dim,0);

  // now we ready to check
  proved=1;
  for(int i=1;i<=Pmap.dim;i++){
    if(abs(PN[i]).right_b()<d[i]){
      report << "i=" << i << "  ok " << endl;
    } else{
      report << "i=" << i << "  bad " << endl;
      proved=0;
    }
  }
  if(proved){
    report << "PROVED" << endl;
    report.flush();
    cout << "PROVED" << endl;
  }

  report << endl << "Estimates: " << endl;
  writeGlEstm(report,&estm,Pmap.igetorder(),fabs(Pmap.igettimestep()));

  Pmap.set_param(param);
  return proved;
}




int maponeshotattempt(FPoincareMap &Pmap,double xb[],double delta[],
            ofstream &report,int coord,int dimdiag)
//  the function computes the map <Pmap> and checks Brouwer theorem
//  xb - our approximate fixed point
//  delta[i] - starting tolerance for i-th coordinate, we will try to
//        perform proof inside these bounds, this serves also
//        as geometric indicator of shape when computing heuristic
//        constants
//  coord - possible values:
//         - standardcoord
//         - orthogonalizedeigenvectors;
//         - eigenvectors;
//         - partialeigenbasis
//         - eigen_subspaces_ortogonalized
{

  FPoincareMap_param param=Pmap.get_param();

  int ok;
  int proved=0;


//  double h=(Pmap.pode)->timestep(Pmap.dgettimestep());
//  Pmap.dsettimestep(h);

  cout << "Limit cycle search " << endl;
  int max_iter=20;
  double err=1e-6;
  ok=improvefixpoint(Pmap,1,xb,max_iter,err);

  if(ok)
    cout << "found" << endl;
  else{
    cout << "not found " << endl;
    Pmap.set_param(param);
    return ok;
  }


  interval ixp[MAX_DIM+1];
  for(int i=1;i<=Pmap.dim;i++)
    ixp[i]=interval(xb[i]);

  // periodic point was found



  int decomposeeps=1; // when guessing ratio and time step; it tells whether to decompose eps

//  int newfpmethod=middleofPN;   // take center of PN
  int newfpmethod=NPN_midlept;  // take center of intersection of N and PN


  int dptype=dpuppertriangular;
  if(coord==eigenvectors) dptype=dpdiagonal;


  // reporting the settings
  const int buflen=80;
  char buffer[buflen];
  cout << osa_coordsname(buffer,buflen,coord) << endl;
  cout << osa_dptypename(buffer,buflen,dptype) << endl;
  cout << osa_shiftmethodname(buffer,buflen,newfpmethod) << endl;


  report << endl << "---------------------------" << endl;
  report << osa_coordsname(buffer,buflen,coord) << endl;
  report << osa_dptypename(buffer,buflen,dptype) << endl;
  report << osa_shiftmethodname(buffer,buflen,newfpmethod) << endl;
  // end of reporting  the settings


   // now we have to compute parameters of our computational model
   // dp(model)
  //  bounds on P - dP(model) - this is the Lipschitz part of the error
  // and bounds for numerical scheme error (x0 - Px0) - \epsilon in my notes


  interval intvDp[MAX_DIM+1][MAX_DIM+1]; // computed bounds for Dp - not necessarily
               // rigorous - just estimate, but better pesimistic one
  interval eps[MAX_DIM+1];  // this is estimates (not neccesarily rigorous)
                    // x0 - P(x0)

  Pmap.param.par_interval.write(report);
  report << "delta - the set used to get  estimates for Dp" << endl;
  saveAllCoord(report,delta,Pmap.dim,0);

  cout << "C1 computation to get estimates for Dp " << endl;
  estimate_dp(Pmap,xb,delta,intvDp,0);

  estimate_eps(Pmap,xb,eps);


  double Dp[MAX_DIM+1][MAX_DIM+1];
  for(int i=1;i<=Pmap.dim;i++)
    for(int j=1;j<=Pmap.dim;j++)
      Dp[i][j]=mid(intvDp[i][j]).left_b();
  // now DP - contains the derivative of the Poincare map

  double dpmodel[MAX_DIM+1][MAX_DIM+1];

  LinCC *coordch=NULL;
  ok=buildcoordchange(coord,Dp,ixp,Pmap.dim,&coordch,dimdiag);
  if(!ok){
    return 0;
  }

  // coordch: is the transition from standard variables to the new ones

  computemodeldp(dptype,Dp,*coordch,dpmodel);  // dpmodel - is given in new
                               // coordinates

  // now we have to transform to new coordinates
  transLinearPart(intvDp,*coordch,intvDp);
  multAV(coordch->intv_m,eps,eps,Pmap.dim);   // now eps is expressed in new coordinates

  interval ddp[MAX_DIM+1][MAX_DIM+1];
  for(int i=1;i<=Pmap.dim;i++)
    for(int j=1;j<=Pmap.dim;j++)
      ddp[i][j]=intvDp[i][j] - dpmodel[i][j];

  report << endl << "Bounds on dP " << endl;
  saveArray(report,intvDp,Pmap.dim);


  report << endl << "dpmodel" << endl;
  saveArray(report,dpmodel,Pmap.dim);

  report << endl << "Bounds on dp - dpmodel: " << endl;
  saveArray(report,ddp,Pmap.dim);
  report << endl << "Epsilon: " << endl;
  saveAllCoord(report,eps,Pmap.dim,1);
  report << endl;

  double l=lipschitzpartoferror(intvDp,dpmodel,Pmap.dim);
  report << "L=" << l << endl;



  double d[MAX_DIM+1]; // this is our set
  double tstep; // this is proposed time step

  // this part below make sens only in 2D - upper triangular
  if(dptype==dpuppertriangular){
    if(Pmap.dim ==2){
      ok=dpuptrian_findproofpar(report,Pmap.dim,dpmodel,eps,l,delta,Pmap.igettimestep(),
               Pmap.igetorder(),d,&tstep,decomposeeps);
      if(tstep < 1e-3) tstep=0.5*Pmap.igettimestep();

    } else{
      cout << "not ready yet " <<endl;
      #if defined(WIN95) && defined(KRAK)
         waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
  }
  if(dptype==dpdiagonal){
    if(Pmap.dim ==2){
      ok=dpdiag_findproofpar(report,Pmap.dim,dpmodel,eps,intvDp,delta,Pmap.igettimestep(),
               Pmap.igetorder(),d,&tstep);
      if(tstep < 1e-3) tstep=0.5*Pmap.igettimestep();

    } else{
      cout << "not ready yet " <<endl;
      #if defined(WIN95) && defined(KRAK)
         waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
  }
  if(!ok){
    cout << "cannot find the good set " << endl;
    report << "cannot find the good set " << endl;
  }


  // now one has to do actual verification
  if(ok){
    int attempt=1;
    int maxnumattempts=5;
    double toolong=1e5;
    while(!proved && (attempt<=maxnumattempts) && (Pmap.gettmax()/tstep < toolong) && ok){
      interval PN[MAX_DIM+1];
      proved=checkBrouwer(report,Pmap,tstep,d,coordch,PN);
          // on output PN=Pmap.map(\Pi[-d[i],d[i]) , where PN and \Pi[\pmd[i]] are
          //     both expressed in coordinates <coordch>
      if(!proved){
        // we would like to update two things - time step and eventually
        //  shift the origin
        interval newfixp[MAX_DIM+1];
        ok=newguess(d,PN,Pmap.dim,newfixp,newfpmethod);
        if(!ok){
          cout << "Unable to construct the new center point " << endl;
          continue;
        }
        // newfixp - is given in new coordinates, on section - we have to go back
        //    to full standard coordinates
        //   we have to find out, the standard coordinates of this point
        coordch->coordinvtran(newfixp,newfixp); // now we have standard coordinates
                  // on section for <newfixp>

        coordch->shiftorigin(newfixp);
        if(attempt>1) tstep=tstep/2.0; // we will do second attempt with
                                       // an updated guess for fixed point
        attempt++;
      }
    }
  }

  delete coordch;
  coordch=NULL;

  Pmap.set_param(param);
  return proved;
}



void mapmanysizesattempt(int testfunc(FPoincareMap &,double[],double [],
   ofstream &,int,int), FPoincareMap &Pmap,double xb[],ofstream &report,
  double maxdelta,double mindelta,int coord,int dimdiag)
//  checks for various sizes that a map has a fixed point (testfunc)
//    PMap  - C1 map
//  xb - our approximate fixed point
//  maxdelta  and mindelta - bounds on the size of the sides of N
//  coord - possible values:
//         - standardcoord
//         - orthogonalizedeigenvectors;
//         - eigenvectors;
//         - eigen_subspaces_ortogonalized
//  parameters of testfunc: PMap, xcenter, delta[], report, coord,dimdiag

{

  double deltasize=maxdelta;
  double sizegrowth=0.5;
  double delta[MAX_DIM+1];

  int atleastonceproven=0;
  int lastproven=0;

  int theend=0;

  while(!theend){
    for(int i=1;i<=Pmap.dim;i++)
      delta[i]=deltasize;
    #if defined(WIN95) && defined(KRAK)
       rootFrame.Clear();
    #endif
    cout << "delta size=" << deltasize << endl;
    if(testfunc(Pmap,xb,delta,report,coord,dimdiag)){
      atleastonceproven=1;
      lastproven=1;
    } else{
      if(lastproven){
        // we try to increase order
        Pmap.isetorder(Pmap.igetorder()+1);
        if(testfunc(Pmap,xb,delta,report,coord,dimdiag)){
          atleastonceproven=1;
          lastproven=1;
        } else{
          lastproven=0;
          theend=1;
        }
      }
    }

    deltasize*=sizegrowth;
    if(deltasize < mindelta) theend=1;

  }
  if(atleastonceproven){
    cout << "something was proven" << endl;
    report << "THERE IS A PROOF FOR SOME DELTA" << endl;
  }
}





int C1computefixP(FPoincareMap &Pmap,double xb[],double delta[],
            ofstream &report,int coord,int dimdiag)
//  the function computes P(xb + delta) and dP(xb) - rigorously
//  xb - our approximate fixed point
//  delta[i] - starting tolerance for i-th coordinate
//  coord - possible values:
//         - standardcoord
//         - orthogonalizedeigenvectors;
//         - eigenvectors;
//         - partialeigenbasis
//         - eigen_subspaces_ortogonalized
{

  FPoincareMap_param param=Pmap.get_param();

//  double h=(Pmap.pode)->timestep(Pmap.dgettimestep()); // need to take a closer
              // look how I control the time step
//  Pmap.dsettimestep(h);

  cout << "Limit cycle search, improving initial guess " << endl;
  int max_iter=20;
  double err=1e-6;
  int ok=improvefixpoint(Pmap,1,xb,max_iter,err);

  if(ok)
    cout << "found" << endl;
  else{
    cout << "not found " << endl;
    Pmap.set_param(param);
    return ok;
  }

  int useothercoord=1;

  estimates estmC1;

 // interval intvDp[MAX_DIM+1][MAX_DIM+1]; // computed bounds for Dp - rigorous
  intervalmatrix1 *intvDp;
  intvDp=new intervalmatrix1[Pmap.dim+1];
  LinCC* coordch=NULL;
  interval ixb[MAX_DIM+1];
  tointerval(xb,ixb,Pmap.dim);

  // building coordinate change
  double Dp[MAX_DIM+1][MAX_DIM+1];
  if(coord!=standardcoord){
    cout << "C1 computation to get estimates for Dp - to build coordinate change " << endl;
    double zero[MAX_DIM+1];
    for(int i=1;i<=Pmap.dim;i++)
      zero[i]=0.0;

    estimate_dp(Pmap,xb,zero,intvDp,1);
    for(int i=1;i<=Pmap.dim;i++)
       for(int j=1;j<=Pmap.dim;j++)
         Dp[i][j]=mid(intvDp[i][j]).left_b();
     // now Dp - contains the derivate of the Poincare map
  }

  ok=buildcoordchange(coord,Dp,ixb,Pmap.dim,&coordch,dimdiag);
  // coordch: is the transition from standard variables to the new ones
  // our candidate for the fixed point is shifted to the origin in new coordinates
  if(!ok){
    delete [] intvDp;
    return 0;
  }

  const int buflen=80;
  char buffer[buflen];
  report << osa_coordsname(buffer,buflen,coord) << endl;
  Pmap.param.par_interval.write(report);
  report << "Initial guess " << endl;
  saveAllCoord(report,xb,Pmap.dim);
  report << endl;
  report << "delta - the initial set " << endl;
  saveAllCoord(report,delta,Pmap.dim,0);

  interval d[MAX_DIM+1],dn[MAX_DIM+1]; // this is our set



  for(int i=1;i<=Pmap.dim;i++)
    d[i]=interval(-delta[i],delta[i]);



  interval Px0[MAX_DIM+1]; // return value of Poincare map

  report << endl << endl ;
  report << "---------------- "<< endl;
  Pmap.param.par_interval.write(report);

  cout << "tstep=" << Pmap.igettimestep() << "  order=" << Pmap.igetorderdf() <<
    " size of the set=" << delta[1] <<  endl;
  cout << "C1 computation " << endl;


  time_t before,after;
  before=time(NULL);


  ok=rigorous_dp(Pmap,d,Px0,intvDp,&estmC1,useothercoord,coordch);

  after=time(NULL);



  if(ok){


    double compt_time=difftime(after,before);
    report << "C1 - comp. time="  << compt_time << "  sec " << endl;
    cout <<   "C1 - comp. time="  << compt_time << "  sec " << endl;


    report << endl << "P(d)" << endl;
    saveAllCoord(report,Px0,Pmap.dim,1);
    report << "max diameter=" << (diameter(Px0,Pmap.dim).right_b()) << endl;
    writediam(report,Px0,Pmap.dim);

    report << endl;

    report << endl << "DP: " << endl;
    saveArray(report,intvDp,Pmap.dim);
    report << "diameter=" << (diameter(intvDp,Pmap.dim).right_b()) << endl;
    report << endl;

    double bdp[MAX_DIM+1][MAX_DIM+1];
    for(int i=1;i<=Pmap.dim;i++)
      for(int j=1;j<=Pmap.dim;j++)
        bdp[i][j]=abs(intvDp[i][j]).right_b();
    report << endl << "fabs(DP)=" << endl;
    saveArray(report,bdp,Pmap.dim);
  } else{
    report << "unable to finish a C1-computation " << endl;
    cout << "unable to finish a C1-computation " << endl;
  }


  report << endl << endl << endl << endl;
  report.flush();


  if(coordch){
    delete coordch;
  }

  delete [] intvDp;
  Pmap.set_param(param);
  return ok;
}


//// Brouwer - via iteration

int simpleintervaliteration(FPoincareMap &PMap,double xb[],double delta[],
       ofstream &report,int coord,int dimdiag,int iternum,int divtable[],
                int dimdiv)
// xb - a good approximation the attracting fixed point
// iternum - the number of iterates
// this function takes the set N=xb + \\Pi_i[-delta[i],delta[i]]
//            (in coordinates specified by by <coord>)
//  if dimdiv >0 , then N_i is divided into divtable[i] equal parts
//         hence the set N is divided into \Pi_{1<= i<=divdim} divtable[i]}
// it computes P(N) ; for the next iterate N=P(N) \\cap N and so on
//// returns 1 - when PROVED P(N) \subset N
{
  cout << " ------------------------ " << endl;
  #if defined(WIN95) && defined(KRAK)
    rootFrame.Clear();
  #endif

  int proven=0;
  int ok=1;



  double d=delta[1];

  double Dp[MAX_DIM+1][MAX_DIM+1];
  //  int dimdiag=PMap.dim;

  if(coord!=standardcoord) {
    cout << "Real computation of Dp " << endl;
    double trash[MAX_DIM+1];

    ok=PMap.map(xb,trash,Dp);
    if(ok){
      const int buflen=80;
      char buffer[buflen];
      report << "Dp (nonrigorous) computed using h=" <<   PMap.dgettimestep() <<
                  "  order=" << PMap.dgetorder() << "  method=" <<
                   num_method_name(buffer,buflen,PMap.dgetmethod()) << endl;
      report <<  "Diagonalization dimension=" <<  dimdiag << endl;
    }else{
      cout << "not computed " << endl;
      return 0;
    }
  }

  interval ixp[MAX_DIM+1];
  tointerval(xb,ixp,PMap.dim);
  LinCC* coordch=NULL;
  if(ok)
    ok=buildcoordchange(coord,Dp,ixp,PMap.dim,&coordch,dimdiag);

  if(!ok){
    delete coordch;
    return  0;
  }


  report << "the starting point in initial section coordinates: " << endl;
  saveAllCoord(report,xb,PMap.dim);


  const int buflen=80;
  char buffer[buflen];
  report << osa_coordsname(buffer,buflen,coord) << endl;



  interval ixb[MAX_DIM+1],fx[MAX_DIM+1];
  zero(ixb,PMap.dim); // in our coordinate frame xb is the origin

  intervalmatrix1 *C;
  C=new intervalmatrix1[PMap.dim+1];
  identity(C,PMap.dim); // C - must be an identity for the intersection
                       // r0 \\cap Pmap.map(x0+C*r0)
                       // to work correctly
  interval r0[MAX_DIM+1];
  for(int i=1;i<=PMap.dim;i++)
    r0[i]=interval(-delta[i],delta[i]);

  PMap.param.par_interval.write(report);

  cout << "the size of the set " << d << "   ";
  cout << "h=" << PMap.igettimestep() << "  ";
  cout << "order=" << PMap.igetorder() << endl;

  estimates estm;

  int iterno=1;
  while(ok && (iterno<=iternum)){

    report << endl << "iterate=" << iterno << endl;
    report << " x0 + r " << endl;
    report << "x0: " << endl;
    saveAllCoord(report,ixb,PMap.dim,0);
    report << "r0:" << endl;
    saveAllCoord(report,r0,PMap.dim,1);

    cout << "iterate=" << iterno << "/" << iternum << endl;


    time_t before,after;
    before=time(NULL);



    int useothercoord=1;
    if(PoincMap_map_divide(PMap,divtable,dimdiv,ixb,C,r0,fx,useothercoord,
              coordch,&estm)){

      after=time(NULL);
      double compt_time=difftime(after,before);
      if(compt_time > 1){
        report << "Computation time="  << compt_time << "  sec " << endl;
        cout <<  "Computation time="  << compt_time << "  sec " << endl;
      }


      report << endl;
      report << "P(x+r)=" << endl;
      saveAllCoord(report,fx,PMap.dim,1);
      report << endl;

      report << "diameters(P(x+r)):" << endl;
      writediam(report,fx,PMap.dim);
      report << endl;


      if(d!=0){ // d - is an initial 1/2-diameter of r0
        report << "Ratios: diam(P(x+r))/diam(r): " << endl;
        int numbigratios=0;
        double maxratio=0;
        for(int i=1;i<=PMap.dim;i++){
          double r=diameter(fx[i]).right_b()/diameter(r0[i]).right_b();
          if(r > 1) numbigratios++;
          report << r << "  ";
          if(r > maxratio) maxratio=r;
        }
        report << endl;
        report << "Largest ratio = " << maxratio << endl;
        report << "Number of ratios > 1 is = " << numbigratios << endl;
        report << endl;
        cout << "Largest ratio = " << maxratio << endl;
        cout << "Number of ratios > 1 is = " << numbigratios << endl;
      }

      // now I would like to take intersection  of ixb+r0 \\cap fx
      interval x[MAX_DIM+1],cp[MAX_DIM+1];
      for(int i=1;i<=PMap.dim;i++)
        x[i]=ixb[i]+r0[i];

      int locproved=1;
      for(int i=1;((i<=PMap.dim) && locproved);i++){
        locproved=fx[i].subset(x[i]);
      }
      if(locproved){
        proven=1;
        report << "BROUWER - PROVED" << endl;
        cout << "BROUWER - PROVED" << endl;
      } else{
        report << "P(x) is not contained in x " << endl;
      }


      ok=commonpart(x,fx,cp,PMap.dim);
      if(!ok){
        report << "x \\cap f(x) = empty " << endl;
        cout << "empty intersection of x and fx" << endl;
      } else{
        copyvector(cp,ixb,PMap.dim);
        DecomposeIMulti(ixb,r0,PMap.dim);
      }

    }else{
      ok=0;
      report << "not computed " << endl;
      cout << "not computed " << endl;
    }
    iterno++;
  }

  if(ok){
    report << endl << "Estimates: " << endl;
    writeGlEstm(report,&estm,PMap.igetorder(),fabs(PMap.igettimestep()));

    if(coord!=standardcoord){
      // write down the matrix
      report << endl << "the coordinate change matrix (from old to new):" << endl;
      saveArray(report,coordch->intv_m,coordch->dim,0);
    }
  }
  delete coordch;
  delete [] C;
  return proven;
}

/// END - Brouwer by iteration




void testdiagfixedp(ostream &report,FMap *PMap,int mapiter,double xb[],int dimdiag)
{
  double dp[MAX_DIM+1][MAX_DIM+1],fx[MAX_DIM+1];
  if(PMap->iterate(mapiter,xb,fx,dp)){

    double xdiff[MAX_DIM+1];
    for(int i=1;i<=PMap->dim;i++){
      xdiff[i]=xb[i]-fx[i];
    }
    double normdiff=normEucl(xdiff,PMap->dim);
    report << "xb - in section coordinates " << endl;
    saveAllCoord(report,xb,PMap->dim);
    if(mapiter > 1){
      report << "P^" << mapiter ;
    } else{
      report << "P" ;
    }
    report << "(xb) - the image of xb  - in section coordinates " << endl;
    saveAllCoord(report,fx,PMap->dim);
    report << "Euclidean norm of the difference xb - P(xb) " << normdiff << endl;

    report << endl << "Approximate norm of DP" << endl;
    report << "euclidean="  << normEucl(dp,PMap->dim) << endl;
    report << "max="  << normMax(dp,PMap->dim) << endl << endl;

    report << "DP" ;
    if(mapiter > 1){
      report << "^" << mapiter;
    }
    report << endl;
    saveArray(report,dp,PMap->dim);





    report << "Number of coordinates used in diagonalization " <<
                       dimdiag << endl;

    // a diagonalisation
    cout << "diagonalization" << endl;
    wrdiag(report,dp,dimdiag);


    // the square of the above matrix
    report << endl << "now for tests we look at DP^2 " << endl;
    double dp2[MAX_DIM+1][MAX_DIM+1];
    ArrArr(dp,dp,dp2,PMap->dim);
    report << "DP" ;
    report << "^" << 2*mapiter;
    report << endl;
    saveArray(report,dp2,PMap->dim);

    report << "diagonalization of DP^2 " << endl;
    wrdiag(report,dp2,dimdiag);
    report << "end of DP^2 " << endl << endl;



    // additionally we write dp   in eigenvectors coordinates

    double dptran[MAX_DIM+1][MAX_DIM+1];
    interval ixb[MAX_DIM+1];
    tointerval(xb,ixb,PMap->dim);
    LinCC* coordch=NULL;

    int coord=eigenvectors;
    if(buildcoordchange(coord,dp,ixb,PMap->dim,&coordch,dimdiag)){
      transLinearPart(dp,*coordch,dptran);
      delete coordch;
      coordch=NULL;
      const int buflen=80;
      char buffer[buflen];
      report << endl << "DP in " << osa_coordsname(buffer,buflen,coord) << endl;
      saveArray(report,dptran,PMap->dim);
    }
    delete coordch;
  }
}



