#ifndef qrdecmp_h
#define qrdecmp_h


int qrdecomp(const double a[][MAX_DIM+1],double q[][MAX_DIM+1],
                   double r[][MAX_DIM+1],int pdim);
// on output Q  and R define a QR-decomposition for A
// if procedure fails Q=id



int iQRDecomp(const interval A[][MAX_DIM+1],interval Q[][MAX_DIM+1], int pdim);
// on output Q is an orthogonal factor in QR-decomposition for A
// if the procedure fails, then Q=id
// this a special QR-decomposition - required for Lohner algorthm


int qrdecomposition(const double a[][MAX_DIM+1],interval q[][MAX_DIM+1],
     int pdim,int maxiterno,double bias,double relbias);
// procedure performs QR-algorithm to find an orthogonal matrix
//  q such that
//   q^t a q - is almost upper triangular (with eventually 2x2 blocks
//       on the diagonal)


int partialqrdiagonalize(const double a[][MAX_DIM+1],double p[][MAX_DIM+1],
     double pinv[][MAX_DIM+1],double eigenval[][3],
     int pdim,int dimdiag,int maxiterno,double bias,double relbias,
     int normalize =1);
//   we diagonalize <a> using QR-decomposition
//   on output:
//       pinv = p^{-1}
//     p - is a coordinate change diagonalizing <a> - in the following (partial)
//            sense
//    pinv * a * p - is close to diagonal (complex eigenvalues give rise
//           to 2x2 cell) - in upper dimdiag X dimdiag - corner
//         the remaining part is upper triangular and is obtained from
//                 QR-decomposition
//     eigenvalues are given by eigenval[k][1] + i* eigenval[k][2]
//     we assume that each Jordan cell is simple
//  if <normalize> then we will try to have <p> balanced (if possible
//     the length of each column in <p> will be equal to 1 (for a 2x2 cell the
//     longer vector)), euclidean norm is used for this normalization


int qrdiagonalize(const double a[][MAX_DIM+1],double p[][MAX_DIM+1],
     double pinv[][MAX_DIM+1],double eigenval[][3],
     int pdim,int maxiterno,double bias,double relbias,int normalize=1);
//   we diagonalize <a> using QR-decomposition
//   on output:
//       pinv = p^{-1}
//     p - is a coordinate change diagonalizing <a>.
//    pinv * a * p - is close to diagonal (complex eigenvalues give rise
//           to 2x2 cell)
//     eigenvalues are given by eigenval[k][1] + i* eigenval[k][2]
//     we assume that each Jordan cell is simple
//    columns of <p> are 'eigenvectors' of a
//  if <normalize> then we will try to have <p> balanced (if possible
//     the length of each column in <p> will be equal to 1 (for a 2x2 cell the
//     longer vector)),  euclidean norm is used for this normalization


int ipartialqrdiagonalize(const double a[][MAX_DIM+1],interval p[][MAX_DIM+1],
     interval pinv[][MAX_DIM+1],double eigenval[][3],
     int pdim,int dimdiag,int maxiterno,double bias,double relbias, int normalize =1);
//   we diagonalize <a> using QR-decomposition
//   on output:
//       pinv = p^{-1}
//     p - is a coordinate change diagonalizing <a> - in the following (partial)
//            sense
//    pinv * a * p - is close to diagonal (complex eigenvalues give rise
//           to 2x2 cell) - in upper dimdiag X dimdiag - corner
//         the remaining part is upper triangular and is obtained from
//                 QR-decomposition
//     eigenvalues are given by eigenval[k][1] + i* eigenval[k][2]
//     we assume that each Jordan cell is simple
//  if <normalize> then we will try to have <p> balanced (if possible
//     the length of each column in <p> will be equal to 1 (for a 2x2 cell the
//     longer vector)), euclidean norm is used for this normalization


int iqrdiagonalize(const double a[][MAX_DIM+1],interval p[][MAX_DIM+1],
     interval pinv[][MAX_DIM+1],double eigenval[][3],
     int pdim,int maxiterno,double bias,double relbias,int normalize=1);
//   we diagonalize <a> using QR-decomposition
//   on output:
//       pinv = p^{-1}
//     p - is a coordinate change diagonalizing <a>.
//    pinv * a * p - is close to diagonal (complex eigenvalues give rise
//           to 2x2 cell)
//     eigenvalues are given by eigenval[k][1] + i* eigenval[k][2]
//     we assume that each Jordan cell is simple
//  if <normalize> then we will try to have <p> balanced (if possible
//     the length of each column in <p> will be equal to 1 (for a 2x2 cell the
//     longer vector)),  euclidean norm is used for this normalization


int calcnounstable(const double df[][MAX_DIM+1],int pdim,int type,int *no_unstabledir,
    double *stabfactor);
//  df - is pdim x pdim matrix - describing linearization for flow or a map
//  type = 1 - flow
//  type = 0 - a map
//  on output *no_ustabledir - is equal to the number of unstable directions
//            *stabfactor - max Real part of eigenvalue when flow
//                        - max absolute value of eigenvalue when flow
// returns 1 - when diagonalization procedure went fine
//         0 - otherwise



int qrinverse(const double a[][MAX_DIM+1],double c[][MAX_DIM+1],int dim);
//  the function computes c=a^{-1} using the QR-decomposition
//  1 - is returned if inversion was possible
//  0 - otherwise

int wrdiag(ostream &report,const double df[][MAX_DIM+1],int dim,int dimpartdiag =0,
    long maxiterno =10000 ,double bias = 1e-8,double relbias =1e-4);
// the results of the diagonalization of <df> are sent to the file <report>
//  dim - the dimension of the square matrix <df>
// if dim==2 - the diagonalization is performed using explicit formulas
//  otherwise we use the QR-decomposition algorithm


int wrdiag(const char fname[],const double df[][MAX_DIM+1],int dim);
// the results of the diagonalisation of <df> are sent to the file <fname>
//  dim - the dimension of the square matrix <df>
// if dim==2 - the diagonalisation is performed using explicite formulas
//  otherwise we use QR-decomposition algorithm

int iwrdiag(ostream &report,const double df[][MAX_DIM+1],int dim,int dimpartdiag =0,
    long maxiterno =1000 ,double bias = 1e-8,double relbias =1e-4 );
// the results of the diagonalization of <df> are sent to the file <report>
//  dim - the dimension of the square matrix <df>
// if dim==2 - the diagonalization is performed using explicit formulas
//  otherwise we use the QR-decomposition algorithm



void QR_rearrange(int dim,const interval M[][MAX_DIM+1],const interval r[],const interval z[],
         interval Bn[][MAX_DIM+1],interval rn[],bool look_at_shape_r);
// our task is to find Bn*rn, such that:  M*x.r + z \subset Bn*rn
// and Bn - is orthogonal
//// it is ok if r,z,rn are pointers to the same array
///     but M and Bn CANNOT be the same pointers
//  if look_at_shape_r==1, then the shape of r  is taken into account



int test_orthogonal(const interval b[][MAX_DIM+1],int dim);
// returns 1 - if B^T*B contains identity
//         0 - otherwise
//// 1 on return, this is does not implicate that B contains the orthogonal matrix


bool EigenVectOrthogonalizeInClusters(int dim,const double p[][MAX_DIM+1],
        const double eigenval[][3],  interval pout[][MAX_DIM+1] );
 // we orthogonalize columns in p  according to cluster-decomposition
 // the cluster is defined as a transitive closure of the relation
 //    abs(p[i]|p[j])/|p[i]|*|p[j]| > 0.9 (or some other constant)
 //  <eigenval> - serves only to decide if we have a pair corresponding to complex 
 //             eigenvalue or single eigenvector
 //            this is currently not used, but later it may turn out that it 
 //             makes sense to take it into account
//  <pdim> - the dimension of matrices
 // On output: pout - contains the 'clusterorthgonalized' basis   as columns


#endif
