
#include<iostream>
#include<fstream>
using namespace std;

#include<cstdlib>

#include"defo.h"
#include"rigkserr.h"
#include"multiint.h"
#include"kserr.h"
#include"vfkso.h"
#include"zgfiles.h"

const int inf=0;
const int sup=1;



interval bound_from_gl_norm_estm(int j,interval imu,interval rho,interval rho1)
  // JKT bounds for <j> -th mode
{
  interval ret;
  ret=4*sqrt(2*pi*rho*rho1*rho1*rho1);
//  ret=ret/(imu*lambda(j) - sqrt(lambda(j)));
  ret=ret/(imu*power(interval(j),4) - power(interval(j),2));
  return(ret);
}

void inittaberr(interval err[],int big_m,interval imu, interval rho,interval rho1,
    interval* c, int* s)
  //initiates bounds for trajectory using JKT approach
  // first from ball in L_2 (rho) and in H_1 (rho1)
  // and then Gronwall type bounds:
  //     4sqrt(2 * pi rho rho1^{3}}/(mu  - k^{-2})
{
  interval terr=0.0;
  for(int n=1;n <=big_m; n++){
    err[n]=rho;
    if(err[n]<rho1/n)
      err[n]=rho1/n;

//    if( imu > 1.0/sqrt(lambda(n))){
    if( imu > power(interval(n),-2)){
      terr=bound_from_gl_norm_estm(n,imu,rho,rho1);
      if(terr<err[n])
        err[n]=terr;
    }
    err[n]=interval(-err[n].right_b(),err[n].right_b());
  }
  interval temp=1.0/(interval(big_m+1)*(big_m+1)); // 1/M^2
  *c=4*sqrt(2*pi*rho*rho1*rho1*rho1)/(imu - temp);
  *c=Right(*c);
  *s=4;
}

void rigboundsinfsum(interval a[],interval C, int s,int m, int big_m,
       interval sum[], interval& D)
// On input: a[i] - a bound for i-th coordinate for i<=big_m
//            for i > big_m the bound is given by: C/i^s*[-1,1]
// On output:  \Sum_{ n \in \mathbb{Z}} a_n a_{k-n} \subset sum[k] for k > m, k<=big_m
//  for k > M (big_m)  |\Sum_{n \in \mathbb{Z}} a_n a_{k-n}| <= D/k^{s-1]
/////////////////////////
/// \Sum_{n \in \mathbb{Z}} a_n a_{k-n} =  FS(k) - 2 * IS(k)
//         where  FS(k)=\sum_{n=1}^{k-1} a_n a_{k-n}
//                IS(K)=\sum_{n=1}^\infty a_n a_{n+k}
/////// the need for this function arise from the following representation
//        of  the KS-equation with odd periodic boundary conditions
//          a_k'=\lambda_k a_k - k*(\Sum_{n \in \mathbb{Z}} a_n a_{k-n})
//      but since a_{-k}=-a_k, then we obtain
//            \Sum_{n \in \mathbb{Z}} a_n a_{k-n} =  FS(k) - 2 * IS(k)
{

  interval *invpowers;
  invpowers=new interval[2*big_m+2];

  for(int k=1;k<=2*big_m+1;k++)
     invpowers[k]=power(interval(k),-s);


  // BEGIN OF COMPUTATION of sum[k] , for  m<k <= M
  for(int k=m+1;k<=big_m;k++){

    // computation of fsk=\sum_{n=1}^{k-1} a_n a_{k-n} =
    //   2 \sum_{n=1,n<k/2}a_n a_{k-n} + e(k) a_{k/2}^2, where e(k)=1, when k is even
    //                                                        e(k)=0, when k is odd
    interval fsk=0.0;
    for(int n=1;2*n<k;n++) fsk+=(a[n]*a[k-n]);
    fsk*=2;
    if(k%2==0)  fsk+=power(a[k/2],2);



    // computation of IS(k) - as in Lemma 3.1 in paper with Mischaikow

    // t1=\sum_{n=1}^{M-k} a_n a_{k+n} -
    interval t1=0.0;
    for(int n=1;n<=big_m-k;n++)
      t1+=a[n]*a[k+n];



    // computation of C * [-1,1] * \sum_{n=M-k+1}^{M} |a_n|/(k+n)^s
    interval t2=0.0;
    for(int n=big_m - k + 1;n<=big_m;n++){
      t2+=Right(abs(a[n])*invpowers[k+n]);
    }
    t2=Right(C*t2);
    t2=t2*interval(-1.0,1.0);

    // C^2/((k+M+1)^s (s-1) M^{s-1})  * [-1,1]
    interval t3;
    t3=invpowers[k+big_m+1]*invpowers[big_m]*big_m*power(C,2)/(s-1);
    t3=Right(t3);
    t3=t3*interval(-1.0,1.0);

    interval isk=t1+t2+t3;
    sum[k]=fsk - 2*isk;


  }
  // END OF   COMPUTATION of sum[k] , for  m<k <=M



  // the computation of the tail of sum

  interval absansum=0.0; // \sum_{n=1}^{M} |a_n|
  for(int n=1;n<=big_m; n++)
    absansum=Right(abs(a[n]) + absansum);


  // form lemmas 3.4 and 3.5  in Zgliczynski + Mischaikow paper
  // we estimate the FS(k) - so that
  // |FS(k)| <= dfsk/k^{s-1}
  interval maxfsTimesK; // here we will have max_{M < k <= 2M} FS(k)*k^{s-1}
  int kmax=0; // the index where the above max is achieved
  for(int k=big_m+1;k<=2*big_m;k++){

    interval fskTksm1;

    interval fsk1=0;  // fsk1=2\sum_{n=1}^{k/2} a_n a_{k-n} + e(k) a^2_{k/2}
    for(int n=k-big_m;2*n < k;n++){
      fsk1+=a[n]*a[k-n];
    }
    fsk1*=2;
    if(k%2==0) fsk1+=power(a[k/2],2);

    interval fsk2=0; // fsk2=abs(2C*[-1,1]\sum_{n=1}^{k-M-1} \frac{|a_n|}{(k-n)^s})
    for(int n=1;n<=k-big_m-1;n++){
      fsk2+=abs(a[n])*invpowers[k-n];
    }
    fsk2*=2*C;
    fsk2=Right(fsk2);

    fskTksm1=(abs(fsk1)+fsk2)*power(interval(k),s-1);
    fskTksm1=Right(fskTksm1);

    if(k==big_m+1){
      maxfsTimesK=fskTksm1;
      kmax=big_m+1;
    }else {
      if( maxfsTimesK < fskTksm1){
        maxfsTimesK=fskTksm1;
        kmax=k;
      }
    }

  }
  // at this moment we have : |FS(k)| <= maxfsTimesK/k^{s-1}, for M < k <= 2M


  // from Lemma 3.5 we have an estimate for FS(k)  for k > 2M
  interval dfsk;
  dfsk=power(interval(2),s+1)*absansum/interval(2*big_m+1);
  dfsk+=C*power(interval(4),s)*invpowers[2*big_m+1]/(2*big_m+1);  // in the formula it is (2M+1)^{-(s+1)},
                           //   but invpowers[2*M+1]=(2*M+1)^{-s}, hence it is ok
  dfsk+=C*power(interval(2),s)*invpowers[big_m]/(s-1);
  dfsk*=C;
  dfsk=Right(dfsk);  // now we have |FS(k)| <= dsfk/k^{s-1}, for k > 2M

  if(maxfsTimesK > dfsk) dfsk=maxfsTimesK;

  // now we have |FS(k)| <= dsfk/k^{s-1}, for k > M


  interval dIS; // from Lemma 3.6 we will obtain IS(k) <= dIS/k^{s-1}
  // where dIS=C/(M+1) *(\frac{C}{(M+1)^{s-1} (s-1)}  + \sum_{n=1}^M|a_n|)
  dIS=C*invpowers[big_m+1]*interval(big_m+1)/(s-1);
  dIS+=absansum;
  dIS*=C/(big_m+1);
  dIS=Right(dIS);

  D=Right(2*dIS+dfsk);

  delete [] invpowers;
}




void   selfconsbounds(const interval inerr[],interval err[],interval c1,interval &c2,
        int s1,int *s2,bool iso[],bool *isotail,int m,int big_m, interval imu,
        const interval b[],interval d)
//  Computation of isolation for KS equation:
// ON INPUT:
//  inerr[k] - k=m+1,..,big_m - current bounds
//    for k>big_m - the bound is c1/k^s1
//  b[k] - a bilinear term   for m < k <= big_m
//   and  d/k^{s1 -2} for k > big_m
//   m - is the number of frozen modes
//   iso[k], m<k <= big_m - tells that we have an isolation in k-th directon
//       *isotail - an isolation for all k>big_m
// ON OUTPUT:
//    err[k] - contains new bounds for m <k <=big_m
//      for k > big_m - the bound is given by c2/k^s2
//   iso[k], m<k <= big_m - tells wether we have an isolation in k-th directon
//       *isotail - an isolation for all k>big_m
////////////////
//  we compute the isolation for
//     - l_k  q_k + Left(b[k])  <
//            \frac{d q_k}{dt} < - l_k q_k + Right(b[k])
///  where l_k=imu *k^4 - k^2
//     then:
//       u_k= Right(b[k])/ l_k
//       d_k= Left(b[k])/ l_k
//  u_k and d_k are new isolating bounds, err[k]=interval(d_k,uk_k)
///////////////
//  as for far tail (for k > big_m) it may happen then when we iterate bounds
//   and obtained higer power in the denomnator, the our estimate get worse and
//   we cannot verify the isolation with the present formulas, but we know that
// it is there (by the construction). In this case the value of *isotail -
//       is unchanged and if it was true (on input), then is also true on output
{
  interval terr;
  for(int k=m+1; k<=big_m; k++){
    err[k]=inerr[k];
    if(imu > power(interval(k),-2)){
        terr=b[k]/(imu*power(interval(k),4) - power(interval(k),2));

      if((terr.right_b() <= err[k].right_b()) &&
            (terr.left_b()>=err[k].left_b()))
        iso[k]=true;

      err[k]=interval(max(err[k].left_b(),terr.left_b()),
         min(err[k].right_b(),terr.right_b()));
    }
  }

  // updating for k > big_m
  c2=c1;
  *s2=s1;

  interval big_m2=power(interval(big_m),2);
  interval temp=d/(big_m2*(imu - 1.0/big_m2));
  if(Right(temp) <= Right(c1)){ // this is equivalent to C1/(M+1)^{s1} >= d/((mu - 1/(M+1)^2)(M+1)^{s1+2} )
                         // with the term on the right hand side representing the candidate for
                         // the new isolation
                         // bounds, which decay faster we have the isolation for the tail
                         // C2/k^{s1+2}, where C2=   d/(mu - 1/(M+1)^2)
    *isotail=true;
    c2=d/(imu - 1.0/big_m2);
    *s2=*s2+2;
  }
}


void  ksgalerr_point(const interval q[],const interval& C,const int s,
          const int m,const int big_m, interval err[])
// AN RIGOROUS FUNCTION
// computes the difference between the galerkin projection and the actual
//  vector field after projection. i.e.
//   Q_k F(p + q) - Q_k F(p)  \subset err[k]
//  where F -  KS vector field,
//   p - is defined by q[1..m]
//   and q[m+1,\dots,big_m] - is the near tail in self-consistent apriori bounds:
//      |q[k]| < C/k^s for k > big_m - is the far tail
//  ON OUTPUT:
//   err[k] \supset 2k \sum_{n=m-k+1}^{\infty} q_n q_{k+n}, for k=1,\dots,m
/////////// in my notes  big_m=M
{
// the sum is split in three parts: P_1, P_2 , P_3
// P_1=\sum_{n=m-k+1}^{M-k} q_n q_{n+k} - both terms are in q[1,\dots,m]
// P_2=\sum_{n=M-k+1}^M q_n q_{n+k} - the first term is q[1,\dots,m]
// P_3= \sum_{n=M+1}^M q_n q_{n+k} - boths terms are from the far tail
//// terms P1,P_2 - are computed by strightforward computation
// to estimate P_3 we use Lemma 3.1 from Zgliczynski,Mischaikow[2001]
//     we use the estimate for third term for IS(K) in this lemma


//  invpowers[k]=k^{-s}
  interval *invpowers;
  invpowers=new interval[m+big_m+2];

  for(int k=1;k<=m+big_m+1;k++)
     invpowers[k]=power(interval(k),-s);

  interval C2mod; // C2mod=C^2/((s-1)M^{s-1})
  C2mod=(C*C*invpowers[big_m]*big_m)/(s-1);
  C2mod=Right(C2mod); // it will be used in the estimate for
                 // the term   \sum_{n=M+1}^\infty |q_n| * |q_{n+k}| \leq
                 //     C^2 \sum_{n=M+1}^\infty 1/(n^s (n+k)^s) \leq
                 // \frac{C^2}{(k+M+1)^s (s-1) M^{s-1}} = \frac{C2mod}{(k+M+1)^s}

  for(int k=1;k<=m;k++){
    interval t1=0.0;
    //  computation of the \sum_{n=m-k+1}^{M-k} q_n q_{k+n}
    for(int n=m-k+1;n <= big_m-k;n++)
      t1+=q[n]*q[n+k];


    // computation \sum_{n=M-k+1]^M |q_n|/(k+n)^s
    interval t2=0.0;
    for(int n=big_m-k+1;n<=big_m; n++)
      t2+=Right(abs(q[n]))*invpowers[k+n];

    t2*=C;
    t2=Right(t2);


    // computation of: \sum_{n=M+1}^\infty |q_n|*|q_{n+k}| \leq
    //          \frac{C^2}{(k+M+1)^s (s-1) M^{s-1}}
    interval t3=C2mod*invpowers[k+big_m+1];
    t3=Right(t3);

    err[k]=t1+(t2+t3)*interval(-1.0,1.0);
  }

  for(int k=1;k<=m;k++){
    err[k]= 2*k*err[k];
  }
  delete [] invpowers;
}


void riggalerrks(interval err[],int galdim,const interval q[],int m,int big_m,interval c,
    int s,const interval boxes[][MAX_DIM+1],int boxnum)
// computes  difference between the galerkin projection and the actual
//  vector field after projection, the projection is <galdim>-dimensional i.e.
//   Q_k F(p + q) - Q_k F(p)  \subset err[k], for k=1..<galdim>
//  where F -  KS vector field,
//  the set, S, over which it is computed is given
//   \sum_{i=1}^{boxnum} boxes[i] \times \Pi q[m < i \leq big_M ] \times_{i > M} [\pm C/i^s]
//  q[k>m],C,s -  the tail  for selfconsistent apriori bounds:
//      q[m+1,...,big_m] and |q[k]| < C/k^s for k > M(=big_m)
////////////
//  galdim - the dimension Galerkin projection
///  p \in P_{galdim}S, q \in (I-P_{galdim})S
////////
/////////////////
  // it must hold : <galdim> <= big_m, dim(boxes[i])=m
  // usually galdim==m
{

  interval *lq;
  lq=new interval[big_m+1];
  for(int k=m+1;k<=big_m;k++)
    lq[k]=q[k];


  interval *lerr;
  lerr=new interval[galdim+1];

  for(int bno=1;bno<=boxnum;bno++){
    for(int k=1;k<=m;k++)
      lq[k]=boxes[bno][k];

    ksgalerr_point(lq,c,s,galdim,big_m,lerr);
    if(bno==1){  // first loop
      for(int k=1;k<=galdim;k++)
        err[k]=lerr[k];
    }
    else{
      for(int k=1;k<=galdim;k++)
        err[k]=Ihull(lerr[k],err[k]);
    }
  }
  delete [] lerr;
  delete [] lq;
}





void rigkscompbilinearterm(interval b[],const interval err[],int m, int big_m,interval c,
               int s,interval &d,const interval boxes[][MAX_DIM+1],int boxnum)
  // estimates bilinear term  -k * Sum_{i=-\infty}^{i=\infty} q_i q_{k-i}
  // on OUTPUT b[k] contains bounds for bilinear term for m <k <=big_m
  //                  d/k^{s-2} is a bound for bilinear term for k
  // including tail - RIGOROUS PROCEDURE
  //  INPUT PARAMETERS:
  // <big_m> - dimension of more carrefull estimates ; err[k]  k <=big_m
  // <m> - which represents the number of first
  //   modes for which bounds are fixed in advance (master variables)
  // <err[k]> - bound for i-th coordinate (when big_m => k> m)
  //  for k > big_m, error estimate is <c>/k^<s>
  // boxes[1..<boxnum>] - is <n> dimensional covering covering of the invariant
  //     set for <m>-dimensional Galerkin projection
{

//  interval lerr[rks_big_dim];
  interval *lerr;
  lerr=new interval[big_m+1];
  for(int i=m+1;i<=big_m;i++)  lerr[i]=err[i];

  interval ld;   // this is 'local D' i.e. part of the bound b[i]\leq ld/i^{s-2}
  // interval lb[rks_big_dim]; // the local bound (for a given box) for bilinear term
  interval *lb;
  lb=new interval[big_m+1];


  int firsttime=1;


  int pointno=0;
  while(pointno < boxnum){
    pointno++;

    for(int i=1;i<=m;i++)
      lerr[i]=boxes[pointno][i];

    rigboundsinfsum(lerr,c,s,m,big_m,lb,ld);

    for(int k=m+1;k<=big_m;k++)
         lb[k]=-k*lb[k];


    if(firsttime){
      for(int k=m+1;k<=big_m;k++)
        b[k]=lb[k];
      d=ld;
      firsttime=0;
    } else{
      for(int k=m+1;k<=big_m;k++)
        b[k]=Ihull(lb[k],b[k]);
      if(Right(ld) > Right(d))
        d=Right(ld);
    }
  }
  delete [] lb;
  delete [] lerr;
}

void rigkscompbilinearterm(interval b[],const interval err[],int m, int big_m,interval c,
               int s,interval &d,const interval x[])
  // estimates bilinear term  -k * Sum_{n=-\infty}^{n=\infty} q_n q_{k-n}
  // on OUTPUT b[k] contains bounds for bilinear term for m <k <=big_m
  //                  d/k^{s-2} is a bound for bilinear term for k
  // including tail - RIGOROUS PROCEDURE
  //  INPUT PARAMETERS:
  // <big_m> -  BIG_M ; err[k]  k <=big_m
  // <m> - which represents the number of main modes
  // the set over which the nonlinear term is computed is given by
  // x - first m- variables
  // <err[k]> - bound for i-th coordinate (when M => k> m)
  //  for k > big_m, error estimate is <c>/k^<s>
{
//  interval lerr[rks_big_dim];
  interval *lerr;
  lerr=new interval[big_m+1];
  for(int i=1;i<=m;i++)
    lerr[i]=x[i];
  for(int i=m+1;i<=big_m;i++)  lerr[i]=err[i];

  rigboundsinfsum(lerr,c,s,m,big_m,b,d);

  for(int k=m+1;k<=big_m;k++)
    b[k]=-k*b[k];

  delete [] lerr;
}



void guessL2bounds(const interval boxes[][MAX_DIM+1],int boxnum,int dim,
           interval *rho, interval *rho1)
  // computes L_2 (euclidean) and H1 norm for set covered by <dim>-dimensional
  //    boxes <boxes>
  //  boxnum - the number of sets in <boxes>, the index run from 1
  //  ON OUTPUT
  //     rho - contains L_2 norm
  //     rho1 - L_2 norm of the gradient
{
  *rho=0.0;
  for(int bno=1;bno <= boxnum;bno++){
    interval t=Right(normEucl(boxes[bno],dim));
    if(t>*rho)  *rho=t;
  }

  *rho1=0.0;
  for(int bno=1;bno <= boxnum;bno++){
    interval t=Right(gradNormEucl(boxes[bno],dim));
    if(t>*rho1)  *rho1=t;
  }
}


void initdatabounds(interval datab[],int dim,const interval boxes[][MAX_DIM+1],
   int boxnum)
 //   for set covered by <dim>-dimensional
  //    boxes <boxes>
  //  <boxnum> - the number of sets in <boxes>, the index run from 1
  //  ON OUTPUT: datab[i] - contains bounds for i-th coordinate; i=1...dim
{
  for(int i=1;i<=dim;i++)
    datab[i]=boxes[1][i];
  for(int bno=2;bno<=boxnum;bno++){
    for(int i=1;i<=dim;i++)
      datab[i]=Ihull(datab[i],boxes[bno][i]);
  }

}



void rigkserr(bool usecurrvaltail,interval mu,const interval boxes[][MAX_DIM+1],
      int boxnum, int m, int big_m,  bool *tail_isol_ok, interval galerr[],
        interval e[],interval *c,  int *s,int iternum /* = 10 */ )
  // computes bounds for higher modes by iteration of Growall-type bounds
  // the set on which we perform computations is covered by <boxes>[1..<n>]
  //  the number of boxes is equal to <boxnum>
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]>
  //  <big_m>  (the variable M in my papers)-- defining the dimension
  //        in tail computations , must be      big_m>= m, but usually <big_m> is few times <m>,
  //  if (usecurrvaltail) then the computation starts with the values
  //              e,c,s - from input, in this case the input value
  //              of <tail_isol_ok> matters - tells if with current boxes,e,c,s
  //               we have already an isolation for k>m - this could have been
  //               done using some other approach..
  //       the tail is represented as follows:   e[k] - for m < k <= big_m
  //                                        |e[k]| \leq  C/k^s  for  big_m < k
  //  else
  //    otherwise bounds are generated from some guesses based on the norm estimates
  // ---------------------
  // on output the following data are send to a file:
  //   a tail description and the Galerkin projection errors in the vector field
  //   resulting from the tail
  //  tail_isol_ok -  is set to true - when self-consistent apriori bounds
  //               were found
  //    and then galerr[k] - contains the error term of the k-th coordinate
  //      comming from neglected modes via galerkin projection
  //   The new tail is returned in  e,c,s as follows
  //   e[k] - for m < k <= big_m
  //   |e[k]| \leq  C/k^s  for  big_m < k
{
  #ifdef WIN95
//    rootFrame.Clear();
  #endif


  char reportfname[80];
  strcpy(reportfname,catwork);
  strcat(reportfname,"arksb.txt");

//  interval e1[rks_big_dim],e2[rks_big_dim],inite[rks_big_dim]; // bounds for k <=big_m
  interval *e1,*e2,*inite;
  e1=new interval[big_m+1];
  e2=new interval[big_m+1];
  inite=new interval[big_m+1];

  interval initc,c1,c2;  // bounds for the rest of the tail
  int s1,s2,inits;    // are given by c1/k^{s1}  or c2/k^{s2}
//  int iso[rks_big_dim];  // iso[k]==true , iff we have an isolation for k-th coordinate
  bool *iso;
  iso=new bool[big_m+1];

  bool isotail;       //  isotail ==true, iff we have an isolation for the tail k>big_m

  if(usecurrvaltail){
    c1=*c;
    s1=*s;
    initc=*c;
    inits=*s;
    for(int k=m+1;k<=big_m;k++){
      e1[k]=e[k];
      inite[k]=e[k];
    }
  } else{
    interval rho=0.0;   // L_2 bound for attractor
    interval rho1=0.0;  // L_2 bound for grad on attractor
    guessL2bounds(boxes,boxnum,m,&rho,&rho1);
//    cout << "rho="<< rho  <<   "   rho1="  << rho1  << endl;
    inittaberr(e1,big_m,mu,rho,rho1,&c1,&s1);
    for(int i=1;i<=big_m;i++)
      inite[i]=e1[i];
    initc=c1;
    inits=s1;
  }

   // interval datab[rks_big_dim];
  interval *datab;
  datab=new interval[m+1];

  initdatabounds(datab,m,boxes,boxnum); // datab[i] - is the bound for i-th coordinate
                                // for i=1,\dots,m

  for(int k=1;k<=m;k++){
    e1[k]=datab[k];
    inite[k]=datab[k];
  }
  delete [] datab;
  datab=NULL;

  if((!usecurrvaltail) || (usecurrvaltail && !(*tail_isol_ok) )){
    for(int k=m+1;k<=big_m;k++)
      iso[k]=false;
    isotail=false;
  }
  if(usecurrvaltail && (*tail_isol_ok)){
     for(int k=m+1;k<=big_m;k++)
      iso[k]=true;
    isotail=true;
  }

  ofstream report(reportfname);
  writeCompInfo(report);
  writedatetime(report);
  report << " errors for " << m
           << " first coordinates  are read from input data  " << endl;
  report
    << " errors for other variables - computed by iterating Gronwall bounds "
      << endl;
  report << " Initial tail : " << c1.right_b() << "/k^" << s1 << endl;
  report << "-----------------------------------------" << endl;


//  interval b[rks_big_dim+1];
  interval *b;
  b=new interval[big_m+1];
  interval d;
  for(int it=1;it <= iternum; it++){
    rigkscompbilinearterm(b,e1,m,big_m,c1,s1,d,boxes,boxnum); // generates new bounds
                     //for bilinear term and puts them in b[m+1,\dots,big_m] and d/k^{s1-2}
    selfconsbounds(e1,e2,c1,c2,s1,&s2,iso,&isotail,m,big_m,mu,b,d);

    for(int k=m+1;k<= big_m;k++) e1[k]=e2[k];
    c1=c2;
    s1=s2;
  }
  delete [] b;
  b=NULL;

  writeCompTailestm(report,inite,e1,m,big_m,c1,s1,iso,isotail);


  *tail_isol_ok=(isotail && all_true_in_tab(iso,m,big_m));

  if(*tail_isol_ok){

    // filling-in output data about  the self-consistent tail
    for(int i=m+1;i<=big_m;i++)
      e[i]=e1[i];

    *c=c1;
    *s=s1;
    //  end of filling-in output data

    int galdim=m;

  //  cout << "Computing Galerkin projection errors in vector field" << endl;
    riggalerrks(galerr,galdim,e1,m,big_m,c1,s1,boxes,boxnum);
    writeestm(report,galerr,galdim,"Galerkin errors");
  }
  else {
    #if defined(WIN95) && defined(KRAK)
      rootFrame.Clear();
    #endif
    cout << "NO - Isolation for tail" << endl;
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
  }

  delete [] iso;
  delete [] e1;
  delete [] e2;
  delete [] inite;
}






interval sum_ak(int l,int Big_M,const interval e[],interval c,int s)
// this function returns upper bound for \sum_{k \geq l} |a_k|
//  the bounds are: e[k] for 1 \leq k \leq Big_M
//                  |e[k]| \leq C/k^s  for k >= Big_M+1
{
  interval ret=0;
  if(l<=Big_M){
    for(int i=l;i<=Big_M;i++)
      ret+=abs(e[i]).right_b();
  }
  int startl=l;
  if(startl < Big_M +1) startl=Big_M +1;
  ret+=c/((s-1)*power(interval(startl-1),s-1));
  return(ret);
}

interval sum_j_akn(int l,int n,int Big_M,const interval e[],interval c,int s)
// this function returns upper bound for \sum_{j \geq l} |ja_{j+n}|
//  the bounds are: e[j] for 1 \leq j \leq Big_M
//                  |e[j]| \leq C/j^s  for j >= Big_M+1
{
   interval ret=0;
   for(int j=l;j<=Big_M-n;j++)
     ret+=j*abs(e[j+n]);
   int r=l;
   if((Big_M-n+1) > r) r=Big_M-n+1;
   ret+=c*(interval(1)/((s-2)*power(interval(r+n-1),s-2)) -
      interval(1)/((s-1)*power(interval(r+n),s-1)));
   return Right(ret);
}



interval ks_d2B(int k,int i,int j)
// calculates D^2f_k/Dx_iDx_j (x)
{
  interval retval=0.0;
  if ((-i+j==k) || (i-j==k)) retval=2.0;
  if (i+j==k) retval=-2.0;

  retval*=k;
  return(retval);
}



interval ks_dB(int i, int j,const interval a[],int Big_M,interval c,int s)
//  returns the partial derivative of the bilinear part of KS
//  dB_i/da_j (a)
//    a - is Big_M-dimensional vector
//     a_k \in a[k] for k=1...Big_M  and |a_k|<= C/|k|^s  for k > Big_M
{
  interval ret;
  if(i==j){
    if(2*i <= Big_M)
      ret=a[2*i];
    else{
      ret=c/power(interval(2*i),s);
      ret=Ihull(ret,-ret);
    }
  }
  if(j<i){
   // ret=-a[i-j]+a[i+j];
   if( i-j <= Big_M)
     ret=-a[i-j];
   else {
     interval t=c/power(interval(i-j),s);
     ret=Ihull(-t,t);
   }
   if(i+j <= Big_M)
     ret+=a[i+j];
   else{
     interval t=c/power(interval(i+j),s);
     ret+=Ihull(-t,t);
   }
  }
  if(j>i){
    // ret=a[j-i] + a[i+j];
    if(j-i <= Big_M)
      ret=a[j-i];
    else{
      interval t=c/power(interval(j-i),s);
      ret=Ihull(-t,t);
    }
    if(i+j <= Big_M)
      ret+=a[i+j];
    else{
      interval t=c/power(interval(i+j),s);
      ret+=Ihull(-t,t);
    }

  }
  ret=2*i*ret;
  return(ret);
}


interval ks_db_nc(int i,int j,const interval a[],int m,int Big_M,interval c,int s,
               const LinCC &coch)
// returns \frac{\partial {\tilde B}_i}{\partial x_j}(a) - of the bilinear term
//  in KS -equations  expressed in the new coordinates
//   a - is Big_M - dimensional vector
//     a_k \in a[k] for k=1...Big_M  and |a_k|<= C/|k|^s  for k > Big_M
//   coch - is the change of coordinates in m-first variables
//      for higher variables the coordinate change is diagonal
//  MUST BE:    (i>m or j>m)
//  the formula is:
//      if   i<=m and j>m   \sum_{k<m} A_{ik} db_k/dx_j
//      if  j>m  and i >m - db_i/da_j
//      if  i>m and  j<=m  - \sum_{l<=m} db_i/da_l * A^{-1}_{lj}
//               where A is 'm'-matrix in coch
{
  interval ret;

  if((i<=m) && (j<=m)){
    cerr << "ks_db_nc - i,j <= m  i=" << i << "  j=" << j << "  m=" << m << endl;
    #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  if((j>m) && (i>m))
    ret=ks_dB(i,j,a,Big_M,c,s);

  if((i>m) && (j<=m)){
    ret=0;
    for(int l=1;l<=m;l++){
      ret+=ks_dB(i,l,a,Big_M,c,s)*coch.intv_mInv[l][j];
    }
  }
  if((i<=m) && (j>m)){
    ret=0;
    for(int k=1;k<=m;k++){
      ret+=coch.intv_m[i][k]*ks_dB(k,j,a,Big_M,c,s);
    }
  }
  return(ret);
}







int findcell(int /*cellnum */,int cells[][MAX_DIM+1],int coord)
  // the function returns -  the index of the cell
  //  containing 'coord'
  //  cellnum - number of blocks in linearization - in dim 1--m
  //       a block is 1-dimensional for a real eigenvalue
  //       and is 2-dimensional for a complex eigenvalue
  //     degenerate situations are not treated here - number of eigenvectors matches the number
  //            of eigenvalues
  //  cells[i][1]..cells[i][cells[i][0]] - contain the indices of variables
  //       in the block
{
  int c=0;
  int found=0;
  while(!found){
    c++;
    for(int i=1;((i<=cells[c][0]) && !found);i++)
      found=(coord==cells[c][i]);
  }
  return(c);
}

////// DerData - functions

DerData::DerData(int pdim_tab)
{
  unique_test_on=false;
  lognorm_computed=false;
  instability_test_on=false;
  cc_test_on=false;

  dim_tab=pdim_tab;
  tab_l=new double[dim_tab+1];
  tab_a=new double[dim_tab+1];
  tab_u=new double[dim_tab+1];
  tab_offdiag=new double[dim_tab+1];
  tab_du=new double[dim_tab+1];
  tab_dl=new double[dim_tab+1];
  tab_offdiag_unstb=new double[dim_tab+1];
  dfnc=new intervalmatrix1[MAX_DIM+1];

  m_dim=0;
}

DerData::~DerData()
{
  delete [] tab_l;
  delete [] tab_a;
  delete [] tab_u;
  delete [] tab_offdiag;
  delete [] tab_du;
  delete [] tab_dl;
  delete [] tab_offdiag_unstb;
  delete [] dfnc;
}

int DerData::setnewdim(int pdim_tab)
    // dim_tab=pdim_tab; returns previous value
   // deletes old tables and assigns  new
{
  int retval=dim_tab;
  delete [] tab_l;
  delete [] tab_a;
  delete [] tab_u;
  delete [] tab_offdiag;
  delete [] tab_du;
  delete [] tab_dl;
  delete [] tab_offdiag_unstb;

  dim_tab=pdim_tab;
  tab_l=new double[dim_tab+1];
  tab_a=new double[dim_tab+1];
  tab_u=new double[dim_tab+1];
  tab_offdiag=new double[dim_tab+1];
  tab_du=new double[dim_tab+1];
  tab_dl=new double[dim_tab+1];
  tab_offdiag_unstb=new double[dim_tab+1];
  return(retval);
}

void DerData::assign_tabl(int i, double val)
//  tab_l[i]=val
{
  if(i<=dim_tab) tab_l[i]=val;
}

void DerData::assign_taba(int i, double val)
//  tab_a[i]=val
{
  if(i<=dim_tab) tab_a[i]=val;
}

void DerData::assign_tabu(int i, double val)
//  tab_u[i]=val
{
  if(i<=dim_tab) tab_u[i]=val;
}


void DerData::assign_taboffdiag(int i, double val)
//  tab_u[i]=val
{
  if(i<=dim_tab) tab_offdiag[i]=val;
}

void DerData::assign_tabdu(int i, double val)
//  tab_du[i]=val
{
  if(i<=dim_tab) tab_du[i]=val;
}


void DerData::assign_tabdl(int i, double val)
//  tab_dl[i]=val
{
  if(i<=dim_tab) tab_dl[i]=val;
}

void DerData::assign_taboffdiag_unstb(int i, double val)
 // tab_offdiag_unstb[i]=val
{
  if(i<=dim_tab)  tab_offdiag_unstb[i]=val;
}



void DerData::wr_derdata(ostream &out,bool fullreport) const
{

  bool write_coord=(dim_tab > 5);
  if(unique_test_on){
    if(unique){
      out << "uniqueness - proved ";
      if(complexeigenvalues){
        out << "  complex eigenvalue,  im=" << complex_off_diag.right_b();
      }
    }
    else{
      out << "uniqueness - inconclusive, for " << uniq_worstco <<
        "-th coord =" << -uniq_badmargin.right_b() ;
      if(complexeigenvalues){
        out << "  complex eigenvalue,  im=" << complex_off_diag.right_b();
      }
    }
    out <<  "  alpha=" << alpha ;
    out << endl;




    if(fullreport){
      out <<  " u_i :" << endl;
      for(int i=1; i<=dim_tab;i++){
        if(write_coord) out << "(" << i <<")";
        out << tab_u[i] << "  " ;
      }
      out << endl;

      out <<  " diag terms for u_i :" << endl;
      for(int i=1; i<=dim_tab;i++){
        if(write_coord) out << "(" << i <<")";
        out << tab_du[i] << "  " ;
      }
      out << endl;



      out << " offdiagonal_i :" << endl;
      for(int i=1; i<=dim_tab;i++){
        if(write_coord) out << "(" << i <<")";
        out << tab_offdiag[i] << "  " ;
      }
      out << endl;
    }
  }

  if(lognorm_computed){
    out << "  Log norm (max)=" << lognorm.right_b() << "   for coord="
      << maxlognorm << endl;

    if(fullreport){
      out << "l_i :" << endl;
      for(int i=1; i<=dim_tab;i++){
        if(write_coord) out << "(" << i <<")";
        out << tab_l[i] << "  " ;
      }
      out << endl;

      out <<  " diag terms for l_i :" << endl;
      for(int i=1; i<=dim_tab;i++){
        if(write_coord) out << "(" << i <<")";
        out << tab_dl[i] << "  " ;
      }
      out << endl;
    }
  }

  //  the results from instability test
  if(instability_test_on){
    if(unstable) out << "instability - proved" << endl;
    else out << "instability - inconclusive" << endl;
    out << "d=" << d_unstable << "  should be > 0" << endl;
    out << "worst 'a_i' for " << minunstable << "-th coordinate " << tab_a[minunstable] << "  should be > 0 " << endl;
    if(fullreport){
      out << "a_i :" <<endl ;
      for(int i=1; i<=dim_tab;i++){
        if(write_coord) out << "(" << i <<")";
        out << tab_a[i] << "  " ;
      }
      out << endl;
      out << "off-diagonal terms_i:" << endl;
      for(int i=1; i<=dim_tab;i++){
        if(write_coord) out << "(" << i <<")";
        out << tab_offdiag_unstb[i] << "  " ;
      }
      out << endl;
    }
  }
  // results from cone-conditions computations
  if(cc_test_on){
    if(cc_ok) out << "cone condition - proved" << endl;
    else out << "cone condition - inconclusive" << endl;
    out << "mingap=" << cc_mingap << "  should be > 0" << endl;
    out << "the 'worst' direction " << cc_index_mingap << "-th coordinate "
         << endl;

  }

  // end - results from cone-conditions computations



  if(fullreport){
    if(m_dim !=0){
      out << "dF in new coordinates: " << endl;
      int write_row_col=(m_dim >=6);
      saveArray(out,dfnc,m_dim,1,write_row_col);
    }
  }

}

#if defined(WIN95) && defined(KRAK)
void DerData::wr_derdata(FRAME &out) const
{
  if(unique_test_on){
    if(unique)
      out << "uniqueness - proved ";
    else{
      out << "uniqueness - inconclusive, for " << uniq_worstco <<
        "-th coord  error margin=" << uniq_badmargin.right_b() ;
      if(complexeigenvalues){
        out << "  complex eigenvalue,  im=" << complex_off_diag.right_b();
      }
    }
    out << " alpha = " << alpha << endl;
  }

  if(lognorm_computed){
    out << "  Log norm (max)=" << lognorm.right_b() << "   for coord="
      << maxlognorm << endl;
  }

   //  the results from instability test
  if(instability_test_on){
    if(unstable) out << "instability - proved" << endl;
    else out << "instability - inconclusive" << endl;
    out << "d=" << d_unstable << "  should be > 0" << endl;
    out << "worst a for " << minunstable << "-th coordinate " << tab_a[minunstable] << endl;
  }

 if(cc_test_on){
    if(cc_ok) out << "cone condition - proved" << endl;
    else out << "cone condition - inconclusive" << endl;
    out << "mingap=" << cc_mingap << "  should be > 0" << endl;
    out << "the 'worst' direction " << cc_index_mingap << "-th coordinate "
         << endl;

  }

}
#endif

// end of DerData functions


interval sum_nondiagks(int i,int cellno,LinCC &coch,int cellnum,
     int cells[][MAX_DIM+1],int m,int Big_M,interval e[],interval c,int s,
     interval dF_NC[][MAX_DIM+1])
// computes off diagonal terms for KS-equation for uniqueness and stability/instability
//   computation for a block which is specified as follows
//   if <cellno <=celnum> then it is computed for cells[cellno] - block
//   otherwise is 1-D is given by <i>
//   dF_NC = dF in new block coordinates
{
  interval offdiag;
  if(cellno<=cellnum){
    int celldim=cells[cellno][0]; // the dimension in current block
    offdiag=0;
    //  off-diagonal term have three parts
    //   \sigma_{j<=m, j \not cells[cellno]} sqrt(\sum_{l} (dF_NC_{i_l}/dxj)^2 )
    //   \sigma_{m < j <=Big_M} sqrt(\sum_{l} (dF_NC_{i_l}/dxj)^2 )
    //         but here we have to call dB(in new coord)
    //   \sigma_{j>Big_M} =  \sum_l \sum_{k<=m} 2k|A_{i_l,k}|(S(Big_M+1 -k)  +  S(Big_M+1 +k) )
    //


    for(int j=1;j<=m;j++){
      int idummy;
      if(!intab(j,cells[cellno],1,celldim,&idummy)){
        interval sum2(0.0);
        for(int l=1;l<=celldim;l++){
          sum2+=power(dF_NC[cells[cellno][l]][j],2);
        }
        offdiag+=sqrt(sum2);
      }
    }

    for(int j=m+1;j<=Big_M;j++){
      interval sum2(0.0);
      for(int l=1;l<=celldim;l++){
        sum2+=power(ks_db_nc(cells[cellno][l],j,e,m,Big_M,c,s,coch),2);
      }
      offdiag+=sqrt(sum2);
    }

    for(int l=1;l<=celldim;l++){
      for(int k=1;k<=m;k++){
        offdiag+=2*k*abs(coch.intv_m[cells[cellno][l]][k])*(sum_ak(Big_M+ 1 - k,Big_M,e,c,s) +
                       sum_ak(Big_M + 1 + k,Big_M,e,c,s));
      }
    }
  }
  if((i>m) && (i<=Big_M)){
    // here the off-diagonal term is:
    //   \sum_{j <= Big_M} |d N(in new coord)_i/dx_j |  + 2i(S(M+1 -i) + S(i+M+1))
    offdiag=0;
    for(int j=1;j<=Big_M;j++){
      offdiag+=abs(ks_db_nc(i,j,e,m,Big_M,c,s,coch));
    }
    offdiag+=2*i*(sum_ak(Big_M+ 1 - i,Big_M,e,c,s) +
                          sum_ak(Big_M + 1 + i,Big_M,e,c,s));
  }

  if(i> Big_M){
    // here the off-diagonal term is:
    //   2*i*m*(S(i-m)+S(i+1))*max A^{-1}[k,l]   +
    //           2*i (S(i+m+1)+2S(1))

    interval maxAinv;  // this will be max_{k,l=1..m}A^{-1}_{kl}
    maxAinv=coch.intv_mInv[1][1];
    for(int k=1;k<=coch.dim;k++)
      for(int j=1;j<=coch.dim;j++)
        maxAinv=Ihull(maxAinv,coch.intv_mInv[k][j]);

    maxAinv=Right(abs(maxAinv));

    offdiag=2*i*m*(sum_ak(i-m,Big_M,e,c,s)+ sum_ak(i+1,Big_M,e,c,s))*maxAinv;
    offdiag+=2*i*(sum_ak(i+m+1,Big_M,e,c,s)+ 2*sum_ak(1,Big_M,e,c,s));
  }
  return offdiag;
}

void diag_terms_ks(int i,int cellno,int cellnum,
     int cells[][MAX_DIM+1],int m,int Big_M, interval dF_NC[][MAX_DIM+1],
     interval imu,interval & diagt, interval &diagtlognorm)
// computes  diagonal terms for KS-equation for uniqueness and logarithmic norm
//   computation for a block which is specified as follows
//   if <cellno <=celnum> then it is computed for cells[cellno] - block
//   otherwise is 1-D is given by <i>
//   dF_NC = dF in new block coordinates
//  on output:
//     diagt - diagonal term for uniqueness computation
//     diagtlognorm -    diagonal term for logarithmic nomr computation
////////
// WARNING: Can handle only 1D and 2-D blocks
//           in 2-D block a complex eigenvalue is assumed
{
  if(cellno<=cellnum){
   if(cells[cellno][0]==1){
    i=cells[cellno][1];
    diagt=dF_NC[i][i];
    diagtlognorm=diagt;
   }else{ // we have a two-dimensional cell
    int i0=cells[cellno][1];
    int i1=cells[cellno][2];
    i=i0;
    interval alpha,beta;   //   we want the matrix
                       //  alpha beta
                       //  -beta alpha
                       //  in the block (i0,i1)
    interval eps; // it will contain the difference - between the
                    //  the actual matrix   dF_NC and this block
    alpha=(dF_NC[i0][i0]+dF_NC[i1][i1])/2;
    beta=(dF_NC[i0][i1]-dF_NC[i1][i0])/2;
    alpha=mid(alpha);
    beta=mid(beta);
    eps=dF_NC[i0][i0] - alpha;
    eps=Ihull(eps,dF_NC[i1][i1] - alpha);
    eps=Ihull(eps,dF_NC[i0][i1]-beta);
    eps=Ihull(eps,dF_NC[i1][i0]+beta);
    eps=Right(abs(eps));

    diagt=sqrt(alpha*alpha + beta*beta) - 2*eps;  //  2*eps is a matrix norm
                           // of two-dimensional matrix with all its entries
                           // less or equal to <eps>



    //  computation of lognorm of the cell
    diagtlognorm=Ihull(dF_NC[i0][i0],dF_NC[i1][i1]);
    diagtlognorm=Right(diagtlognorm);
    eps=(dF_NC[i0][i1] + dF_NC[i1][i0])/2;
    eps=Right(abs(eps));
    diagtlognorm+=eps;
    // end of comp. of lognorm
   }
  }
  if((i>m) && (i<=Big_M)){
    //  the diagonal term: i^2(1-\nu i^2)
    diagt=i*i*(1 - imu*i*i);
    diagtlognorm=diagt;
  }


  if(i> Big_M){
    //  the diagonal term: i^2(1-\nu i^2)
    diagt=i*i*(1 - imu*i*i);
    diagtlognorm=diagt;
  }
}

int uniqueness_lognorm(KuraShi &kurashi,LinCC &coch,int cellnum,
     int cells[][MAX_DIM+1],const interval boxes[][MAX_DIM+1],
    int boxnum,int m,int Big_M,interval e[],interval c,int s,
    DerData *dffxipdata)
//  the uniqueness and logarithmic norm  computation around fixed point
// the set on which we perform computations is covered by <boxes>[1..<m>]
//  the number of boxes is equal to <boxnum>
  //  coch - a linearlizing coordinates  ; warning: boxes are in standard coordinates
  //      dim(boxes)=dim(coch)= <m>
  //  cellnum - number of blocks in linearization - in dim 1--m
  //       a block is 1-dimensional for a real eigenvalue
  //       and is 2-dimensional for a complex eigenvalue
  //     degenerate situations are not treated here - the number
  //    of eigenvectors matches the number  of eignevalues
  //  cells[i][1]..cells[i][cells[i][0]] - contain the indices of variables
  //       in the block
  //
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]> and it must coincide with kurashi.dim
  //  <Big_M> - M - defining the dimension in tail computations , must be
  //     Big_M>= m, but usually <Big_M> is a few times <m>
  //   the tail is defined by  e,c,s as follows
  //   e[k] - for m < k <= Big_M
  //   |e[k]| \leq  C/k^s  for  Big_M < k
  // -----------------------------------
  //  the function returns 1 - when we have uniqueness of fixed point
  //                       0 - otherwise
  //   on output *dffxipdata - contains both logarithmic norms and uniqueness data
  //                         in the block-decomposition infinity norm
  //      ; IT IS ASSUMED that all blocks are at most 2_D-dimensional
  //      this assumption is explictly used in computations of diagonal terms only
{

  if(kurashi.dim!=m){

    cerr << "uniqueness_lognorm - Bad input data - kurashi.dim != m " << endl;
    #if defined(WIN95) && defined(KRAK)
      cout << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  // filling in e[k] for 1<=k <= m
  e[0]=0.0;
  for(int k=1;k<=m;k++){
    e[k]=boxes[1][k];
    for(int boxno=2; boxno <= boxnum;boxno++)
      e[k]=Ihull(e[k],boxes[boxno][k]);
  }
  //  now e - contains bounds for 1<=k <= Big_M
//  interval dF[MAX_DIM+1][MAX_DIM+1];  //
  intervalmatrix1 *dF;
  dF=new intervalmatrix1[m+1];
  for(int i=1;i<=m;i++)
    for(int j=1;j<=m;j++){
      dF[i][j]=ks_dB(i,j,e,Big_M,c,s);
      if(i==j) dF[i][j]+= i*i*(1.0 - kurashi.imu*i*i);
    }


//  interval dF_NC[MAX_DIM+1][MAX_DIM+1];  //  dF in new coordinates
  intervalmatrix1 *dF_NC;
  dF_NC=new intervalmatrix1[m+1];
  transLinearPart(dF,coch,dF_NC);
  delete [] dF;
  dF=NULL;

  dffxipdata->m_dim=m;
  copymatrix(dF_NC,dffxipdata->dfnc,m);

  dffxipdata->setnewdim(Big_M+1);

  int maxlognorm=0;  // here will be the coordinate for which the logartithmic
                 // norm was produced

  bool uniq_ok=true;
  interval lognorm;

  double maxbad;   // here is the margin by which we fail to establish uniqueness
  int comaxbad=0;  //  here is the coordinate for which this bad margin was realized
  int firsttime=1;
  int finished=0;
  int i=1;
  int cellno=1;
  while(!finished){
    interval loclognorm;
    int uniq_passed;
    interval uniq_margin;
    interval offdiag;
    interval diagt;    // a diagonal term used for uniqueness comp.
    interval diagtlognorm; // a diagonal term used for logarithmic norm
                          //   these terms differ for complex eigenvalues

    if(cellno<=cellnum){
      i=cells[cellno][1];
    }

    diag_terms_ks(i,cellno,cellnum,cells,m,Big_M,dF_NC,kurashi.imu,diagt,diagtlognorm);

    offdiag= sum_nondiagks(i,cellno,coch,cellnum,cells,m,Big_M,e,c,s,dF_NC);
    offdiag=Right(offdiag);
    loclognorm=diagtlognorm + offdiag;
    loclognorm=Right(loclognorm);
    diagt=Left(abs(diagt));

    //  updating the uniqness data
    uniq_margin= diagt - offdiag;
    if(!(uniq_margin >0)){
      uniq_passed=0;
    } else uniq_passed=1;

    if(!uniq_passed){
      uniq_ok=0;
      if(comaxbad==0){
        comaxbad=i;
        maxbad=abs(uniq_margin).right_b();
      } else{
        if(maxbad < abs(uniq_margin).right_b()){
          comaxbad=i;
          maxbad=abs(uniq_margin).right_b();
        }
      }
    }
    double uniq_ratio= (offdiag/diagt).right_b();
    if(firsttime){
       dffxipdata->alpha=uniq_ratio;
    } else{
      if(uniq_ratio > dffxipdata->alpha)  // we want max ratio in <alpha>
                                          // this ratio is decreasing for n > m_dim
                                          //  hence it is ok.
        dffxipdata->alpha=uniq_ratio;
    }





    // taking care of logarithmic norm
    if(firsttime){
      lognorm=loclognorm.right_b();
      maxlognorm=i;
      firsttime=0;
    }
    if(loclognorm.right_b() > lognorm){
      maxlognorm=i;
      lognorm=loclognorm.right_b();
    }



    // updating tab_l, tab_u , tab_offdiag ...
    if(cellno<=cellnum){
      int cell_dim=cells[cellno][0];
      for(int k=1;k<=cell_dim;k++){
        dffxipdata->assign_tabl(cells[cellno][k],loclognorm.right_b());
        dffxipdata->assign_tabu(cells[cellno][k],uniq_margin.left_b());
        dffxipdata->assign_taboffdiag(cells[cellno][k],offdiag.right_b());
        dffxipdata->assign_tabdl(cells[cellno][k],diagtlognorm.right_b());
        dffxipdata->assign_tabdu(cells[cellno][k],diagt.right_b());
      }
    } else{ // we are in the tail
      dffxipdata->assign_tabl(i,loclognorm.right_b());
      dffxipdata->assign_tabu(i,uniq_margin.left_b());
      dffxipdata->assign_taboffdiag(i,offdiag.right_b());
      dffxipdata->assign_tabdl(i,diagtlognorm.right_b());
      dffxipdata->assign_tabdu(i,diagt.right_b());
    }


    if(i>=Big_M +1){
      finished=(loclognorm < 0); // the ending condition for lognorm comp. is stronger
              // than for uniqueness - in fact this is the same condition
    }
    cellno++;
    if(cellno == cellnum +1)  i=m+1; //  we just finished with cells from diagonalization
              //  and we start with near tail
    else i++;
  } // while !finished


  dffxipdata->unique_test_on=true;
  dffxipdata->complexeigenvalues=false;
  dffxipdata->unique=uniq_ok;
  if(!uniq_ok){
    dffxipdata->uniq_worstco=comaxbad;
    dffxipdata->uniq_badmargin=maxbad;
    if(cellnum < m){ //  we have complex eigenvalues here
                     //  and we want to check if it was a reason for the failure
                     //  of the uniqueness
      int cblock=findcell(cellnum,cells,comaxbad);  // cblock - is the index of the cell
                             //  containing 'comaxbad'
      if(cells[cblock][0] > 1){   // this means that our 'bad' coordinate is in
                             //  the 'complex' cell
        dffxipdata->complexeigenvalues=true;
        dffxipdata->complex_off_diag=0;
        if(comaxbad>1){
          dffxipdata->complex_off_diag=Right(abs(dF_NC[comaxbad-1][comaxbad]));
        }
        if(comaxbad<m){
          if( Right( abs(dF_NC[comaxbad][comaxbad+1]))  >
               dffxipdata->complex_off_diag)
            dffxipdata->complex_off_diag=Right(abs(dF_NC[comaxbad][comaxbad+1]));
        }
      }
    }
  }
  dffxipdata->lognorm_computed=true;
  dffxipdata->lognorm=lognorm;
  dffxipdata->maxlognorm=maxlognorm;

  delete [] dF_NC;
  return uniq_ok;
}


interval unstable_mu(int infsup,int i,int cellno,int cellnum,
       int cells[][MAX_DIM+1],interval dF_NC[][MAX_DIM+1],interval imu)
// computes  mu(inf or sup) for diagonal term
//   <inf> for <infsup>==<inf>
//  <sup>  for <infsup>==<sup>
// it returns an estimate from below for mu_inf
//                        and from above for mu_sup
{
  interval retval=0;
  if(cellno<=cellnum){
    int celldim=cells[cellno][0];
    if(celldim==1){
      int k=cells[cellno][1];
      retval=dF_NC[k][k];
    }else{  // dim >=2
    //  interval a[MAX_DIM+1][MAX_DIM+1];
      intervalmatrix1 *a;
      a=new intervalmatrix1[celldim+1];
      for(int i=1;i<=celldim;i++)
        for(int j=1;j<=celldim;j++){
          if(i==j) a[i][j]=dF_NC[cells[cellno][i]][cells[cellno][i]];
          else  a[i][j]=(dF_NC[cells[cellno][i]][cells[cellno][j]] +
                     dF_NC[cells[cellno][j]][cells[cellno][i]])/2.0;
        }

      if(inf==infsup){
        for(int j=1;j<=celldim;j++){
          interval temp=0;
          for(int k=1;k<=celldim;k++){
            if(k==j) temp+=a[j][j];
            else temp-=abs(a[j][k]);
          }
          temp=Left(temp);
          if(j==1) retval=temp;
          else {
            if(temp< retval) retval=temp;
          }
        }
      }
      if(sup==infsup){
        for(int j=1;j<=celldim;j++){
          interval temp=0;
          for(int k=1;k<=celldim;k++){
            if(k==j) temp+=a[j][j];
            else temp+=abs(a[j][k]);
          }
          temp=Right(temp);
          if(j==1) retval=temp;
          else {
            if(temp > retval) retval=temp;
          }
        }
      }
      delete [] a;
      a=NULL;
    }
  } else{
    retval=i*i*(1.0- imu*i*i);
  }

  if(infsup==inf) retval=Left(retval);
  if(infsup==sup) retval=Right(retval);

  return retval;
}


int ks_instability_fp(KuraShi &kurashi,LinCC &coch,int cellnum,
     int cells[][MAX_DIM+1],int unstbcell,const interval boxes[][MAX_DIM+1],
    int boxnum,int m,int Big_M,interval e[],interval c,int s,
    DerData *dffixpdata)
//  an instability computation around fixed point
// the set on which we perform computations is covered by <boxes>[1..<m>]
//  the number of boxes is equal to <boxnum>
  //  coch - a linearlizing coordinates  ; warning: boxes are in standard coordinates
  //      dim(boxes)=dim(coch)= <m>
  //  cellnum - number of blocks in linearization - in dim 1--m
  //       a block is 1-dimensional for a real eigenvalue
  //       and is 2-dimensional for a complex eigenvalue
  //     degenerate situations are not treated here - the number
  //    of eigenvectors matches the number  of eignevalues
  //  cells[i][1]..cells[i][cells[i][0]] - contain the indices of variables
  //       in the block
  //   unstbcell - this is an unstable cell
  //
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]> and it must coincide with kurashi.dim
  //  <Big_M> - M - defining the dimension in tail computations , must be
  //     Big_M>= m, but usually <Big_M> is a few times <m>
  //   the tail is defined by  e,c,s as follows
  //   e[k] - for m < k <= Big_M
  //   |e[k]| \leq  C/k^s  for  Big_M < k
  // -----------------------------------
  //  the function returns 1 - when we have established an instability of fixed point
  //   on output *dffixpdata - contains data

{



  if(kurashi.dim!=m){

    cerr << "ks_instability_fp - Bad input data - kurashi.dim != m " << endl;
    #if defined(WIN95) && defined(KRAK)
      cout << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  // filling in e[k] for 1<=k <= m
  e[0]=0.0;
  for(int k=1;k<=m;k++){
    e[k]=boxes[1][k];
    for(int boxno=2; boxno <= boxnum;boxno++)
      e[k]=Ihull(e[k],boxes[boxno][k]);
  }
  //  now e - contains bounds for 1<=k <= Big_M
//  interval dF[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *dF;
  dF=new intervalmatrix1[m+1];
  for(int i=1;i<=m;i++)
    for(int j=1;j<=m;j++){
      dF[i][j]=ks_dB(i,j,e,Big_M,c,s);
      if(i==j) dF[i][j]+= i*i*(1.0 - kurashi.imu*i*i);
    }

//  interval dF_NC[MAX_DIM+1][MAX_DIM+1];  //  dF in new coordinates
  intervalmatrix1 *dF_NC;
  dF_NC=new intervalmatrix1[m+1];
  transLinearPart(dF,coch,dF_NC);
  delete [] dF;
  dF=NULL;



//  dffixpdata->setnewdim(Big_M+1); this will destroy tab_l



  int unstable_ok=1;

  int i=1;
  //  we compute d_{unstbcell} first
  interval dunstbl=unstable_mu(inf,i,unstbcell,cellnum,cells,dF_NC,kurashi.imu);
  interval offdiag=sum_nondiagks(i,unstbcell,coch,cellnum,cells,m,Big_M,e,c,s,dF_NC);
  dunstbl-=offdiag;
  dunstbl=Left(dunstbl);

  int cell_dim=cells[unstbcell][0];
  for(int k=1;k<=cell_dim;k++){
    dffixpdata->assign_taboffdiag_unstb(cells[unstbcell][k],offdiag.right_b());
  }

  if(dunstbl < 0)
    unstable_ok=0;


  int indminunstble=0;
  double a_min=0;

  int finished=0;
  int cellno=1;
  while(!finished){

    interval a=0.0;
    interval diagt=0.0;;
    interval offdiag=0.0;

    if(cellno<=cellnum){
      i=cells[cellno][1];
    }

    if(cellno!=unstbcell){
      diagt=unstable_mu(sup,i,cellno,cellnum,cells,dF_NC,kurashi.imu);

      offdiag= sum_nondiagks(i,cellno,coch,cellnum,cells,m,Big_M,e,c,s,dF_NC);

      a=dunstbl - diagt - offdiag;
      a=Left(a);
      if(!(a > 0))
        unstable_ok=0;



      if(indminunstble==0){
        indminunstble=i;
        a_min=a.left_b();
      } else{
        if(a_min > a.left_b()){
          indminunstble=i;
          a_min=a.left_b();
        }
      }

    }


    // updating tables
    if(cellno<=cellnum){
      int cell_dim=cells[cellno][0];
      for(int k=1;k<=cell_dim;k++){
        dffixpdata->assign_taba(cells[cellno][k],a.left_b());
        if(cellno!=unstbcell)
          dffixpdata->assign_taboffdiag_unstb(cells[cellno][k],offdiag.right_b());
      }
    } else{ // we are in the tail
      dffixpdata->assign_taba(i,a.left_b());
      dffixpdata->assign_taboffdiag_unstb(i,offdiag.right_b());
    }


    if(i>=Big_M +1){
      finished=(a > 0);
    }
    cellno++;
    if(cellno == cellnum +1)  i=m+1; //  we just finished with cells from diagonalization
              //  and we start with near tail
    else i++;
  } // while !finished

  dffixpdata->instability_test_on=true;
  dffixpdata->unstable=unstable_ok;
  dffixpdata->d_unstable=dunstbl.left_b();
  dffixpdata->minunstable=indminunstble;

  delete [] dF_NC;
  return unstable_ok;
}



int ks_instability_full_test(KuraShi &kurashi,LinCC &coch,int cellnum,
     int cells[][MAX_DIM+1],const interval boxes[][MAX_DIM+1],
    int boxnum,int m,int Big_M,interval e[],interval c,int s,
    DerData *dffixpdata)
//  an instability computation around fixed point
// the set on which we perform computations is covered by <boxes>[1..<m>]
//  the number of boxes is equal to <boxnum>
  //  coch - a linearlizing coordinates  ; warning: boxes are in standard coordinates
  //      dim(boxes)=dim(coch)= <m>
  //  cellnum - number of blocks in linearization - in dim 1--m
  //       a block is 1-dimensional for a real eigenvalue
  //       and is 2-dimensional for a complex eigenvalue
  //     degenerate situations are not treated here - the number
  //    of eigenvectors matches the number  of eignevalues
  //  cells[i][1]..cells[i][cells[i][0]] - contain the indices of variables
  //       in the block
  //
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]> and it must coincide with kurashi.dim
  //  <Big_M> - M - defining the dimension in tail computations , must be
  //     Big_M>= m, but usually <Big_M> is a few times <m>
  //   the tail is defined by  e,c,s as follows
  //   e[k] - for m < k <= Big_M
  //   |e[k]| \leq  C/k^s  for  Big_M < k
  // -----------------------------------
  //  the function returns 1 - when we have established an instability of fixed point
  //   on output *dffixpdata - contains data

{



  if(kurashi.dim!=m){

    cerr << "ks_instability_full_test - Bad input data - kurashi.dim != m " << endl;
    #if defined(WIN95) && defined(KRAK)
      cout << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  // filling in e[k] for 1<=k <= m
  e[0]=0.0;
  for(int k=1;k<=m;k++){
    e[k]=boxes[1][k];
    for(int boxno=2; boxno <= boxnum;boxno++)
      e[k]=Ihull(e[k],boxes[boxno][k]);
  }
  //  now e - contains bounds for 1<=k <= Big_M
//  interval dF[MAX_DIM+1][MAX_DIM+1];  //
  intervalmatrix1 *dF;
  dF=new intervalmatrix1[m+1];
  for(int i=1;i<=m;i++)
    for(int j=1;j<=m;j++){
      dF[i][j]=ks_dB(i,j,e,Big_M,c,s);
      if(i==j) dF[i][j]+= i*i*(1.0 - kurashi.imu*i*i);
    }


  // interval dF_NC[MAX_DIM+1][MAX_DIM+1];  //  dF in new coordinates
  intervalmatrix1 *dF_NC;
  dF_NC=new intervalmatrix1[m+1];
  transLinearPart(dF,coch,dF_NC);
  delete [] dF;
  dF=NULL;

  int unstable=0;


  // we look for unstable cells
  int unstb_cells[MAX_DIM+1];
  int unstb_c_count=0;
  double unstb_factor[MAX_DIM+1];
  for(int i=1;i<=cellnum;i++){
    int k=cells[i][1];
    if(Left(dF_NC[k][k])>0){
      unstb_c_count++;
      unstb_cells[unstb_c_count]=i;
      unstb_factor[unstb_c_count]=(dF_NC[k][k]).left_b();
    }
  }
  if(unstb_c_count==0) return 0;

  // we sort from the most unstable cells to the least one
  for(int k=1;k<=unstb_c_count-1;k++){
    for(int j=1;j<=unstb_c_count-1;j++){
      if(unstb_factor[j] < unstb_factor[j+1]){
        int temp_cell=unstb_cells[j];
        double temp_fact=unstb_factor[j];
        unstb_cells[j]=unstb_cells[j+1];
        unstb_factor[j]=unstb_factor[j+1];
        unstb_cells[j+1]=temp_cell;
        unstb_factor[j+1]=temp_fact;
      }
    }
  }
  //

  int finished=0;
  int num_unstb=1;
  while(!finished){

    int ncells[MAX_DIM+1][MAX_DIM+1];
    int unstblcell=1;
    //  we will glue together cells : unstb_cells[1],.., unstb_cells[num_unstb]
    int ncells_num=1;
    ncells[1][0]=0;
    for(int i=1;i<=num_unstb;i++){
      int curr_cell=unstb_cells[i];
      int cdim=cells[curr_cell][0];
      for(int j=1;j<=cdim;j++){
        ncells[1][0]+=1;
        ncells[1][ncells[1][0]]=cells[curr_cell][j];
      }
    }

    for(int cno=1;cno<=cellnum;cno++){
      int addcell=1;
/*      int no_in_unstb=intab(cno,unstb_cells,1,unstb_c_count);
      if(no_in_unstb !=0) addcell=(no_in_unstb > num_unstb); */
      int no_in_unstb;
      if(intab(cno,unstb_cells,1,unstb_c_count,&no_in_unstb)){
        addcell=(no_in_unstb > num_unstb);
      }
      if(addcell){
        ncells_num++;
        for(int i=0;i<=cells[cno][0];i++)
          ncells[ncells_num][i]=cells[cno][i];
      }
    }


    unstable=ks_instability_fp(kurashi,coch,ncells_num,ncells,unstblcell,boxes,
      boxnum,m,Big_M,e,c,s,dffixpdata);

    if(!unstable) num_unstb++;
    finished=(num_unstb > unstb_c_count) || unstable;
  }
  delete [] dF_NC;
  return unstable;

}


interval ks_cc_ft_ffunc(int i,const LinCC &coch,int m,int Big_M,const interval e[],
  interval c,int s)
// for the cone-condition computation - the value of f-function is returned
{
  if(i<=Big_M){
    cerr << "ks_cc_ft_ffunc - Bad input data - i <= Big_M" << endl;
    #if defined(WIN95) && defined(KRAK)
      cout << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  intervalmatrix1* Atilde=new intervalmatrix1[m+1];
  for(int l=1;i<=m;l++)
    for(int k=1;k<=m;k++)
      Atilde[l][k]=Right(abs(coch.intv_m[l][k])*k);

  interval nAtilde=Right(norm(Atilde,m,NORM_SUM));
  delete [] Atilde;

  interval nAInv=Right(norm(coch.intv_mInv,m,NORM_MAX));

  interval ret=(2*nAtilde/i + 2*nAInv)*(sum_ak(i-m,Big_M,e,c,s)+sum_ak(i+1,Big_M,e,c,s));
  ret+=2*sum_ak(i+m+1,Big_M,e,c,s) + 8*sum_ak(1,Big_M,e,c,s);
  ret+=2*(sum_j_akn(m+1,i,Big_M,e,c,s)+ sum_j_akn(1,0,Big_M,e,c,s))/interval(i);

  return Right(ret);
}


interval ks_cc_sum_offdiag(int i,const interval q[], intervalmatrix1 dF_NC[],
       const LinCC &coch,   int m,int Big_M,const  interval e[],interval c,int s)
// computation of the 'off-diagonal' terms for the cone-conditions computations
//          for KS-equation; warning: for i>Big_M a part of diagonal term
//            \frac{\partial N_i}{\partial x_i} is also treated as off-diagonal
// i - the coordinate for which we perferm the sum for off-diagonal terms
// q - the digaonal of our quadratic form
// dF_NC \in \mathbb{R}^{m \tims m} - the matrix dF in the new coordinates defined by <coch>
// coch - coordinate transformation,  almost diagonalizing dF,
//        in my notes:  A=coch.intv_m,  A^{-1}=coch.intv_mInv
//  m - the dimension of 'main' variables
//  Big_M - the dimension of main variables and near tail (together)
//  our set:  x[i] \in  e[k] for 1 <=k  \leq Big_M
//            x[i] \in [-1,1]C/k^s  for k > Big_M
{
  interval offdiag=0;

  if(i<=m){
     // contribution from main variables
    for(int j=1;j<=m;j++){
      if(i!=j){
        offdiag+=abs(q[j]*dF_NC[j][i] + q[i]*dF_NC[i][j]);
      }
    }

    // contribution from near tail
    for(int j=m+1;j<=Big_M;j++){
      if(i!=j){
        interval dfji=ks_db_nc(j,i,e,m,Big_M,c,s,coch);
        interval dfij=ks_db_nc(i,j,e,m,Big_M,c,s,coch);

        offdiag+=abs(q[j]*dfji+q[i]*dfij);
      }
    }

    // contribution from far tail
    for(int k=1;k<=m;k++){
      offdiag+=2*k*abs(coch.intv_m[i][k])*
               (sum_ak(Big_M+1-k,Big_M,e,c,s)+sum_ak(Big_M+1+k,Big_M,e,c,s));
    }
    for(int k=1;k<=m;k++){
      offdiag+=2*abs(coch.intv_mInv[k][i])*
             (sum_j_akn(Big_M+1,-k,Big_M,e,c,s)+sum_j_akn(Big_M+1,k,Big_M,e,c,s));
    }
  }
  if((m <i ) && (i<=Big_M)){
    for(int j=1;j<=Big_M;j++){
      if(i!=j){
        interval dfji=ks_db_nc(j,i,e,m,Big_M,c,s,coch);
        interval dfij=ks_db_nc(i,j,e,m,Big_M,c,s,coch);

        offdiag+=abs(q[j]*dfji+q[i]*dfij);
      }
    }

    // the far-tail influence
    offdiag+=2*i*(sum_ak(Big_M+1-i,Big_M,e,c,s)+sum_ak(Big_M+1+i,Big_M,e,c,s));
    offdiag+=2*(sum_j_akn(Big_M+1,-i,Big_M,e,c,s)+sum_j_akn(Big_M+1,i,Big_M,e,c,s));

  }
  if(i>Big_M){
    offdiag=i*ks_cc_ft_ffunc(i,coch,m,Big_M,e,c,s) - 2*sum_ak(1,Big_M,e,c,s);
  }

  return Right(offdiag);
}

bool ks_cone_condition(KuraShi &kurashi,const LinCC &coch,
    const interval boxes[][MAX_DIM+1], int boxnum,int m,int Big_M,
    interval e[],interval c,int s,  DerData *dffixpdata)
//  a cone condition computation around fixed point
// the set on which we perform computations is covered by <boxes>[1..<m>]
//  the number of boxes is equal to <boxnum>
  //  coch - a linearlizing coordinates  ; warning: boxes are in standard coordinates
  //      dim(boxes)=dim(coch)= <m>
  //
  //  <m> - is the dimension of Galerkin projection and also the dimension
  //      of <boxes[i]> and it must coincide with kurashi.dim
  //  <Big_M> - M - defining the dimension in tail computations , must be
  //     Big_M>= m, but usually <Big_M> is a few times <m>
  //   the tail is defined by  e,c,s as follows
  //   e[k] - for m < k <= Big_M
  //   |e[k]| \leq  C/k^s  for  Big_M < k
  // -----------------------------------
  //  the function returns true - when we have established the cone-conditions
  //  in the neighborhood covered by <box> of fixed point
  //   on output *dffixpdata - contains data

{



  if(kurashi.dim!=m){

    cerr << "ks_cone_condition - Bad input data - kurashi.dim != m " << endl;
    #if defined(WIN95) && defined(KRAK)
      cout << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  dffixpdata->cc_test_on=true;


  // filling in e[k] for 1<=k <= m
  e[0]=0.0;
  for(int k=1;k<=m;k++){
    e[k]=boxes[1][k];
    for(int boxno=2; boxno <= boxnum;boxno++)
      e[k]=Ihull(e[k],boxes[boxno][k]);
  }
  //  now e - contains bounds for 1<=k <= Big_M
//  interval dF[MAX_DIM+1][MAX_DIM+1];  //
  intervalmatrix1 *dF;
  dF=new intervalmatrix1[m+1];
  for(int i=1;i<=m;i++)
    for(int j=1;j<=m;j++){
      dF[i][j]=ks_dB(i,j,e,Big_M,c,s);
      if(i==j) dF[i][j]+= i*i*(1.0 - kurashi.imu*i*i);
    }


  // interval dF_NC[MAX_DIM+1][MAX_DIM+1];  //  dF in new coordinates
  intervalmatrix1 *dF_NC;
  dF_NC=new intervalmatrix1[m+1];
  transLinearPart(dF,coch,dF_NC);
  delete [] dF;
  dF=NULL;

  bool cc_verified=true;


  // we look for unstable directions
  //  and define the quadratic form Q(x)=\sum q[i]*x[i]^2
  //   with q[i]=1 for unstable directions
  //        q[i]=-1 for the stable ones
  interval *q=new interval[Big_M+1];
  for(int i=1;i<=Big_M;i++)
    q[i]=interval(-1);

  int unstb_c_count=0;
  for(int i=1;i<=m;i++){
    if(Left(dF_NC[i][i])>0){
      unstb_c_count++;
      q[i]=interval(1);
    }
  }

  double mingap;
  int index_mingap=0;

  for(int i=1;i<=Big_M;i++){
    interval diagt;
    if(i<=m)
      diagt=Left(2*abs(dF_NC[i][i]));
    if(i>m){
      //  the diagonal term is: i^2(1-\mu i^2) + 2*i a_{2i}
      diagt=power(interval(i),2)*(1-kurashi.imu*power(interval(i),2));
      interval a2i;
      if(2*i > Big_M) a2i=interval(-1,1)*c*power(interval(2*i),-s);
      else a2i=e[2*i];
      diagt=2*q[i]*(diagt+2*i*a2i);
      diagt=Left(diagt);
    }
    interval offdiag=Right(ks_cc_sum_offdiag(i,q,dF_NC,coch,m,Big_M,e,c,s));

    if(i==1){
      mingap=(diagt-offdiag).left_b();
      index_mingap=1;
    } else{
      double t= (diagt-offdiag).left_b();
      if(t < mingap){
        mingap=t;
        index_mingap=i;
      }
    }

    if(!(diagt - offdiag >0)) cc_verified=false;


  }

  // now we treat 'uniformly' the case of i >  Big_M
  int i=Big_M+1;
  interval ffunc=ks_cc_ft_ffunc(i,coch,m,Big_M,e,c,s);
  if(2*(kurashi.imu*power(interval(i),3)-i) > ffunc) ;
    // the diagonal term is dominant for all i >= Big_M+1
  else cc_verified=false;

  // for the 'reporting reasons' we still want to compute the margin we
  // have for i=Big_M+1
  interval diagt=-2*power(interval(i),2)*(1-kurashi.imu*power(interval(i),2));
  interval offdiag= ks_cc_sum_offdiag(i,q,dF_NC,coch,m,Big_M,e,c,s);
  double t= (diagt-offdiag).left_b();
  if(t < mingap){
    mingap=t;
    index_mingap=i;
  }



  dffixpdata->cc_ok=cc_verified;
  dffixpdata->cc_mingap=mingap;  // the worst value of diag(i) - offdiag(i)
  dffixpdata->cc_index_mingap=index_mingap;


  delete [] q;
  delete [] dF_NC;
  return cc_verified;

}

