
#include<iostream>
#include<fstream>
using namespace std;

#include<cstdlib>


#include"defo.h"
#include"lohner.h"
#include"section.h"
#include"zdynsys.h"
#include"multiint.h"
#include"func.h"
#include"coordc.h"
#include"qrdecmp.h"
#include"zset.h"
#include"zgfiles.h"
#include"rigpoinc.h"




void poincUpdateEstm(estimates *estm,Ode &ode,const C1Func &section,
     int order,interval h0,Set &x,interval PoincVal[],
     interval tPoinc,interval remMax,interval lmarrmax[],
     int usediffcoord ,LinCC*  coordch ,interval  CPoincVal[])
// x =\varphi(t,initial condition), where t >tPoinc - used here to gauge
//              the size of what we can get without the section errors
//     also for the doubleton sets - the representation x.r0 - contains
//          the size of initial data as x=0 + C*r0 + B*r
//            and x_0,C,B,r - are changed during the evolution
//            r0 - stays constant
// PoincVal - the value of P(initial condition) - always expressed in standard
//             coordinates
// CPoincVal - the value of P(initial condition) -  expressed in new coordinates
//             if usediffcoord==1, otherwise is not used
// coordch - the new coordinates
/////
//  when used for a differential inclusion ,then it has a tendency to underestimate
//   the width of the passing through the section time

{
   // we will try to compute the projection of <xd> and <PoincVal>
  // onto the section (assuming the section is linear or almost linear)
  // and we compare the diameters of both things, with the projected
  //  section error -
  //  the function saliently assumes that if <usediffcoord==1>,
  //  then the section is is in fact y1=C, where (y1,y2,\dots) are the new
  //  coordinates; otherwise the section perpendicular to the vector field
  //  ON OUTPUT: (under above silient assumption aboout section
   //      <trPoincVal> - sizes(diameters) of PoincaVal(or CpoincVal)
   //      <trxd> - sizes(diameters) of x



  estm->rettime=tPoinc;

  // estimates for section error
  estm->seccrossingtime=diameter(tPoinc);

  interval   scprod_f_grad_s;
  interval *valvfield;
  valvfield=new interval[ode.dim+1];
  interval *grad_of_sec_func;
  grad_of_sec_func=new interval[ode.dim+1];


  ode.field(PoincVal,valvfield);  // for differential inclusions it should be 'incl_field'
                               // the size of the vector field is slightly different
                               // this influences the crossing time through the section
  if(!usediffcoord){
    section.gradient(PoincVal,grad_of_sec_func);
  } else{
    coordch->vectcoordtran(valvfield,valvfield);
    zero(grad_of_sec_func,ode.dim);
    grad_of_sec_func[1]=1.0;
  }
  scprod_f_grad_s=scprod(grad_of_sec_func,valvfield,ode.dim);
  //  scprod_f_grad_s - gives me   d s/dt, where s is the function defining
  //        the section
  delete [] grad_of_sec_func;
  grad_of_sec_func=NULL;

  interval size=diameter(section.eval(x));
  interval res_poinctime=abs(size/scprod_f_grad_s);  // this is a reasonable
            // guess, what the crossing time should be

  estm->overestm_crosstime=estm->seccrossingtime/abs(res_poinctime).right_b();


  GeneralLinearSection *gls;
  int destroy_gls=0;
  LinCC *loc_cc=NULL;



  if(!usediffcoord){
    interval *cp,*v, *temp;
    cp=new interval[ode.dim+1];
    v=new interval[ode.dim+1];
    temp=new interval[ode.dim+1];
    x.intvencl(cp);
    DecomposeIMulti(cp,temp,ode.dim); // cp=mid(x)
    copyvector(valvfield,v,ode.dim);
    DecomposeIMulti(v,temp,ode.dim);
    gls=createPerpendicularSection(ode.dim,cp,v);
    delete [] cp;
    cp=NULL;
    delete [] v;
    v=NULL;
    delete [] temp;
    temp=NULL;

    destroy_gls=1;
    loc_cc=gls->plincc;

  }else{
    loc_cc=coordch;
  }


  interval *trPoincVal, *trxd;
  trPoincVal=new interval[ode.dim+1];
  trxd=new interval[ode.dim+1];


  x.intvencl(trxd,1,loc_cc);
  if(usediffcoord){
    copyvector(CPoincVal,trPoincVal,ode.dim); // CPoincVal is already expressed
                               // in new coordinates
    // valvfield - is already expressed in new coordinates
  }else{
    loc_cc->coordtran(PoincVal,trPoincVal);
    loc_cc->vectcoordtran(valvfield,valvfield);
  }



  interval *expected_sizes; // [MAX_DIM+1]; // this is an estimate of the size
           //in i-th direction in <c> coordinates of the
           // which is the size of <xd> + expected_crossing_time*vect_field
  expected_sizes=new interval[ode.dim+1];
  for(int i=2;i<=ode.dim;i++)
    expected_sizes[i]=diameter(trxd[i]) + res_poinctime*Right(abs(valvfield[i]));


  trPoincVal[1]=0.0;
  trxd[1]=0;
  estm->sec_ratio=diameter(trPoincVal,ode.dim)/diameter(trxd,ode.dim);
  for(int i=2;i<=ode.dim;i++){
    trPoincVal[i]=diameter(trPoincVal[i]);
    trxd[i]=diameter(trxd[i]);
  }

  // now in <trPoincVal> - sizes computed by the algorithm
  //    <expected_sizes> - the sizes from reasonable geometric picture
  //              of the box moving with constant speed and crossing
  //              the section


  for(int i=2;i<=ode.dim;i++){
    int update_max,update_d;
    if(i==2) {
      update_max=1;
      update_d=1;
    }else{
      update_max=((trPoincVal[i] - trxd[i])>estm->max_sec_error);
      update_d=(abs(trPoincVal[i]-expected_sizes[i]).right_b() > estm->d_max_sec_error);
    }
    if(update_max){
      estm->max_sec_error=trPoincVal[i] - trxd[i];
      estm->x_sec_error=trPoincVal[i]/expected_sizes[i];
    }
    if(update_d)
      estm->d_max_sec_error=trPoincVal[i]-expected_sizes[i];

    estm->ratios_sec_errors[i]= trPoincVal[i]/expected_sizes[i];
  }
  delete [] expected_sizes;
  expected_sizes=NULL;

  // now we try to compute the Lipschitz constant and
  //  the error which came from time discretization
  //  such that
  //    diam(x) = LipConst * diam(r0) + IntRkError * h0^{order}
  //       in the doubleton representation - we assume that x.r0
  //      is the initial set;
  //         the error term from numerical method is assumed to be
  //         B*r, but since B is orthogonal we use only <r>
  //         to measure the size of this error


  interval *xd; // interval xd[MAX_DIM+1];
  xd=new interval[ode.dim+1];
  x.intvencl(xd);


  if(x.snumLohner()!=LOHNER1){
    /// trxd - contains the diameters of xd projected on section
    int norm_used=NORM_MAX;
    double diamxd=MultiIDiam(xd,ode.dim,norm_used);
    double sec_diamxd= norm(trxd,ode.dim,norm_used).right_b();
    double sizeRK=0;
    double sec_sizeRK=0;
//    interval Br[MAX_DIM+1];
//    interval sec_Br[MAX_DIM+1];
    interval *Br, *sec_Br, *xr;
    intervalmatrix1* xB=new intervalmatrix1[ode.dim+1];
    Br=new interval[ode.dim+1];
    sec_Br=new interval[ode.dim+1];
    xr=new interval[ode.dim+1];

    x.get_Bmatrix(xB);
    x.get_r(xr);

    multAV(xB,xr,Br,ode.dim);
    sizeRK=MultiIDiam(Br,ode.dim,norm_used);

    // on section:
    intervalmatrix1 *t;   //      interval t[][MAX_DIM+1];
    t=new intervalmatrix1[ode.dim+1];
    ArrArr(loc_cc->intv_m,xB,t,ode.dim);
    multAV(t,xr,sec_Br,ode.dim);
    delete [] t;
    t=NULL;
    sec_Br[1]=0;
    sec_sizeRK= MultiIDiam(sec_Br,ode.dim,norm_used);

    if(x.doubletonRep()){
      interval *xr0=new interval[ode.dim+1];
      x.get_r0(xr0);
      if (MultiIDiam(xr0,ode.dim,norm_used)==0){
        sizeRK=diamxd;
        sec_sizeRK=sec_diamxd;
      }

      if (MultiIDiam(xr0,ode.dim,norm_used)>0){
        estm->LipConst=(diamxd - sizeRK)/MultiIDiam(xr0,ode.dim,norm_used);
        estm->sec_LipConst=(sec_diamxd - sec_sizeRK)/MultiIDiam(xr0,ode.dim,norm_used);
      }else {
        estm->LipConst=0.0;
        estm->sec_LipConst=0.0;
      }
      delete [] xr0;
    }
    estm->IntRkError=sizeRK/power(h0,order);
    estm->sec_IntRkError=sec_sizeRK/power(h0,order);

    delete [] xr;
    delete [] xB;
    delete [] sec_Br;
    delete [] Br;

  } else{
    estm->LipConst=0.0;
    estm->IntRkError=0.0;
    estm->sec_LipConst=0.0;
    estm->sec_IntRkError=0.0;
  }

  estm->Rkerror = remMax;
  estm->LMVal=lmarrmax[1];
  for(int i=2;i<=ode.dim;i++)
    estm->LMVal=Ihull(lmarrmax[i],estm->LMVal);


  delete [] xd;

  if(destroy_gls)
    delete gls;

  delete [] trPoincVal;
  delete [] trxd;

  delete [] valvfield;


}








void OkGen_Poincare(int *ok,Set &x,interval xd[],interval xdn[],int dim,interval t,
       interval h,const char *funcname)
// this function is a  test for errors and sends error messages correspeonding to
//  function <funcname>
{

   if(!(*ok)){
     ofstream errorlog(zglerrorlogname,ios::app);
     entriesinerrorlog=1;
     errorlog <<"----------------------------------" << endl;
     writeCompInfo(errorlog);
     writedatetime(errorlog);

     cerr << funcname << " - OneStep function failed " << endl;
     cerr << "t=" << t << "   h=" << h <<  endl;

     errorlog << funcname << " - OneStep function failed " << endl;
     errorlog << "t=" << t << "   h=" << h <<  endl;
     errorlog << "the set causing problem: " ;
   //  saveAllCoord(errorlog,xd,dim,1,(dim > 4),4);
     x.savetxt(errorlog);
   }
   if(*ok){
     *ok=!MultiIn(xd,xdn,0,dim);
     if(!(*ok)){
       #if defined(WIN95) && defined(KRAK)
       if(dim >3)   rootFrame.Clear();
       #endif

       ofstream errorlog(zglerrorlogname,ios::app);
       entriesinerrorlog=1;
       errorlog <<"----------------------------------" << endl;
       writeCompInfo(errorlog);
       writedatetime(errorlog);


       cerr << funcname << "- image stabilized " << endl;
       cerr << "t=" << (t+h) << endl;
       cerr << "xd[i]    xdn[i] " << endl;

       errorlog << funcname << "- image stabilized " << endl;
       errorlog << "t=" << (t+h) << endl;
       errorlog << "xd[i]    xdn[i] " << endl;

       for(int i=1;i<=dim;i++){
         cerr << xd[i] << "   " << xdn[i] << endl;
         errorlog << xd[i] << "   " << xdn[i] << endl;
       }
     }
   }
   if(!(*ok)){
     #if defined(WIN95) && defined(KRAK)
//       cerr << "press any key " << endl;
//       waitBt();
     #endif
   }
}

void  c0updates_in_sec_reg(Ode &ode,interval h,Set &x,interval xd[],
    Set &xn, interval xdn[],interval xencl[],int usediffcoord,LinCC *coordch,
    interval lmarrmax[],interval PoincVal[],interval CPoincVal[])
// The function updates PoincVal and CPoincVal , which are enclosures of
//  the intersection of the current trajectory with current section
// it is called in the 'section region', (not present in parameters here)
// Input data:
//     ode - diff. eq. - defining the flow \varphi
//      h - time step
//      xd=x,  xdn=xn;   are such that \varphi(h,x) \subset xn
//      and \varphi([0,h],x) \subset xecnl
//    if <usediffcoord> then we are interested in in the intersection expressed
//      in other coordinates <coordch>
//     PoincVal, CPoincVal - current values of the intersection
//  Output data:
//     PoincVal and CPoincVal - will contain all [0,h]-trajectories emanating
//           from x, we estimate them more effectively than by taking just
//           xencl
//     lmarrmax - a control variable used to monitor the size of the section
//           error due to difference between actual trajctory and the segment
//           joining x and \varphi(h,x)
{
  interval lmarr[MAX_DIM+1];
  interval xdnLM[MAX_DIM+1],xdLM[MAX_DIM+1];

  ode.fromsegdist(xencl,lmarr,h);
  for(int i=1;i<=ode.dim;i++)
    lmarr[i]=Ihull(lmarr[i],-lmarr[i]);

  for(int i=1;i<=ode.dim;i++){
    xdnLM[i]=xdn[i]+lmarr[i];
    xdLM[i]=xd[i]+lmarr[i];
  }

  if(!usediffcoord)
    MultiIhull(lmarrmax,lmarr,lmarrmax,ode.dim);

  MultiIhull(xdLM,PoincVal,PoincVal,ode.dim);
  MultiIhull(xdnLM,PoincVal,PoincVal,ode.dim);

  if(usediffcoord){
   // we compute the Poincare map in other coordinates <coordch>
    interval CxdnLM[MAX_DIM+1],CxdLM[MAX_DIM+1];
    ode.fromsegdist(xencl,lmarr,h,coordch->intv_m);
    for(int i=1;i<=ode.dim;i++)
      lmarr[i]=Ihull(lmarr[i],-lmarr[i]);

    xn.intvencl(CxdnLM,usediffcoord,coordch);
    x.intvencl(CxdLM,usediffcoord,coordch);
    for(int i=1;i<=ode.dim;i++){
      CxdnLM[i]=CxdnLM[i]+lmarr[i];
      CxdLM[i]=CxdLM[i]+lmarr[i];
    }

    MultiIhull(CxdLM,CPoincVal,CPoincVal,ode.dim);
    MultiIhull(CxdnLM,CPoincVal,CPoincVal,ode.dim);
    MultiIhull(lmarrmax,lmarr,lmarrmax,ode.dim);
  }
}



void c1updates_in_sec_reg(int firsttime,Ode &ode,interval h,interval xencl[],
   interval dp[][MAX_DIM+1],interval dpn[][MAX_DIM+1],interval dpencl[][MAX_DIM+1],
   interval DPVal[][MAX_DIM+1])
// this function should be called   in the section region  to complete
//  C1 part of the computation of Poincare map
// The function updates DPVal, which is an enclosure of
//  the intersection of the D\varphi(t,*) on the current trajectory
//    with the current section
// it is called in the 'section region', (not present in parameters here)
// INPUT DATA:
//     ode - the diff. eq. - defining the flow \varphi
//      h - a time step
//      \varphi(t+[0,h],x) \subset xecnl  ; x - is a starting  point (set) for
//                   current trajectory
//      dp = dp(t), dpn=dp(t+h),  t - is the current time
//      dp(t+[0,h],x) \subset dpencl=[W4]  - in my notes 'C1-Lohner algorithm'
//     DPVal - current values of DP on the intersection
//  OUTPUT DATA:
//     DPVal - contains   DP(t+[0,h],x),
//           this estimate is better than  just taking   <dpencl>
{
  interval trash[MAX_DIM+1];
//  interval lmdarr[MAX_DIM+1][MAX_DIM+1];
//  interval DpnLM[MAX_DIM+1][MAX_DIM+1];
//  interval DpLM[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *lmdarr;
  lmdarr=new intervalmatrix1[ode.dim+1];
  intervalmatrix1 *DpnLM;
  DpnLM=new intervalmatrix1[ode.dim+1];
  intervalmatrix1 *DpLM;
  DpLM=new intervalmatrix1[ode.dim+1];

  ode.varfromsegdist(xencl,trash,dpencl,lmdarr,h);
//  ode.fromsegdist(xencl,trash,dpencl,lmdarr,h);

  for(int i=1;i<=ode.dim;i++)
    for(int j=1;j<=ode.dim;j++){
      DpnLM[i][j]=dpn[i][j]+
           interval(-lmdarr[i][j].right_b(),lmdarr[i][j].right_b());
      DpLM[i][j]=dp[i][j]+
         interval(-lmdarr[i][j].right_b(),lmdarr[i][j].right_b());
    }
  if(firsttime){
    MultiIhull(DpLM,DpnLM,DPVal,ode.dim);   // DPVal=DpLM \cup DpnLM
  }else {
    MultiIhull(DpLM,DPVal,DPVal,ode.dim);
    MultiIhull(DpnLM,DPVal,DPVal,ode.dim);   // DPVal=DPval \cup DpnLM
  }
  delete [] lmdarr;
  delete [] DpnLM;
  delete [] DpLM;
}

double newtimestep(double ha,double minh,double maxh,double biasratio)
  // the function returns a new proposed value for <ha>
  // used in the searches for the good time shift along the trajectory
  // biasratio  should less than 1/2
{
  if((biasratio <= 0) || (biasratio >= 0.5)){
    cerr << "newtimestep - biasratio out of range (0,0.5) - " << biasratio <<
      endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  // it can happen that our heurisitic estimates are not too good
  // hence :
  if(ha < minh) ha=(minh + maxh)/2;
  if(ha > maxh) ha=(minh + maxh)/2;

  // we reject too small changes in the size of (minh,maxh)
  if((ha-minh)/(maxh - minh) < biasratio) {
    ha=(1-biasratio)*minh + biasratio*maxh;
  }

  if((maxh-ha)/(maxh - minh) < biasratio){
     ha=(1-biasratio)*maxh + biasratio*minh;
  }
  return ha;
}


void near_section_problems_report(const char comment[],int dim,int maxitnum,
   interval xd[],interval xdn[],interval sec_on_x,interval sec_on_xn,
   double c,interval dsec_dt,double ds_dt,double h,double minh, double maxh,
   double ha )
  // the error diagonosis function for functions: 'get_close_to_section' and
  //   'get_past_section'
{
  ofstream report(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  report <<"************************" << endl;
  writeCompInfo(report);
  writedatetime(report);
  report << comment  << endl;
  report << "this is not an error - just a warning about a suboptimal "
             "performance" << endl;
  report << "maxitnum=" << maxitnum << endl;
  report << "the initial condition xd: " << endl;
  saveAllCoord(report,xd,dim,2,(dim > 5),5);
  report << endl;
  report << "the current value of xdn: " << endl;
  saveAllCoord(report,xdn,dim,2,(dim > 5),5);
  report << endl;
  report << "the section value for initial condition=" <<
               sec_on_x << endl;
  double diamSx=diameter(sec_on_x).right_b();
  report << "c*diamSx=" << c*diamSx << endl;
  report << "ds/dt=" << dsec_dt << "  selected speed ds/dt=" <<
                       ds_dt << endl;
  report << "the section value in last iteration=" << sec_on_xn << endl;
  report << "h=" << h ;
  report << "  minh=" <<  minh << "   maxh=" << maxh << endl;
  report << "ha=" << ha << endl;
  report.flush();
}


int get_close_to_section(const C1Func &section,int secdir,Ode &ode,OneStepParamType &onesteppar,
     interval h,Set &x,interval xd[], Set &xn,interval xdn[], interval xencl[],
     interval rem[],interval *h_used,interval *new_h)
// We assume that on input:
//   x \subset xd - xd is an enclosure for x,  does not intersect the section
//   \varphi(h,x) \subset xn \subset xdn=enclosure(xn)  intersects the
//        the section <section> or is already on the other side
//  On output:
//    x,xd - are unchanged
//    h_used - is the time step found to get close to section
//            but to be on the same side as <x>
//   if 1 is returned, then:
//    *h_used - is computed so that we are still before the section
//    and if *h_used==0, then xn,xdn,xencl- are undefined,
//    if *h_used > 0, then xn,xdn,xecnl - is changed so that
//    \varphi(h_used,x) \subset xn \subset xdn=enlosure(xn)
//    \varphi([0,h_used],x) \subset xencl
//    new_h  - a time step suggested for the computation of crossing
// if 0 is returned, then xn, xecnl - are changed, but for all purposes
//       we may treat these variables as undefined
/////////////////
///  Idea:
//     new_h - shall be related to the section crossing time of the set
//                    <x>
//     We can estimate this crossing time by estimating
//        ds/dt=d section / dt over <x+xn> \subset xd + xdn
//     Assuming that the size of <x> or <xn> does not change much
//     during the passing of <x> through the section, we can estimate
//       the passing time by  t_passsing= diam(sec(x))/|ds/dt|
//     new_h - will be defined then in terms of t_passing;
//          new_h=t_passing/(k > 2)
//     We compute heuristically <h_used> so that
//       |sec(\varphi(h_used,x))|_{min} > diam(sec(x))*c; where c - is some
//                     small number c=0.01 or c=0.1
//     Due to these choices we can expect the computed passing time to be
//      only slightly bigger than the heuristic estimates
{
  interval sec_on_x=section.eval(x);
  if(sec_on_x.contains(0)){
    cerr << "get_close_to_section - bad initial condition - already intersects section" << endl;
    #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  double c=0.01;
  int ok=1;

  interval xxn[MAX_DIM+1];
  MultiIhull(xd,xdn,xxn,ode.dim);  //  x+xn \subset xxn

  interval valvfield[MAX_DIM+1];
  ode.field(xxn,valvfield);

  interval grad_of_sec_func[MAX_DIM+1];
  section.gradient(xxn,grad_of_sec_func);

  interval dsec_dt=scprod(grad_of_sec_func,valvfield,ode.dim); // dsec_dt = dsec/dt
  if(dsec_dt.contains(0)){
    cerr << "get_close_to_section - Bad direction of vect. f. in section region" << endl;
    #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  dsec_dt=abs(dsec_dt);
  double ds_dt=mid(dsec_dt).left_b();


  double diamSx=diameter(sec_on_x).right_b();
  double t_passing=diamSx/ds_dt;  // this is an estimate for the crossing time

  double sec_min_on_x=abs(sec_on_x).left_b(); // this gives me a minimal value of the
         // section function on x
  //  to estimate the time to section we use  sec_min_on_x=ds/dt * h
  //  but to leave some margin for wrapping (from interval comp. ) we want
  //        sec_min_x  - c*diamSx = ds/dt *h
  double ha=(sec_min_on_x - c*diamSx)/ds_dt;


  double minh=0; // because x - is on the correct side of the section
  double maxh=h.left_b(); //  minh - should be ok, but too far before section
               // maxh - allways too big, we cross the section


  ha=newtimestep(ha,minh,maxh,0.1);

  int done=0;

  if(secdir==UP)
    done=(sec_on_x.left_b() > -c*diamSx);
  if(secdir==DOWN)
    done=(sec_on_x.right_b() < c*diamSx);


  Set cxn=xn; // just to initiate the dimension
  interval cxdn[MAX_DIM+1];
  interval cxencl[MAX_DIM+1];
  interval crem[MAX_DIM+1];


  int improvement=0;
  interval h_impr;
  Set xn_impr=xn; // just to initiate the  dimension
  interval xdn_impr[MAX_DIM+1];
  interval xencl_impr[MAX_DIM+1];
  interval rem_impr[MAX_DIM+1];

  int itno=0;
  int maxitnum=10;

  while(!done && ok  && (itno < maxitnum)){
    itno++;

    ok=OneStepLohner(ode,onesteppar,ha,x,xd,cxn,cxdn,cxencl,crem);
    if(ok){
      int saveimpr=0;
      interval sec_on_xn=section.eval(cxn);
      if(secdir==UP){
        if(sec_on_xn<0){
          saveimpr=1;
          if(sec_on_xn.right_b() > -c*diamSx) done=1;
        }
      }
      if(secdir==DOWN){
        if(sec_on_xn>0){
          saveimpr=1;
          if(sec_on_xn.left_b() < c*diamSx)  done=1;
        }
      }
       // saving improved values
      if(saveimpr){
        improvement=1;
        h_impr=ha;
        xn_impr=cxn;
        copyvector(cxdn,xdn_impr,ode.dim);
        copyvector(cxencl,xencl_impr,ode.dim);
        copyvector(crem,rem_impr,ode.dim);
      } // endif saveimpr




      if((!done) && (itno < maxitnum)){ // the inequality tells that there
         // will be next iteration, if not we want to keep
                // the current values to report on them
        double dha;
        int overshot;
        if(secdir==UP) overshot=(sec_on_xn.right_b() > 0);
        else overshot=(sec_on_xn.left_b() < 0);
        if(overshot){  // I overshot, I need a better guess
          maxh=ha;
          double sec_min_xn=(secdir==UP)?sec_on_xn.right_b():fabs(sec_on_xn.left_b());
          dha=-(sec_min_xn+c*diamSx/2.0)/ds_dt;

        }else{ //  I'm  before section
          minh=ha;
          double sec_min_xn=abs(sec_on_xn).left_b();
          dha=(sec_min_xn-c*diamSx/2.0)/ds_dt;

        } // endif overshot
        ha+=dha;

        ha=newtimestep(ha,minh,maxh,0.1);

      } // endif !done
    } //endif ok
  } // end while

  double divfact=4;
  if(ode.divsecreg > divfact) divfact=ode.divsecreg;
  *new_h=Left(t_passing/divfact);
  if(*new_h > h) *new_h=h;
  if(improvement){
    *h_used=h_impr;
    xn=xn_impr;
    copyvector(xdn_impr,xdn,ode.dim);
    copyvector(xencl_impr,xencl,ode.dim);
    copyvector(rem_impr,rem,ode.dim);
  }else{
    if(done){ // this happens only if our initial data was satisfatory
      *h_used=0;
    }else{    // this means that the function above failed to produce
      // something closer to section, but we can still continue computations
      *h_used=0;
    }
  }
  if(ok && !done){ // it means there were no errors, but we failed with guessing
            // and we would like to report it
    char comment[]="Problem with the deterministic get_close_to_section";
    near_section_problems_report(comment,ode.dim,maxitnum,xd,xdn,
       sec_on_x,section.eval(xn),c,dsec_dt,ds_dt,
                                   h.right_b(),minh,maxh,ha);
  }

  return(ok);
}



int get_past_section(const C1Func &section,int secdir,Ode &ode,OneStepParamType &onesteppar,
     interval h,Set &x,interval xd[],Set &xn,interval xdn[], interval xencl[],
     interval rem[], interval *h_used)
// We assume that on input:
//   x \subset xd=enclosure(x), is either intersecting the section or is
//                               before the section
//   \varphi(h,x) \subset xn \subset xdn=enclosure(xn) is on the other side of
//        the section <section>
//  On output:
//   1- is returned when we find, h_used,xn,xdn,xencl such that:
//    h_used - the time step, possibly as small as possible, such that
//             we can move to the  the other side of the section
//    \varphi(h_used,x) \subset xn \subset xdn=enclosure(xn)
//    \varphi([0,h_used],x) \subset xencl
/////////////////
///  Idea:
//     We can estimate this crossing time by estimating
//        ds/dt=d section / dt over <x \cup xn> \subset xd \cup xdn
//     Assuming that the size of <x> or <xn> does not change much
//     during the passing of <x> through the section, we can estimate
//     the time needed to flow past section but not further than
//        diam(sec(x))*c; where c - is some
//                     small number c=0.01 or c=0.1
{

  interval sec_on_xn=section.eval(xn);
  if(sec_on_xn.contains(0)){
    cerr << "get_past_section - bad xn - intersects section" << endl;
    #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  double c=0.01;
  int ok=1;

  interval xxn[MAX_DIM+1];
  MultiIhull(xd,xdn,xxn,ode.dim);  //  x \cup xn \subset xxn

  interval valvfield[MAX_DIM+1];
  ode.field(xxn,valvfield);

  interval grad_of_sec_func[MAX_DIM+1];
  section.gradient(xxn,grad_of_sec_func);

  interval dsec_dt=scprod(grad_of_sec_func,valvfield,ode.dim); // dsec_dt = dsec/dt
  if(dsec_dt.contains(0)){
    cerr << "get_past_section - Bad direction of vect. f. in section region" << endl;
    #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
    exit(1);
  }
  dsec_dt=abs(dsec_dt);
  double ds_dt=mid(dsec_dt).left_b();
  interval sec_on_x=section.eval(x);

  double diamSx=diameter(sec_on_x).right_b();

  double dist_x;  // this tells how far before the section is
                   //  the 'rear' end of <x>
  if(secdir==DOWN) dist_x=sec_on_x.right_b();
  else dist_x=-sec_on_x.left_b();

  double ha=(dist_x + c*diamSx)/ds_dt; // assuming constant speed, this is
        // an estimate of how long does it take to fly through the section


  double minh=0;
  double maxh=h.left_b();    // the interval (minh,maxh) - gives
            // an reasonable interval
               // for tests ; minh - is bad, maxh - is ok. but too big

  ha=newtimestep(ha,minh,maxh,0.1);

  int done;


  if((secdir==UP) && (sec_on_xn >0))
    done=(sec_on_xn.left_b() < c*diamSx);
  if((secdir==DOWN) && (sec_on_xn <0))
    done=(sec_on_xn.right_b() > -c*diamSx);


  Set cxn=xn; // just to correctly initiate the dimension
  interval cxdn[MAX_DIM+1];
  interval cxencl[MAX_DIM+1];
  interval crem[MAX_DIM+1];



  int improvement=0;
  interval h_impr;
  Set xn_impr=xn; // just to correctly initiate the dimension
  interval xdn_impr[MAX_DIM+1];
  interval xencl_impr[MAX_DIM+1];
  interval rem_impr[MAX_DIM+1];

  int itno=0;
  int maxitnum=10;
  while(!done && ok  && (itno < maxitnum)){
    itno++;
    ok=OneStepLohner(ode,onesteppar,ha,x,xd,cxn,cxdn,cxencl,crem);
    if(ok){
      int saveimpr=0;
      interval sec_on_xn=section.eval(cxn);
      if((secdir==UP) && (sec_on_xn >0)){
        saveimpr=1;
        done=(sec_on_xn.left_b() < c*diamSx);
      }
      if((secdir==DOWN) && (sec_on_xn <0)){
        saveimpr=1;
        done=(sec_on_xn.right_b() > -c*diamSx);
      }

      // saving improved values
      if(saveimpr){
        improvement=1;
        h_impr=ha;
        xn_impr=cxn;
        copyvector(cxdn,xdn_impr,ode.dim);
        copyvector(cxencl,xencl_impr,ode.dim);
        copyvector(crem,rem_impr,ode.dim);
      } // endif saveimpr


      if(!done && (itno<maxitnum)){
        double dha;
        if(((secdir==UP) && (sec_on_xn > c*diamSx))
          || ((secdir==DOWN)  && sec_on_xn < -c*diamSx)  ){ // I overshot
          double toofar;
          if(secdir==UP) toofar=sec_on_xn.left_b() - c*diamSx;
          if(secdir==DOWN) toofar=fabs(sec_on_xn.right_b()) - c*diamSx;
          dha=-toofar/ds_dt;

          if(maxh > ha) maxh=ha; // we know that with maxh - we are
                  // on the correct side but we overshot


        } else{ // we need a bigger time step
          double adddist;
          if(secdir==UP) adddist=c*diamSx - sec_on_xn.left_b() ;
          if(secdir==DOWN) adddist=sec_on_xn.right_b() + c*diamSx;
          dha=adddist/ds_dt;

          if(ha > minh) minh=ha;

        }
        ha+=dha;
        ha=newtimestep(ha,minh,maxh,0.1);


      } // endif !done

    } // endif ok
  } // end while (!done ...

  if(improvement){
    *h_used=h_impr;
    xn=xn_impr;
    copyvector(xdn_impr,xdn,ode.dim);
    copyvector(xencl_impr,xencl,ode.dim);
    copyvector(rem_impr,rem,ode.dim);
  } else{
    if(done){   // this is possible only if our initial data were satisfactory
      *h_used=h;
    } else{ // we must accept the xn - even if we think it is not optimal
      *h_used=h;
    }
  }
  if(!done && ok){
    char comment[]="get_past_section - deterministic - problems";
    near_section_problems_report(comment,ode.dim,maxitnum,xd,xdn,sec_on_x,
        section.eval(xn),c,dsec_dt,ds_dt,h.right_b(),minh,maxh,ha);
  }
  return(ok);
}


int in_sec_region(int secdir,interval secnew,int *startPoinc,Ode &ode,
    interval h,Set &x,interval xd[],Set &xn, interval xdn[],
    interval xencl[],int usediffcoord,LinCC *coordch,
    interval lmarrmax[],interval PoincVal[],interval CPoincVal[])
//  the function handles the section region for General_Lo_Poincare
//  It is called close to section (in 'Section region')
//  returns 1 - when we crossed the section (i.e. we got on the other
//               'desirable' side of the section), in this case we don't
//             update PoincVal and CPoincVal
//  secdir - UP or DOWN
//  secnew = section(new set)
//  *startPoinc - tells if we started to save the image of the Poincare map
//    in <PoincVal> and <CPoincVal>
//  h - time step
//  xencl=enclosure for \varphi([0,h],x)
//  xd - interval enclosure of x
//  xdn - interval enclosure of xn
//  \varphi(h,x) \subset xn
//  ON OUTPUT:
//     we update *startPoinc,lmarrmax, PoincVal, CPoincVal
{

  int poinc_finished=0;  // this is the returned variable
  int oldstartPoinc=*startPoinc;


  if (!(*startPoinc)){ // we did'nt started yet to build the Poincare image
    if ((secdir == DOWN)? (secnew.left_b() <= 0.0): (secnew.right_b() >= 0.0)){
       // here  we start to build the Poincare image
       //  we have to initalize PoincVal,Px,CBr,CDx,Clmarr
      *startPoinc=1;
    }
  } // endif if(!(*startPoinc))

  if ((secdir == DOWN)? (secnew < 0): (secnew > 0)){
    poinc_finished=1;
  }


  // Making various updates
  if(*startPoinc){
    if(!oldstartPoinc){  // we have to initialize PoincVal and CPoincVal
      copyvector(xd,PoincVal,ode.dim);
      if(usediffcoord)
        x.intvencl(CPoincVal,usediffcoord,coordch);
    }
    if(!poinc_finished)
      c0updates_in_sec_reg(ode,h,x,xd,xn,xdn,xencl,usediffcoord,coordch,
                              lmarrmax,PoincVal,CPoincVal);
  }

  return poinc_finished;
}



void repbaddir(bool c1compt,Ode &ode,interval PoincVal[],interval valvfield[],
   interval grad_of_sec_func[],interval pval[],interval s_grad[])
// c1compt - should be true, iff the problem was found in computation of
//          the Jacobian matrix, only in this case <pval> and <s_grad>
//             will be used
{
  ofstream errorlog(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  errorlog <<"----------------------------------" << endl;
  writeCompInfo(errorlog);
  writedatetime(errorlog);
  ode.writeparameters(errorlog);

  errorlog << " General_Lo_Poincare - the problem in section region" << endl;
  if(c1compt) errorlog << " with modification of varaiatioal matrix " << endl;

  errorlog << endl << "PoincVal - enclosure in section region of result of Poincare map " << endl;
  saveAllCoord(errorlog,PoincVal,ode.dim,2,(ode.dim > 4),4);

  errorlog << endl << "Vfield(PoincVal)" << endl;
  saveAllCoord(errorlog,valvfield,ode.dim,1,(ode.dim > 4),4);

  errorlog << endl << "Gradient(Section)" << endl;
  saveAllCoord(errorlog,grad_of_sec_func,ode.dim,1,(ode.dim > 4),4);

  interval sc=scprod(grad_of_sec_func,valvfield,ode.dim);

  errorlog << endl << "Gradient * Vfield= " << sc << endl;
  if(sc.contains(0)) errorlog << "Contains ZERO !!!" << endl;
  else  errorlog << " direction in section region ok " << endl;

  if(c1compt){
    errorlog << endl << "In C1-computation "<< endl;
    errorlog << endl << "pval - intersection of PoincVal and section " << endl;
    saveAllCoord(errorlog,pval,ode.dim,2,(ode.dim > 4),4);

    errorlog << endl << "Gradient(Section)" << endl;
    saveAllCoord(errorlog,s_grad,ode.dim,1,(ode.dim > 4),4);

    interval gradRf=0.0;
    for(int i=1;i<= ode.dim; i++)
      gradRf+=s_grad[i]*(ode.vfield(i,pval));

    errorlog << endl << "Gradient * Vfield(pval)= " << gradRf << endl;
    if(gradRf.contains(0)) errorlog << "Contains ZERO !!!" << endl;
    else  errorlog << " direction in section region ok " << endl;
  }
}



int General_Lo_Poincare(bool c1computations,Ode &ode,  Par_Poincare &param,Set & x,
             const C1Func &section,int secdir,estimates  *estm,interval Px[],
             interval dp[][MAX_DIM+1], int usediffcoord,LinCC *coordch,int c1numLohner/*=0 */)
//  c1computations==true - P and dP is computed
//       otherwise P - only
//  computes rigorously the Poincare map of <x> using the Lohner
//                      algorithm
//  1 if returned, when Poincare map can be computed and the return time
//          is <= tmax,
//     then P(set(numLohner,x0,C0,r0)) \subset Px  (see below for the description
//     of the representation)
//  equation for section is given by  section.eval(x)=0 , but we are interested in the intersection
//     in the correct direction: if secdir==DOWN, then section(x) should decrease along the trajectory
//                               if secdir==UP,  then section(x) should increase along the trajectory
//
//  if !c1computations
//  dp - a  matrix C - just after passing the section
//          this  a Lohner matrix   C - for controlling a Lipschitz part
//             of error, used elsewhere
//             for diagnosing properties, sooner or latter this type of information
//         shall become part of the estm - class, which will collect what is
//         needed for a particular computations
//  if c1computations - then <dp> is a derivative with of the return map
//
//
//  usediffcoord - tells if you different coordinates in the section region
//  coordch  - coordinates in which we express the result <Px>
//     (these are new coordinates), dp - is always expressed in OLD coordinates


{



 interval h=param.h;  // this is a value of time step used from the begining,
      // but close to the section the time-step is chosen to
      // assure reasonable section error


 interval rem[MAX_DIM+1];
 interval h0=h;
 double remMax=0;

 interval xencl[MAX_DIM+1];  // used for enclosures of \varphi([0,h],x)

 interval lmarrmax[MAX_DIM+1];    // to deal with the section
                                //errors  componentwise
 zero(lmarrmax,ode.dim);
 interval CPoincVal[MAX_DIM+1]; // used if <usediffcoord> - to compute
                // the Poincare map in other coordinates


 // the variables used in C1-computations
 C1Set *V, *Vn;
// interval dpn[MAX_DIM+1][MAX_DIM+1];  // new value of V(after time step)
 intervalmatrix1 *dpn;
// interval dpencl[MAX_DIM+1][MAX_DIM+1];  //  enclosure for V([t,t+h]) - during the
                 // time step
 intervalmatrix1 *dpencl;


// interval dprem[MAX_DIM+1][MAX_DIM+1];  //  remainder part for V
 intervalmatrix1 *dprem;


// interval DPVal[MAX_DIM+1][MAX_DIM+1];
 intervalmatrix1 *DPVal;
 //  end - the variables used in C1-computations



 Set xn=x;   // this is just to initialize <xn> with something
 interval xd[MAX_DIM+1],xdn[MAX_DIM+1]; // x \subset xd
                                      // xn \subset xdn

 x.intvencl(xd);
 if(c1computations){
   dpencl=new intervalmatrix1[ode.dim+1];
   dpn=new intervalmatrix1[ode.dim+1];
   dprem=new intervalmatrix1[ode.dim+1];
   DPVal=new intervalmatrix1[ode.dim+1];
   V=new C1Set(ode.dim,c1numLohner);
   Vn=new C1Set(ode.dim,c1numLohner);
   V->intvencl(dp);
 }


 initestm(estm,ode.dim);
 interval secnew=section.eval(x);
 interval secold=secnew;

 int insecregion=0;
 int startPoinc=0;
 interval PoincVal[MAX_DIM+1];

 interval tPoinc;

 int num_time_steps_in_sec_reg=0;

 int ok=1;
 int firststep=1;
 interval t=0;
 interval new_h=h;
 int poinc_finished=0;
 while((t<=param.tmax) && (!poinc_finished) && ok){
   int dosubst=1;
   h=new_h;
   if(!c1computations){
     ok=OneStepLohner(ode,param.onesteppar,h,x,xd,xn,xdn,xencl,rem);
      // xencl=enclosure for \varphi([0,h],x)
      // xd=enclosure for x = eval(x)
      // xdn = enclosure for image set xn
   }else{


     ok=C1OneStepLohner(ode,param.onesteppar,h,x,xd,xn,xdn,xencl,rem,*V,dp,
       *Vn,dpn,dpencl,dprem);

       // xencl=enclosure for \varphi([0,h],x)
      // xd=enclosure for x; xd=eval(x)
      // xdn = enclosure for image set xn
      // dpencl = enclosure for  V(t+[0,h],x_0)
      // dp - enclosure for *V=V(t,x_0))
      // dpn - enclosure for *Vn=V(t+h,x_0)
   }

   OkGen_Poincare(&ok,x,xd,xdn,ode.dim,t,h,"General_Lo_Poincare");  // a test for errors
   if(!ok) continue;


   secnew=section.eval(xn);



   if (!firststep){
     if (insecregion){
       num_time_steps_in_sec_reg++;
       int oldstartPoinc=startPoinc;
       poinc_finished=in_sec_region(secdir,secnew,&startPoinc,ode,h,
          x,xd,xn,xdn,xencl,usediffcoord,coordch,lmarrmax,PoincVal,CPoincVal);
       if(poinc_finished){
         // we try to make the time-step as small as possible and
         //  still pass through section:
         interval h_used;
         if(get_past_section(section,secdir,ode,param.onesteppar,h,x,xd,xn,xdn,
                                xencl,rem,&h_used)){
           h=h_used;
           c0updates_in_sec_reg(ode,h,x,xd,xn,xdn,xencl,usediffcoord,coordch,
                              lmarrmax,PoincVal,CPoincVal);

           if(c1computations){
             //  in principle there is no need to worry about x,xd,xn,xdn
             // below, because C1One... should update it exactly as get_close...
             // did, but this can  change in the future
             //  and OneStepLohner and C1OneStepLohner may be different
             // on <x> variables, hence I prefer
             // rather to work on copies of <x>,<xd>,<xn>,<xdn>,<xencl>,<rem>
             Set cxn;
             Set cx=x;
             interval cxd[MAX_DIM+1],crem[MAX_DIM+1],
                       cxencl[MAX_DIM+1],cxdn[MAX_DIM+1];
             copyvector(xd,cxd,ode.dim);
             C1OneStepLohner(ode,param.onesteppar,h,cx,cxd,cxn,cxdn,cxencl,crem,
                    *V,dp,*Vn,dpn,dpencl,dprem);
           }
         } else{
           cerr << "General_Lo_Poincare - get_past... returned error" << endl;
           #if defined(WIN95) && defined(KRAK)
             cerr << "press any key" << endl;
             waitBt();
           #endif
           ok=0;
           poinc_finished=0;
           continue;
         }
       }

       if(c1computations){
         if(startPoinc){
           c1updates_in_sec_reg((oldstartPoinc!=startPoinc),ode,h,xencl,dp,dpn,
              dpencl,DPVal);
         }
       }

       // keeping track of the return time
       if((!oldstartPoinc) && startPoinc)
         tPoinc=t;
       if(startPoinc) tPoinc=Ihull(tPoinc,t+h);

     } // endif insecregion
     if (!insecregion){
       if ((secdir == DOWN ) ? ((secnew.left_b() <= 0.0) && (secold.left_b() > 0.0) ):
          ((secnew.right_b() >= 0.0) && (secold.right_b() < 0.0) )){
              // we have reached the section
         insecregion=1;
         dosubst=0;   // if !dosubst at the end of the loop, then
             // we will repeat the calculations for this time 't'
                       // with a smaller time step h

         interval h_used;
         if(get_close_to_section(section,secdir,ode,param.onesteppar,h,x,xd,xn,
                    xdn,xencl,rem,&h_used,&new_h)){

           if(new_h < 0.01*h0){
             new_h=0.01*h0;  // there is no point is using too small time step
              // because  it happens (in case of small sets) the
                   // proposed time set is many orders of magnitude smaller
                   // than h0
           }
           if(h_used >0){  // if <x> was already
                           // close enough, then h_used=0;
                           // in this case we are interested only in new_h
                           // and we don't do the substitution x=xn for
                           // the next iterate
             h=h_used;
             if(c1computations){
               //  in principle there is no need to worry about x,xd,xn,xdn
               // below, because C1One... should update them exactly as get_close...
               // did, but  this can change in the future
               //  and OneStepLohner and C1OneStepLohner may be different
               // on <x> variables, hence I prefer
               // rather to work on copies of <x>,<xd>,<xn>,<xdn>,<xencl>,<rem>
               Set cxn;
               Set cx=x;
               interval cxd[MAX_DIM+1],crem[MAX_DIM+1],
                         cxencl[MAX_DIM+1],cxdn[MAX_DIM+1];
               copyvector(xd,cxd,ode.dim);
               C1OneStepLohner(ode,param.onesteppar,h,cx,cxd,cxn,cxdn,cxencl,crem,
                    *V,dp,*Vn,dpn,dpencl,dprem);
             }
             dosubst=1;
             secnew=section.eval(xn);
           } // endif (h_used >0)
         } // get_close_...
         else{
           cerr << "General_Lo_Poincare - get_close... returned error" << endl;;
           #if defined(WIN95) && defined(KRAK)
             cerr << "press any key" << endl;
             waitBt();
           #endif
           ok=0;
           continue;
         }
       }
     }
   }  // endif !firststep

   if (dosubst){
     //  These things are less important - keeping track of various estimates
     // we calculate various quantities along trajectory
     UpdateBounds(xencl,estm,firststep);
     if(arrfabs(rem,ode.dim) > remMax) remMax=arrfabs(rem,ode.dim);
     //  END OF - These things are less important - keeping track of various
     //                          estimates

     t+=h;
     secold=secnew;
     x=xn;

     copyvector(xdn,xd,ode.dim);
     if(x.rearrangement(xd))
        estm->num_of_rep_switches=estm->num_of_rep_switches + interval(1.0);


     if(c1computations){
       *V=*Vn;
       copymatrix(dpn,dp,ode.dim);
       V->rearrangement(dp);
     }

     firststep=0;
   } // endif dosubst


 } // end while



 if(poinc_finished){
   interval valvfield[MAX_DIM+1];
   ode.field(PoincVal,valvfield);

   interval grad_of_sec_func[MAX_DIM+1];
   section.gradient(PoincVal,grad_of_sec_func);

   interval scprod_f_grad_s=scprod(grad_of_sec_func,valvfield,ode.dim);

   poinc_finished=((secdir == DOWN)?scprod_f_grad_s<0:scprod_f_grad_s>0 );
   if (!poinc_finished){
     repbaddir(0,ode,PoincVal,valvfield,grad_of_sec_func,(interval *)NULL,
             (interval *)NULL);
     cerr << "General_Lo_Poincare - Bad direction of vect. f. in section region" << endl;
     #if defined(WIN95) && defined(KRAK)
       cerr << "press any key " << endl;
       waitBt();
     #endif
   }

   if(poinc_finished){
     if(!usediffcoord)  copyvector(PoincVal,Px,ode.dim);
     else   copyvector(CPoincVal,Px,ode.dim);

     if(!c1computations)
       xn.get_Cmatrix(dp);

     if(c1computations){
       interval s_grad[MAX_DIM+1];
       interval pval[MAX_DIM+1];
       section.intersection(PoincVal,pval);
       section.gradient(pval,s_grad);
       if(!ode.JacPoincMod(s_grad,pval,DPVal,ode.dim)){
         poinc_finished=0;
         repbaddir(1,ode,PoincVal,valvfield,grad_of_sec_func,pval,s_grad);
       }
       copymatrix(DPVal,dp,ode.dim);
     }

     poincUpdateEstm(estm,ode,section,param.onesteppar.order,h0,x,PoincVal,tPoinc,
                   remMax,lmarrmax,usediffcoord,coordch,CPoincVal);
     estm->sec_timestepnumber=num_time_steps_in_sec_reg;

   }
 } // endif poinc_finished

 if(c1computations){
   delete V;
   delete Vn;
   delete [] DPVal;
   delete [] dprem;
   delete [] dpencl;
   delete [] dpn;
 }


 return(poinc_finished);
}




int LogNormPoincMap(Ode &ode, int method, int order, interval h, interval xb[],
             double delta, interval x[] ,double tmax,const C1Func &section,
             int secdir, estimates  *estm )
// calculates rigorously the Poincare map on (xb + [-delta,delta]^dim)
//   using logarithmic norms
//  equation for section is given by formula:
//     section.eval(x)=0
{

  cerr << "not all logarithmic norms are supported for all equations" << endl;
  #if defined(WIN95) && defined(KRAK)
    rootFrame << "press any key " << endl;
    waitBt();
  #endif


  int oldwaysecregion=0;
  interval t=0;
  interval rem[MAX_DIM+1];
  interval idelta,ideltan,deltan;
  interval hdo;

  double remMax=0;
  double diam;
  interval xn[MAX_DIM+1],y[MAX_DIM+1],y1[MAX_DIM+1],trash[MAX_DIM+1];

  interval xd[MAX_DIM+1],xdn[MAX_DIM+1]; // computer evaluated C*x + B*z
                                     // or whatever represent our set
  interval xdnLM[MAX_DIM+1]; //  used in the section region in estimation where
                      // trajectory was between time steps
  interval xdLM[MAX_DIM+1]; //  used in the section region


  interval loclip;
  interval loglip;
  interval logN;
  interval intRkErr,explogn;
  interval LMVal;  // value of |f||df| in section region
  interval LMValMax;

  interval lmarr[MAX_DIM+1];     // in new method - section errors are dealt
                                // componentwise
  interval lmarrmax[MAX_DIM+1];


  int i;
  int poinc_finished;
  int firststep=1;

  interval secold,secnew;

  int insecregion=0;
  int startPoinc=0;
  interval PoincVal[MAX_DIM+1];
  interval tPoinc;

  initestm(estm,ode.dim);

  hdo=1.0;
  for(i=1;i<=order+1;i++)
    hdo*=h;
  hdo=hdo.right_b();

  for(i=1; i <= ode.dim; i++) x[i]=xb[i];

  if (method==RUNGEKUTTA)
    order=4;

  intRkErr=0.0;
  loclip=1.0;

  secnew=section.eval(x);
  secold=secnew;

  idelta=interval(-delta,delta);
  for(i=1;i<=ode.dim;i++)
   xd[i]=x[i]+ idelta;

  int ok=1;

  poinc_finished=0;
  while((t<=tmax) && (!poinc_finished)  && ok){
    int dosubst=1;

    // estimation of the set y,  \varphi([0,h],x) \subset Y
    // for error terms
    ok=ode.enclosure(x,h,y);
    if(!ok){
      cerr << "LogNormPoincMap: bad enclosure for center point " << endl;
      #if defined(WIN95) && defined(KRAK)
        cerr << "press any key " << endl;
        waitBt();
      #endif
      break;
    }

    // estimation of the set y1,  \varphi([0,h],x+K(0,\delta)) \subset Y1
    // for Lipschitz constant calculations (or variations )
    ok=ode.enclosure(xd,h,y1);
    if(!ok){
      cerr << "LogNormPoincMap: bad enclosure for xd " << endl;
      #if defined(WIN95) && defined(KRAK)
        cerr << "press any key " << endl;
        waitBt();
      #endif
      break;
    }



    if (method==TAYLORM){
      taylor(ode,order,h,x,xn,0,rem);   // to calulate xn=\varphi(h,x)
      taylor(ode,order,h,y,trash,1,rem); // to calculate remainder term
    }


    if (method==RUNGEKUTTA){
      cerr << "Runge-Kutta not supported " << endl;
       #if defined(KRAK) && defined(WIN95)
                     waitBt();
    #endif
      exit(EXIT_FAILURE);
    }


    loglip=ode.lognorm(y1);

    explogn=exp(h*loglip);

    MultiISplit(xn,&diam,ode.dim,ode.lognormused);
    deltan=interval(delta)*explogn+norm(rem,ode.dim,ode.lognormused)*hdo + diam;
    deltan=Right(abs(deltan));
    ideltan=Ihull(-deltan,deltan);
    for(i=1;i<=ode.dim;i++)
      xdn[i]=xn[i]+ ideltan;



    ok=!MultiIn(xd,xdn,0,ode.dim);
    if(!ok){
      #if defined(WIN95) && defined(KRAK)
        rootFrame.Clear();
      #endif
      cerr << "LogNormPoincMap - image stabilized " << endl;
      cerr << "t=" << (t+h) << endl;
      cerr << "xd[i]    xdn[i] " << endl;
      for(int i=1;i<=ode.dim;i++)
        cerr << xd[i] << "   " << xdn[i] << endl;
    }
    if(!ok){
      #if defined(WIN95) && defined(KRAK)
        cerr << "press any key " << endl;
        waitBt();
      #endif
      break;
    }


    // compute some quatities for a trajectory


    if(arrfabs(rem,ode.dim) > remMax) remMax=arrfabs(rem,ode.dim);


    UpdateBounds(y1,estm,firststep);


    if (firststep) logN=loglip;
    else logN=Ihull(loglip,logN);




    // the end of computation of some auxilary data


    secnew=section.eval(xdn);

    if (!firststep){
      if(insecregion){
        if(oldwaysecregion){
          LMVal=ode.lmfunc(y);
          MultiINeigh(xdn,LMVal*h*h/2,xdnLM,ode.dim); // xdnLM=xdn+[-+LMVal*h^2/2]^ode.dim
          MultiINeigh(xd,LMVal*h*h/2,xdLM,ode.dim); // xdLM=xd+[-+LMVal*h^2/2]^ode.dim
        } else {
          ode.fromsegdist(y,lmarr,h);
          for(int i=1;i<=ode.dim;i++){
            xdnLM[i]=xdn[i]+interval(-lmarr[i].right_b(),lmarr[i].right_b());
            xdLM[i]=xd[i]+interval(-lmarr[i].right_b(),lmarr[i].right_b());
          }
        }

        if (!startPoinc){
          if ((secdir == DOWN)? (secnew.left_b() <= 0.0): (secnew.right_b() >= 0.0)){
            startPoinc=1;
            if(oldwaysecregion){
              LMValMax=LMVal;
            } else {
              for(int i=1;i<=ode.dim;i++){
                lmarrmax[i]=lmarr[i];
              }
            }
            MultiIhull(xdLM,xdnLM,PoincVal,ode.dim);   // PoincVal=xdLM \cup xdnLM
                              // in this way we know that whole trajectories
                              //  started in our set <xd> which after time <h>
                              // are mapped into <xdn> are contained in <PoincVal>
            tPoinc=t;
          }
        }else {  // form if (!startPoinc)
          if(oldwaysecregion){
            LMValMax=Ihull(LMValMax,LMVal);
          } else {
            for(int i=1;i<=ode.dim;i++)
              lmarrmax[i]=Ihull(lmarrmax[i],lmarr[i]);
          }
          MultiIhull(xdLM,PoincVal,PoincVal,ode.dim);
          MultiIhull(xdnLM,PoincVal,PoincVal,ode.dim);
        } // endif !startPoinc
        if(startPoinc){
          if ((secdir == DOWN)? (secnew.right_b() < 0): (secnew.left_b()> 0)){
            poinc_finished=1;
            tPoinc=Ihull(tPoinc,t+h);
          }
        }
      } // endif insecregion
      if (!insecregion){
        if ((secdir == DOWN ) ? ((secnew.left_b() <= 0.0) && (secold.left_b() > 0.0) ):
           ((secnew.right_b() >= 0.0) && (secold.right_b() < 0.0) )){
          insecregion=1;
          h/=ode.divsecreg;
          dosubst=0;   // we will repeat the calculations for this time 't'
                       // with smaller time step h
        }
      }
    } // endif firststep
    firststep=0;
    if (dosubst){
      t+=h;
      loclip=loclip*explogn;
      intRkErr=arrfabs(rem,ode.dim)*h + intRkErr*explogn;

      secold=secnew;
      delta=abs(deltan).right_b();

      for(i=1; i <= ode.dim; i++){
        x[i]=xn[i];
        xd[i]=xdn[i];
      }
    }
  }  // end while



  if (poinc_finished){
    interval valvfield[MAX_DIM+1];
    ode.field(PoincVal,valvfield);


    interval grad_of_sec_func[MAX_DIM+1];
    section.gradient(PoincVal,grad_of_sec_func);

    interval scprod_f_grad_s=scprod(grad_of_sec_func,valvfield,ode.dim);

    poinc_finished=((secdir == DOWN)?scprod_f_grad_s<0:scprod_f_grad_s>0 );


    if (!poinc_finished) {
      cerr << "Bad direction of vector field in section region" << endl;
      #if defined(WIN95) && defined(KRAK)
         cerr << "press any key " << endl;
         waitBt();
      #endif
      return(0);
    }

    for(i=1;i<=ode.dim;i++)
      x[i]=PoincVal[i];

    estm->rettime=t;
    estm->LipConst=loclip;
    estm->IntRkError=intRkErr;
    estm->Rkerror = remMax;
    estm->LogN=logN;
    if(oldwaysecregion)
      estm->LMVal=LMValMax*h*h/2;
    else {
      estm->LMVal=lmarrmax[1];
      for(int i=2;i<=ode.dim;i++)
        estm->LMVal=Ihull(lmarrmax[i],estm->LMVal);
    }
  }
  return(poinc_finished);
}





int tosectiontime(Ode &ode,double x[],double xn[],double *poinctime,double h,
        const C1Func &section, int secdir, int method,int order, int numberofiterates)
// the function computes the time <x>  needs to reach the section <section>
//    - the evolution is governed by the <ode>
//// On input:
//     x - is our point of interest
//     xn=\varphi(h,x)
//     x and xn must be on different sides of the section
//     section  - specifies the Poincare section and the direction
//      in which we cross the section
//     <method> and <order> - specify the numerical method
//     numberofiterates - technical parameter tells us how many times
//           we will try to improve xn and the reaching time
//  On output:
//     xn - in the point on section (approximately)
//     *poinctime - an approximate time of reaching the section
//  1 - is returned if everything was ok
//  0 - otherwise, this happens when the section is not transversal to the flow
{

  int ok=1;
  double xl[MAX_DIM+1],xr[MAX_DIM+1],xm[MAX_DIM+1];
  double sl,sr,sm;

  double trash[MAX_DIM+1];

  copyvector(x,xl,ode.dim);
  copyvector(xn,xr,ode.dim);
  sl=section.eval(xl);
  sr=section.eval(xr);
  if(sl*sr >0){
    cerr << "tosectiontime - both points are on the same side of the section" << endl;
    #if defined(WIN95) && defined(KRAK)
       waitBt();
       exit(EXIT_FAILURE);
    #endif
  }


  int found=0;
  int iter=1;

  if(sl==0.0){ // never happens, but ...
    found=1;
    copyvector(xl,xn,ode.dim);
    *poinctime=0.0;
  }
  if(!found){
    if(sr==0.0){
      found=1;
      copyvector(xr,xn,ode.dim);
      *poinctime=h;
    }
  }
  double tstep=h;
  if(!found)  *poinctime=0.0;
  while(!found && (iter<=numberofiterates ) && ok){

    // to guess the correct Poincare time, we assume that the section function
    //  changes linearly along the segment connecting  xl and xr (which
    //  are on the same trajectory - shifted in time  - by <tstep>)
    //  and we assume that the flow is parallel along this segment with a uniform   speed
    //
    //      segment is given by xl + c(xr-xl) for c \in [0,1]
    //      then  c= - section(xl)/(section(xr) - section(xl))
    //      and the  Poincare time will be c*tstep
    double c;
    c=-sl/(sr-sl);

    if (method==TAYLORM) taylor(ode,order,c*tstep,xl,xm,0,trash);
    if (method==RUNGEKUTTA) rk4(ode,c*tstep,xl,xm);
       // to calculate xm=\varphi(c*tstep,xl)

    sm=section.eval(xm);
    *poinctime+=c*tstep;
    if(sm==0.0){ // never happens, but ...
      found=1;
      copyvector(xm,xn,ode.dim);
    } else{
      //  xm - is not on the section
      // we need to find a point on the other side  with
      // the distance from the section of the same order

      double gradsection[MAX_DIM+1];
      section.gradient(xm,gradsection);

      double vectfield[MAX_DIM+1];
      ode.field(xm,vectfield);

      double secXfield=scprod(gradsection,vectfield,ode.dim); // this is d section/dt

      if (secXfield==0) ok=0; // the section is not transversal
      double hs=-sm/secXfield;
      if(fabs(hs) > fabs(h)) ok=0;  // the section appears to be not transversal
      if(ok){
        double hh=hs;
        double st;
        double xt[MAX_DIM+1];
        do{
          hh=hh + hs;
          if (method==TAYLORM) taylor(ode,order,hh,xm,xt,0,trash);
          if (method==RUNGEKUTTA) rk4(ode,hh,xm,xt);
           // to calculate xt=\varphi(hh,xm)
           st=section.eval(xt);
        }   while (st*sm > 0);   // this should be violated usually in the first step

        if(((sm<0) && (secdir==UP) ) || (sm>0) && (secdir==DOWN)){
           // xm - is before the section, hh - is positive
           copyvector(xm,xl,ode.dim);
           sl=sm;
           copyvector(xt,xr,ode.dim);
           sr=st;
        } else{
           // xm - went on the other side of the section
           // xt - is now before the section and hh - is negative
           *poinctime+=hh;         //  it was 18.VII - 16:34 ; ERROR:  *poinctime-=hh;
           copyvector(xt,xl,ode.dim);
           sl=st;
           copyvector(xm,xr,ode.dim);
           sr=sm;
        }
        tstep=fabs(hh);
      }
    }
    iter++;
  }
  if(!found)
    copyvector(xm,xn,ode.dim);

  return ok;
}







int NonRigPMap(int calcvar,Ode &ode,const Par_Poincare &param, const double xb[],
      double x[],double* tPoinc,const C1Func &section,
      int secdir,estimates *estm,double
        Dx[][MAX_DIM+1],double Dx0[][MAX_DIM+1],int dimVar)
// calculates the Poincare map of (xb) nonrigorously:
// On output: x=P(xb)
// calcvar - (L) compute Jacobian of the flow (Dx - Jacobian of Dp on output)
// dimVar - the dimension for Jacobian calculations - in case of invariant
//          manifold we may want it to be less than ode.dim
// Dx -  on output - Jacobian matrix of the flow, treated as
//     the function in R^{ode.dim} - to make this the jacobian of Poincare
//     map it is necessary to introduce coordinates on the section and
//     extract DP
//  the equation for section is given by formula:
//       section.eval(x)=0
//// the end part of the procedure (adjusting energy) silently assumes that the
//    section is linear, when we investigate a system with a first integral of
//    motion, for example Hamiltonian equations
// Dx0 - the initial condition for Dx, it is ok to to send Dx0=NULL; then
//   Dx is initialized with the identity
{

 int method=param.onesteppar.method;
 int order=param.onesteppar.order;
 double h=param.h;
 double tmax=param.tmax;
 int ordervar=param.onesteppar.orderdf;

 int ok=1;


 double rem[MAX_DIM+1];
 double xn[MAX_DIM+1];

 int poinc_finished;
 int firststep=1;

 double secold,secnew;
 double PoincVal[MAX_DIM+1],Dxn[MAX_DIM+1][MAX_DIM+1];

 for(int i=1; i <= ode.dim; i++) x[i]=xb[i];

 if(method==RUNGEKUTTA)
   order=4;

 if (calcvar){
   if(dimVar==0) dimVar=ode.dim;
   if(dimVar > ode.dim) dimVar=ode.dim;
   if(Dx0 != NULL){
     copymatrix(Dx0,Dx,dimVar);
   }else{
     identity(Dx,dimVar);
   }
 }

// secnew=scprod(x,c_sec,ode.dim) + c_sec[0];
 secnew=section.eval(x);
 secold=secnew;

 initestm(estm,ode.dim);
 UpdateBounds(x,estm,1);

 double t=0;
 poinc_finished=0;
 while((t<=tmax) && (!poinc_finished) && ok){
   if (method==TAYLORM) taylor(ode,order,h,x,xn,0,rem);
   if (method==RUNGEKUTTA) rk4(ode,h,x,xn);
       // to calulate xn=\varphi(h,x)

   ode.adjustenergy(xn);

   if (calcvar) OneStepVar(ode,h,x,Dx,Dxn,dimVar,ordervar);

   secnew=section.eval(xn);

   if (!firststep){
     poinc_finished=(secdir==UP)?((secold <=0) && (secnew > 0)):((secold >0) && (secnew <= 0));
     if(poinc_finished){
       double tstep,tonsec;
       ok=tosectiontime(ode,x,xn,&tonsec,h,section,secdir,method,order,3);
       if(ok){
         copyvector(xn,PoincVal,ode.dim);
         *tPoinc=t+ tonsec;
         tstep=tonsec;


         if(calcvar)
           OneStepVar(ode,tstep,x,Dx,Dxn,dimVar,ordervar);
       }else{
         poinc_finished=0;
         ofstream errorlog(zglerrorlogname,ios::app);
         entriesinerrorlog=1;
         errorlog <<"----------------------------------" << endl;
         writeCompInfo(errorlog);
         writedatetime(errorlog);

         cerr <<  " NonRigorousPoincMap - tosectiontime function failed " << endl;
         cerr << "t=" << t << "   h=" << h <<  endl;

         errorlog <<  " NonRigorousPoincMap - tosectiontime function failed " << endl;
         errorlog << "t=" << t << "   h=" << h <<  endl;

         saveAllCoord(errorlog,x,ode.dim,1,(ode.dim > 4));
       }
     }
   }
   if(ok){
     firststep=0;

     t+=h;
     secold=secnew;

     for(int i=1; i <= ode.dim; i++) x[i]=xn[i];
     if(calcvar)  copymatrix(Dxn,Dx,dimVar); // Dx=Dxn

     UpdateBounds(xn,estm,0);
   }
 }
 if(poinc_finished){
   for(int i=1;i<=ode.dim;i++)  x[i]=PoincVal[i];

   //// the part below of the procedure (adjusting energy) silently assumes that the
   //    section is linear, when we investigate a system with a first integral of
   //    motion, for example Hamiltonian equations

   double gradsec[MAX_DIM+1];
   section.gradient(x,gradsec);

   ode.adjustenergy(x,gradsec);
   if(calcvar){
     ode.JacPoincMod(gradsec,x,Dx,dimVar);
   }
 }
 return(poinc_finished);
}


int NonRigorousPoincMap(Ode &ode,const Par_Poincare &param ,const double xb[],
      double x[],double* tPoinc,const C1Func &section,
      int secdir,estimates *estm,double
      Dx[][MAX_DIM+1], int calcvar/* =0*/,int dimVar/*=0*/)
// calculates the Poincare map of (xb) nonrigorously:
// On outpout: x=P(xb)
// calcvar - (L) compute Jacobian of the flow
// dimVar - dimension for Jacobian calculations - in case of invariant
//          manifold we may want it to be less than ode.dim
// Dx -  on output - Jacobian matrix of the folow, treated as
//     the function in R^{ode.dim} - to make this a jacobian of Poincare
//     map it is necessary to introduce coordinates on the section and
//     extract DP
//  the equation for section is given by formula:
//       section.eval(x)=0
//// the end part of the procedure (adjusting energy) silently assumes that the
//    section is linear, when we investigate a system with a first integral of
//    motion, for example Hamiltonian equations
{
  doublematrix1* Dx0=NULL;
  return NonRigPMap(calcvar,ode,param ,xb,x,tPoinc,section,secdir,estm,Dx,Dx0,dimVar);
}



int iterate_nonrigpoinc(int itno,Ode &ode,Par_Poincare param,double x[],
         double xn[],double *tpoinc,const C1Func &section,int secdir, estimates *estm)
//  nonrigorously computes <itno>-th iterate of Poincare map for <ode>
//  Poincare map is defined by
//  section.eval(x)=0
//  if secdir==UP - then <section.eval()> is
//      increasing on section
//  x - starting point
//  xn=P^{itno}(x) - on output
{

 double dummy[MAX_DIM+1][MAX_DIM+1];
 double ltpoinc;
 *tpoinc=0;
 double y[MAX_DIM+1];
 for(int i=1;i<=ode.dim ;i++){
   xn[i]=x[i];
 }
 int ok=1;
 int iter=1;


 while((iter <= itno) && ok ){
   for(int i=1;i<=ode.dim ;i++){
     y[i]=xn[i];
   }
   ok= NonRigorousPoincMap(ode,param,y,xn,&ltpoinc,section,
                      secdir,estm,dummy);
   if(ok) *tpoinc+=ltpoinc;
   iter++;
 }
 return(ok);
}






