#ifndef _rigpoinc_h_
#define _rigpoinc_h_

#include"defo.h"
#include"zdynsys.h"
#include"metnum.h"
#include"func.h"
#include"coordc.h"



void poincUpdateEstm(estimates *estm,Ode &ode,const C1Func &section,
     int order,interval h0,Set &x,interval PoincVal[],
     interval tPoinc,interval remMax,interval lmarrmax[],
     int usediffcoord ,LinCC*  coordch ,interval  CPoincVal[]);
// x =\varphi(t,initial condition), where t >tPoinc - used here to gauge
//              the size of what we can get without the section errors
//     also for the doubleton sets - the representation x.r0 - contains
//          the size of initial data as x=0 + C*r0 + B*r
//            and x_0,C,B,r - are changed during the evolution
//            r0 - stays constant
// PoincVal - the value of P(initial condition) - always expressed in standard
//             coordinates
// CPoincVal - the value of P(initial condition) -  expressed in new coordinates
//             if usediffcoord==1, otherwise is not used
// coordch - the new coordinates
/////
//  when used for a differential inclusion ,then it has a tendency to underestimate
//   the width of the passing throough the section time

double newtimestep(double ha,double minh,double maxh,double biasratio);
  // the function returns new proposed value for <ha>
  // used in the searches for good time shift along the trajectory
// biasratio  should less than 1/2

void near_section_problems_report(const char comment[],int dim,int maxitnum,
   interval xd[],interval xdn[],interval sec_on_x,interval sec_on_xn,
   double c,interval dsec_dt,double ds_dt,double h,double minh, double maxh,
      double ha );
  // the error diagonosis function for functions: 'get_close_to_section' and
  //   'get_past_section'


int General_Lo_Poincare(bool c1computations,Ode &ode,  Par_Poincare &param,Set & x,
             const C1Func &section,int secdir,estimates  *estm,interval Px[],
             interval dp[][MAX_DIM+1], int usediffcoord,LinCC *coordch,int c1numLohner=0 );
//  c1computations==true - P and dP is computed
//       otherwise P - only
//  computes rigorously the Poincare map of <x> using the Lohner
//                      algorithm
//  1 if returned, when Poincare map can be computed and the return time
//          is <= tmax,
//     then P(set(numLohner,x0,C0,r0)) \subset Px  (see below for the description
//     of the representation)
//  equation for section is given by  section.eval(x)=0 , but we are interested in the intersection
//     in the correct direction: if secdir==DOWN, then section(x) should decrease along the trajectory
//                               if secdir==UP,  then section(x) should increase along the trajectory
//
//  if !c1computations
//  dp - a  matrix C - just after passing the section
//          this  a Lohner matrix   C - for controlling a Lipschitz part
//             of error, used elsewhere
//             for diagnosing properties, sooner or latter this type of information
//         shall become part of the estm - class, which will collect what is
//         needed for a particular computations
//  if c1computations - then <dp> is a derivative with of the return map
//
//
//  usediffcoord - tells if you different coordinates in the section region
//  coordch  - coordinates in which we express the result <Px>
//     (these are new coordinates), dp - is always expressed in OLD coordinates



int LogNormPoincMap(Ode &ode, int method, int order, interval h, interval xb[],
             double delta, interval x[] ,double tmax,const C1Func &section,
             int secdir, estimates  *estm );
// calculates rigorously the Poincare map on (xb + [-delta,delta]^dim)
//   using logarithmic norms
//  equation for section is given by formula:
//     section.eval(x)=0


int NonRigPMap(int calcvar,Ode &ode,const Par_Poincare &param ,const double xb[],
      double x[],double* tPoinc,const C1Func &section,
      int secdir,estimates *estm,double
        Dx[][MAX_DIM+1],double Dx0[][MAX_DIM+1],int dimVar);
// calculates the Poincare map of (xb) nonrigorously:
// On output: x=P(xb)
// calcvar - (L) compute Jacobian of the flow (Dx - Jacobian of Dp on output)
// dimVar - the dimension for Jacobian calculations - in case of invariant
//          manifold we may want it to be less than ode.dim
// Dx -  on output - Jacobian matrix of the flow, treated as
//     the function in R^{ode.dim} - to make this the jacobian of Poincare
//     map it is necessary to introduce coordinates on the section and
//     extract DP
//  the equation for section is given by formula:
//       section.eval(x)=0
//// the end part of the procedure (adjusting energy) silently assumes that the
//    section is linear, when we investigate a system with a first integral of
//    motion, for example Hamiltonian equations
// Dx0 - the initial condition for Dx, it is ok to to send Dx0=NULL; then
//   Dx is initialized with the identity


int NonRigorousPoincMap(Ode &ode,const Par_Poincare & param ,const double xb[],
        double x[],double* tPoinc,const C1Func &section,int secdir,
        estimates *estm,double Dx[][MAX_DIM+1], int calcvar =0,int dimVar=0);
// calculates the Poincare map of (xb) nonrigorously:
// On outpout: x=P(xb)
// calcvar - (L) compute Jacobian of the flow
// dimVar - dimension for Jacobian calculations - in case of invariant
//          manifold we may want it to be less than ode.dim
// Dx -  on output - Jacobian matrix of the folow, treated as
//     the function in R^{ode.dim} - to make this a jacobian of Poincare
//     map it is necessary to introduce coordinates on the section and
//     extract DP
//  the equation for section is given by formula:
//       section.eval(x)=0
//// the end part of the procedure (adjusting energy) silently assumes that the
//    section is linear, when we investigate a system with a first integral of
//    motion, for example Hamiltonian equations

int iterate_nonrigpoinc(int itno,Ode &ode,Par_Poincare param,double x[],
         double xn[],double *tpoinc,const C1Func &section,int secdir, estimates *estm);
//  nonrigorously computes <itno>-th iterate of Poincare map for <ode>
//  Poincare map is defined by
//  section.eval(x)=0
//  if secdir==UP - then <section.eval()> is
//      increasing on section
//  x - starting point
//  xn=P^{itno}(x) - on output






#endif //  _rigpoinc_h
