#include<iostream>
#include<fstream>
using namespace std;

#include<cstdlib>

#include"defo.h"
#include"coordc.h"
#include"multiint.h"
#include"rigpoinc.h"
#include"qrdecmp.h"
#include"func.h"
#include"section.h"
#include"zgfiles.h"




char* osa_coordsname(char* name,int maxlen,int coords)
{
  strcpy(name,"");
  if(coords==standardcoord)
    strncpy(name,"standard coordinates",maxlen-1);
  if(coords==orthogonalizedeigenvectors)
    strncpy(name,"orthogonalized eigenvectors",maxlen-1 );
  if(coords==eigenvectors)
     strncpy(name,"normalized eigenvectors",maxlen-1 );

  if(coords==partialeigenbasis)
     strncpy(name,"leading eigenvectors + orthogonalized rest",maxlen-1 );

  if(coords==eigen_subspaces_ortogonalized)
     strncpy(name,"clusters of eigenvectors  orthogonalized ",maxlen-1 );

  return name;
}

char *namehsetstrategy(int strategy, char *stratname,int maxlen)
{
  strcpy(stratname,"");
  if(strategy==gen_new_intersect)
    strncpy(stratname,"intersections",maxlen-1);
  if(strategy==gen_new_shift)
    strncpy(stratname,"shift",maxlen-1);
  if(strategy==gen_from_point)
    strncpy(stratname,"from point",maxlen-1);
  return stratname;
}

int teststrategyVal(int strategy)
// returns 1, when strategy has a correct value
{
  int ret=0;
  if(strategy==gen_new_intersect)
    ret=1;
  if(strategy==gen_new_shift)
    ret=1;
  if(strategy==gen_from_point)
    ret=1;
  return ret;
}


chnsecparType::chnsecparType()
{
  zero();
}

void chnsecparType::writetxt(ostream &report,int secdataonly /*=0 */) const
{
  if(sec_perpendicular){
    report << "sections are perpendicular " << endl;
    if(dimsecper > 0){
      report << dimsecper << "  main coordinates used in the construction of  sections" << endl;
    }
  }else
    report << "sections probably not perpendicular " << endl;
  const int buflen=80;
  char buffer[buflen];
  report << "coordinates =" << coordtype  << "   " <<  osa_coordsname(buffer,buflen,coordtype) << endl;

  if(dimdiag > 0)
    report << "the number of coordinates used in diagonalization=" <<  dimdiag << endl;
  else
    report << "no info about the number of coordinates in the diagonalization " << endl;

  if(!secdataonly){
    if(period > 0){
      report << "period=" << period << endl;
    }else{
      report << "no info about the period " << endl;
    }
    if(strategy!=0){
      report << "strategy for generation of h-sets =" <<
            namehsetstrategy(strategy,buffer,buflen) << endl;
    }else{
      report << "no info about the strategy " << endl;
    }
  }
}

void chnsecparType::writebin(FILE *f) const
{
  fwrite(&sec_perpendicular,sizeof(sec_perpendicular),1,f);
  fwrite(&dimsecper,sizeof(dimsecper),1,f);
  fwrite(&coordtype,sizeof(coordtype),1,f);
  fwrite(&dimdiag,sizeof(dimdiag),1,f);
  fwrite(&period,sizeof(period),1,f);
  fwrite(&strategy,sizeof(strategy),1,f);
}

void chnsecparType::readbin(FILE *f,int version)
{
  if(version <=-2){
    fread(&sec_perpendicular,sizeof(sec_perpendicular),1,f);
    fread(&dimsecper,sizeof(dimsecper),1,f);
    fread(&coordtype,sizeof(coordtype),1,f);
    fread(&dimdiag,sizeof(dimdiag),1,f);
    fread(&period,sizeof(period),1,f);
    fread(&strategy,sizeof(strategy),1,f);
  }else{
    zero();
  }
}

void chnsecparType::zero()
{
  sec_perpendicular=0;
  dimsecper=0;
  coordtype=0;
  dimdiag=0;
  period=0;
  strategy=0;
}




LinearF* createLinearF(const GeneralLinearSection& gls)
// creates (extracts) a LinearFunc from GeneralLinearSection
{
  LinearF* plf=new LinearF(gls.dim,gls.isec);
  return plf;
}


int Section::sectiontofull(const interval xcsec[],const interval Csec[][MAX_DIM+1],
              const interval rsec[],
              interval xcfull[],interval Cfull[][MAX_DIM+1],interval rfull[]) const
  // transition from the section  coordinates to  the ambient space coordinates
  // for the set   <xcsec + Csec*rsec> on section
  //    in the ambient space we seek for the following representation
  //            xcfull + Cfull*rfull
  /////  diam(xcsec)=0 and diam(xcfull)=0
  // This is a very crude function here, it should be always overloaded with
  // something better
{
  interval xsec[MAX_DIM+1];
  multAV(Csec,rsec,xsec,dimsec);
  for(int i=1;i<=dimsec;i++)
    xsec[i]+=xcsec[i];

  // xsec=xcsec+Csec*rsec
  sectiontofull(xsec,xcfull);
  DecomposeIMulti(xcfull,rfull,dim);
  identity(Cfull,dim);

  return 1;
}

int  Section::sectiontofull(const Set &xonsec,Set &xfull) const
    // the transition from the section  coordinates to  the ambient space coordinates
 // This is a very crude function here, it should be always overloaded with
  // something better
{
  xfull.getsettings(xonsec);
  xfull.setnewdim(dim);

  interval xsec[MAX_DIM+1];
  interval *rfull,*xcfull;
  intervalmatrix1 *C;
  rfull=new interval[xfull.sdim()+1];
  xcfull=new interval[xfull.sdim()+1];
  C=new intervalmatrix1[xfull.sdim()+1];
  identity(C,xfull.sdim());

  xonsec.intvencl(xsec);

  sectiontofull(xsec,xcfull);
  DecomposeIMulti(xcfull,rfull,xfull.sdim());
  xfull.newvalues(xfull.sdim(),xfull.snumLohner(),xcfull,C,rfull);

  delete [] C;
  delete [] xcfull;
  delete [] rfull;

  return 1;
}


void Section::intersection(const interval x[],interval x_on_level[]) const
    // on output   x \\cap \\{z| f(z) ==c \\} \subset x_on_level
    // used by Poincare maps (really needed in C1 computations)
{
  interval xsec[MAX_DIM+1];
  fulltosection(x,xsec);
  sectiontofull(xsec,x_on_level);
  if(!commonpart(x,x_on_level,x_on_level,dim)){
    cerr << "in section.intersection - an empty intersection" << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
}


// GENERAL LINEAR SECTION
// this is our old section, just a table of coefficients
//  sec[0] + \Sigma_{i=1}^{dim} x[i]*sec[i]=0
//  PLUS linear coordinates on section

GeneralLinearSection::GeneralLinearSection()
// the 'empty' constructor
{
  dim=0;
  dimsec=0;
  destroy_lincc=1;
  plincc=NULL;
}

GeneralLinearSection::GeneralLinearSection(int pdim,const double psec[],
    LinCC *cchange,int pdestroy_lincc)
{
  dim=pdim;
  dimsec=dim-1;
  for(int i=0;i<=dim;i++){
    sec[i]=psec[i];
    isec[i]=sec[i];
  }
  destroy_lincc=pdestroy_lincc;
  plincc=cchange;
}

GeneralLinearSection::GeneralLinearSection(int pdim,const interval psec[],
   LinCC *cchange,int pdestroy_lincc)
{
  dim=pdim;
  dimsec=dim-1;
  for(int i=0;i<=dim;i++){
    isec[i]=psec[i];
    sec[i]=mid(psec[i]).left_b();
  }
  destroy_lincc=pdestroy_lincc;
  plincc=cchange;
}


GeneralLinearSection::GeneralLinearSection(int pdim,int selected_coord,
             interval ivalsection)
// the section is given by x[selected_coord]+ivalsection=0
{
  dim=pdim;
  dimsec=dim-1;
  for(int i=0;i<=dim;i++){
    sec[i]=0.0;
    isec[i]=0.0;
  }
  isec[0]=ivalsection;
  sec[0]=mid(isec[0]).left_b();
  isec[selected_coord]=1.0;
  sec[selected_coord]=1.0;

  interval originshift[MAX_DIM+1];
  for(int i=1;i<=dim;i++)
    originshift[i]=0.0;

  originshift[selected_coord]=-ivalsection;

  destroy_lincc=1;
  plincc=new LinCC(dim,originshift);

  int per[MAX_DIM+1];
  for(int i=1;i<=dim;i++)
    per[i]=i;

  per[1]=selected_coord;
  for(int i=2;i<=selected_coord;i++)
    per[i]=i-1;

  permutecoordc(plincc,per);
}

GeneralLinearSection::GeneralLinearSection(LinCC *cchange,int pdestroy_lincc)
{
  dim=cchange->dim;
  dimsec=dim-1;
  for(int i=1;i<=dim;i++){
    isec[i]=cchange->intv_m[1][i];
    sec[i]=cchange->m[1][i];
  }
  interval t=0;
  for(int i=1;i<=dim;i++)
    t+=cchange->intv_m[1][i]*cchange->intv_origin_shift[i];
  isec[0]=-t;
  sec[0]=mid(isec[0]).left_b();
  destroy_lincc=pdestroy_lincc;
  plincc=cchange;
}

GeneralLinearSection::GeneralLinearSection(const GeneralLinearSection &gls)
{
  dim=gls.dim;
  dimsec=gls.dimsec;
  for(int i=0;i<=dim;i++){
    isec[i]=(gls.isec)[i];
    sec[i]=(gls.sec)[i];
  }
  destroy_lincc=0;
  if(gls.plincc != NULL){
    plincc=new LinCC(*(gls.plincc));
    destroy_lincc=1;
  }
}


int GeneralLinearSection::new_coord_on_sec(const LinCC *pcoordch)
   // <pcoordch> is dimsec-dimensional
    // introduces new coordinates on section ;  this a coordinate change to new coordinates on section;
  //  the original coordinates on section are used as the "old" coordinates in this composition
{
  if(!destroy_lincc){
    cerr << "an attempt to redefine the coordinate change in GeneralLinearSection " << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }


  // we have to extend pcoordch - to (dimsec+1) - dimensions
  // by inserting a new coordinate
  LinCC *extcoordch=insertcoord(pcoordch,1,0.0);

  // now we compose extcoordch and pcc
  LinCC *pcc=composeLinCC(extcoordch,plincc);
  delete extcoordch;
  delete plincc;
  plincc=pcc;


  return 1;
}


GeneralLinearSection::~GeneralLinearSection()
 // destructor
{
  if(destroy_lincc){
    delete plincc;
  }
}


int GeneralLinearSection::savebin(FILE *fbin,int whatsave) const
    // saves the section to a binary file
    // whatsave=0; only double's are saved
    //         =1; only intervals
    //         =2; everything
// the structure of the file:
//  whatsave,dim,dimsec,isec[0,dim] or sec[0,dim] or both(depending on whatsave),
//  *plincc
{
  fwrite(&whatsave,sizeof(int),1,fbin);
  fwrite(&dim,sizeof(int),1,fbin);
  fwrite(&dimsec,sizeof(int),1,fbin);

  // saving  sec
  if((whatsave==0) || (whatsave==2))  // saving doubles
    fwrite(sec,sizeof(double),dim+1,fbin);

  // saving isec
  if((whatsave==1) || (whatsave==2))  // saving intervals
    fwrite(isec,sizeof(interval),dim+1,fbin);

  plincc->savebin(fbin,whatsave);
  return 1;
}


int GeneralLinearSection::savebin(const char *filename,int whatsave) const
{
  FILE* fhandle;
  fhandle=fopen(filename,"wb");
  int ret=(fhandle!=NULL);
  if(ret){
    ret=savebin(fhandle,whatsave);
    fclose(fhandle);
  }else{
    cerr << "cannot open " << filename << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  return ret;
}

int GeneralLinearSection::savetext(ostream &ftext,int whatsave) const
    // saves the section to a text file
    // whatsave=0; only double's are saved
    //         =1; only intervals
    //         =2; everything
    // the structure of the file:
    //  whatsave,dim,dimsec,isec[0,dim] or sec[0,dim] or both(depending on whatsave),
    //  *plincc
{
  ftext << whatsave << "   what save " << endl;
  ftext << dim  << "  dim " << endl;
  ftext << dimsec << "  dimsec " << endl;

  // saving  sec
  if((whatsave==0) || (whatsave==2)){  // saving doubles
    ftext << " section - double: sec[0}+ \\sum sec[i]*x[i]=0 " << endl;
    savesection(ftext,sec,dim);
  }
  // saving isec
  if((whatsave==1) || (whatsave==2)){  // saving intervals
     ftext << " section - interval: sec[0}+ \\sum sec[i]*x[i]=0  " << endl;
     savesection(ftext,isec,dim);
  }
   ftext << " The coordinate change: y (new coordinate)=m(x-shift)" << endl;
  plincc->savetext(ftext,whatsave);

  return 1;
}

int GeneralLinearSection::savetext(const char *filename ,int whatsave) const
{
  ofstream ftext(filename);
  int ret=(ftext!=NULL);
  if(ret){
    savetext(ftext,whatsave);
  }
  return ret;
}

int GeneralLinearSection::readbin(FILE *fbin,bool recomp_inverse/*=false*/)
 // reads (really reinitiates) the section from the file <fbin>
  // if  recomp_inverse, then the inverse in *plincc is recomputed
   //  to make sure that computation is rigorous, can be of use, when
   // computing things first on the computer on which we don't have correct rounding
{

  if((!destroy_lincc) && plincc){
    cerr << " can't read GL-section from file - can't destroy LinCC" << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  int whatsave;
  fread(&whatsave,sizeof(int),1,fbin);
  fread(&dim,sizeof(int),1,fbin);
  fread(&dimsec,sizeof(int),1,fbin);

  if((whatsave==0) || (whatsave==2)){  // reading doubles
    fread(sec,sizeof(double),dim+1,fbin);
  }

  if((whatsave==1) || (whatsave==2)){  // reading intervals
    fread(isec,sizeof(interval),dim+1,fbin);
  }

  if(whatsave==1){
    for(int i=0;i<=dim;i++)
      sec[i]=mid(isec[i]).left_b();
  }
  if(whatsave==0){
    for(int i=0;i<=dim;i++)
      isec[i]=sec[i];
  }
  destroy_lincc=1;
  if(plincc==NULL){ // this may happen if in fact we are reading the section
                    // from the file
    plincc=new LinCC();
  }
  plincc->readbin(fbin,recomp_inverse);
  return 1;
}



int GeneralLinearSection::readbin(const char *filename,bool recomp_inverse /*=false*/)
 // reads (really reinitiates) the section from the file <filename>
  // if  recomp_inverse, then the inverse in *plincc is recomputed
   //  to make sure that computation is rigorous, can be of use, when
   // computing things first on the computer on which we don't have correct rounding
{
  FILE* fhandle;
  fhandle=fopen(filename,"rb");
  int ret=(fhandle!=NULL);
  if(ret){
    ret=readbin(fhandle,recomp_inverse);
    fclose(fhandle);
  }else{
    cerr << "cannot open " << filename << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  return ret;
}

int GeneralLinearSection::readtext(istream &ftext)
 // reads (really reinitiates) the section from the file ftext
{

  if((!destroy_lincc) && plincc){
    cerr << " can't read GL-section from file - can't destroy LinCC" << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  char trash[240];

  int whatsave;
  ftext >> whatsave;
  ftext.getline(trash,240); // to eat endl
  ftext >> dim;
  ftext.getline(trash,240); // to eat endl
  ftext >> dimsec;
  ftext.getline(trash,240); // to eat endl

  if((whatsave==0) || (whatsave==2)){  // reading doubles
    ftext.getline(trash,240); //  to eat "section - double "
    readsection(ftext,sec,dim);
  }

  if((whatsave==1) || (whatsave==2)){  // reading intervals
    ftext.getline(trash,240); //  to eat "section - intervals "
    readsection(ftext,isec,dim);
  }
  ftext.getline(trash,240); //  to eat "The coordinate change "

  if(whatsave==1){
    for(int i=0;i<=dim;i++)
      sec[i]=mid(isec[i]).left_b();
  }
  if(whatsave==0){
    for(int i=0;i<=dim;i++)
      isec[i]=sec[i];
  }
  destroy_lincc=1;
  if(plincc==NULL){ // this may happen if in fact we are reading the section
                    // from the file
    plincc=new LinCC();
  }
  plincc->readtext(ftext);
  return 1;
}

int GeneralLinearSection::readtext(const char *filename)
 // reads (really reinitiates) the section from the file <filename>
{
  ifstream ftext(filename);
  int ret=(ftext!=NULL);
  if(ret){
    readtext(ftext);
  }
  return ret;
}



double GeneralLinearSection::eval(const double x[]) const
{
  double retval=sec[0];
  for(int i=1;i<=dim;i++)
    retval+=sec[i]*x[i];
  return(retval);
}

interval GeneralLinearSection::eval(const interval x[]) const
{
  interval retval=sec[0];
  for(int i=1;i<=dim;i++)
    retval+=isec[i]*x[i];
  return(retval);
}



interval GeneralLinearSection::eval(const Set &x) const
{
  interval res;
  if(x.snumLohner()!=LOHNER1){
    res=eval(x.xc);
    //  QR-part
    res+=smarteval_vAv(isec,x.B,x.r,dim);
    if(x.doubletonRep())
      res+=smarteval_vAv(isec,x.C,x.r0,dim);
  }else{
    interval *xencl;
    xencl=new interval[dim+1];
    x.intvencl(xencl);
    res=eval(xencl);
    delete [] xencl;
  }
  return res;
}


void  GeneralLinearSection::gradient(const double /*x*/[],double grad[]) const
{
  copyvector(sec,grad,dim);
}

void  GeneralLinearSection::gradient(const interval /*x*/[],interval grad[])  const
{
  copyvector(sec,grad,dim);
}


int GeneralLinearSection::fulltosection(const double x[],double xsec[]) const
// transition from the ambient space to section for points
//                dim (x) = dim; dim(xsec)=dimsec
//    on output <xsec> is <x> in section coordinates
{
  double *xcoords;
  xcoords=new double[plincc->dim+1];
  plincc->coordtran(x,xcoords); // we can't use xsec as the second argument
                    //  because the dimensions of x and xsec may differ
  // we have to drop the first coordinate
  for(int i=1;i<=dimsec;i++)
    xsec[i]=xcoords[i+1];

  delete [] xcoords;
  return 1;
}

int GeneralLinearSection::fulltosection(const interval x[],interval xsec[]) const
    // transition from the ambient space to section for points
    //                dim (x) = dim; dim(xsec)=dimsec
    //    on output <xsec> is <x> in section coordinates
{
  interval *xcoords;
  xcoords=new interval[plincc->dim+1];
  plincc->coordtran(x,xcoords);    // we can't use xsec as the second argument
                    //  because the dimensions of x and xsec may differ

  // we have to drop the first coordinate
  for(int i=1;i<=dimsec;i++)
    xsec[i]=xcoords[i+1];

  delete [] xcoords;
  return 1;
}


int GeneralLinearSection::vectfulltosection(const double x[],const double vect[],
                 double xsec[],double vectsec[]) const
    // transition from the ambient space to section for points and tangent vectors
    //                dim (x)=dim(vect) = dim; dim(xsec)=dim(vectsec)=dimsec
    //    on output <xsec> is <x> in section coordinates
    //              <vectsec> is a projection of <vect> onto section in <x>
    //               expressed in section coordinates
{
  fulltosection(x,xsec);

  double *vcs;
  vcs=new double[plincc->dim+1];

  plincc->vectcoordtran(vect,vcs);
  for(int i=1;i<=dimsec;i++)
    vectsec[i]=vcs[i+1];

  delete [] vcs;

  return 1;
}


int GeneralLinearSection::vectfulltosection(const interval x[],const interval vect[],
                 interval xsec[],interval vectsec[]) const
    // transition from the ambient space to section for points and tangent vectors
    //                dim (x)=dim(vect) = dim; dim(xsec)=dim(vectsec)=dimsec
    //    on output <xsec> is <x> in section coordinates
    //              <vectsec> is a projection of <vect> onto section in <x>
    //               expressed in section coordinates
{
  fulltosection(x,xsec);

  interval *vcs;
  vcs=new interval[plincc->dim+1];
  plincc->vectcoordtran(vect,vectsec);
  for(int i=1;i<=dimsec;i++)
    vectsec[i]=vcs[i+1];

  delete [] vcs;
  return 1;
}




int  GeneralLinearSection::sectiontofull(const double xsec[],double x[]) const
  // transition from the section  coordinates to  the ambient space coordinates
  //                for points dim (x) = dim; dim(xsec)=dimsec
  //    on output <x> is <xsec> in the ambient space coordinates
{
   x[1]=0.0;
   for(int i=2;i<=dim;i++)
     x[i]=xsec[i-1];
   plincc->coordinvtran(x,x);
   return 1;
}


int  GeneralLinearSection::sectiontofull(const interval xsec[],interval x[]) const
  // transition from the section  coordinates to  the ambient space coordinates
  //                for points dim (x) = dim; dim(xsec)=dimsec
  //    on output <x> is <xsec> in the ambient space coordinates
{
   x[1]=0.0;
   for(int i=2;i<=dim;i++)
     x[i]=xsec[i-1];
   plincc->coordinvtran(x,x);
   return 1;
}


int  GeneralLinearSection::vectsectiontofull(const  double xsec[],const double vectsec[],
                 double x[],double vect[])  const
  // transition from the section  coordinates to  the ambient space coordinates
  //     for points and tangent vectors
  //       dim (x) =dim(vect)= dim; dim(xsec)=dim(vectsec)=dimsec
  //    on output <x> is <xsec> in the ambient space coordinates
  //      and <vect> is <vectsec> in the ambient space coordinates
{
  sectiontofull(xsec,x);
  vect[1]=0.0;
  for(int i=2;i<=dim;i++)
    vect[i]=vectsec[i-1];
  plincc->vectcoordinvtran(vect,vect);

  return 1;
}

int  GeneralLinearSection::vectsectiontofull(const interval xsec[],const interval vectsec[],
                 interval x[],interval vect[]) const
  // transition from the section  coordinates to  the ambient space coordinates
  //     for points and tangent vectors
  //       dim (x) =dim(vect)= dim; dim(xsec)=dim(vectsec)=dimsec
  //    on output <x> is <xsec> in the ambient space coordinates
  //      and <vect> is <vectsec> in the ambient space coordinates
{
  sectiontofull(xsec,x);
  vect[1]=0.0;
  for(int i=2;i<=dim;i++)
    vect[i]=vectsec[i-1];
  plincc->vectcoordinvtran(vect,vect);

  return 1;
}


int GeneralLinearSection::sectiontofull(const interval xcsec[],
              const interval Csec[][MAX_DIM+1],const interval rsec[],
              interval xcfull[],interval Cfull[][MAX_DIM+1], interval rfull[]) const
  // transition from the section  coordinates to  the ambient space coordinates
  // for the set   <xcsec + Csec*rsec> on section
  //    in the ambient space we seek for the following representation
  //            xcfull + Cfull*rfull
{
  // coordinates on section are "new coordinates" given by <plincc>
  //  y[2],y[3],...,y[dim]
  //  hence we have to do the following:
  //  1) insert '0' on first coordinate to make the set xcsec+CSec*rsec
  //        dim- dimensional
  //  2) apply the inverse transformation in <plincc>


//  interval ycsec[MAX_DIM+1];
  interval *ycsec;
  ycsec=new interval[dim+1];

  intervalmatrix1 *y_Csec;
  y_Csec=new intervalmatrix1[dim+1];

  insertcoord(xcsec,1,ycsec,0.0,dim);
  insertcoord(rsec,1,rfull,0.0,dim);
  insertcoord(Csec,1,y_Csec,dim);

  // now we have to apply inverse transformation in <plincc> to
  //  ycsec+y_Csec*rfull

  // the inverse transformation is given by :
  //  x=mInv*y + origin_shift
  //  hence: ycsec+y_Csec*rfull -> mInv*ycsec + mInv*yCsec*rfull + origin_shift=
  //                           x(ycsec) + (mInv*yCsec)*yrsec
  plincc->coordinvtran(ycsec,xcfull);  // xcfull=mInv*ycsec + origin_shift
  ArrArr(plincc->intv_mInv,y_Csec,Cfull,dim);

  delete [] y_Csec;
  delete [] ycsec;
  return 1;
}


int GeneralLinearSection::sectiontofull(const Set & xonsec, Set &xfull) const
  // the transition from the section  coordinates to  the ambient space coordinates
{
  // coordinates on section are "new coordinates" given by <plincc>
  //  y[2],y[3],...,y[dim]
  //  hence we have to do the following:
  //  1) insert '0' on first coordinate to make the x-set
  //        dim- dimensional
  //  2) apply the inverse transformation in <plincc>

  Set *xt;
  xt=new Set;
  Setinsertcoord(xonsec,1,0.0,*xt);
  int inversecc=1;
  transformSet(*xt,plincc,inversecc,xfull);
  xfull.set_B_identity(); // the term B*r is so small, that the error from discretization
          // will be considerably larger, hence using a nontrival B will introduce
          // an unnecessary wrapping
  delete xt;


  return 1;
}





int GeneralLinearSection::integritytest() const
 // test if the section is given by y[1]=0 in the new coordinates
 //   (given by *plincc)
{
// the conditions to check are: eval(intv_origin_shift)=0
//                   and scalar product of isec[1...dim] with
//                       2,3,..-th column of the transition matrix from
//                      new to old coordinates must be zero

  int retval;
  interval sec_on_p=eval(plincc->intv_origin_shift);
  interval v[MAX_DIM+1];
  retval=sec_on_p.contains(0.0);
  int colnum=2; //
  while(retval && (colnum<=dim)){
    for(int i=1;i<=dim;i++)
      v[i]=plincc->intv_mInv[i][colnum];

    interval scpr=scprod(isec,v,dim);
    retval=scpr.contains(0.0);
    colnum++;
  }
  return(retval);
}

GeneralLinearSection* createPerpendicularSection(int pdim,
           const interval x[],const interval vect[])
      // the creation of the section passing thru the point <x>
      //  and perpendicular to the vector <vect>
      //  the coordinate system is given as follows:
      //    - 1-st coordinate - the direction of <vect>
      //  the coordinates on section are obtained by the orthogonalization
      //   of the following set of vectors:  <vect> \cup \{e_1,e_2, \dots,
      //    we drop the one for which  |vect[i]| is maximal }
{
  GeneralLinearSection* pgns=NULL;

  interval lvect[MAX_DIM+1];
  for(int i=1;i<=pdim;i++)
    lvect[i]=mid(vect[i]);

  // looking for the largest coordinate in vect
  int maxi=1;
  double maxc=abs(lvect[maxi]).right_b();
  for(int i=2;i<=pdim;i++){
    if(abs(lvect[i]).right_b() > maxc){
      maxi=i;
      maxc=abs(lvect[maxi]).right_b();
    }
  }

//  interval m_vect[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *m_vect;
  m_vect=new intervalmatrix1[pdim+1];
  zero(m_vect,pdim);
  copyvector(lvect,m_vect[1],pdim);
  for(int i=2;i<=pdim;i++){
    if(i<=maxi)
      m_vect[i][i-1]=1.0;
    else
      m_vect[i][i]=1.0;
  }

  int ok=orthogonalize(m_vect,pdim,pdim);
  if(!ok){
    cerr << "orthogonalize failed in the creation of perpendicular section" << endl;
    #if defined(WIN95) && defined(KRAK)
      waitBt();
    #endif
  }
  if(ok){
    interval psec[MAX_DIM+1];
    copyvector(m_vect[1],psec,pdim);
    psec[0]=-scprod(psec,x,pdim);

 //   interval invm[MAX_DIM+1][MAX_DIM+1];
    intervalmatrix1 *invm;
    invm=new intervalmatrix1[pdim+1];
    transpose(m_vect,invm,pdim);


    LinCC* plincc=new LinCC(pdim,x,m_vect,invm);

    int pdestroy_lincc=1;
    pgns=new GeneralLinearSection(pdim,psec,plincc,pdestroy_lincc);
    delete [] invm;
  }
  delete [] m_vect;
  return pgns;
}


GeneralLinearSection* createPerpendicularSection(int pdim,
          const interval x[],Ode &ode)
    // the creation of the section  passing through the point <x>
      //  and perpendicular to the vector <ode.vfield(x)>
      //  the coordinate system is given as follows:
      //    - 1-st coordinate - the direction of <vect>
         //  the coordinates on section are obtained by the orthogonalization
      //   of the follwoing set of vectors:  <ode.vfield(x)> \cup \{e_1,e_2, \dots,
      //    we drop the one for which  |ode.vfield(x)[i]| is maximal }
      // when pdim < ode.dim - this is a section for the <pdim>-dimensional
      // Galerkin projection
      // On input: dim(x)=ode.dim
{
  interval vect[MAX_DIM+1];
  ode.field(x,vect);
  return  createPerpendicularSection(pdim,x,vect);
}

GeneralLinearSection* createPerpendicularSection(int pdim,
            const double x[],Ode &ode)
  // the creation of the section passing through the point <x>
      //  and perpendicular to the vector <ode.vfield(x)>
      //  the coordinate system is given as follows:
      //    - 1-st coordinate - the direction of <vect>
         //  the coordinates on section are obtained by the orthogonalization
      //   of the following set of vectors:  <ode.vfield(x)> \cup \{e_1,e_2, \dots,
      //    we drop the one for wich  |ode.vfield(x)[i]| is maximal }
       // when pdim < ode.dim - this is a section for the <pdim>-dimensional
      // Galerkin projection
      // On input: dim(x)=ode.dim

{
  interval vect[MAX_DIM+1],ix[MAX_DIM+1];
  tointerval(x,ix,ode.dim);
  ode.field(ix,vect);
  return  createPerpendicularSection(pdim,ix,vect);
}



GeneralLinearSection* GLSincreasedim(const GeneralLinearSection *gls,int newdim)
// this function imbeds the section  gls into higher dimensional phase space
//  nothing  changes - the section form is the same and the coordinate frame
// on section is an identity on new variables
//  <newdim> - is the dimension of the ambient space
// *gls - remains unchanged
{
  LinCC *cc=LinCCincreasedim(gls->plincc,newdim);
  GeneralLinearSection *secn;
  int destroycc=1;
  secn=new GeneralLinearSection(cc,destroycc);
  return secn;
}


GeneralLinearSection* gal_createPerpendicularSection(int galdim,
            const double x[],Ode &odeeq)
 // the creation of the section  in dimension <odeeq.dim>
 //  which is buildt from the perpendicular section for the <galdim>-dimensional
 //  projection of the vector field at x (but not its galdim-projection)
 // On input: dim(x)=ode.dim
{
  GeneralLinearSection *psec=createPerpendicularSection(galdim,x,odeeq);
  if(galdim < odeeq.dim){
    GeneralLinearSection* nsec=GLSincreasedim(psec,odeeq.dim);
    delete psec;
    psec=nsec;
  }
  return psec;
}

void transOnSec_to_full(GeneralLinearSection *psection,int other_coord_on_sec,
                LinCC* pcoordch, Set &xonsec, Set &xfull)
// this function transforms the set <xonsec> to the full space and places
// the result in <xfull>;
// if <other_coord_on_sec> then <xonsec> (on input) is expressed in coordinate frame <pcoordch>
{
  Set *tonsec;
  tonsec=new Set(xonsec);

  if(other_coord_on_sec){
    int inverseCC=1;
    transformSet(xonsec,pcoordch,inverseCC,*tonsec);
  }
  // now our set is expressed in section coordinates


  psection->sectiontofull(*tonsec,xfull);
  delete tonsec;
}


int GLS_thesame(const GeneralLinearSection *psec1,const GeneralLinearSection* psec2,int
  whattocheck)
// 1 - is returned when the sections (together with the coordinate changes) are the same
// 0 - otherwise
//   if whattocheck==0 - we check doubles (also in coordinate change)
//                1 - we check intervals
//                2 - we check both doubles and intervals
{
  if((whattocheck>2) || (whattocheck<0)) whattocheck=2;

  int thesame=1;
  thesame=thesame && (psec1->dim==psec2->dim);
  thesame=thesame && (psec1->dimsec==psec2->dimsec);
  if(thesame){
    if((whattocheck==0) || (whattocheck==2)){ // checking double: sec
      for(int i=0;i<=psec1->dim;i++)
        thesame=thesame && (psec1->sec[i]==psec2->sec[i]);
    }
    if((whattocheck==1) || (whattocheck==2)){  // checking intervals
      for(int i=0;i<=psec1->dim;i++)
        thesame=thesame && (psec1->isec[i]==psec2->isec[i]);
    }
  }
  if(thesame){
    if(psec1->plincc!=psec2->plincc){
      if((psec1->plincc==NULL) || (psec1->plincc==NULL)){
        thesame=0;
      }else{
        thesame=LinCC_thesame(psec1->plincc,psec2->plincc,whattocheck);
      }
    }
  }
  return thesame;
}



void from_sec_to_ambient_space(const GeneralLinearSection *psection,const interval xc_sec[],
  const interval r0_sec[],interval xfull[],int useothercoord,LinCC *coordch)
// the set xc_sec + r0_sec on the section &psection, which is eventually
//  expressed (if useothercoord) in coordinates <coordch> is transformed
// to the full space and enclosed in xfull
{
  interval x_sec[MAX_DIM+1];
  for(int i=1;i<=psection->dimsec;i++)
    x_sec[i]=xc_sec[i]+r0_sec[i];
  if(useothercoord)
    coordch->coordinvtran(x_sec,x_sec);

  psection->sectiontofull(x_sec,xfull);
}



///POINCARE MAP
FPoincareMap::FPoincareMap()
// default constructor
{
  destroy_Insection=0;
  pInsection=NULL;
  destroy_Outsection=0;
  pOutsection=NULL;
}


void FPoincareMap::setdefaultval(double ph,double ptmax,int porder,int pmethod)
  // default values
{
  param.par_double.tmax=ptmax;
  param.par_interval.tmax=ptmax;

  param.par_double.h=ph;
  param.par_interval.h=ph;

  param.par_interval.onesteppar.method=pmethod;
  param.par_interval.onesteppar.order=porder;


  param.par_double.onesteppar.orderdf=porder;
  param.par_interval.onesteppar.orderdf=porder; // the order in C1-computations for the derivative

  param.par_double.onesteppar.method=RUNGEKUTTA; // the numerical scheme for real (nonrigorous computations)
  param.par_double.onesteppar.order=4;  // the order of numerical scheme for real (nonrigorous computations)
  if(porder > 4){  //  we have runge-kutta only for order 4
    param.par_double.onesteppar.method=pmethod;
    param.par_double.onesteppar.order=porder;
  }
  param.par_double.onesteppar.doC0intersection=0;
  param.par_interval.onesteppar.doC0intersection=0;
}

int FPoincareMap::testdim()
 // we check whether various dimensions are in mutual agreement
 // used for 'standard sections'
{
  int ok=1;
  if(dim!=pode->dim-1){
    ok=0;
  }
  if(pode->dim != pInsection->dim)
    ok=0;
  if(pode->dim != pOutsection->dim)
    ok=0;

  if(!ok){
    cerr << "problem with dimensions in FPoincareMap " << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  return ok;
}



FPoincareMap::FPoincareMap(Ode *ppode,GeneralLinearSection *ppsection,int psecdir,
              const Par_Poincare &pparam,int pdestroy_section /*=0*/)
{
  dim=ppode->dim-1;
  pode=ppode;
  pInsection=ppsection;
  destroy_Insection=pdestroy_section;
  pOutsection=ppsection;
  destroy_Outsection=0;  // because both pInsection, pOutSection are pointing to
                      // the same object
  secdir=psecdir;
  param.par_double=pparam;
  if(param.par_double.onesteppar.order==4){
    param.par_double.onesteppar.method=RUNGEKUTTA;
  }
  param.par_double.onesteppar.doC0intersection=0; //this setting
                      // does not make sense for double computation
  param.par_interval=pparam;

  testdim();
}


FPoincareMap::FPoincareMap(Ode *ppode,GeneralLinearSection *ppsection,int psecdir,
              const Par_Poincare &pparaminterval,const Par_Poincare &pparamdouble,
              int pdestroy_section /*=0*/)
{
  dim=ppode->dim-1;
  pode=ppode;
  pInsection=ppsection;
  destroy_Insection=pdestroy_section;
  pOutsection=ppsection;
  destroy_Outsection=0;  // because both pInsection, pOutSection are pointing to
                      // the same object
  secdir=psecdir;
  param.par_double=pparamdouble;
  param.par_double.onesteppar.doC0intersection=0; //this setting
                      // does not make sense for double computation
  param.par_interval=pparaminterval;

  testdim();
}





FPoincareMap::FPoincareMap(Ode *ppode,GeneralLinearSection *ppsection,
        int psecdir,double ph,double ptmax,int porder,int pmethod,
        int pdestroy_section /*=0*/)
{
  dim=ppode->dim-1;
  pode=ppode;
  pInsection=ppsection;
  destroy_Insection=pdestroy_section;
  pOutsection=ppsection;
  destroy_Outsection=0;  // because both pInsection, pOutSection are pointing to
                      // the same object
  secdir=psecdir;

  setdefaultval(ph,ptmax,porder,pmethod);  // default values

  testdim();
}

FPoincareMap::FPoincareMap(Ode *ppode,GeneralLinearSection *ppInsection,
                   GeneralLinearSection *ppOutsection,int psecdir,double ph,
                   double ptmax,int porder,int pmethod,
                   int pdestroy_Insection/*=0*/,int  pdestroy_Outsection/*=0*/)
{
  dim=ppode->dim-1;
  pode=ppode;
  pInsection=ppInsection;
  destroy_Insection=pdestroy_Insection;
  pOutsection=ppOutsection;
  destroy_Outsection=pdestroy_Outsection;
  if(destroy_Insection && destroy_Outsection){
    if(pInsection==pOutsection)
      destroy_Outsection=0;
  }
  secdir=psecdir;

  setdefaultval(ph,ptmax,porder,pmethod);  // default values

  testdim();
}


FPoincareMap::FPoincareMap(Ode *ppode,GeneralLinearSection *ppInsection,
                   GeneralLinearSection *ppOutsection,int psecdir,
                   const Par_Poincare &pparam,
                   int pdestroy_Insection/*=0*/,int  pdestroy_Outsection/*=0*/)
{
  dim=ppode->dim-1;
  pode=ppode;
  pInsection=ppInsection;
  destroy_Insection=pdestroy_Insection;
  pOutsection=ppOutsection;
  destroy_Outsection=pdestroy_Outsection;
  if(destroy_Insection && destroy_Outsection){
    if(pInsection==pOutsection)
      destroy_Outsection=0;
  }
  secdir=psecdir;

  param.par_double=pparam;
  if(param.par_double.onesteppar.order==4){
    param.par_double.onesteppar.method=RUNGEKUTTA;
  }
  param.par_double.onesteppar.doC0intersection=0; //this setting
                      // does not make sense for double computation
  param.par_interval=pparam;

  testdim();
}


FPoincareMap::FPoincareMap(Ode *ppode, GeneralLinearSection *ppInsection,
                   GeneralLinearSection *ppOutsection,int psecdir,
                   const Par_Poincare &pparaminterval,const Par_Poincare &pparamdouble,
                   int pdestroy_Insection/*=0*/,int  pdestroy_Outsection/*=0*/)
{
  dim=ppode->dim-1;
  pode=ppode;
  pInsection=ppInsection;
  destroy_Insection=pdestroy_Insection;
  pOutsection=ppOutsection;
  destroy_Outsection=pdestroy_Outsection;
  if(destroy_Insection && destroy_Outsection){
    if(pInsection==pOutsection)
      destroy_Outsection=0;
  }
  secdir=psecdir;

  param.par_double=pparamdouble;
  param.par_double.onesteppar.doC0intersection=0; //this setting
                      // does not make sense for double computation
  param.par_interval=pparaminterval;

  testdim();
}




FPoincareMap::FPoincareMap(Ode *ppode,int pselcoord,interval pivalsec,
        int psecdir,double ph,double ptmax,int porder,int pmethod)
      // the section will be given by x[pselcoord]+pivalsec=0
{
  dim=ppode->dim-1;
  pode=ppode;

  pInsection=new GeneralLinearSection(dim+1,pselcoord,pivalsec);
  destroy_Insection=1;

  pOutsection=pInsection;
  destroy_Outsection=0;  // because both pInsection, pOutSection are pointing to
                      // the same object


  secdir=psecdir;

  setdefaultval(ph,ptmax,porder,pmethod);  // default values
  testdim();
}

FPoincareMap::~FPoincareMap()
{
  if(destroy_Outsection)
    destroy_Outsection=(pInsection!=pOutsection);

  if(destroy_Insection)
    delete pInsection;
  if(destroy_Outsection)
    delete pOutsection;
}


double FPoincareMap::dsettimestep(double ph)
  //  returns  previous timestep for double computation, and sets timesptep=ph
{
  double ret=param.par_double.h;
  param.par_double.h=ph;
  return ret;
}

double FPoincareMap::isettimestep(double ph)
  //  returns  previous for interval computations timestep, and sets timesptep=ph
{
  double ret=param.par_interval.h;
  param.par_interval.h=ph;
  return ret;
}


double FPoincareMap::dgettimestep()
{
  return param.par_double.h;
}

double FPoincareMap::igettimestep()
{
  return param.par_interval.h;
}


double FPoincareMap::settmax(double ptmax)
{
  double ret=param.par_double.tmax;
  param.par_double.tmax=ptmax;
  param.par_interval.tmax=ptmax;
  return ret;
}


int FPoincareMap::isetorder(int porder)
  // returns previous value of order; and set order=porder and orderdf=order
  // interval computations
{
  int ret=param.par_interval.onesteppar.order;
  param.par_interval.onesteppar.order=porder;
  param.par_interval.onesteppar.orderdf=porder;
  return(ret);
}


int FPoincareMap::igetorder()
{
  return param.par_interval.onesteppar.order;
}

int FPoincareMap::igetorderdf()
{
  return param.par_interval.onesteppar.orderdf;
}


int FPoincareMap::dsetorder(int porder)
  // returns previous value of order; and set order=porder and orderdf=order
  // double computations
{
  int ret=param.par_double.onesteppar.order;
  param.par_double.onesteppar.order=porder;
  param.par_double.onesteppar.orderdf=porder;
  return(ret);
}


int FPoincareMap::dgetorder()
{
  return param.par_double.onesteppar.order;
}

int FPoincareMap::dgetorderdf()
{
  return param.par_double.onesteppar.orderdf;
}

double FPoincareMap::gettmax()
{
  return param.par_double.tmax;
}


int FPoincareMap::dgetmethod()
{
  return param.par_double.onesteppar.method;
}

int FPoincareMap::igetmethod()
{
  return param.par_interval.onesteppar.method;
}


int FPoincareMap::map( double x[],double fx[])
 // on output fx=f(x), 1  is returned when computation sucessfull
{

  double xfull[MAX_DIM+1],fxfull[MAX_DIM+1];
  estimates estm;
  pInsection->sectiontofull(x,xfull);

  doublematrix1* dxtrash;
  dxtrash=new doublematrix1[pode->dim+1];
  double tpoinc;
  int ret=NonRigorousPoincMap(*pode,param.par_double,xfull,fxfull,&tpoinc,*pOutsection,
      secdir,&estm,dxtrash,0,0);

  delete [] dxtrash;

  if(ret){
    pOutsection->fulltosection(fxfull,fx);
  }
  return(ret);
}


int FPoincareMap::transporttosection(double x0full[],double fxsec[])
    // on output fxsec is the point on the section (in the section coordinates)
    //    which is on first intersection of the trajectory through the point
    //   <x0full> which is given in standard ambient space coordinates
    //         if <x0full> is very close to the section, then we just project
    //           it onto the section and return 1,
    //  otherwise 1 - is returned if the Poincare time for x0full < tmax
{
  double tpoinc;
  double fxfull[MAX_DIM+1];
  estimates estm;

  int ret=0;
  // we first check whether x0full is very close section
  // if yes then we simply project it on section
  if(fabs(pInsection->eval(x0full))<1e-8){
    double *vectfield, *gradsec;
    vectfield=new double[pode->dim+1];
    gradsec=new double[pode->dim+1];
    pode->field(x0full,vectfield);
    pInsection->gradient(x0full,gradsec);
    double vectf_x_grad=scprod(vectfield,gradsec,pode->dim);
    delete [] gradsec;
    gradsec=NULL;
    delete [] vectfield;
    vectfield=NULL;
    if(fabs(vectf_x_grad)>1e-6){
      if((vectf_x_grad >0) && secdir==UP)
        ret=1;

      if((vectf_x_grad <0) && secdir==DOWN)
        ret=1;

      if(ret)
        pInsection->fulltosection(x0full,fxsec); // this realizes the projection
                    // onto the section
    }
  }

  if(!ret){
    doublematrix1 *dxtrash;
    dxtrash=new doublematrix1[pode->dim+1];
    ret=NonRigorousPoincMap(*pode,param.par_double,x0full,fxfull,&tpoinc,
               *pInsection,secdir,&estm,dxtrash,0,0);
    delete [] dxtrash;
    if(ret){
      pInsection->fulltosection(fxfull,fxsec);
    }
  }
  return(ret);
}



int FPoincareMap::imapDE( interval xcenter[], interval C[][MAX_DIM+1],interval r0[],
                 interval fx[],estimates *estm)
       // on output fx=f(xcenter + C*r0)
{
  return imapDE(xcenter,C,r0,fx,0,(LinCC*) NULL,0,(LinCC *) NULL,estm);
}


int FPoincareMap::imapDE(Set &x,interval fx[],estimates *estm)
       // on output fx=f(x)
{
  return imapDE(x,fx,0,(LinCC*) NULL,0,(LinCC *) NULL,estm);
}



int FPoincareMap::imapDE(interval xcenter[], interval C[][MAX_DIM+1],interval r0[],
                 interval fx[],interval df[][MAX_DIM+1],int use_other_inputcoord,
                 LinCC* inpcoord,int use_other_outcoord,LinCC *outpcoord,
                 estimates *estm)
      // this function computes the map: usually this is the Poincare map
       //  and it makes sense sometimes to play with different coordinate
       //  systems
       //   if (use_other_inputcoord) then the input (xcenter + C*r0) is expressed
       //   using the  <inpcoord> coordinate   system
       //   if(use_other_outcoord) the the output - fx is expressed in coordinates
       //  <outpcoord>
       //    df - is the enclosure for the derivative
       //     in coordinates taking into account <use_other_inputcoord> and
       //   <use_other_outcoord>
       // this function  assumes that the coordinates on sections (both domian and
       //  range are as follows:
       // y[1]=0, (y[2],...) - where y - are the coordinates this->pIn(Out)section->plincc

{
  Set x(this->dim,xcenter,C,r0);
  return this->imapDE(x,fx,df,use_other_inputcoord,inpcoord,use_other_outcoord,outpcoord,estm);
}


int FPoincareMap::imapDE(Set &x,interval fx[],interval df[][MAX_DIM+1],int use_other_inputcoord,
                 LinCC* inpcoord,int use_other_outcoord,LinCC *outpcoord,
                 estimates *estm)
      // this function computes the map: usually this is the Poincare map
       //  and it makes sense sometimes to play with different coordinate
       //  systems
       //   if (use_other_inputcoord) then the input <x> is expressed
       //   using the  <inpcoord> coordinate   system
       //   if(use_other_outcoord) the the output - fx is expressed in coordinates
       //  <outpcoord>
       //    df - is the enclosure for the derivative
       //     in coordinates taking into account <use_other_inputcoord> and
       //   <use_other_outcoord>
       // this function  assumes that the coordinates on sections (both domian and
       //  range are as follows:
       // y[1]=0, (y[2],...) - where y - are the coordinates this->pIn(Out)section->plincc

{



  if(this->dim!=x.sdim()){
    wrongdiminfo(this->dim,x.sdim(),"FPoincareMap::imapDE");
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  Set *xfull;
  xfull=new Set;

  transOnSec_to_full(this->pInsection,use_other_inputcoord,inpcoord,x,*xfull); // transforms our set
                // on section to the ambient space coordinates


  int c1numLohner=LOHNER5;


  interval fxfull[MAX_DIM+1];


  int delpccOut=0;
  LinCC* pccOut=(this->pOutsection)->plincc;
  if(use_other_outcoord){
    // we have to extend <outpcoord> - to (dim+1) - dimensions
    // by inserting a new coordinate
    LinCC *extcoordch=insertcoord(outpcoord,1,0.0);

    // now we compose extcoordch and pccOut
    pccOut=composeLinCC(extcoordch,pccOut);

    delpccOut=1;
    delete extcoordch;
  }


  int delpccIn=0;
  LinCC *pccIn=(this->pInsection)->plincc;
  if(use_other_inputcoord){
    // we have to extend <inpcoord> - to (dim+1) - dimensions
    // by inserting a new coordinate
    LinCC *extcoordch=insertcoord(inpcoord,1,0.0);

    // now we compose extcoordch and pccIn
    pccIn=composeLinCC(extcoordch,pccIn);

    delpccIn=1;
    delete extcoordch;
  }


  intervalmatrix1 *dffull;
  dffull=new intervalmatrix1[pode->dim+1];
  int onOutputUse_pcc=1;




  int retval=General_Lo_Poincare(1,*(this->pode),param.par_interval,*xfull,*(this->pOutsection),
             secdir,estm,fxfull,dffull,onOutputUse_pcc,pccOut,c1numLohner);
      // dffull is expressed in the ambient space coordinates


  if(retval){
    // we have to do the projection on section coordinates
    dropcoord(fxfull,1,fx,dim+1);

    // df - is now in the ambient space coordinates , we have to transform it
    //      to section coordinates
    transLinearPart(dffull,*pccIn,*pccOut,dffull);
    dropcoord(dffull,1,df,dim+1);
  }

  if(delpccOut)
    delete pccOut;

  if(delpccIn)
    delete pccIn;

  delete [] dffull;



  delete xfull;
  return retval;
}


int FPoincareMap::imapDE(interval xcenter[], interval C[][MAX_DIM+1],interval r0[],
                 interval fx[],int use_other_inputcoord,LinCC* inpcoord,
                 int use_other_outcoord,LinCC *outpcoord,estimates *estm)
// the function computes f(xcenter +C*r0) and the result is placed in fx
//  when the computation terminates successfully, in this case 1 is returned
// otherwise we return 0
/// both the input values and output values are given in terms of section
//  coordinates
////// if use_other_inputcoord - then the input values are expressed in inpcoord
///    similarly if use_other_outcoord, the the output value (fx) is expressed
//          using outpcoord
///////// it should be stressed that both these linear transformations inpcoord
//              and outpcoord are are working on section, and expressed in section
///              coordinates
/////////
// this function  assumes that the coordinates on sections
// y[1]=0, (y[2],...) - where y - are the coordinates this->psection->plincc
{
  Set x(this->dim,xcenter,C,r0);
  return this->imapDE(x,fx,use_other_inputcoord,inpcoord,use_other_outcoord,outpcoord,estm);
}


int FPoincareMap::imapDE(Set &x,interval fx[],int use_other_inputcoord,
                 LinCC* inpcoord,int use_other_outcoord,LinCC *outpcoord,
                 estimates *estm)
      // this function computes the map: usually this is the Poincare map
       //  and it makes sense sometimes to play with different coordinate
       //  systems
       //   if (use_other_inputcoord) then the input <x> is expressed
       //   using the  <inpcoord> coordinate   system
       //   if(use_other_outcoord) the the output - fx is expressed in coordinates
       //  <outpcoord>
       //
       // this function  assumes that the coordinates on sections (both domian and
       //  range are as follows:
       // y[1]=0, (y[2],...) - where y - are the coordinates this->pIn(Out)section->plincc

{

  if(this->dim!=x.sdim()){
    wrongdiminfo(this->dim,x.sdim(),"FPoincareMap::imapDE");
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  Set *xfull;
  xfull=new Set;

  transOnSec_to_full(this->pInsection,use_other_inputcoord,inpcoord,x,*xfull); // transforms our set
                // on section to the ambient space coordinates


  int c1numLohner=LOHNER5;


  interval fxfull[MAX_DIM+1];


  int delpccOut=0;
  LinCC* pccOut=(this->pOutsection)->plincc;
  if(use_other_outcoord){
    // we have to extend <outpcoord> - to (dim+1) - dimensions
    // by inserting a new coordinate
    LinCC *extcoordch=insertcoord(outpcoord,1,0.0);

    // now we compose extcoordch and pccOut
    pccOut=composeLinCC(extcoordch,pccOut);

    delpccOut=1;
    delete extcoordch;
  }





  intervalmatrix1 *dffull;
  dffull=new intervalmatrix1[pode->dim+1];
  int onOutputUse_pcc=1;




  int retval=General_Lo_Poincare(0,*(this->pode),param.par_interval,*xfull,*(this->pOutsection),
             secdir,estm,fxfull,dffull,onOutputUse_pcc,pccOut,c1numLohner);
      // dffull is expressed in the ambient space coordinates


  if(retval){
    // we have to do the projection on section coordinates
    dropcoord(fxfull,1,fx,dim+1);
  }

  if(delpccOut)
    delete pccOut;



  delete [] dffull;




  delete xfull;
  return retval;
}


int FPoincareMap::imapDE(interval xcenter[],interval C[][MAX_DIM+1],interval r0[],
             interval fx[],interval df[][MAX_DIM+1],estimates *estm)
{
  return imapDE(xcenter,C,r0,fx,df,0,(LinCC *)NULL,0,(LinCC *)NULL,estm);
}

int FPoincareMap::imapDE(Set &x,interval fx[],interval df[][MAX_DIM+1],estimates *estm)
{
  return imapDE(x,fx,df,0,(LinCC *)NULL,0,(LinCC *)NULL,estm);
}




int FPoincareMap::map(double x[],double fx[],double df[][MAX_DIM+1])
{
  double xfull[MAX_DIM+1],fxfull[MAX_DIM+1];
  double dffull[MAX_DIM+1][MAX_DIM+1];
  estimates estm;
  pInsection->sectiontofull(x,xfull);
  double tpoinc;
  int ret=NonRigorousPoincMap(*pode,param.par_double,xfull,fxfull,
             &tpoinc,*pOutsection,secdir,&estm,dffull,1,pode->dim);

  if(ret){
    pOutsection->fulltosection(fxfull,fx);
    transLinearPart(dffull,*(pInsection->plincc),*(pOutsection->plincc),dffull);;
    dropcoord(dffull,1,df,pode->dim);
  }
  return(ret);
}








FPoincareMap_param FPoincareMap::get_param()
{
  return param;
}

FPoincareMap_param FPoincareMap::set_param(const FPoincareMap_param &lparam)
{
  FPoincareMap_param old_param;
  old_param=param;

  if(lparam.par_double.onesteppar.order>0)
    param.par_double.onesteppar.order=lparam.par_double.onesteppar.order;

  if(lparam.par_double.onesteppar.method)
    param.par_double.onesteppar.method=lparam.par_double.onesteppar.method;

  if(lparam.par_double.tmax >0)
    param.par_double.tmax=lparam.par_double.tmax;

  if(lparam.par_double.h >0)
    param.par_double.h=lparam.par_double.h;

  if(lparam.par_double.onesteppar.orderdf >0)
    param.par_double.onesteppar.orderdf=lparam.par_double.onesteppar.orderdf;


  if(lparam.par_interval.onesteppar.order>0)
    param.par_interval.onesteppar.order=lparam.par_interval.onesteppar.order;

  if(lparam.par_interval.onesteppar.method)
    param.par_interval.onesteppar.method=lparam.par_interval.onesteppar.method;

  if(lparam.par_interval.tmax >0)
    param.par_interval.tmax=lparam.par_interval.tmax;

  if(lparam.par_interval.h >0)
    param.par_interval.h=lparam.par_interval.h;

  if(lparam.par_interval.onesteppar.orderdf >0)
    param.par_interval.onesteppar.orderdf=lparam.par_interval.onesteppar.orderdf;

  param.par_interval.onesteppar.doC0intersection=lparam.par_interval.onesteppar.doC0intersection;


  return old_param;
}




// END - POINCARE MAP


int PoincMap_map_divide(FPoincareMap& pmap,int divisiontable[],
       int divdim, interval ixb[],interval C[][MAX_DIM+1],interval r0[],
       interval fx[],int useothercoord,LinCC *coordch,estimates *estm)
// the function computes pmap.imapD(ixb + C*r0)
//  for this computation   the set ixb + C*r0  is first divided
//     into \Pi_{1<= i <=divdim}  divisiontable[i] parts
 //  the i-th direction in <r0> is divided into divisiontabl[i] equal parts
{
  int ok=1;
  int the_end=0;

  interval pr0[MAX_DIM+1];
  interval pixb[MAX_DIM+1];
  interval pfx[MAX_DIM+1], pfy[MAX_DIM+1];

  int seq[MAX_DIM+1];
  for(int i=1;i<=divdim;i++)
    seq[i]=0;
  int firsttime=1;
  while(ok && !the_end){
    // generate new inital condition; we have to divide
    //  the set ixb + C*r0 into \Pi_{1<= i <=divdim}  divisiontable[i] parts
    //  the i-th direction in <r0> is divided into divisiontabl[i] equal parts
    divide_set(pmap.dim,seq,divisiontable,divdim,ixb,C,r0,pixb,pr0);

    ok=pmap.imapDE(pixb,C,pr0,pfx,useothercoord,coordch,useothercoord,coordch,
             estm);
    if(ok){

      // update the results:
      if(firsttime){
        copyvector(pfx,fx,pmap.dim);
      } else{
        MultiIhull(fx,pfx,fx,pmap.dim);
      }
      the_end=!gennewseqi(seq,divisiontable,1,divdim);
    }

    firsttime=0;
  }
  return ok;
}





int iterate_C1nonrigpoinc(int itno,Ode &ode,Par_Poincare param,double x[],
         double xn[],double *tpoinc,const GeneralLinearSection &section,int secdir,
         estimates *estm,double df[][MAX_DIM+1])
//  nonrigorously computes <itno>-th iterate of Poincare map for <ode>
//  Poincare map is defined by
//  section.eval(x)=0
//  if secdir==UP - then <section.eval()> is
//      increasing on section
//  x - starting point , dim(x) = ode.dim - is expressed in ambient space coordinates
//  xn=P^{itno}(x) - on output , dim(xn) = ode.dim - is expressed in ambient space coordinates
//  df=dP^{itno}(x) - in the section coordinates
{

 doublematrix1 *dfi;
 dfi=new doublematrix1[ode.dim+1];
 double ltpoinc;
 *tpoinc=0;
 double y[MAX_DIM+1];
 for(int i=1;i<=ode.dim ;i++){
   xn[i]=x[i];
 }
 int ok=1;
 int iter=1;

 identity(df,ode.dim-1);

 while((iter <= itno) && ok ){
   for(int i=1;i<=ode.dim ;i++){
     y[i]=xn[i];
   }
   ok= NonRigorousPoincMap(ode,param,y,xn,&ltpoinc,section,
                      secdir,estm,dfi,1,ode.dim);
   if(ok){
     *tpoinc+=ltpoinc;
     transLinearPart(dfi,*(section.plincc),*(section.plincc),dfi);;
     dropcoord(dfi,1,dfi,ode.dim);
     ArrArr(dfi,df,df,ode.dim-1);
   }
   iter++;
 }

 delete [] dfi;
 return(ok);
}



bool buildcoordchange(int coordtype,const double Dp[][MAX_DIM+1],const interval ixp[],
             int dim,LinCC **coordch,int dimdiag/* =0 */)
//  we use matrix <Dp> to build a coordinate change:
// On output : **coordch - is the coordinate change is build as follows:
// 1) the origin is shifted to <ixp>
// 2) linear part  = identity, when <coordtype>==standardcoord
//                 = orthogonal change to orthogonalized eigenvectors basis
//                      when, <coordtype>==orthogonalizedeigenvectors
//                 = change to the eigenvector basis, when
//                     <coordtype>==eigenvectors
//                 = change to mixed basis - the dominant eigenvalues
//                      and orthogonalized remaining directions
//                     <coordtype>==partialeigenbasis
//                 =  change to mixed basis - subspaces of eigenvectors
//                     which are close to parallel are  orthogonalized
//                      <coordtype>==eigen_subspaces_ortogonalized
//  dimdiag - tells how many first coordinates are used in diagonalization
//          attempt
//   if dimdiag==0 then it is ignored - all coordinates are used
{

  if(*coordch!=NULL)
    delete *coordch;

  *coordch=NULL;

  if(dimdiag==0) dimdiag=dim;

  int dimpartialdiag=dimdiag;


  bool ok=true;

  double eigenval[MAX_DIM+1][3];
//  interval p[MAX_DIM+1][MAX_DIM+1];
//  interval pinv[MAX_DIM+1][MAX_DIM+1];
  intervalmatrix1 *p,*pinv;
  p=new intervalmatrix1[dim+1];     // columns are eigenvectors of Dp
  pinv=new intervalmatrix1[dim+1];

  //  we diagonalize
  int maxiterno=10000;
  double bias=1e-8;
  double relbias=1e-4;


  if(coordtype==eigenvectors) {
    if(!iqrdiagonalize(Dp,p,pinv,eigenval,dimdiag,maxiterno,bias,relbias)){
      // if successful then,   pinv * Dp * p - is close to diagonal
      //  (complex eigenvalues give rise to 2x2 cell)
      ok=false;


      cerr << "buildcoordchange - the diagonalization problem, bias="
    << bias << " iter=" << maxiterno << endl;

      // trying again
      cout << "repeated diagonalization attempt with truncation via read/write operation " << endl;
      char Dpfname[80];
      strcpy(Dpfname,catwork);
      strcat(Dpfname,"dptemp.txt");
      saveArray(Dpfname,Dp,dim);

      doublematrix1* Dptrunc=new doublematrix1[dim+1];


      readmatrix(Dpfname,Dptrunc,dim);

      if(!iqrdiagonalize(Dptrunc,p,pinv,eigenval,dimdiag,maxiterno,bias,relbias)){
        cerr << "buildcoordchange - the diagonalization problem, bias="
          << bias << " iter=" << maxiterno << endl;
      }else ok=true;
      delete [] Dptrunc;
    }
  }
  if(coordtype==eigen_subspaces_ortogonalized){
    // first we try to diagonalize - using double's; only after this
    // we use the 'double' pd - to cluster-orthogonalize it
    doublematrix1 *pd =new doublematrix1[dimdiag+1];
    doublematrix1 *pdinv =new doublematrix1[dimdiag+1];

    if(!qrdiagonalize(Dp,pd,pdinv,eigenval,dimdiag,maxiterno,bias,relbias)){
       // if successful then,   pinv * Dp * p - is close to diagonal
      //  (complex eigenvalues give rise to 2x2 cell)
      ok=false;

      cerr << "buildcoordchange - the diagonalization problem, bias="
               << bias << " iter=" << maxiterno << endl;

      // trying again
      cout << "repeated diagonalization attempt with truncation via read/write operation " << endl;
      char Dpfname[80];
      strcpy(Dpfname,catwork);
      strcat(Dpfname,"dptemp.txt");
      saveArray(Dpfname,Dp,dimdiag);

      doublematrix1* Dptrunc=new doublematrix1[dimdiag+1];


      readmatrix(Dpfname,Dptrunc,dimdiag);

      if(!qrdiagonalize(Dptrunc,pd,pdinv,eigenval,dimdiag,maxiterno,bias,relbias)){
        cerr << "buildcoordchange - the diagonalization problem, bias="
          << bias << " iter=" << maxiterno << endl;
      }else ok=true;

      delete [] Dptrunc;

    }
    if(ok){
     if(EigenVectOrthogonalizeInClusters(dimdiag,pd,eigenval,p)){
       ok=inverse(p,pinv,dimdiag);
       if(!ok){
        cerr << "buildcoordchange - inverse matrix failed"<< endl;
       }
     }else{
       ok=false;
       cerr << "buildcoordchange - EigenVectOrthogonalizeInClusters failed"<< endl;
     }
    }
    delete [] pd;
    delete [] pdinv;
  }
  if(coordtype==orthogonalizedeigenvectors){
    if(!qrdecomposition(Dp,p,dimdiag,maxiterno,bias,relbias)){
        //   p^t Dp p - is almost upper triangular (with eventually 2x2 blocks
       //       on the diagonal), p - is orthogonal
      ok=false;
      cerr << "buildcoordchange - a problem with the qrdecomposition " << endl;
    }else{
      transpose(p,pinv,dimdiag);
    }
  }
  if(coordtype==partialeigenbasis){
    if(!ipartialqrdiagonalize(Dp,p,pinv,eigenval,dim,dimpartialdiag,maxiterno,bias,relbias)){
      cerr << "buildcoordchange - the diagonalization problem, bias="
    << bias << " iter=" << maxiterno << endl;
      ok=false;
    }
  }
  if(ok){
    if((coordtype==orthogonalizedeigenvectors) || (coordtype==eigenvectors)
      || (coordtype==eigen_subspaces_ortogonalized)){
      // because in this case we have dom(p)=dim(pinv)=dimdiag
      if(dimdiag < dim){
        for(int i=dimdiag+1;i<=dim;i++)
          for(int j=1;j<=dim;j++){
            if(i==j) {
              p[i][j]=1.0;
              pinv[i][j]=1.0;
            } else{
              p[i][j]=0.0;  p[j][i]=0.0;
              pinv[i][j]=0.0; pinv[j][i]=0.0;
            }
          }
      }
    }


    if(coordtype==standardcoord){
      *coordch=new LinCC(dim,ixp);   // simply the shift of the origin to xp
    }else{
      *coordch=new LinCC(dim,ixp,pinv,p); // the order is ok.
                         // pinv - is the transition from standard coordinates
                         //    to the eigenvectors basis
    }
  }
  if(!ok){
   #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
    #endif
  }

  delete [] pinv;
  delete [] p;

  return ok;
}



int generate_sections_traj(Ode &ode,Par_Poincare &param,const double xs[],
       GeneralLinearSection* psec[],int secnum,int secdir,
                       double tabxs[][MAX_DIM+1])
// the function computes  the trajectory  of xs;
//   dim(xs)=ode.dim
//  the trajectory is defined by transitions between psec[i-1]->psec[i];
//   for i=1,...,secnum;
///On output the points form trajectory  are placed
//        in tabxs[i],  with tabxs[0]=xs
{
  int ok=1;

  doublematrix1 *dftrash;
  dftrash=new doublematrix1[ode.dim+1];
  double tpoinc;

  estimates estm;

  copyvector(xs,tabxs[0],ode.dim);
  int i=1;
  while(ok && (i<=secnum)){
    ok=NonRigorousPoincMap(ode,param,tabxs[i-1],tabxs[i],&tpoinc,*(psec[i]),
                   secdir,&estm,dftrash);
    i++;
  }
  delete [] dftrash;
  return ok;
}




int  create_perpen_sections_along_traj(Ode &ode,Par_Poincare &param,const double xs[],
       GeneralLinearSection* psec[],int secnum,int secdir,int dimpersec)
// the function creates a sequence of perpendicular sections along the trajectory
//  of xs;
//  the trajectory is defined by transitions between psec[i]->psec[i+1];
//  dimpersec - is the dimension of Galerkin projection for which we construct
//            the sections
////
// On output: psec[0,\dots,secnum-1] -  are pointing to  the perpendicular
//     sections; the memory occupied by psec[i] on input is released
{
  if(dimpersec==0){
    dimpersec=ode.dim;
  }
  int ok=1;

  double *xn,*x;
  x=new double[ode.dim+1];
  xn=new double[ode.dim+1];

  doublematrix1 *dftrash;
  dftrash=new doublematrix1[ode.dim+1];
  double tpoinc;
  estimates estm;

  delete psec[0];
  psec[0]=gal_createPerpendicularSection(dimpersec,xs,ode);



  copyvector(xs,xn,ode.dim);
  int i=1;
  while(ok && (i<=secnum-1)){
    copyvector(xn,x,ode.dim);
    ok=NonRigorousPoincMap(ode,param ,x,xn,&tpoinc,*(psec[i]),secdir,&estm,dftrash);
    if(ok ){
      delete psec[i];
      psec[i]=gal_createPerpendicularSection(dimpersec,xn,ode);
    }
    i++;
  }
  delete [] dftrash;
  delete [] x;
  delete [] xn;
  return ok;
}



int createFromFileSections(const char filename[],bool recompute_inverse,int *odedim,
   chnsecparType &chnpar,GLS_Pointer_Type**  psec,int *secnum)
// (*psec) - are indexed from 0 to *secnum-1
// chnpar - is used just for reporting
// recompute_inverse - tells whether  to recompute the inverse corddinate transformation
//     on sections,     to make sure that computation is rigorous, can be of use, when
// computing things first on the computer on which we don't have correct rounding
// this is 'the inverse function'  to <save_generated_sections>
{
  FILE* fbin;
  fbin=fopen(filename,"rb");
  int ret=(fbin!=NULL);

  chnpar.zero();
  if(ret){
    int version=0;
    fread(odedim,sizeof(int),1,fbin);
    if(*odedim<0){
      version=*odedim;
      fread(odedim,sizeof(int),1,fbin);
    }
    if(version==0){
      fread(&(chnpar.sec_perpendicular),sizeof(int),1,fbin);
      fread(&(chnpar.coordtype),sizeof(int),1,fbin);
    }else{
      chnpar.readbin(fbin,version);
    }
    fread(secnum,sizeof(int),1,fbin);
    *psec=new GLS_Pointer_Type[*secnum];
    if(ret){
      for(int i=0;i<=*secnum-1;i++){
         (*psec)[i]=new GeneralLinearSection();
         (*psec)[i]->readbin(fbin,recompute_inverse);
      }
    }
    fclose(fbin);
  }else{
    cerr << "createFromFileSections - cannot open " << filename << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }

  return ret;
}







int save_generated_sections(const char filename[],int odedim,int secnum,
      GeneralLinearSection* psec[],const chnsecparType &chnpar)
// in the binary format we save:
// odedim,chnpar,secnum,psec[0],..,psec[secnum-1]
///// the parameter: chnpar - is just for the record
/////
// this is 'the inverse function' to  <createFromFileSections>
// currently the version ==-2
{
  int version=-2;
  FILE* fbin;
  fbin=fopen(filename,"wb");
  int ret=(fbin!=NULL);
  if(ret){
    fwrite(&version,sizeof(int),1,fbin);
    fwrite(&odedim,sizeof(int),1,fbin);
    chnpar.writebin(fbin);
    fwrite(&secnum,sizeof(int),1,fbin);
    for(int i=0;i<=secnum-1;i++){
      psec[i]->savebin(fbin,1); // 1- only intervals are saved
    }
    fclose(fbin);
  }else{
    cerr << "cannot open " << filename << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  return ret;
}

int text_save_generated_sections(const char filename[],int odedim,int secnum,
       GeneralLinearSection* psec[],const chnsecparType &chnpar,int whatsave)
// in the text format we save:
// odedim,chnpar,secnum,psec[0],..,psec[secnum-1]
//  how the sections are save is specified by the parameter <whatsave>
    // whatsave=0; only double's are saved
    //         =1; only intervals
    //         =2; everything
///// the parameter:chnpar - are here just for the record
/////
{
  ofstream  report(filename);
  int ret=(report!=NULL);
  if(ret){
    report << odedim  << "  the dimension " << endl;
    chnpar.writetxt(report);
    report << secnum  << "   the number of sections " << endl;
    for(int i=0;i<=secnum-1;i++){
      psec[i]->savetext(report,whatsave);
    }
  }else{
    cerr << "cannot open " << filename << endl;
     #if defined(KRAK) && defined(WIN95)
                   waitBt();
    #endif
    exit(EXIT_FAILURE);
  }
  return ret;
}


int testReadingSections(ostream &report,const char filename[],int odedim,
       const chnsecparType &chnpar , int secnum,   GeneralLinearSection* psec[])
// we compare the sections in file <filename> with the sections
//       in psec[0,..,secnum-1]
//   and chnpar with the 'chnpar' in the file
{
  int nodedim,nsecnum;
  chnsecparType nchnpar;
  GLS_Pointer_Type *nsec;
  bool recompute_inverse=false;
  createFromFileSections(filename,recompute_inverse,&nodedim,nchnpar,&nsec,&nsecnum);

  int testok=1;
  if(nodedim!=odedim){
    testok=0;
    cout << "not equal odedim " << endl;
  }
  if(nchnpar.sec_perpendicular!=chnpar.sec_perpendicular){
    testok=0;
    cout << "not equal sec_perpendicular " << endl;
  }
  if(nchnpar.coordtype!=chnpar.coordtype){
    testok=0;
    cout << "not equal coordtype " << endl;
  }
  if(nchnpar.dimdiag!=chnpar.dimdiag){
    testok=0;
    cout << "not equal dimdiag " << endl;
  }
  if(nchnpar.period!=chnpar.period){
    testok=0;
    cout << "not equal period " << endl;
  }
  if(nchnpar.strategy!=chnpar.strategy){
    testok=0;
    cout << "not equal strategy " << endl;
  }
  if(nsecnum!=secnum){
    testok=0;
    cout << "not equal secnum " << endl;
  }
  for(int i=0;((i<=nsecnum-1) && (i<=secnum-1));i++){
    if(!GLS_thesame(psec[i],nsec[i],1)){ // 1 - we check intervals
      testok=0;
      cout << "i=" << i << " sections differ " << endl;
    }
  }

  // computation  of the center points
  report << "Computation of the center points on sections " << endl;
  interval *x,*xfull;
  x=new interval[odedim];
  xfull=new interval[odedim+1];
  zero(x,odedim-1);
  for(int i=0;i<=nsecnum-1;i++){
    nsec[i]->sectiontofull(x,xfull);
    report << "section= " << i << endl;
    saveAllCoord(report,xfull,odedim,0);
  }
  delete [] xfull;
  delete [] x;


  for(int i=0;i<=nsecnum-1;i++)
    delete nsec[i];

  return testok;
}




int gen_sections_along_traj_poincare(ostream &report,Ode &ode,
        GeneralLinearSection* psec[],int secnum,int secdir,Par_Poincare &param,
 const double xs[],const chnsecparType &chnpar,int datasavebin,const char binfilename[],
        int datasavetext,const char textfn[])
// psec[0,1,...,secnum-1] - the initial sections
//  on output: here are the generated sections
// chnpar.sec_perpendicular - if 1, then we generate the sections perpedicular to the vector field
// chnpar.period - a variable telling what is the iterate of the full Poincare map
//          resulting from all transitions between psec[0]->psec[1]->...->
//             psec[secnum-1]->psec[0]
{
  if(chnpar.sec_perpendicular){
    create_perpen_sections_along_traj(ode,param,xs,psec,secnum,secdir,chnpar.dimsecper);
  }


  // the generation of trajectory on sections
  doublematrix1 *tabxs;
  tabxs=new doublematrix1[secnum];
  generate_sections_traj(ode,param,xs,psec,secnum-1,secdir,tabxs);
       // tabxs[i]= is the point from the trajectory of <xs> on section psec[i]
       //            it is expressed in the ambient space coordinates

  /// Test lines
  report << endl << "testing the trajectory on sections " << endl;
  for(int i=0;i<=secnum-1;i++){
    report << "section=" << i << endl;
    report << "Point x: " << endl;
    saveAllCoord(report,tabxs[i],ode.dim);
    double* vx;
    vx=new double[ode.dim+1];
    ode.field(tabxs[i],vx);
    report << "v(x)=" << endl;
    saveAllCoord(report,vx,ode.dim);
    vect_normalize(vx,ode.dim,NORM_EUCL);
    report << "normalized v(x)=" << endl;
    saveAllCoord(report,vx,ode.dim);
    delete [] vx;
  }
  report << endl;
  /////


  // now the diagonalization on sections - we compute
  // independently eigenvectors for each section
  //  one might hope for a more optimal approach, like the computation
  // between sections and using this data


  double *xn;
  xn=new double[ode.dim+1];
  doublematrix1 *df;
  df=new doublematrix1[ode.dim+1];
  double tpoinc;
  estimates estm;

  pLinCC *coordch;
  coordch=new pLinCC[secnum];
  for(int i=0;i<=secnum-1;i++)
    coordch[i]=NULL;

  interval *ixfull;
  ixfull=new interval[ode.dim+1];

  interval *ixb;
  ixb=new interval[ode.dim];


  for(int i=0;i<=secnum-1;i++){
    if(chnpar.coordtype==standardcoord){
      identity(df,ode.dim-1);
    }else{
      if(!iterate_C1nonrigpoinc(chnpar.period,ode,param,tabxs[i],xn,&tpoinc,*(psec[i]),
          secdir,&estm,df)){
         // df=dP^{chnpar.period} is expressed in the section coordinates psec[i]->plincc
        cerr << "gen_sections_along_traj_poincare - can't compute C1-Poincare map " << endl;
         #if defined(KRAK) && defined(WIN95)
                       waitBt();
    #endif
        exit(EXIT_FAILURE);
      }
    }

    tointerval(tabxs[i],ixfull,ode.dim);
    psec[i]->fulltosection(ixfull,ixb);

    if(!buildcoordchange(chnpar.coordtype,df,ixb,ode.dim-1,&(coordch[i]),chnpar.dimdiag)){
      cerr << "gen_sections_along_traj_poincare - can't diagonalize " << endl;
       #if defined(KRAK) && defined(WIN95)
                     waitBt();
    #endif
      exit(EXIT_FAILURE);
    }
  }
  delete [] xn;
  xn=NULL;



  delete [] ixb;
  ixb=NULL;

  delete [] ixfull;
  ixfull=NULL;


  for(int i=0;i<=secnum-1;i++){
    psec[i]->new_coord_on_sec(coordch[i]); // here we transform section coordinates
                     // the section remains the same, but the other variables
                     // describing the section are changed
  }



  if(datasavebin){
    save_generated_sections(binfilename,ode.dim,secnum,psec,chnpar);
    report << "the sections and coordinate changes are in [" << binfilename <<  "]" << endl;

  }
  if(datasavetext){
     int whatsave=0; // only doubles
     text_save_generated_sections(textfn,ode.dim,secnum,psec,chnpar,whatsave);
     report << "the sections and coordinate changes (doubles) are in [" << textfn <<  "]" << endl;

  }

  for(int i=0;i<=secnum-1;i++){
    delete coordch[i];
  }
  delete [] coordch;


  if(datasavebin){
    if(testReadingSections(report,binfilename,ode.dim,chnpar,secnum,psec)){
      cout << " Reading sections test - ok " << endl;
    }else{
      cout << " Reading sections test - ERRORS " << endl;
    }
  }

  delete [] df;
  delete [] tabxs;
  return 1;
}



int SectionsExpFromFile(const char secfilename[],int dim,bool recompute_inverse,
        chnsecparType& chnpar, GLS_Pointer_Type** psections,int *secnum)
// (*psections)[i] - are indexed from 0 to *secnum-1
// chnpar - is just for reporting
// dim - this is the dimension of the ambient space, we require
//   that dim will be equal to or larger than the dimension of sections
// in file <secfilename>, if we have the strong inequality then the sections
//  are expanded into dim-dimensional space in the natural way
////
// Warning: since the section file does not contain the value for chnpar.strategy
//   then in this function it is allways set to 0
{
  int ok=1;
  int dimInSecFile;

  if(createFromFileSections(secfilename,recompute_inverse,&dimInSecFile,chnpar,
           psections,secnum)){

    if(dimInSecFile < dim){
      GeneralLinearSection* psec;
      for(int i=0;i<*secnum;i++){
        psec=GLSincreasedim((*psections)[i],dim);
        delete (*psections)[i];
        (*psections)[i]=psec;
      }
    }
    if(dimInSecFile > dim){
      ok=0;
      cerr << "<SectionsExpFromFile> - the dimension is section files is " << dimInSecFile << endl;
      cerr << " it should be less than or equal to " << dim << endl;
      #if defined(WIN95) && defined(KRAK)
      cerr << "press any key " << endl;
      waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
  } else{
    ok=0;
  }
  return ok;
}

int  chn_create_pmaps_ff(const char secfilename[],int closedloop,
          bool recompute_inverse, GLS_Pointer_Type**  psections,int *secnum,Ode &ode,
          pFPoincareMap** PMap,int *pMapnum,const Par_Poincare &param,
          chnsecparType &chnpar)
// prepares PMap for the proofs of the existence
// proofs based on chains of covering relations
//  the sections - are contained in <secfilename> - this is a binary file
//  closedloop - 1 or 0 - tell whether we have the relation N_{k-1} ->N_0
//               i.e. from the last section to the first one
//  psections[0],...,psection[secnum-1] - are consequtive sections
//    the last Poincare map is defined as the transition
//              psections[secnum-1] -> psections[0]
//  others PMap[i]= the transition: psections[i] -> psections[i + 1 mod secnum]
//  all sections are to be crossed in UP - direction
{

  int ok=1;

  int oldstrategy=chnpar.strategy;
  //  reading sections
  if(!SectionsExpFromFile(secfilename,ode.dim,recompute_inverse,chnpar,psections,secnum)){
    ok=0;
    exit(EXIT_FAILURE);
  }
  chnpar.strategy=oldstrategy;

  // defining Poincare maps
  *pMapnum=*secnum;
  if(!closedloop)
    *pMapnum=*secnum -1;
  *PMap=new pFPoincareMap[*pMapnum];
  if(ok){
    for(int i=0;i<*pMapnum;i++){
      (*PMap)[i]=new FPoincareMap(&ode,(*psections)[i],
                   (*psections)[(i + 1) % (*secnum)],UP,param,0,0);
    }
  }
  return ok;
}


