#ifndef _section_h
#define _section_h
#include"defo.h"
#include"func.h"
#include"coordc.h"
#include"rigpoinc.h"


  const int standardcoord=1;
  const int orthogonalizedeigenvectors=2;
  const int eigenvectors=3;
  const int partialeigenbasis=4;
  const int eigen_subspaces_ortogonalized=5;



// constants identifying the strategies of the generation of new h-sets
const int gen_new_intersect=1;  // we take an 'intelligent intersection'
                             // and shift in case of an empty intersection
const int gen_new_shift=2; // we rather shift the h-set and focus on the diameter

const int gen_from_point=3;  // we generate h-sets starting from zero width


char *namehsetstrategy(int strategy, char *stratname,int maxlen);

int teststrategyVal(int strategy);
// returns 1, when strategy has a correct value


char* osa_coordsname(char* name,int maxlen,int coords);




class chnsecparType{
  public:
    int sec_perpendicular;
    int dimsecper;
    int coordtype;
    int dimdiag; // the number of coordinates used in diagonalization process
                    // in the creation of good coordinates
    int period; // if 0 -then either we don't know whether is comes
                // from the periodic orbit; or it is not from periodic orbit
    int strategy; // the strategy of the generation (update) of the h-set
               // in the next iteration



    void writetxt(ostream &report,int secdataonly = 0) const;
    void writebin(FILE *f) const;
    void readbin(FILE *f,int version);
    void zero();

    chnsecparType();

};



// class Section - the section of the flow: we need coordinates on section
//     and transition from section coordinates to the ambient space coordinates
//     and back, both for points and tangent vectors
//      the section is given by the equation eval(x)=0
class Section: public C1Func
{
 public:
  char infostr[40];
  int dimsec;
 // int dim; from C1Func
  double virtual eval(const double x[]) const=0;  // dim(x) = dim
  interval virtual eval(const interval x[]) const=0; // dim(x) = dim
  interval virtual eval(const Set &x) const =0; // dim(x)=x
  void virtual gradient(const double x[],double grad[]) const =0;
  void virtual gradient(const interval x[],interval grad[]) const=0;

  void virtual intersection(const interval x[],interval x_on_level[]) const;
    // on output x_on_level \supset x \\cap \\{z| f(z) ==0 \\}
    // used by Poincare maps (really needed in C1 computations)

  int virtual fulltosection(const double x[],double xsec[]) const =0;
  int virtual fulltosection(const interval x[],interval xsec[]) const =0;
  // transition from the ambient space to section for points
  //                dim (x) = dim; dim(xsec)=dimsec
  //    on output <xsec> is <x> in section coordinates

  int virtual vectfulltosection(const double x[],const double vect[],
                 double xsec[],double vectsec[]) const=0;
  int virtual vectfulltosection(const interval x[],const interval vect[],
                 interval xsec[],interval vectsec[]) const=0;
  // transition from the ambient space to section for points and tangent vectors
  //                dim (x)=dim(vect) = dim; dim(xsec)=dim(vectsec)=dimsec
  //    on output <xsec> is <x> in section coordinates
  //              <vectsec> is a projection of <vect> onto section in <x>
  //               expressed in section coordinates


  int virtual sectiontofull(const double xsec[],double x[]) const=0;
  int virtual sectiontofull(const interval xsec[],interval x[]) const=0;
  // transition from the section  coordinates to  the ambient space coordinates
  //                for points dim (x) = dim; dim(xsec)=dimsec
  //    on output <x> is <xsec> in the ambient space coordinates
  int virtual sectiontofull(const interval xcsec[],const interval Csec[][MAX_DIM+1],
              const interval rsec[], interval xcfull[],
              interval Cfull[][MAX_DIM+1],interval rfull[]) const;
  // transition from the section  coordinates to  the ambient space coordinates
  // for the set   <xcsec + Csec*rsec> on section
  //    in the ambient space we seek  the following representation
  //            xcfull + Cfull*rfull
  /////  diam(xcsec)=0 and diam(xcfull)=0

   int virtual sectiontofull(const Set &xonsec,Set &xfull) const;
    // transition from the section  coordinates to  the ambient space coordinates


  int virtual vectsectiontofull(const double xsec[],const double vectsec[],
                 double x[],double vect[]) const =0;
  int virtual vectsectiontofull(const interval xsec[],const interval vectsec[],
                 interval x[],interval vect[]) const=0;
  // transition from the section  coordinates to  the ambient space coordinates
  //     for points and tangent vectors
  //       dim (x) =dim(vect)= dim; dim(xsec)=dim(vectsec)=dimsec
  //    on output <x> is <xsec> in the ambient space coordinates
  //      and <vect> is <vectsec> in the ambient space coordinates

};


class GeneralLinearSection: public Section
// this is our old section, just a table of coefficients
//  sec[0] + \Sigma_{i=1}^{dim} x[i]*sec[i]=0
//  PLUS linear coordinates on section
//    these linear coordinates on section are imbeded in the coordinate
//    change in the ambient space such such that in these coordinates
//  the section is given by y[1]=0
{
  public:
    double sec[MAX_DIM+1];
    interval isec[MAX_DIM+1];


    LinCC *plincc; // this is the coordinate change in the ambient space
                   // such that in the new coordinates section
                   //  is given by y[1]=0
                   // coordinates y[2],y[3],... - provide then the coordinate
                   //              frame on the section
                   //  x - the coordinates of ambient space - are
                    //   the old coordinates for *plincc
                    ///////////////////
                    //  plincc->coordtran(?,?) - is the transformation from the ambient space
                    //                        to the section coordinates
                    //  plincc->coordinvtran(?,?) - is the transformation from (0,y) - the section coordinates
                    //                            to the full space


    double  eval(const double x[]) const;  // dim(x) = dim
    interval  eval(const interval x[]) const; // dim(x) = dim
    interval eval(const Set &x) const; // dim(x)=dim
    void  gradient(const double /*x*/[],double grad[]) const;
    void  gradient(const interval /*x*/[],interval grad[]) const;


    GeneralLinearSection();
      // the 'empty' constructor

    GeneralLinearSection(int pdim,const double psec[],LinCC *cchange,
           int pdestroy_lincc=0);
    GeneralLinearSection(int pdim,const interval psec[],LinCC *cchange,
            int pdestroy_lincc=0);
    // the  variable <destroy_lincc> tells whether the destructor for
    // GeneralLinearSection should delete cchange

    GeneralLinearSection(int pdim,int selected_coord,interval ivalsection);
       // the section is given by x[selected_coord]+ivalsection=0


    GeneralLinearSection(LinCC *cchange,int pdestroy_lincc);
       // the section is given by y[1]=0, where y - are the new coordinates
       //          defined by *cchange

    GeneralLinearSection(const GeneralLinearSection &gls);


    ~GeneralLinearSection(); // destructor



    int fulltosection(const double x[],double xsec[]) const;
    int fulltosection(const interval x[],interval xsec[]) const;
    // transition from the ambient space to section for points
    //                dim (x) = dim; dim(xsec)=dimsec
    //    on output <xsec> is <x> in section coordinates



    int vectfulltosection(const double x[],const double vect[],
                 double xsec[],double vectsec[]) const;
    int vectfulltosection(const interval x[],const interval vect[],
                 interval xsec[],interval vectsec[]) const;
    // transition from the ambient space to section for points and tangent vectors
    //                dim (x)=dim(vect) = dim; dim(xsec)=dim(vectsec)=dimsec
    //    on output <xsec> is <x> in section coordinates
    //              <vectsec> is a projection of <vect> onto section in <x>
    //               expressed in section coordinates



    int  sectiontofull(const double xsec[],double x[]) const;
    int  sectiontofull(const  interval xsec[],interval x[]) const;
    // the transition from the section  coordinates to  the ambient space coordinates
    //                for points dim (x) = dim; dim(xsec)=dimsec
    //    on output <x> is <xsec> in the ambient space coordinates
    int  sectiontofull(const interval xcsec[],const interval Csec[][MAX_DIM+1],
              const interval rsec[],interval xcfull[],interval Cfull[][MAX_DIM+1],
              interval rfull[]) const;
    // the transition from the section  coordinates to  the ambient space coordinates
    // for the set   <xcsec + Csec*rsec> on section
    //    in the ambient space we seek for the following representation
    //            xcfull + Cfull*rfull
    /////  diam(xcsec)=0 and diam(xcfull)=0
     int  sectiontofull(const Set &xonsec,Set &xfull) const;
    // transition from the section  coordinates to  the ambient space coordinates


    int vectsectiontofull(const  double xsec[],const double vectsec[],
                 double x[],double vect[]) const;
    int vectsectiontofull(const interval xsec[],const interval vectsec[],
                 interval x[],interval vect[]) const;
    // transition from the section  coordinates to  the ambient space coordinates
    //     for points and tangent vectors
    //       dim (x) =dim(vect)= dim; dim(xsec)=dim(vectsec)=dimsec
    //    on output <x> is <xsec> in the ambient space coordinates
    //      and <vect> is <vectsec> in the ambient space coordinates


    int new_coord_on_sec(const LinCC *pcoordch);
   // <pcoordch> is dimsec-dimensional
  // introduces new coordinates on section ;  this a coordinate change to new coordinates on section;
  //  the original coordinates on section are used as the "old" coordinates in this composition

    int savebin(const char *filename,int whatsave) const;

    int savebin(FILE *fbin,int whatsave) const;
        // saves the section to a binary file
    // whatsave=0; only double's are saved
    //         =1; only intervals
    //         =2; everything
// the structure of the file:
//  whatsave,dim,dimsec,isec[0,dim] or sec[0,dim] or both(depending on whatsave),
//  *plincc

    int readbin(const char *filename,bool recomp_inverse/*=false*/);
    int readbin(FILE *fbin,bool recomp_inverse/*=false*/);
 // reads (really reinitiates) the section from the file <filename>
   // if  recomp_inverse, then the inverse in *plincc is recomputed
   //  to make sure that computation is rigorous, can be of use, when
   // computing things first on the computer on which we don't have correct rounding


    int savetext(ostream &ftext,int whatsave) const;
    // saves the section to a text file
    // whatsave=0; only double's are saved
    //         =1; only intervals
    //         =2; everything
    // the structure of the file:
    //  whatsave,dim,dimsec,isec[0,dim] or sec[0,dim] or both(depending on whatsave),
    //  *plincc
    int savetext(const char *filename ,int whatsave) const;

    int readtext(istream &ftext);
    int readtext(const char *filename);


  private:
    int destroy_lincc; //  if the coordinate transformation
                       // was created by this object, then the destructor
                       // will free the memory (destroy_lincc==1)
    int integritytest() const; // test if the section is given by y[1]=0
                         // in new coordinates (given by *plincc)

};


typedef GeneralLinearSection *GLS_Pointer_Type;


GeneralLinearSection* GLSincreasedim(const GeneralLinearSection *gls,int newdim);
// this function imbeds the section  gls into higher dimensional phase space
//  nothing  changes - the section form is the same and the coordinate frame
// on section is identity on new variables
//  <newdim> - is the dimension of the ambient space
// *gls - remains unchanged

GeneralLinearSection* createPerpendicularSection(int pdim,
          const interval x[],Ode &ode);
 // the creation of the section passing thru the point <x>
      //  and perpendicular to the vector <ode.vfield(x)>
      //  the coordinate system is given as follows:
      //    - 1-st coordinate - the direction of <vect>
         //  the coordinates on section are obtained by the orthogonalization
      //   of the following set of vectors:  <ode.vfield(x)> \cup \{e_1,e_2, \dots,
      //    we drop the one for which  |ode.vfield(x)[i]| is maximal }
      // when pdim < ode.dim - this is a section for the <pdim>-dimensional
      // Galerkin projection
      // On input: dim(x)=ode.dim

GeneralLinearSection* createPerpendicularSection(int pdim,const double x[],Ode &ode);
 // the creation of the section passing thru the point <x>
      //  and perpendicular to the vector <ode.vfield(x)>
      //  the coordinate system is given as follows:
      //    - 1-st coordinate - the direction of <vect>
         //  the coordinates on section are obtained by the orthogonalization
      //   of the following set of vectors:  <ode.vfield(x)> \cup \{e_1,e_2, \dots,
      //    we drop the one for which  |ode.vfield(x)[i]| is maximal }
      // when pdim < ode.dim - this is a section for the <pdim>-dimensional
      // Galerkin projection
      // On input: dim(x)=ode.dim


GeneralLinearSection* createPerpendicularSection(int pdim,const interval x[],
        const interval vect[]);
     // the creation of the section passing thru the point <x>
      //  and perpendicular to the vector <vect>
      //  the coordinate system is given as follows:
      //    - 1-st coordinate - the direction of <vect>
      //  the coordinates on section are obtained by the orthogonalization
      //   of the following set of vectors:  <vect> \cup \{e_1,e_2, \dots,
      //    we drop the one for wich  |vect[i]| is maximal }


GeneralLinearSection* gal_createPerpendicularSection(int galdim,
            const double x[],Ode &odeeq);
 // the creation of the section  in dimension <odeeq.dim>
 //  which is build from the perpendicular section for the <galdim>-dimensional
 //  projection of the vector field at x (but not its galdim-projection)
 // On input: dim(x)=ode.dim

void transOnSec_to_full(GeneralLinearSection *psection,int other_coord_on_sec,
                LinCC* pcoordch, Set &xonsec, Set &xfull);
// this function transforms the set <xonsec> to the full space and places
// the result in <xfull>;
// if <other_coord_on_sec> then <xonsec> (on input) is expressed in coordinate frame <pcoordch>


LinearF* createLinearF(const GeneralLinearSection& gls);
// creates (extracts) a LinearFunc from GeneralLinearSection


int GLS_thesame(const GeneralLinearSection *psec1,const GeneralLinearSection* psec2,int
  whattocheck);
// 1 - is returned  when the sections (together with the coordinate changes) are the same
// 0 - otherwise
//   if whattocheck==0 - we check doubles (also in coordinate change)
//                1 - we check intervals
//                2 - we check both doubles and intervals



void from_sec_to_ambient_space(const GeneralLinearSection *psection,const interval xc_sec[],
  const interval r0_sec[],interval xfull[],int useothercoord,LinCC *coordch);
// the set xc_sec + r0_sec on the section &psection, which is eventually
//  expressed (if useothercoord) in coordinates <coordch> is transformed
// to the full space and enclosed in xfull


/// POINCARE MAP


class FPoincareMap_param{
  public:
    Par_Poincare par_double,par_interval;
};


class FPoincareMap: public  FMap
{
  protected:

   int destroy_Insection; // is 1, when the constructor created the section
                     // pInsection (the domain) and then the destructor has to destroy it
   int destroy_Outsection; // as above but for the output section (the range)
  public:
    // int dim; // inherited from Fmap (and Map)
    Ode *pode;

    FPoincareMap_param param; // param.par_double - used for nonrigorous
                       // computations;
                       //  param.par_interval - for rigorous computations

    GeneralLinearSection *pInsection; // the section for the domain
    GeneralLinearSection *pOutsection; // the section for the range

    int secdir;  // UP  or DOWN    // the direction in which we intersect the section
        // if UP - then the section(x(t)) increases, if ==DOWN then it decreases

    int virtual map( double x[],double fx[]);  // on output fx=f(x)


    int virtual imapDE( interval xcenter[], interval C[][MAX_DIM+1],interval r0[],
                 interval fx[],estimates *estm);
       // on output fx=map(xcenter + C*r0)
    int virtual imapDE(Set &x,interval fx[],estimates *estm);



    int virtual imapDE(interval xcenter[], interval C[][MAX_DIM+1],interval r0[],
                 interval fx[],int use_other_inputcoord,LinCC* inpcoord,
                 int use_other_outcoord,LinCC *outpcoord,estimates *estm);
// the function computes f(xcenter +C*r0) and the result is placed in fx
//  when the computation terminates successfully, in this case 1 is returned
// otherwise we return 0
/// both the input values and output values are given in terms of section
//  coordinates
////// if use_other_inputcoord - then the input values are expressed in inpcoord
///    similarly if use_other_outcoord, the the output value (fx) is expressed
//          using outpcoord
///////// it should be stressed that both these linear transformations inpcoord
//              and outpcoord are are working on section, and expressed in section
///              coordinates
/////////
// this function  assumes that the coordinates on sections
// y[1]=0, (y[2],...) - where y - are the coordinates this->psection->plincc


  int virtual imapDE(Set &x, interval fx[],int use_other_inputcoord,LinCC* inpcoord,
                 int use_other_outcoord,LinCC *outpcoord,estimates *estm);
// the function computes f(x) and the result is placed in fx
//  when the computation terminates successfully, in this case 1 is returned
// otherwise we return 0
/// both the input values and output values are given in terms of section
//  coordinates
////// if use_other_inputcoord - then the input values are expressed in inpcoord
///    similarly if use_other_outcoord, the the output value (fx) is expressed
//          using outpcoord
///////// it should be stressed that both these linear transformations inpcoord
//              and outpcoord are are working on section, and expressed in section
///              coordinates
/////////
// this function  assumes that the coordinates on sections
// y[1]=0, (y[2],...) - where y - are the coordinates this->psection->plincc



   int imapDE(interval xcenter[], interval C[][MAX_DIM+1],interval r0[],
                 interval fx[],interval df[][MAX_DIM+1],int use_other_inputcoord,
                 LinCC* inpcoord,int use_other_outcoord,LinCC *outpcoord,
                 estimates *estm);
      // this function computes the map: usually this is the Poincare map
       //  and it makes sense sometimes to play with different coordinate
       //  systems
       //   if (use_other_inputcoord) then the input (xcenter + C*r0) is expressed
       //   using the  <inpcoord> coordinate   system
       //   if(use_other_outcoord) the the output - fx is expressed in coordinates
       //  <outpcoord>
       //    df - is the enclosure for the derivative
       //     in coordinates taking into account <use_other_inputcoord> and
       //   <use_other_outcoord>
       // this function  assumes that the coordinates on sections (both domain and
       //  range are as follows:
       // y[1]=0, (y[2],...) - where y - are the coordinates this->pIn(Out)section->plincc


   int imapDE(Set &x,interval fx[],interval df[][MAX_DIM+1],int use_other_inputcoord,
                LinCC* inpcoord,int use_other_outcoord,LinCC *outpcoord,
                estimates *estm);
      // this function computes the map: usually this is the Poincare map
       //  and it makes sense sometimes to play with different coordinate
       //  systems
       //   if (use_other_inputcoord) then the input (x) is expressed
       //   using the  <inpcoord> coordinate   system
       //   if(use_other_outcoord) the the output - fx is expressed in coordinates
       //  <outpcoord>
       //    df - is the enclosure for the derivative
       //     in coordinates taking into account <use_other_inputcoord> and
       //   <use_other_outcoord>
       // this function  assumes that the coordinates on sections (both domain and
       //  range are as follows:
       // y[1]=0, (y[2],...) - where y - are the coordinates this->pIn(Out)section->plincc



    int virtual map(double x[],double fx[],double df[][MAX_DIM+1]);


    int virtual imapDE( interval xcenter[],interval C[][MAX_DIM+1],interval r0[],
             interval fx[],interval df[][MAX_DIM+1],estimates *estm);
    int virtual imapDE(Set &x,interval fx[],interval df[][MAX_DIM+1],estimates *estm);


    int virtual transporttosection(double x0full[],double fxsec[]);
   // on output fxsec is the point on the section (in the section coordinates)
    //    which is on first intersection of the trajectory through the point
    //   <x0full> which is given in standard ambient space coordinates
    //         if <x0full> is very close to the section, then we just project
    //           it onto the section and return 1 otherwise
    // 1 - is returned if the Poincare time for x0full < tmax

    double dsettimestep(double ph);
     //  returns  previous timestep for double computation, and sets timesptep=ph

    double isettimestep(double ph);
    //  returns  previous   timestep for interval computations, and sets timesptep=ph


    double dgettimestep();

    double igettimestep();

    double settmax(double ptmax);
    double gettmax();


    int isetorder(int porder);
  // returns previous value of order; and set order=porder and orderdf=order
  // interval computations


    int igetorder();

    int igetorderdf();

    int dgetorderdf();

    int dsetorder(int porder);
  // returns previous value of order; and set order=porder and orderdf=order
  // double computations


    int dgetorder();

    int dgetmethod();

    int igetmethod();


    FPoincareMap(Ode *ppode,GeneralLinearSection *ppInsection,
                   GeneralLinearSection *ppOutsection,int psecdir,double ph,
                   double ptmax,int porder,int pmethod,
                   int pdestroy_Insection=0,int  pdestroy_Outsection=0);

    FPoincareMap(Ode *ppode,GeneralLinearSection *ppInsection,
               GeneralLinearSection *ppOutsection,int psecdir,const Par_Poincare &param,
               int pdestroy_Insection=0,int  pdestroy_Outsection=0);

    FPoincareMap(Ode *ppode, GeneralLinearSection *ppInsection,
                GeneralLinearSection *ppOutsection,int psecdir,
               const Par_Poincare &pparaminterval,const Par_Poincare &pparamdouble,
               int pdestroy_Insection=0,int  pdestroy_Outsection=0);


    FPoincareMap(Ode *ppode,GeneralLinearSection *ppsection,int psecdir,double ph,
        double ptmax,int porder,int pmethod,int pdestroy_section=0);

    FPoincareMap(Ode *ppode, GeneralLinearSection *ppsection,int psecdir,
              const Par_Poincare &pparam,int pdestroy_section=0);

    FPoincareMap(Ode *ppode, GeneralLinearSection *ppsection,int psecdir,
              const Par_Poincare &pparaminterval,const Par_Poincare &pparamdouble,
              int pdestroy_section=0);


    FPoincareMap(Ode *ppode,int pselcoord,interval pivalsec,int psecdir,double ph,
        double ptmax,int porder,int pmethod);
      // the section will be given by x[pselcoord]+pivalsec=0

    void setdefaultval(double ph,double ptmax,int porder,int pmethod);
   // default values

    FPoincareMap();
    // default constructor - for heirs to be created without problems

    virtual ~FPoincareMap();
    // destructor



    FPoincareMap_param get_param();
    FPoincareMap_param set_param(const FPoincareMap_param &lparam);

  private:
    int testdim(); // we check whether various dimensions are in mutual agreement

};

typedef FPoincareMap  *pFPoincareMap;

int PoincMap_map_divide(FPoincareMap& pmap,int divisiontable[],
       int divdim, interval ixb[],interval C[][MAX_DIM+1],interval r0[],
       interval fx[],int useothercoord,
       LinCC *coordch,estimates *estm);
// the function computes pmap.map(ixb + C*r0)
//  for this computation   the set ixb + C*r0  is first divided
//     into \Pi_{1<= i <=divdim}  divisiontable[i] parts
 //  the i-th direction in <r0> is divided into divisiontabl[i] equal parts

int iterate_C1nonrigpoinc(int itno,Ode &ode,Par_Poincare param,double x[],
         double xn[],double *tpoinc,const GeneralLinearSection &section,int secdir,
         estimates *estm,double df[][MAX_DIM+1]);
//  nonrigorously computes <itno>-th iterate of Poincare map for <ode>
//  the Poincare map is defined by
//  section.eval(x)=0
//  if secdir==UP - then <section.eval()> is
//      increasing on section
//  x - starting point , dim(x) = ode.dim - is expressed in ambient space coordinates
//  xn=P^{itno}(x) - on output , dim(xn) = ode.dim - is expressed in ambient space coordinates
//  df=dP^{itno}(x) - in the section coordinates


bool buildcoordchange(int coordtype,const double Dp[][MAX_DIM+1],const interval ixp[],
             int dim,LinCC **coordch,int dimdiag=0 );
//  we use matrix <Dp> to build a coordinate change:
// On output : **coordch - is the coordinate change is build as follows:
// 1) the origin is shifted to <ixp>
// 2) linear part  = identity, when <coordtype>==standardcoord
//                 = orthogonal change to orthogonalized eigenvectors basis
//                      when, <coordtype>==orthogonalizedeigenvectors
//                 = change to the eigenvector basis, when
//                     <coordtype>==eigenvectors
//                 = change to mixed basis - the dominant eigenvalues
//                      and orthogonalized remaining directions
//                     <coordtype>==partialeigenbasis
//                 =  change to mixed basis - subspaces of eigenvectors
//                     which are close to parallel are  orthogonalized
//                      <coordtype>==eigen_subspaces_ortogonalized
//  dimdiag - tells how many first coordinates are used in diagonalization
//          attempt
//   if dimdiag==0 then it is ignored - all coordinates are used

int gen_sections_along_traj_poincare(ostream &report,Ode &ode,
        GeneralLinearSection* psec[],int secnum,int secdir,Par_Poincare &param,
 const double xs[],const chnsecparType &chnpar,int datasavebin,const char binfilename[],
         int datasavetext,const char textfn[]);
// psec[0,1,...,secnum-1] - the initial sections
//  on output: here are the generated sections
// chnpar.sec_perpendicular - if 1, the we generate the sections perpedicular to the vector field
// chnpar.period - a variable telling what is the iterate of the full Poincare map
//          resulting from all transitions between psec[0]->psec[1]->...->
//             psec[secnum-1]->psec[0]


int save_generated_sections(const char filename[],int odedim,int secnum,
      GeneralLinearSection* psec[],const chnsecparType &chnpar);
// in the binary format we save:
// odedim,chnpar,secnum,psec[0],..,psec[secnum-1]
///// the parameter: chnpar - is just for the record
/////
// this is 'the inverse function' to  <createFromFileSections>
// currently the version ==-2



int testReadingSections(ostream &report,const char filename[],int odedim,
   const chnsecparType &chnpar , int secnum,   GeneralLinearSection* psec[]);
// we compare the sections in file <filename> with the sections
//       in psec[0,..,secnum-1]
//   and chnpar with the 'chnpar' in the file


int text_save_generated_sections(const char filename[],int odedim,int secnum,
       GeneralLinearSection* psec[],const chnsecparType &chnpar,int whatsave);
// in the text format we save:
// odedim,chnpar,secnum,psec[0],..,psec[secnum-1]
//  how the sections are save is specified by the parameter <whatsave>
    // whatsave=0; only double's are saved
    //         =1; only intervals
    //         =2; everything
///// the parameter:chnpar - are here just for the record
/////

int SectionsExpFromFile(const char secfilename[],int dim,bool recompute_inverse,
       chnsecparType& chnpar, GLS_Pointer_Type** psections,int *secnum);
// (*psections)[i] - are indexed from 0 to *secnum-1
// chnpar - is just for reporting
// dim - this is the dimension of the ambient space, we require
//   that dim will be equal to or larger than the dimension of sections
// in file <secfilename>, if we have the strong inequality then the sections
//  are expanded into dim-dimensional space in the natural way
////
// Warning: since the section file does not contain the value for chnpar.strategy
//   then in this function it is allways set to 0


int  chn_create_pmaps_ff(const char secfilename[],int closedloop, bool recompute_inverse,
          GLS_Pointer_Type**  psections,int *secnum,Ode &ode,
          pFPoincareMap** PMap,int *pMapnum,const Par_Poincare &param,
          chnsecparType &chnpar);
// prepares PMap for the proofs of the existence
// proofs based on chains of covering relations
//  the sections - are contained in <secfilename> - this is a binary file
//  closedloop - 1 or 0 - tell whether we have the relation N_{k-1} ->N_0
//               i.e. from the last section to the first one
//  psections[0],...,psection[secnum-1] - are consequtive sections
//    the last Poincare map is defined as the transition
//              psections[secnum-1] -> psections[0]
//  others PMap[i]= the transition: psections[i] -> psections[i + 1 mod secnum]
//  all sections are to be crossed in UP - direction

#endif /* _section_h */
