#ifndef _vfkso_h_
#define _vfkso_h_
#include"zdynsys.h"

// #define ENCL_TEST  // this variable should be used when we test
// the enclosure function for the tail and we want a detailed report to the file


#if defined(DIFF_INCL)
class KuraShi_Tail;
#endif

// constants defining the type of the rough enclosure function
extern const int ks_ode_enclosure;   // standard ODE
extern const int ks_ode_iso_enclosure; //  standard ODE + (if previous succeds)
                              //   the improvement via isolation computation

extern const int ks_mixed_enclosure;
                             //  the mixture of the standard encl. and
                             // isolation from the begining
                            //  this appears to be slighty worse than ode_iso_encl..
                            //  but for 'large' time steps it works were the other fails



char* ksencl_name(int enclfunctype,char buffer[],int bufflen);

class KuraShi: public Ode {
 public:
   double mu;
   interval imu;
   int rough_encl_func_type;  // identifies the rough enclosure function
                              //
                              // ks_ode_enclosure (=0) - standard ODE
                              // ks_ode_iso_enclosure(=1) - standard ODE + (if previous succeds)
                              //             the improvement via isolation computation
                              // ks_mixed_enclosure (=2) - mixture of standard and isolation from the begining
                              //       this appears to be slighty worse than 2
                              //       but for 'large' time steps it works were 1 fails
   int c1rough_encl_func_type; // as above but for the C1-computations

   interval lambda(int k);    // returns k^2(1-\mu k^2)
   double dlambda(int k);     // returns k^2(1-\mu k^2)
   interval virtual  vfield(int k,const interval x[]);
   double virtual vfield(int k,const double x[]);

   interval vfieldndiagt(int k,const interval x[]); // returns f_k - k^2(1-mu k^2)a_k
   double vfieldndiagt(int k,const double x[]);  // returns f_k - k^2(1-mu k^2)a_k



   interval virtual dvfield(int k, int i,const interval x[]);  //  d f_k/ dx_i
   double virtual  dvfield(int k, int i,const double x[]);     //  d f_k/ dx_i

   interval virtual d2vfield(int k, int i, int j,const interval x[]);
   double virtual  d2vfield(int k, int i, int j,const double x[]);


   double virtual timestep(double h);  // a proposed dim-depended time step

   interval virtual timestep(interval h);  // a proposed dim-depended time step

   void virtual diagBadEnclosure(const interval d[],interval h0,const interval y[],const interval z[],
      int use_perturbation,const interval perturb_vect[]);
// the function writes to the file <zglerrorlogname>
// the data from failed enlosure procedure
// it tries to give a reason for this


   int virtual enclosure(const interval d[],interval h0, interval y[],int nomessages=0);


   int virtual pert_enclosure(const interval d[],interval h0, interval y[],int useperturbation,
      const interval perturb_vector[],  int nomessages =0);
  // if true is returned then:
  //    \varphi([0,h0],d) \subset y
  //   \varphi is defined by x' \in field(x) + perturb_vector if useperturb
  //   and    \varphi is defined by x' = field(x) if !useperturb
//  h0 - cannot  be  negative if <useperturbation>
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen

  int dissOneStep(const interval x0[],interval h,interval xencl[],interval xstep[],
    int use_perturbation,const interval perturb_vect[]);
    //  if use_perturbation then wee add <perturb_vect> to the vector field of KS
  // on input: \varphi([0,h],x0) \subset xencl
  // On output in  <xstep> - the bounds for \varphi(h,x)_i
  //           obtained from the dissipativity are returned



  int virtual c1dissOneStep(const interval xencl[],interval h,
               interval Vencl[][MAX_DIM+1],interval Vstep[][MAX_DIM+1]);
  // On output: in Vstep  are returned  the bounds for
  //         \frac{\partial \varphi_i}{\partial x}_j(h,x0)
  //           obtained from the dissipativity, x0 is the initial condition
  //            for x, and xencl is such that \varphi([0,h],x0) \subset xencl





   void c1enclosure(interval h,const interval xencl[], interval Vencl[][MAX_DIM+1]);
    // computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  //  i.e starting from identity
  //  this estimate is given  by |Vestm| \leq exp(t*lognorm(df))
  // then we intersect this estimate with Id + [0,h]df(xencl)*Vestm


   KuraShi(int pdim,interval pmu,int prough_encl=ks_mixed_enclosure);




   void writeparameters(ostream &out);
   #ifdef KRAK
     void writeparameters(Frame &out);
   #endif

   double parameter_value(const char par[]);
  // returns the value of the parameter par




   #if defined(DIFF_INCL)
   #if !defined(ENCL_TEST)
   int enclosure_with_tail(interval h,const interval x[],KuraShi_Tail &T0,
            interval enclx[],KuraShi_Tail &T, int Tinitialguess,
            int useappriorigalerror,interval apriorigal_err[],
            int nomessages);
   #else
   int enclosure_with_tail(interval h,const interval x[],KuraShi_Tail &T0,
            interval enclx[],
            KuraShi_Tail &T, int Tinitialguess, int useappriorigalerror,
            interval apriorigal_err[],int nomessages,ostream &report);
   #endif
  // the function used for the dynamic tail evolution
  //  Input: h - time step, x \oplus  T0 - the inital condition
  //         if(useappriorigalerr) then we always intersect locally computed
  //               Galerkin errors with apriorigal_err - USED in some preparatory(test)
  //              computations when set as too large
  //  Output: we find <enclx> and T, such that
  //     if(!useappriorigalerr) \varphi([0,h], x\oplus T0) \subset enclx \oplus T
  //     oterwise we have \varphi([0,h], x\oplus T0) \subset enclx \oplus T
  //                under the additional condition that the local Galerkin
   //               errors
  //                are contained in apriorigal_err
  //     if this is possible then we return 1
  //   otherwise 0 is returned
  //   if Tinitialguess - then inside the function a procedure
  //     building the enclosure for the tail stats with T(input)
  //         otherwise it  starts with T0
  //
    // nomessages=0 -  if the function  fails then the message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in the message to screen
  #endif

  #if defined(DIFF_INCL)
  int move_tail(interval h,interval W2[],KuraShi_Tail &T0,
        KuraShi_Tail &T, KuraShi_Tail &Th);
// moves forward the tail - the resulting tail is Th
// we assume that :
//  T0 - is the initial tail
//   T - is the enclosure for tail for t \in [0,h]
//   W2 - the enclosure for leading variables for t \in [0,h]
// 1- is returned if everything is ok
// 0 - otherwise
  #endif

  #if defined(DIFF_INCL)
  void galerkin_errors(const interval x[],const KuraShi_Tail &T, interval galerr[]);
// computes the Galerkin projection error (proj. dim=kurashi.dim)
// the result is stored in galerr
  #endif

  private:


  void load_lambda_diss_arr(bool coord_diss[],interval arrlambda[]);
   // on output   arrlambda[i]=lambda(i) and (coord_diss[i]=(lambda(i) < 0))

  interval dnonlinearpart(int k, int i,const interval x[]);
// calculates DN_k/Dx_i (x)
//  where f_k= \lambda_k x_k + N_k(x)


  int in_enclmaxiter; // if ==1, then in the enclosure function we iterate
                      // always the same number of times
                      // if==0 the number of iterates in the enclosure functions varies
                     // and the iteration is interupted when the validation is achieved


  void isoimprvvarenclosure(interval h,const interval xencl[],
          interval Vencl[][MAX_DIM+1]);
//  in computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  // hence the initial condition for variational part is V(0)=Id
  // it takes the already validated enclosure Vencl and tries it improve it
  // by using linear inequalities

  bool varmixed_enclosure(interval h,const interval xencl[], interval Vencl[][MAX_DIM+1]);
    // computation of enclosure for variational part
  //  \frac{\partial \varphi}{\partial x} ([0,h],x)=V([0,h],x)
  //// the function uses dissipativity


  void compute_gi_wi(const int &i,const interval lambda[],
          const interval &xi,const interval &h,const interval &N,interval &gi,
          interval &wi);
     // g_i^\pm=(X_i^\pm - b_i^\pm)e^{h\lambda_i} + b_i^\pm
     // w_i=diam(gi)
  interval compute_gi(const interval &h,const interval lambda[],int i,const interval X[],
     const interval y[],bool useperturbation,const interval perturb_vector[]);
  // computes g_i - a bound based on linear inequalites (both for positive and negative \lambda[i]) ;
     //    \varphi_i(h,X) \subset g_i,  provided that y - is an enclosure for \varphi([0,h],X)

  void refineEncl_viaLinUnEq(int maxiter,const interval &h,const interval lambda[],
            const interval X[], interval y[], bool useperturbation,
            const interval perturb_vector[]);
// refines the enclosure using linear inequalities, both for \lambda_i >0 and \lambda_i <0
//  On input: x - initial condition
//            y - enclosure for \varphi([0,h],x)
//  On output: y - a new enclosure
//         returns the value of:  volume(y(output))/volume(y(input))



  void diag_bad_varmixed_enclosure(interval h,const interval xencl[],
           bool coord_dissipative[][MAX_DIM+1]);
// for now just an indication that something went bad




  #if defined(ENCL_TEST)
   int mixodeiso_pert_enclosure(const interval d[],interval h0, interval y[],
               int useperturbation,const interval perturb_vector[],int nomessages,
               ostream& report);
  #else
   int mixodeiso_pert_enclosure(const interval d[],interval h0, interval y[],
               int useperturbation,const interval perturb_vector[],int nomessages);
  #endif
  // if true is returned then:
  //    \varphi([0,h0],d) \subset y
  //   \varphi is defined by x' \in field(x) + perturb_vector if useperturb
  //   and    \varphi is defined by x' = field(x)  if !useperturb
//  h0 - cannot  be  negative if <useperturbation>
 // we want the condition:
   //    y \subset interior(Z)
   //  where: y=d + [0,h]*vfield(Z)( + perturb (if necessary)) on 'non-dissipative coordinates'
   //      and an 'isolation' for the 'dissipative' ones
  // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen

  #if defined(DIFF_INCL)
  void compute_bk_gk(interval h,const interval z[],const KuraShi_Tail &T0,
          const KuraShi_Tail &T, KuraShi_Tail &b,interval gk[]);
// computes the isolation segments over z \oplus T and stores them in  b
// and gk[i] - for   T0.m <=i<= T0.big_m  - used later in the computation
//         of enclosures for T[0,h] and Th
  #endif




  #if defined(DIFF_INCL)
  #if !defined(ENCL_TEST)
     bool validate_fartail(interval h,const KuraShi_Tail &T0,const KuraShi_Tail &T,KuraShi_Tail &b);
  #else
     bool validate_fartail(interval h,const KuraShi_Tail &T0,const KuraShi_Tail &T,KuraShi_Tail &b, ostream &report);
  #endif
  // 1 is returned if T is validated, otherwise 0 is returned.
  #endif

   #if defined(DIFF_INCL)
   #if !defined(ENCL_TEST)
     bool validate_b_fartail(interval h,interval z[],const KuraShi_Tail &T0,const KuraShi_Tail &T);
   #else
     bool validate_b_fartail(interval h,interval z[],const KuraShi_Tail &T0,const KuraShi_Tail &T, ostream &report);
   #endif
   // 1 is returned if T is validated, otherwise 0 is returned.
   #endif


  #if defined(DIFF_INCL)
  #if !defined(ENCL_TEST)
  int validate_tail(interval h,const interval z[],const KuraShi_Tail &T0,KuraShi_Tail &T,
    int gen_new_T);
  #else
  int validate_tail(interval h,const interval z[],const KuraShi_Tail &T0,KuraShi_Tail &T,
    int gen_new_T,ostream &report);
  #endif
  // 1 is returned if T is validated, otherwise 0 is returned.
  // Additionally if gen_new_T , then T is updated as follows : in case it
  //  is validated then we find better T([0,h]) (T is smaller)
  // otherwise we produce the new guess for $T([0,h])$. If !gen_new_T
  // then T is left unchanged.


  void new_cand_tailenclosure(const KuraShi_Tail &T0,KuraShi_Tail &T,const KuraShi_Tail &b,const interval g[],
         int validated,const int i_c_valid[],int fartailvalidated);
  // the function generates a new guess for the enclosure - and on output stores it in T

  #if defined(ENCL_TEST)
  int incl_enclosure_m_var(interval h,const interval x[],KuraShi_Tail &T,interval enclx[],
     int useappriorigalerror,interval apriorigal_err[],int nomessages,ostream &report);
  #else
  int incl_enclosure_m_var(interval h,const interval x[],KuraShi_Tail &T,interval enclx[],
     int useappriorigalerror,interval apriorigal_err[],int nomessages);
  #endif
  // function seeks the enclosure for differential inclusion
  //    x' \in P_m(F(x,T)) for t \in [0,h]; this means that Tail is in T for t \in [0,h]
  // On output: if 1 is returned, then enclx is the desired enclosure
  //    otherwise enclx is unchanged
  // the variable <useappriorigalerror> changes a bit the meaning of the output 1 -
  // namely if it is true, then in the generation of Galerkin errors we take intersection
  //  with locally computed bounds and local bounds - this is very useful in tests
  // runs with large sets
 // nomessages=0 -  if the function  fails then a message is send
  //             to a screen and files
  // nomessages=1 -  the failure does not result in message to screen
  #endif  // DIFF_INCL

};

#if defined(DIFF_INCL)
#if defined(ENCL_TEST)
int ksdiagtailenclosure(interval imu,const interval x[],KuraShi_Tail &T,
                   interval *smax,interval *CL,ostream &report);
#else
int ksdiagtailenclosure(interval imu,const interval x[],
   KuraShi_Tail &T, interval *smax,interval *CL);
#endif
  // this heuristic  function tries find out whether it is possible
  // to find the enclosure for the far tail, for s=T.s
  // over the set defined by a_i \in x[i] for i<=T.m and
  //                         a_i \in T(i) for T.m <i <= T.big_m
  ///////
  // On output:
  //  we update smax and CL, see below for an explation
  //   1 - is returned when the heuristic tells that for C larger than CL
  //     one can hope for the enclosure (in this case also T.s <= smax)
  //   0 - if the tests tells that no enclosure is possible (T.s > smax)
  //   the <smax> is an upper bound for the power 's' it will make
  //   sense to look for the enclosure
#endif



class KS_Exp{
// this class is devised to optimize the computation of exp(h \lambda_k)
// for k such that lambda_k < 0   \lambda_k=k^2(1-mu k^2)
//////////
// the main function e_to_k_l returns a rigorous upper bound for exp(h lambda_k)
  public:
    interval e_to_k_l(interval& h,interval& pmu, int k);
    KS_Exp();
    KS_Exp(int pdimtable);
    ~KS_Exp();
    void inflatetable(int pdim);

  private:
    int par_defined; // tels if h and mu have been defined
    double h; // min (h)=h.left_b();
    double mu; // mu=imu.left_b()
    int tabledim;
    int* inittable;
    double* tabexp;
    void inittabexp(double ph,double pmu);

};




 void KuramotoPoint(double y[],int ldim,int numpt);
 void KuramotoPoint(interval y[],int ldim,int numpt);
  //  numpt=2 or 101 - mu \approx 0.125 - from limit cycle
  //  numpt=3 - attracting fixed point fixed mu=0.125 bimodal stable
  //        31 - a fixed point  mu=0.127 bimodal unstable
  //       5 - an attracting fixed point, mu=0.08 - causing problems
  //          computed in 12D; differes from the one computed in 10D
  //               on third digit
  //       6 - an 'negative' attracting point form the 'giant' branch for
  //           mu=0.06 - may be used as a starting point for the search for
  //           the nearby values
  ///////////
  //  numbers from 3-99 - fixed points
  //               101 - 199 - periodic orbits
  //               201 - 299 - interesting initial cond. for attractor



int chooseksp(double mu);
// the choice of an initial condition for KS-equations
// the returned value should be later feeded to function <KuramotoPoint>
// mu - is suggesting the interesting range of parameters

void KuramotoSection(double c_sec[],int sec_no,int pdim);
void KuramotoSection(interval c_sec[],int sec_no,int pdim);
///// All sections are in 1-3 plane
///// the trajectories rotate in clockwise direction
//  1 - x[1]=0, x'[1] >0,               angle Pi/2
//  2 - x[1]-x[3]=0  ; x'[1]-x[3]' >0   angle Pi/4
//  3 - x[3]=0,   x'[3] <0              angle 0
//  4 - x[1]+x[3]=0, x'[1]+x'[3] <0     angle -Pi/4
//  5 - x[1]=0,  x'[1]<0                      -Pi/2
//  6 - -x[1]+x[3]=0,  -x'[1]+x'[3] >0        -3Pi/4
//  7 - x[3]=0,   x'[3] >0                    -Pi
//  8 - x[1]+x[3]=0, x'[1]+x'[3] >0           -5Pi/4



void KuramotoRadialDir(interval vect[],int numsec,int pdim);
void KuramotoRadialDir(double c_sec[],int sec_no,int pdim);
//  in the plane 1-3 , we produce the radial vector,
//   numsec=1, coresponds  to the angle Pi/2 (upper half of OX axis)
//  and we continue clockwise
//  with an step -Pi/4  for numsec=1,...,8



 void KuramotoColVectDir(double colvect[][MAX_DIM+1],int *numcolvect,
      int sec_no,int pdim,int betweenS);
 // procedure returns directions in which we want the coherent coordinates
  // to point (i.e. to have positive scalar product with them)
  // these directions are given by colvect[1]  and eventually colvect[2]
  // if betweenS - then colvect[1] is the section direction
  //          and colvect[2] is radial direction in the plane (a1,a3)
  //          on output also *numcolvect=2
  // if !betweenS - we set colvect[1] only, as the radial direction
  //    in the plane (a1,a3), in this case *numcolvect=1


// functions converting mu to alpha and vice versa
double mutoalpha(double mu);
double alphatomu(double alpha);
interval mutoalpha(interval mu);
interval alphatomu(interval alpha);




//  symmetry functions

void kssymmetry(interval x[],int dim);
//  the symmetry changing signs on odd modes is applied to <x>

void kssymmetry(interval Dp[][MAX_DIM+1],int dim);
//  the symmetry changing signs on odd modes is applied to the square
//   dim x dim matrix Dp


void kssymmetry(double x[],int dim);
//  the symmetry changing signs on odd modes is applied to <x>

void kssymmetry(double Dp[][MAX_DIM+1],int dim);
//  the symmetry changing signs on odd modes is applied to the square
//   dim x dim matrix Dp

void kssymmmatrix(double R[][MAX_DIM+1],int dim);
// returns the symmetry matrix:
//  the symmetry is changing signs on odd modes


void kssymmmatrix(interval R[][MAX_DIM+1],int dim);
// returns the symmetry matrix:
//  the symmetry is changing signs on odd modes

interval ksl2norm(int normtype,interval a[],int n,interval c, int s);
//  for a set \Pi_{i=1}^n a[k] \times \Pi_{k=n+1}&\infty C/k^s
// representing the functions u=\Sum_{k=1}^{\infty} - 2a_k \sin(kx)
//   this function computes L_2 on (0,2\Pi) if normtype==0
//  or L_2 norm of gradient u, when normtype==1



interval ksuniformnorm(int normtype,interval a[],int n,interval c, int s);
//  for a set \Pi_{i=1}^n a[k] \times \Pi_{k=n+1}&\infty C/k^s
// representing the functions u=\Sum_{k=1}^{\infty} - 2a_k \sin(kx)
//   this function computes C^0 on (0,2\Pi) if normtype==0
//  or C^0 norm of gradient u, when normtype==1

int kssugestedorder(int dim,double mu,double h);
// the function computes a minimal <order> for given times step h - so that
// the numerical scheme becomes stable
// ----------------
// the function is based on the following considerations
//  for equation x'=Lx, where L < 0
//  we look at how many terms are need in the partial sum:
//        exp(Lh)= \sum_{i=0}^n \frac{(Lh)^i}{i!}
//   we want this partial sum + estimate for the remainder (for e^{|Lh|})
//   to be less than 1

int ksstablepar(int dim,double mu,double ih,int iorder,double *newh,int *neworder);
// the function tries to propose the values of the time step (*newh) and order (*neworder)
// which should lead to a stable nonrigorous computation using explicite methods



void ks_input_par(double *mu,int *dimks,bool useinitialval);
  // reads from the console mu and dim for KuraShi
  // if (useinitialval) then the current parameter value are proposed

void ks_input_par(interval *mu,int *dimks);
  // reads from the console mu and dim for KuraShi

void read_dimension(int *dimks,int useinitialval,int maxdim);

#endif
