#ifndef _zgfiles_h_
#define _zgfiles_h_


 #include"defo.h"
 #include"func.h"  // to know class Func;


int fileexists(const char *filename);

int copyfile(const char source[],const char dest[]);
// I don't know why this function does not work under windows (Borland)
// but it works fine under Linux

void readfirstline(const char name[],char line[]);


int readptstable(const char name[],double points[][MAX_DIM+1],int *num_of_pts,
              int pdim,int max_num);
// pdim - is the dimension of <points>, depending on dim_in_file
//                    we truncate points from <name> dim_in_file > pdim
//           or set points[i][pdim >= j > dim_in_file]
// max_num - is the maximum number of points
// it is assumed the the first line in file <name> contains the dimension (dim_in_file)
//  and in the following lines we have coordinates of  points separated by
//       one or more spaces

int readArr2D(const char name[],double tab[][MAX_DIM+1],int *diminfile);
 // reads 2D array from the text file
  // 1st line have to begin with the dimension - that a comment might be put
  // from the second line on we have the array entries

int readVect(const char name[],double x[],int *diminfile);


void delptsfromtable(double ptslist[][MAX_DIM+1],int *num_of_pts,int pts_dim,
           int num_pts_to_del);
// from the table <ptslist> containing <*num_of_pts> points of dimension <pts_dim>
//  we remove first <num_pts_to_del> points;
//  On output also we update <*num_of_pts>
void save_table(const char name[],const double ptslist[][MAX_DIM+1],int num_of_pts,
   int diminfile);
void save_table(ostream& fhandle,const double ptslist[][MAX_DIM+1],int num_of_pts,
   int diminfile);
// we save ptslist[1],...ptslist[num_of_pts]


void sortTab(double points[][MAX_DIM+1],int num_of_pts,int pts_dim,const Func &func);


int  find_pt_inTab(double current_r,const double points[][MAX_DIM+1],int pts_number,
       int pts_dim,const Func &func,double x[],int pdim);
//   in a table <points>  coordinates of <pts_dim>-dimensional vectors
  // we assume that the table <points> is ordered increasingly according
  //    to the value of <func> i.e.  func(points[i]) < func(points[i+1])
  // using table <points> are representing a set parameterized by the
  //  value of func(x) we try solve the equation:
  //    func(x)=current_r,   dim(x)=pdim (we suitabley truncate/expand <points[i]>
  //  1 is returned if success 0 otherwise


int  findClosePtinTab(const double x[],const double points[][MAX_DIM+1],
     int pts_number,int pts_dim,double (*normfunc)(const double[],int),int xdim);
     // dim(x)=xdim
     // dim(points[i])=pts_dim
     // on output we have an index of a point in <points>, which is the closest
     // point to <x>

double normMaxT(const double ptslist[][MAX_DIM+1],int pdim,int num_of_pts);

int addfilespoints(const char dest[],const char name1[],const char name2[],
                     int dim,int bias);
  // function writes to file <dest> points from files <name1>, <name2>
  //   <dim> is the dimension of points (may be different then the one
  //                      in files)
  //   <bias> - number of points on the beginning of both files not included
  //        into <dest>




void wr_interval_center_form(ostream &report, interval x);
// writes to <report> an interval <x> as center + [-d,d]

void writeDiff(ostream &report,const double x1[],const double x2[],int pdim,
        int write_ind =0 , int breakline = 0);
// writes to the report:   x1-x2

void saveAllCoord(const char* report,const double x[],int pdim);

void saveAllCoord(ostream& report,const double xn[],int pdim,int dummy=0 ,
                                     int write_ind =0,int breakline = 0 );
void saveAllCoord(ostream& report,const interval x[],int pdim,int howtowrite =0,int
    write_ind =0 ,int breakline = 0);
// howtowrite == 0 - only the mid-point is written
//                 1 - interval, standard form
//                 2 - interval : center + [-d,d]
//                 3 - in a program format 'x[i]=interval(....,...)
//    if howtowrite ==3 then the other parameters are ignored
// write_ind = 1 - then write an index in parentheses
// breakline !=0 , then we pass to a new row if an index % breakline ==0


void readvector(istream& report,interval xn[],int pdim);
void readvector(istream& report,double xn[],int pdim);

void savesection(ostream& report,const double sec[],int pdim);
void savesection(ostream& report,const interval sec[],int pdim);

void readsection(istream& report,double sec[],int pdim);
void readsection(istream& report,interval sec[],int pdim);

void writediam(ostream& report,const interval xn[],int pdim);
// writes to <report> diameters of each coordinate of <pdim>-dimensional vector

void writediam(ostream& report,const interval Dx[][MAX_DIM+1],int pdim);
// writes to <report> diameters of each entry of <pdim>^2-dimensional table




void writeinterval(ostream &out,const interval &r, int howtowrite);
//  howtowrite=1 - standard interval
//            =2 - center + [-d,d]


void saveArray(ostream& report,const interval Dx[][MAX_DIM+1],int pdim,
   int howtowrite =1,bool write_ij = false);
// howtowrite == 0 - only the mid-point is written
//                 1 - interval, standard form
//                 2 - interval : center + [-d,d]
// if write_ij,  then before each term we write the row and column in parentheses

void saveArray(const char* report,const interval Dx[][MAX_DIM+1],int pdim,
  int howtowrite =1, bool write_ij=false);
// howtowrite == 0 - only the mid-point is written
//                 1 - interval, standard form
//                 2 - interval : center + [-d,d]
// if write_ij, then before each term we write the row and column in parentheses

void saveArray(ostream& report,const double Dx[][MAX_DIM+1],int pdim,
   int dummy=0, bool write_ij = false);

void saveArray(const char* report,const double Dx[][MAX_DIM+1],int pdim,
   int dummy =0, bool write_ij = false);




void readmatrix(istream& report,interval Dx[][MAX_DIM+1]);
void readmatrix(istream& report,double Dx[][MAX_DIM+1]);
void readmatrix(const char *repfname,double Dx[][MAX_DIM+1],int /*pdim*/);

void wreigenval(ostream& report ,const double eigenval[][3],int num);
void wreigenval(const char* name,const double eigenval[][3],int num);

void wrdiag2x2(ostream &report,const double a[][MAX_DIM+1]);
// writes to <report> ; eigenvalues and eigenvectors of <a>
// dim(<a>)=2

void wrdiag2x2(ostream &report,const interval a[][MAX_DIM+1]);
// writes to <report> ; eigenvalues and eigenvectors of <a>
// dim(<a>)=2


void wrmatrix(const double a[][MAX_DIM+1],int dim,const char* fname,
      const char* comment);
//  writes matrix with some comments

void wrmatrix(const double a[][MAX_DIM+1],int dim,ostream& report, const char* comment);
//  writes matrix with some comments

#if defined(BORLAND)
void writetime(ostream &fout);
void writedate(ostream &fout);
#endif
void writedatetime(ostream &fout);
void writeCompInfo(ostream &fout);


void readp(ifstream &tr,int dimintraj,double *t,double p[]);
// the first number in the line is t (time) and then the coordinates follow
int pointfromtrajfile(const char traj[],double xs[],int dim,double bias);
    // reads a point from the trajectory file <traj>
    // <traj> has the following structure
    //  1-line - the dimension of the phase space <dimintraj>
    //  other lines
    //   t(time) a_1 a_2 .... a_{dimintraj}
  /// On output
  //    xs - contains coordinates of the first point for which t>=bias
  //    xs is <dim> dimensional, if dimintraj > dim then we
  //       use only dim coordinates, otherwise
  //        xs[i]=0 for  dimintraj <i <=dim



void   writedivtable(ostream &fout,const int divtable[],int dimdiv);
#ifdef KRAK
  void   writedivtable(FRAME &fout,const int divtable[],int dimdiv);
#endif

int readMapfromFile(const char name[],double points[][3],int *num_of_pts,
               int sizeofpoints);


int writebin(const char* filename,const double p[],int dim);
// writes in binary mode:
//  dim and the vector p[1],\dots,p[dim]

int readbin(const char* filename,double p[],int dim,int *diminfile);
// reads in binary mode:
//  dim and the vector p[1],\dots,p[dim]
//  if the dimension in <filename> is smaller, then
//    p is padded with zeros, otherwise it is truncated


void saveinitialcond(ostream &report,const interval ixb[],const interval r0[],int dim);

void write_compar_quotients(ostream & report,const interval fx[],
       const interval rcomp[],int dim);

void write_empty_intersection(ostream &report,const interval x[],const interval fx[],
      int dim,const char *nx,const char *nfx);


int save_points_bin(const char fbin[],int dim,doublematrix1 p[],int first,int last);
// the structure of the file:
// (int)version=1,(int)dim,(int)number of points,p[first],p[first+1],...,p[last]
// p[i] - is saved as sequence of dim "doubles"
// p[i][1],...,p[i][dim] - the index in p[i] starts from 1
//  version=1 - indicates that we save doubles

int read_points_bin(const char fbin[],int *dim,doublematrix1 **p,int *ptnumber);
// the inverse function to <save_points_bin>
//  on output:
//    *dim - the dimension of points in the file
//    *ptnumber - the number of points
//    (*p)[i] , for i=0,..,*ptnumber-1 - are the <dim>-dimensional points
//                 with the coordinate index starting from 1

void writeratios(ostream &report,const interval n[],const interval d[],int pdim);
// function writes to file the sequence of ratios diameter(n[i])/diameter(d[i] )


interval enterIntervalAsRatio(const char* varname );
//  we return the intrval defined by p/q, where p,q are integers 
//   provided by the user



#endif  /* _zgfiles_h_ */





