
#include<iostream>
using namespace std;

#include<cstdlib>

#include"defo.h"


#include"zgmatrix.h"



void zgmatrixerror(const char* s)
{
  #if defined(WIN95) && defined(KRAK)
    rootFrame.Clear();
  #endif
  cerr << s << endl;
  #if defined(WIN95) && defined(KRAK)
    cerr << "press any key " << endl;
    waitBt();
  #endif
  exit(EXIT_FAILURE);
}

// DOUBLE VERSION
matrix &matrix::operator=(const matrix &a)
{
  if(d1==0){
    d1=a.d1;
    d2=a.d2;
  }
  if((d1!=a.d1) || (d2!=a.d2)){
    zgmatrixerror("wrong dimensions in matrix assignments ");
  }

  for(int i=1;i<=d1;i++)
    for(int j=1;j<=d2;j++) c[i][j]=a.c[i][j];
  return *this;
}

matrix::matrix(int dim1,int dim2)
{
  d1=dim1;
  d2=dim2;
  clear();
}

matrix::matrix()
{
  d1=0;d2=0;
  clear();
}

matrix::matrix(int dim)
{
  d1=dim;
  d2=dim;
  clear();
}


void matrix::clear()
{
  for(int i=1;i<=d1;i++)
    for(int j=1;j<=d2;j++)
      c[i][j]=0.0;
}

int matrix::nonzero()
{
  int ret=0;
  for(int i=1;i<=d1;i++)
    for(int j=1;j<=d2;j++)
      if(fabs(c[i][j]) > 0)  ret=1;
  return(ret);
}

double matrix::trace()
{
  if(d1!=d2){
    zgmatrixerror("matrix::trace called for not square matrix");
  }
  double t=0.0;
  for(int i=1;i<=d1;i++)
    t+=c[i][i];
  return(t);
}


double matrix::determinant()
{
  if(d1!=d2){
    zgmatrixerror("determinant called for an unsquare matrix");
  }
  if(d1 > 2){
    zgmatrixerror("determinant called for a matrix with dimension higher than 2");

  }
  double t=0.0;
  if(d1==1) t=c[1][1];
  if(d1==2){
     t=c[1][1]*c[2][2]-c[1][2]*c[2][1];
  }
  return(t);
}


//  arithmetic operations
matrix operator+(const matrix &a1,const  matrix &a2){
  if((a1.d1!=a2.d1) && (a1.d2!=a2.d2)){

    zgmatrixerror("wrong dimensions in matrix: operator+ ");

  }
  matrix result(a1.d1,a1.d2);
  for(int i=1;i<=a1.d1;i++)
    for(int j=1;j<=a1.d2;j++)
      result.c[i][j]=a1.c[i][j]+a2.c[i][j];
  return result;
}

matrix operator-(const matrix &a1,const  matrix &a2){
  if((a1.d1!=a2.d1) && (a1.d2!=a2.d2)){

    zgmatrixerror("wrong dimensions in matrix2x2:  operator- ");

  }
  matrix result(a1.d1,a1.d2);
  for(int i=1;i<=a1.d1;i++)
    for(int j=1;j<=a1.d2;j++)
      result.c[i][j]=a1.c[i][j]-a2.c[i][j];
  return result;
}

matrix operator*(const matrix &a1,const  matrix &a2)
{
  if(a1.d2!=a2.d1){

    zgmatrixerror("wrong dimensions in matrix: operator* ");

  }
  matrix result(a1.d1,a2.d2);
  for(int i=1;i<=result.d1;i++)
    for(int j=1;j<=result.d2;j++){
      double t=0.0;
      for(int k=1;k<=a1.d2;k++)
        t+=a1.c[i][k]*a2.c[k][j];
      result.c[i][j]=t;
    }
  return result;
}


matrix operator*(double r,const  matrix &a)
{
  matrix result(a.d1,a.d2);
  for(int i=1;i<=result.d1;i++)
    for(int j=1;j<=result.d2;j++){
      result.c[i][j]*=r;
    }
  return result;
}

matrix inverse(const matrix &a)
{
  if(a.d1!=a.d2){
    std::ostringstream errinfo;
    errinfo << "matrix: inverse  - not a square matrix "  << a.d1 << "  " << a.d2
            << endl;
    zgmatrixerror((errinfo.str()).c_str() );
  }
  if(a.d1 > 2){
    std::ostringstream errinfo;
    errinfo  << "Too big dimension in matrix: inverse (<=2 ), here" << a.d1 << endl;
    zgmatrixerror((errinfo.str()).c_str() );
  }
  matrix r(a.d1);
  double det;
  if(r.d1==1) det=a.c[1][1];
  if(r.d1==2) det=a.c[1][1]*a.c[2][2]-a.c[1][2]*a.c[2][1];

  if(det==0.0){

    zgmatrixerror( "Singular matrix: inverse ");

  }

  det=1.0/det;
  if(r.d1==1){
    r.c[1][1]=det;
  }
  if(r.d1==2){
    r.c[1][1]=det*a.c[2][2];
    r.c[1][2]=-det*a.c[1][2];
    r.c[2][1]=-det*a.c[2][1];
    r.c[2][2]=det*a.c[1][1];
  }
  return r;
}

matrix operator-(const matrix &a)
{
  matrix r(a.d1,a.d2);
  for(int i=1;i<=a.d1;i++)
    for(int j=1;j<=a.d2;j++)
      r.c[i][j]=-a.c[i][j];
  return(r);
}


ostream &operator<<(ostream &out,const matrix &a)
{
  for(int i=1; i<=a.d1;i++){
    for(int j=1; j<=a.d2;j++){
      out << a.c[i][j] << "   " ;
    }
    out << endl;
  }
  return out;
}

#if  defined(KRAK)
Frame &operator<<(Frame &f, const matrix &a)
{
  for(int i=1; i<=a.d1;i++){
    for(int j=1; j<=a.d2;j++){
      f << a.c[i][j] << "   " ;
    }
    f << "\n";
  }
  return f;
}
#endif


// INTERVAL VERSION
// conversion between double and interval version
imatrix::imatrix(const matrix& a)
{
  d1=a.d1;
  d2=a.d2;
  for(int i=1;i<=d1;i++)
    for(int j=1;j<=d2;j++)
      c[i][j]=a.c[i][j];

}

imatrix& imatrix::operator=(const matrix& a)
{
  if(d1==0){
    d1=a.d1;
    d2=a.d2;
  }
  if((d1!=a.d1) || (d2!=a.d2)){

    zgmatrixerror("wrong dimensions in imatrix = matrix assignments ");

  }
  for(int i=1;i<=d1;i++)
    for(int j=1;j<=d2;j++)
      c[i][j]=a.c[i][j];
  return *this;
}

imatrix &imatrix::operator=(const imatrix &a)
{
  if(d1==0){
    d1=a.d1;
    d2=a.d2;
  }
  if((d1!=a.d1) || (d2!=a.d2)){

    zgmatrixerror( "wrong dimensions in imatrix  assignments " );

  }

  for(int i=1;i<=d1;i++)
    for(int j=1;j<=d2;j++) c[i][j]=a.c[i][j];
  return *this;
}

imatrix::imatrix(int dim1,int dim2)
{
  d1=dim1;
  d2=dim2;
  clear();
}

imatrix::imatrix()
{
  d1=0;d2=0;
  clear();
}

imatrix::imatrix(int dim)
{
  d1=dim;
  d2=dim;
  clear();
}


void imatrix::clear()
{
  for(int i=1;i<=d1;i++)
    for(int j=1;j<=d2;j++)
      c[i][j]=0.0;
}


int imatrix::nonzero()
{
  int ret=0;
  for(int i=1;i<=d1;i++)
    for(int j=1;j<=d2;j++)
      if(abs(c[i][j]).right_b() > 0)  ret=1;
  return(ret);
}

interval imatrix::trace()
{
  if(d1!=d2){

    zgmatrixerror("trace called for an non-square matrix");

  }
  interval t=0.0;
  for(int i=1;i<=d1;i++)
    t+=c[i][i];
  return(t);
}


interval imatrix::determinant()
{
  if(d1!=d2){
    zgmatrixerror("determinant called for an nonsquare matrix");
  }
  if(d1 > 2){
    zgmatrixerror( "determinant called for a matrix higher than 2");
  }
  interval t=0.0;
  if(d1==1) t=c[1][1];
  if(d1==2){
     t=c[1][1]*c[2][2]-c[1][2]*c[2][1];
  }
  return(t);
}


//  arithmetic operations
imatrix operator+(const imatrix &a1,const  imatrix &a2){
  if((a1.d1!=a2.d1) && (a1.d2!=a2.d2)){

    zgmatrixerror("wrong dimensions in imatrix2x2: operator+ " );
  }
  imatrix result(a1.d1,a1.d2);
  for(int i=1;i<=a1.d1;i++)
    for(int j=1;j<=a1.d2;j++)
      result.c[i][j]=a1.c[i][j]+a2.c[i][j];
  return result;
}

imatrix operator-(const imatrix &a1,const  imatrix &a2){
  if((a1.d1!=a2.d1) && (a1.d2!=a2.d2)){

    zgmatrixerror("wrong dimensions in imatrix2x2:  operator- ");

  }
  imatrix result(a1.d1,a1.d2);
  for(int i=1;i<=a1.d1;i++)
    for(int j=1;j<=a1.d2;j++)
      result.c[i][j]=a1.c[i][j]-a2.c[i][j];
  return result;
}

imatrix operator*(const imatrix &a1,const  imatrix &a2)
{
  if(a1.d2!=a2.d1){

    zgmatrixerror("wrong dimensions in imatrix: operator* ");
  }
  imatrix result(a1.d1,a2.d2);
  for(int i=1;i<=result.d1;i++)
    for(int j=1;j<=result.d2;j++){
      interval t=0.0;
      for(int k=1;k<=a1.d2;k++)
        t+=a1.c[i][k]*a2.c[k][j];
      result.c[i][j]=t;
    }
  return result;
}

imatrix operator*(const interval & r,const  imatrix &a)
{
  imatrix result(a.d1,a.d2);
  for(int i=1;i<=result.d1;i++)
    for(int j=1;j<=result.d2;j++){
      result.c[i][j]*=r;
    }
  return result;
}


imatrix inverse(const imatrix &a)
{
  if(a.d1!=a.d2){
      std::ostringstream errinfo;
      errinfo << "matrix: inverse  - not a square matrix "  << a.d1 << "  " << a.d2
              << endl;
      zgmatrixerror((errinfo.str()).c_str() );
    }
    if(a.d1 > 2){
      std::ostringstream errinfo;
      errinfo  << "Too big dimension in matrix: inverse (<=2 ), here" << a.d1 << endl;
      zgmatrixerror((errinfo.str()).c_str() );
  }


  imatrix r(a.d1);
  interval det;
  if(r.d1==1) det=a.c[1][1];
  if(r.d1==2) det=a.c[1][1]*a.c[2][2]-a.c[1][2]*a.c[2][1];

  if(det.contains(0.0)){
    zgmatrixerror("Singular imatrix: inverse ");
  }

  det=1.0/det;
  if(r.d1==1){
    r.c[1][1]=det;
  }
  if(r.d1==2){
    r.c[1][1]=det*a.c[2][2];
    r.c[1][2]=-det*a.c[1][2];
    r.c[2][1]=-det*a.c[2][1];
    r.c[2][2]=det*a.c[1][1];
  }
  return r;
}

imatrix operator-(const imatrix &a)
{
  imatrix r(a.d1,a.d2);
  for(int i=1;i<=a.d1;i++)
    for(int j=1;j<=a.d2;j++)
      r.c[i][j]=-a.c[i][j];
  return(r);
}


ostream &operator<<(ostream &out,const imatrix &a)
{
  for(int i=1; i<=a.d1;i++){
    for(int j=1; j<=a.d2;j++){
      out << a.c[i][j] << "   " ;
    }
    out << endl;
  }
  return out;
}

#if defined(KRAK)
Frame &operator<<(Frame &f, const imatrix &a)
{
  for(int i=1; i<=a.d1;i++){
    for(int j=1; j<=a.d2;j++){
      f << a.c[i][j] << "   " ;
    }
    f << "\n";
  }
  return f;
}
#endif





////// ARRAY OF MATRIX'es
void Bmatmult(matrix a[][MAX_DIM+1], matrix b[][MAX_DIM+1],
    matrix c[][MAX_DIM+1],int bnum,int bnum2, int bnum3)
//   c=a*b;   dim a= bnum x bnum2
//            dim b= bnum2 x bnum3
{
//  matrix b2[MAX_DIM+1][MAX_DIM+1];
  table_matrix1 *b2;
  b2=new table_matrix1[MAX_DIM+1];
  for(int i=1;i<=bnum;i++)
    for(int j=1;j<=bnum3;j++){
      matrix t;
      t=a[i][1]*b[1][j];
      for(int k=2;k<=bnum2;k++)
        t=t+a[i][k]*b[k][j];
      b2[i][j]=t;
    }

  for(int i=1;i<=bnum;i++)
    for(int j=1;j<=bnum3;j++)  c[i][j]=b2[i][j];

  delete [] b2;
}

void Bmatmult(imatrix a[][MAX_DIM+1], imatrix b[][MAX_DIM+1],
    imatrix c[][MAX_DIM+1],int bnum,int bnum2, int bnum3)
//   c=a*b;   dim a= bnum x bnum2
//            dim b= bnum2 x bnum3
{
//  imatrix b2[MAX_DIM+1][MAX_DIM+1];
  table_imatrix1 *b2;
  b2=new table_imatrix1[MAX_DIM+1];
  for(int i=1;i<=bnum;i++)
    for(int j=1;j<=bnum3;j++){
      imatrix t;
      t=a[i][1]*b[1][j];
      for(int k=2;k<=bnum2;k++)
        t=t+a[i][k]*b[k][j];
      b2[i][j]=t;
    }

  for(int i=1;i<=bnum;i++)
    for(int j=1;j<=bnum3;j++)  c[i][j]=b2[i][j];

  delete [] b2;
}


void Bmatmult(matrix a[][MAX_DIM+1], matrix b[][MAX_DIM+1],
    matrix c[][MAX_DIM+1],int bnum)
//   c=a*b
//  a,b - are square matrices dim=bnum
{
  Bmatmult(a,b,c,bnum,bnum,bnum);
}



void Bmatmult(imatrix a[][MAX_DIM+1], imatrix b[][MAX_DIM+1],
    imatrix c[][MAX_DIM+1],int bnum)
//   c=a*b
//  a,b - are square matrices dim=bnum
{
  Bmatmult(a,b,c,bnum,bnum,bnum);
}


int inverse_uptr(matrix b[][MAX_DIM+1],  matrix d[][MAX_DIM+1],int bnum)
// d=b^{-1}  where <b> is upper diagonal matrix
{
  int ret=1;

//  matrix c[MAX_DIM+1][MAX_DIM+1];
  table_matrix1 *c;
  c=new table_matrix1[MAX_DIM+1];

  for(int i=1;i<=bnum;i++)
    for(int j=1;j<=bnum;j++)
      c[i][j]=matrix(b[i][j].d1,b[i][j].d2);

  // algorithm
  //  1) diagonal elements: c[i][i]=b[i][i]^{-1}
  //  2) over diagonal  c[i,j]=-b[i,i]^{-1}(b[i,i+1]c[i+1,j] + ..+b[ij]c[jj])

  // diagonal elements
  for(int i=1;i<=bnum;i++)
    c[i][i]=inverse(b[i][i]);

  // off diagonal elements
  for(int k=1;k<bnum; k++){
    int i=1;
    int j=i+k;
    while(j<=bnum){
      matrix t(b[i][j].d1,b[i][j].d2);
      for(int l=i+1;l<=j;l++)
        t=t+(b[i][l]*c[l][j]);

      c[i][j]=-c[i][i]*t;
      i++;
      j++;
    }
  }

  for(int i=1;i<=bnum;i++)
    for(int j=1;j<=bnum;j++)
      d[i][j]=c[i][j];

  delete [] c;
  return ret;
}


int inverse_uptr(imatrix b[][MAX_DIM+1],
                 imatrix d[][MAX_DIM+1],int bnum)
// d=b^{-1}  where <b> is upper diagonal imatrix - interval version
{
  int ret=1;

//  imatrix c[MAX_DIM+1][MAX_DIM+1];
  table_imatrix1 *c;
  c=new table_imatrix1[MAX_DIM+1];

  for(int i=1;i<=bnum;i++)
    for(int j=1;j<=bnum;j++)
      c[i][j]=imatrix(b[i][j].d1,b[i][j].d2);

  // algorithm
  //  1) diagonal elements: c[i][i]=b[i][i]^{-1}
  //  2) over diagonal  c[i,j]=-b[i,i]^{-1}(b[i,i+1]c[i+1,j] + ..+b[ij]c[jj])

  // diagonal elements
  for(int i=1;i<=bnum;i++)
    c[i][i]=inverse(b[i][i]);

  // off diagonal elements
  for(int k=1;k<bnum; k++){
    int i=1;
    int j=i+k;
    while(j<=bnum){
      imatrix t(b[i][j].d1,b[i][j].d2);
      for(int l=i+1;l<=j;l++)
        t=t+(b[i][l]*c[l][j]);

      c[i][j]=-c[i][i]*t;
      i++;
      j++;
    }
  }

  for(int i=1;i<=bnum;i++)
    for(int j=1;j<=bnum;j++)
      d[i][j]=c[i][j];

  delete [] c;
  return ret;
}


void upperblocktrangulartoblocks(double a[][MAX_DIM+1],int pdim,
      matrix blocks[][MAX_DIM+1],int *bnum)
//  function transforms <a> - which is block 2x2 upper triangular to
//      <blocks> - which will be upper triangular is strict sense
//   <pdim> - dimension of <a>
//    <*bnum> - number of blocks in <blocks> on output
{
  int tabcblock[MAX_DIM+1];  // tabcblock[i] - tells the coordinate number,
          //  where (i,i)-th block starts

  int i=1;
  *bnum=0;
  while(i<=pdim){
    int block2x2=0;
    if(i<pdim){
      if(fabs(a[i+1][i]) > 0.0){
        block2x2=1;
        (*bnum)++;
        blocks[*bnum][*bnum].d1=2;
        blocks[*bnum][*bnum].d2=2;

        tabcblock[*bnum]=i;
        i=i+2;
      }
    }
    if(!block2x2){
      (*bnum)++;
      blocks[*bnum][*bnum].d1=1;
      blocks[*bnum][*bnum].d2=1;
      tabcblock[*bnum]=i;
      i++;
    }
  }


  for(int bi=1;bi<=*bnum ; bi++)
    for(int bj=1;bj<=*bnum;bj++){
      int ld1=blocks[bi][bi].d1;
      int ld2=blocks[bj][bj].d1;
      blocks[bi][bj].d1=ld1;
      blocks[bi][bj].d2=ld2;

      for(int i1=1;i1<=ld1;i1++)
        for(int j1=1;j1<=ld2;j1++)
          blocks[bi][bj].c[i1][j1]=a[tabcblock[bi]+i1-1][tabcblock[bj]+j1-1];

    }

}


void upperblocktrangulartoblocks(interval a[][MAX_DIM+1],int pdim,
      imatrix blocks[][MAX_DIM+1],int *bnum)
//  function transforms <a> - which is block 2x2 upper triangular to
//      <blocks> - which will be upper triangular is strict sense
//   <pdim> - dimension of <a>
//    <*bnum> - number of blocks in <blocks> on output
{
  int tabcblock[MAX_DIM+1];  // tabcblock[i] - tells the coordinate number,
          //  where (i,i)-th block starts

  int i=1;
  *bnum=0;
  while(i<=pdim){
    int block2x2=0;
    if(i<pdim){
      if(abs(a[i+1][i]).right_b() > 0.0){
        block2x2=1;
        (*bnum)++;
        blocks[*bnum][*bnum].d1=2;
        blocks[*bnum][*bnum].d2=2;

        tabcblock[*bnum]=i;
        i=i+2;
      }
    }
    if(!block2x2){
      (*bnum)++;
      blocks[*bnum][*bnum].d1=1;
      blocks[*bnum][*bnum].d2=1;
      tabcblock[*bnum]=i;
      i++;
    }
  }


  for(int bi=1;bi<=*bnum ; bi++)
    for(int bj=1;bj<=*bnum;bj++){
      int ld1=blocks[bi][bi].d1;
      int ld2=blocks[bj][bj].d1;
      blocks[bi][bj].d1=ld1;
      blocks[bi][bj].d2=ld2;

      for(int i1=1;i1<=ld1;i1++)
        for(int j1=1;j1<=ld2;j1++)
          blocks[bi][bj].c[i1][j1]=a[tabcblock[bi]+i1-1][tabcblock[bj]+j1-1];

    }

}





void blockstostdform(double a[][MAX_DIM+1],int *pdim,
      matrix blocks[][MAX_DIM+1],int bnum)
//  function transforms <blocks> to standard matrix
//    <bnum> - number of blocks in <blocks> on output
//  <a> is on output the standard form of <blocks>
//   <*pdim> - dimension of <a>

{
  int tabcblock[MAX_DIM+1]; // tabcblock[i] - tells the coordinate number,
          //  where (i,i)-th block starts
  *pdim=0;
  for(int i=1;i<=bnum;i++){
    (*pdim)+=blocks[i][i].d1;
    if(blocks[i][i].d1==1)
      tabcblock[i]=*pdim;
    else tabcblock[i]=(*pdim)-1;
  }

  for(int i=1;i<=(*pdim);i++)
    for(int j=1;j<=(*pdim);j++)
      a[i][j]=0.0;


  for(int bi=1;bi<=bnum ; bi++)
    for(int bj=bi;bj<=bnum;bj++){
      int ld1=blocks[bi][bj].d1;
      int ld2=blocks[bi][bj].d2;
      for(int i1=1;i1<=ld1;i1++)
        for(int j1=1;j1<=ld2;j1++)
          a[tabcblock[bi]+i1-1][tabcblock[bj]+j1-1]=blocks[bi][bj].c[i1][j1];
    }

}

void blockstostdform(double a[][MAX_DIM+1],int *rownum,int *colnum,
      matrix blocks[][MAX_DIM+1],int brownum,int bcolnum)
//  function transforms <blocks> to standard matrix
//    <brownum>x<bcolnum> - dimensions of the matrix <blocks>
//  <a> is on output the standard form of <blocks>
//   dimensions of <a>  <rownum> x <colnum>

{
  *rownum=0;
  for(int bi=1;bi<=brownum ; bi++){
    *colnum=0;
    for(int bj=bi;bj<=bcolnum;bj++){
      int ld1=blocks[bi][bj].d1;
      int ld2=blocks[bi][bj].d2;
      for(int i1=1;i1<=ld1;i1++)
        for(int j1=1;j1<=ld2;j1++)
          a[*rownum+i1][*colnum+j1]=blocks[bi][bj].c[i1][j1];
      *colnum+=blocks[bi][bj].d2;
    }
    *rownum+=blocks[bi][1].d1;
  }
}



void blockstostdform(interval a[][MAX_DIM+1],int *pdim,
      imatrix blocks[][MAX_DIM+1],int bnum)
//  function transforms <blocks> to standard matrix
//    <bnum> - number of blocks in <blocks> on output
//  <a> is on output the standard form of <blocks>
//   <*pdim> - dimension of <a>

{
  int tabcblock[MAX_DIM+1]; // tabcblock[i] - tells the coordinate number,
          //  where (i,i)-th block starts
  *pdim=0;
  for(int i=1;i<=bnum;i++){
    (*pdim)+=blocks[i][i].d1;
    if(blocks[i][i].d1==1)
      tabcblock[i]=*pdim;
    else tabcblock[i]=(*pdim)-1;
  }

  for(int i=1;i<=(*pdim);i++)
    for(int j=1;j<=(*pdim);j++)
      a[i][j]=0.0;


  for(int bi=1;bi<=bnum ; bi++)
    for(int bj=bi;bj<=bnum;bj++){
      int ld1=blocks[bi][bj].d1;
      int ld2=blocks[bi][bj].d2;
      for(int i1=1;i1<=ld1;i1++)
        for(int j1=1;j1<=ld2;j1++)
          a[tabcblock[bi]+i1-1][tabcblock[bj]+j1-1]=blocks[bi][bj].c[i1][j1];
    }

}



int inverse_uptr(double a[][MAX_DIM+1],double c[][MAX_DIM+1],
    int pdim)
//  c=a^{-1}
{
  int ret;
//   matrix b[MAX_DIM+1][MAX_DIM+1];
  table_matrix1 *b;
  b=new table_matrix1[MAX_DIM+1];
  int bnum;
  upperblocktrangulartoblocks(a,pdim,b,&bnum);

  ret=inverse_uptr(b,b,bnum);

  if(ret){
    int tdim;
    blockstostdform(c,&tdim,b,bnum);
  }

  delete [] b;
  return(ret);

}


int inverse_uptr(interval a[][MAX_DIM+1],interval c[][MAX_DIM+1],
    int pdim)
//  c=a^{-1}
{
  int ret;
//  imatrix b[MAX_DIM+1][MAX_DIM+1];
  table_imatrix1 *b;
  b=new table_imatrix1[MAX_DIM+1];
  int bnum;
  upperblocktrangulartoblocks(a,pdim,b,&bnum);

  ret=inverse_uptr(b,b,bnum);

  if(ret){
    int tdim;
    blockstostdform(c,&tdim,b,bnum);
  }
  delete [] b;
  return(ret);

}

int diagupbltrian(double a[][MAX_DIM+1],int pdim,
   double eigenvect[][MAX_DIM+1], double eigenval[][MAX_DIM+1])
// diagonalization of upper block triangular matrix <a> - on the diagonal
//    we allow for 2x2-cells with complex eigenvalues
//  a procedure fails when there is an repeating eigenvalue
//  on output k-th column of <eigenvect> is k-th eigenvector of <a>
//          and k-th eigenvalue is given by eigenval[k][1] in case of real eigenvalue
//          for complex eigenvalue  we have a block eigenval[k,k+1][1,2]
//                    such that eigenval[k][1]=alpha,  [k][2]=beta
//                              eigenval[k+1][1]=-beta, eigenval[k+1][2]=alpha
{
  int ret=1;

  double a11ma[MAX_DIM+1][MAX_DIM+1]; // used as upper diagonal block - alpha*Id
  double inva11a[MAX_DIM+1][MAX_DIM+1];  // inverse of a11ma


  for(int i=1;i<=pdim;i++){
    for(int j=1;j<=pdim;j++){
      eigenvect[i][j]=0.0;
      eigenval[i][j]=0.0;
    }
  }


  // diagonal first
  int i=1;
  while(i<=pdim){
    int celldim=1;
    if(i<pdim)
      if(fabs(a[i+1][i])>0) celldim=2;

    if(celldim==1){
      eigenvect[i][i]=1.0;
      eigenval[i][i]=a[i][i];
    }
    if(celldim==2){
       // an  eigenvalue is complex: alpha \pm beta
       //     alpha=trace/2 ; beta = sqrt(-delta)/2; delta=trace^2 - 4*det
       //  eigenvectors: v=(1,(alpha - a11)/a12), w=(0,beta/a12)
       //  Av = alpha * v - beta * w
       //  Aw = beta * v + alpha * w
       double alpha,beta,delta;
       double tr,det;
       tr=a[i][i]+a[i+1][i+1];
       det=a[i][i]*a[i+1][i+1] - a[i][i+1]*a[i+1][i];
       delta=pow(tr,2)-4.0*det;
       alpha=tr/2.0;
       beta=sqrt(-delta)/2;

       //  vector v
       eigenvect[i][i]=1.0;
       eigenvect[i+1][i]=(alpha - a[i][i])/a[i][i+1];

       double normv;
       normv=pow(eigenvect[i][i],2)+pow(eigenvect[i+1][i],2);
       normv=sqrt(normv);

       // vector w
       eigenvect[i][i+1]=0.0;
       eigenvect[i+1][i+1]=beta/a[i][i+1];
       double normw;
       normw=pow(eigenvect[i][i+1],2) + pow(eigenvect[i+1][i+1],2);
       normw=sqrt(normw);


       // normalisation of v and w
       if(normw > normv) normv=normw;
       for(int k=i;k<=i+1;k++)
         for(int l=i;l<=i+1;l++)
           eigenvect[k][l]/=normv;


       eigenval[i][i]=alpha;
       eigenval[i+1][i]=-beta;
       eigenval[i][i+1]=beta;
       eigenval[i+1][i+1]=alpha;

    }
    if(celldim==1)  i++;
    if(celldim==2) i+=2;
  }

  // consider i-th block
  // it is easy to see that the eigenvector will have nonzero coordinates
  // in blocks 1,2..,i;  in part coresponding to block  i - we already
  //     computed them (diagonal part above)
  i=1;
  while(i<=pdim){
    int celldim=1;
    if(i<pdim){
      if(fabs(a[i+1][i]) > 0) celldim=2;
    }
    if(i>1){

      //  creating  (i-1)x(i-1) block from <a>
      //     and substracting  eigenval[i][i]*id
      for(int j=1;j<i;j++){
        for(int k=1;k<i;k++){
          a11ma[j][k]=a[j][k];
          if(j==k) a11ma[j][k]-=eigenval[i][i];
        }
      }


      if(celldim==1){
        // eigenvector has the form:   (x,v)^t ; v - diagonal part already
        //                 computed; we have to find x
        //  a{1..(i-1)}{1..(i-1)}*x + a{1..(i-1}}i *v= aii * x
        //  hence  x= - (a{1(i-1)}{1(i-1)}-aii)^{-1} * a{1(i-1)}i *v

        inverse_uptr(a11ma,inva11a,i-1);

        double lhs[MAX_DIM+1];   // a{1(i-1)}i *v
        for(int j=1;j<=(i-1);j++){
          lhs[j]=-a[j][i]*eigenvect[i][i];
        }

        for(int j=1;j<=(i-1);j++){
          double t=0.0;
          for(int k=1;k<=(i-1);k++){
            t+=inva11a[j][k]*lhs[k];
          }
          eigenvect[j][i]=t;
        }
        // maybe it will be reasonable to normalize this eigenvector

      }
      if(celldim==2){
        // we look for a pair of vectors v=(x1,y1)^t, w=(x2,y2)^t
        //  where x1, x2 - are (i-2)-dimensional
        //        y1,y2  - are 2-dimensional and have been already computed
        //
        //  the cell is 2x2-dimensional with complex eigenvalues
        //        alpha +- i*beta
        //  alpha=eigenval[i][i]=eigenval[i+1][i+1]
        //  beta=eigenval[i][i+1]=-eigenval[i+1][i]
        //   then    a * v = alpha * v - beta *w
        //           a * w = beta *v + alpha *w
        //  this leads to an equation:
        //  a11*x1 + a12 * y1= alpha *x1 - beta *x2
        //  a11*x2 + a12 * y2= beta *x1 + alpha *x2
        //  hence we get
        //  (a11 - alpha)*x1 + beta *x2=  -a12 * y1 (=l1)
        //  (a11 - alpha)*x2 - beta *x1 = - a12 * y2 (=l2)
        //
        //   form the second equation we have:
        //   x1=1/beta *(a11 - alpha)*x2 - 1/beta *l2
        //
        //   after substitution we have:
        //  (a11 - alpha)*(1/beta *(a11 - alpha)*x2 - 1/beta *l2) + beta *x2=l1
        //  (1/beta *(a11-alpha)^2 + beta)* x2 = l1 + 1/beta * (a11-alpha)*l2
        //  if we denote:
        //   c=(1/beta *(a11-alpha)^2 + beta)
        //   l=l1 + 1/beta * (a11-alpha)*l2
        //  then:
        //    c*x2=l  =>   x2=c^{-1}*l
        //   observe that c - is block upper triangular


       // double alpha=eigenval[i][i];
        double beta=eigenval[i][i+1];

        double l1[MAX_DIM+1],l2[MAX_DIM+1];
        //  l1= -a12*y1
        for(int j=1;j<=(i-1);j++){
          double t=0.0;
          for(int k=1;k<=2;k++){
            t-=a[j][i-1+k]*eigenvect[i-1+k][i];
          }
          l1[j]=t;
        }

        //  l2= -a12*y2
        for(int j=1;j<=(i-1);j++){
          double t=0.0;
          for(int k=1;k<=2;k++){
            t-=a[j][i-1+k]*eigenvect[i-1+k][i+1];
          }
          l2[j]=t;
        }


        double c[MAX_DIM+1][MAX_DIM+1],l[MAX_DIM+1];

        // c=(1/beta *(a11-alpha)^2 + beta)
        for(int j=1;j<=(i-1);j++)
          for(int k=1;k<=(i-1);k++){
            double t=0.0;
            for(int s=1;s<=(i-1);s++)
              t+=a11ma[j][s]*a11ma[s][k];

            c[j][k]=t;
          }


        for(int j=1;j<=(i-1);j++){
          for(int k=1;k<=(i-1);k++){
            c[j][k]/=beta;
            if(k==j) c[j][k]+=beta;
          }
        }

        // clearing the matrix c - in case when roundoff errors  produce
        //  something nonzero, where we shall have zero's
        for(int j=1;j<=(i-1);j++){
          for(int k=j+1;k<=(i-1);k++){
            if(k==j+1){
              if(fabs(a[k][j]) == 0.0)  c[k][j]=0.0;
            } else c[k][j]=0.0;

          }
        }


        //   l=l1 + 1/beta * (a11-alpha)*l2
        for(int j=1;j<=(i-1);j++){
          double t=0.0;
          for(int k=1;k<=(i-1);k++){
            t+=a11ma[j][k]*l2[k];
          }
          t/=beta;
          t+=l1[j];
          l[j]=t;
        }


        double x2[MAX_DIM+1],x1[MAX_DIM+1];
        //  x2=c^{-1}*l
        inverse_uptr(c,c,i-1);
        for(int j=1;j<=(i-1);j++){
          double t=0.0;
          for(int k=1;k<=(i-1);k++)
            t+=c[j][k]*l[k];
          x2[j]=t;
        }

        //   x1=1/beta *(a11 - alpha)*x2 - 1/beta *l2
        for(int j=1;j<=(i-1);j++){
          double t=0.0;
          for(int k=1;k<=(i-1);k++)
            t+=a11ma[j][k]*x2[k];
          t-=l2[j];
          t/=beta;
          x1[j]=t;
        }

        for(int j=1;j<=(i-1);j++){
          eigenvect[j][i]=x1[j];
          eigenvect[j][i+1]=x2[j];
        }

      } // end of computations for 2x2 cell
    } // end if (i>1)
    if(celldim==1) i++;
    if(celldim==2) i+=2;
  }  // end while (i<=pdim)


  return ret;
}


