
#include<iostream>
#include<fstream>
using namespace std;

#include<cstdlib>


#include"defo.h"
#include"coordc.h"
#include"multiint.h"
#include"qrdecmp.h"
#include"zgfiles.h"
#include"zset.h"



// #define SWITCH_OLD_VER
//  introduced around 20.V.2008 - using older version of reanrrangement


 // constants enumerating Lohner methods
const int LOHNER1=1;  // just multiplication, interval
const int LOHNER3=3;  // QR - decompositon
const int  LOHNER4=4;  // error from Lipschitz part is represented as A*r_0
                     //   QR-decomposition for rest-term of numerical methods
const int LOHNER5=5;  // as above - but we keep variational matrix like in lohner
                     //  4-th evaluation  (inner enclosure), to avoid wrapping
                     // in this term, any entry in A with a nonzero width is moved to the other
                     //  part of the error
                     //  doubleton + QR
const int  LOHNER33=33;  // mQR : QR - decomposition, taking shape into account
const int  LOHNER53=53;  // doubleton + mQR
const int  LOHNER51=51;  // doubleton + interval




bool doubletonRep(int NumLohnerMethod)
     // returns true when  this is doubleton representation
{
  return ((NumLohnerMethod==LOHNER4) || (NumLohnerMethod==LOHNER5) ||
        (NumLohnerMethod==LOHNER51)  || (NumLohnerMethod==LOHNER53));
}

bool QuboidRep(int NumLohnerMethod)
     // returns true when this a cuboid set
{
   return ((NumLohnerMethod==LOHNER3) || (NumLohnerMethod==LOHNER33));
}

bool usesSmartQR(int NumLohnerMethod)
      // returns true, when the set uses modified QR which
      // takes the shape of r into consideration
{
    return ((NumLohnerMethod==LOHNER33) || (NumLohnerMethod==LOHNER53));
}

bool usesIntervalSet(int NumLohnerMethod)
    // returns true when  the whole or part of the set representation
    //  is an interval set (product of intervals)
{
  return ((NumLohnerMethod==LOHNER1) || (NumLohnerMethod==LOHNER51));
}






//  SET
void Set::setdefaults()
// numLohner - must be allready set when using this function
{
  this->switchingOn=false;
  #ifdef SWITCH_REP_ON
  if(this->doubletonRep())
    this->switchingOn=true;
  #endif
      // if size(r) > bias*size(r0) and switchon
  this->bias=2;

  this->intersect_rep=true;
}


Set::Set()
{
  this->dim=0;
  this->numLohner=LOHNER1;
  this->setdefaults();
}

void Set::newvalues(int pdim,const int pnumLohner,const interval px0[],
      const interval pC0[][MAX_DIM+1],const interval pr0[])
// if pnumLohner==LOHNER1 - xc=px0, r=pr0
// if pnumLohner==LOHNER3x - xc=px0, r=pr0, B=pC0
// if  pnumLohner==LOHNER4 or LOHNER5x - xc=px0, r0=pr0, C=pC0, B=identity, r=0
{
  this->dim=pdim;
  this->numLohner=pnumLohner;



  if(pnumLohner==LOHNER1){
    copyvector(px0,this->xc,this->dim);
    copyvector(pr0,this->r,this->dim);
  }

  if(this->QuboidRep()){
    copyvector(px0,this->xc,this->dim);
    copyvector(pr0,this->r,this->dim);
    copymatrix(pC0,this->B,this->dim);
  }

  if(this->doubletonRep()){
    // initialization of the Lipschitz part   set= pC0*pr0 + ..
    copymatrix(pC0,this->C,this->dim);

    copyvector(px0,this->xc,this->dim);
    copyvector(pr0,this->r0,this->dim);

    //  the initialisation of the QR-part  x=..*r0 + B*r
    identity(this->B,dim);
    zero(this->r,dim);
  }
  if(this->intersect_rep){
    this->lointvencl(this->enclo_intv); // on output enclo_intv=xc + C*r0+B*r (according to the represention)
  }
}


Set::Set(int pdim,int pnumLohner,const interval px0[],const interval pC0[][MAX_DIM+1],
    const interval pr0[],bool pintersect_rep  /*=true */):
          dim(pdim), numLohner(pnumLohner)
    // if pnumLohner==LOHNER1 - xc=px0, r=pr0
    // if pnumLohner==LOHNER3x - xc=px0, r=pr0, B=pC0
    // if  pnumLohner==LOHNER4 or LOHNER5x - xc=px0, r0=pr0, C=pC0, B=identity, r=0
{
  this->setdefaults();    // this must be called after dim and numLohner are set, and before <newvalues>
  this->newvalues(pdim,pnumLohner,px0,pC0,pr0);
  this->intersect_rep=pintersect_rep ;
}



Set::Set(int pdim,const interval px0[],const interval pC0[][MAX_DIM+1],
    const interval pr0[],int pnumLohner /*=LOHNER5*/,bool pintersect_rep  /*=true */):
          dim(pdim), numLohner(pnumLohner)
    // if pnumLohner==LOHNER1 - xc=px0, r=pr0
    // if pnumLohner==LOHNER3x - xc=px0, r=pr0, B=pC0
    // if  pnumLohner==LOHNER4 or LOHNER5x - xc=px0, r0=pr0, C=pC0, B=identity, r=0
{
  this->setdefaults();    // this must be called after dim and numLohner are set, and before <newvalues>
  this->newvalues(pdim,pnumLohner,px0,pC0,pr0);
  this->intersect_rep=pintersect_rep ;
}



Set::Set(int pdim,const interval px0[],int pnumLohner /*=LOHNER5 */,bool pintersect_rep  /*=true */):
        dim(pdim), numLohner(pnumLohner)
    // if pnumLohner==LOHNER1 - xc=center(px0), r=pr0-xc
    // if pnumLohner==LOHNER3xx - xc=center(px0), r=pr0-xc, B=Id
    // if  pnumLohner==LOHNER4 or LOHNER5xx - xc=center(px0), r0=pr0-xc, C=Id, B=identity, r=0
{
  this->setdefaults();   // this must be called after dim and numLohner are set, and before <newvalues>
  this->intersect_rep=pintersect_rep ;

  interval *lxc;
  lxc=new interval[pdim+1];
  interval *lr0;
  lr0=new interval[pdim+1];
  copyvector(px0,lxc,pdim);
  DecomposeIMulti(lxc,lr0,pdim); // px0=lxc + lr0; diam(lxc)=0


  intervalmatrix1 *lC;
  lC=new intervalmatrix1[pdim+1];
  identity(lC,pdim);
  this->newvalues(pdim,pnumLohner,lxc,lC,lr0);

  delete [] lC;
  delete [] lr0;
  delete [] lxc;



}





Set::Set(const Set &s)
// it is a term by term copy, but for the arrays's it
// copy only the necessary terms
{
  this->dim=s.dim;
  this->numLohner=s.numLohner;
  this->switchingOn=s.switchingOn;
  this->bias=s.bias;
  this->intersect_rep=s.intersect_rep;

  copyvector(s.xc,this->xc,s.dim);  // xc
  copyvector(s.r,this->r,s.dim);
  if(s.doubletonRep()){
    copymatrix(s.C,this->C,s.dim);
    copyvector(s.r0,this->r0,s.dim);
  }

  if(!(s.usesIntervalSet())){
    copymatrix(s.B,this->B,s.dim);
  }
  if(s.intersect_rep){
    copyvector(s.enclo_intv,this->enclo_intv,s.dim);
  }
}

Set & Set::operator=(const Set & s) // the assignment/copy operator
// it is a term by term copy, but for the arrays's it
// copy only the necessary terms
{
  this->dim=s.dim;
  this->numLohner=s.numLohner;
  this->switchingOn=s.switchingOn;
  this->bias=s.bias;
  this->intersect_rep=s.intersect_rep;

  copyvector(s.xc,this->xc,s.dim);  // xc
  copyvector(s.r,this->r,s.dim);
  if(s.doubletonRep()){
    copymatrix(s.C,this->C,s.dim);
    copyvector(s.r0,this->r0,s.dim);
  }

  if(!(s.usesIntervalSet())){
    copymatrix(s.B,this->B,s.dim);
  }
  if(s.intersect_rep){
    copyvector(s.enclo_intv,this->enclo_intv,s.dim);
  }
  return *this;
}



int Set::sdim() const
{
  return this->dim;
}

int Set::setnewdim(int pdim)
{
  int olddim=this->dim;
  this->dim=pdim;
  return olddim;
}

int Set::snumLohner() const
{
  return this->numLohner;
}

bool Set::sintersect_rep()   const
{
  return this->intersect_rep;
}


void Set::getsettings(const Set &s)
     // copies numLohner,bias,switchingOn,intersect_rep etc, but <dim> is unchanged
{
  this->numLohner=s.numLohner;
  this->switchingOn=s.switchingOn;
  this->bias=s.bias;
  this->intersect_rep=s.intersect_rep;
}

bool Set::doubletonRep()  const
     // returns true when when this is doubleton representation
{
  return  ::doubletonRep(this->numLohner);
}

bool Set::QuboidRep()  const
     // returns true when this is a cuboid set
{
   return ::QuboidRep(this->numLohner);
}

bool Set::usesSmartQR() const
      // returns true, when the set uses modified QR which
      // takes the shape of r into consideration
{
    return ::usesSmartQR(this->numLohner);
}

bool Set::usesIntervalSet() const
    // returns true when when the whole or part of the set representation
    //  is just interval set (product of intervals)
{
  return ::usesIntervalSet(this->numLohner==LOHNER1);
}


void Set::set_B_identity()
   //  if numLohner!=LOHNER1 we rearange the part B*r, so that B=id
   //         r=interval enclosure of (B(old)*r(old))
   /// this makes sense if r is small when compared of other 'errors'
   //   which will be incorporarated into the term B*r in the next steps
   //  of the evolution, hence having nontrivial B introduces unnnecessary
   // wrapping
{
  if(!(this->usesIntervalSet())){
    multAV(B,r,r,dim);
    identity(B,dim);
  }
  // enclo_intv - stays the same, as this is only a rearrangement
}

void Setinsertcoord(const Set &xonsec,int coord,interval cval,Set &xt)
//  from <xonsec> we create a dim(xonsec)+1 - dimensional set,
// by inserting the <c>-direction  and setting it to <cval>
{
  xt.dim=xonsec.dim+1;
  xt.getsettings(xonsec);

  if(xt.intersect_rep){
    insertcoord(xonsec.enclo_intv,coord,xt.enclo_intv,cval,xt.dim);
  }

  interval xtc=cval;
  interval rcval;
  xtc.split(rcval);

  insertcoord(xonsec.xc,coord,xt.xc,xtc,xt.dim);
  insertcoord(xonsec.r,coord,xt.r,rcval,xt.dim);
  if((xonsec.numLohner!=LOHNER1) && (xonsec.numLohner!=LOHNER51)){
    insertcoord(xonsec.B,coord,xt.B,xt.dim); // we insert  column and row of zeros
    xt.B[coord][coord]=1; // xt.B will be  orthogonal
  }
  if(xonsec.doubletonRep()){
    insertcoord(xonsec.r0,coord,xt.r0,0.0,xt.dim);
    insertcoord(xonsec.C,coord,xt.C,xt.dim); // we insert  column and row of zeros
  }

}


void Set::lointvencl(interval setencl[]) const
//   computes setencl - the interval enclosure of "this" : xc+B*r + C*r0
//              this means we ignore the fact that we may have an 'intersection'
//               representation
{
  for(int i=1;i<=this->dim;i++)
    setencl[i]=xc[i];

  interval rtemp[MAX_DIM+1];

  if((this->numLohner==LOHNER1) || (this->numLohner==LOHNER51)){  // the interval set
                                               // or doubleton + Interval
    for(int i=1;i<=this->dim;i++)
      setencl[i]+=this->r[i];
  }else{
    multAV(this->B,this->r,rtemp,this->dim);
    for(int i=1;i<=this->dim;i++)
      setencl[i]+=rtemp[i];
  }

  if(this->doubletonRep()){
    multAV(this->C,this->r0,rtemp,this->dim);
    for(int i=1;i<=this->dim;i++)
      setencl[i]+=rtemp[i];
  }

}

void Set::intvencl(interval setencl[],int useothercoord,const LinCC *coordch) const
//   computes setencl - the interval enclosure of "this"
//   the result is given in coordinates (*coordch) if <useothercoord>

{

  for(int i=1;i<=this->dim;i++)
    setencl[i]=this->xc[i];

  if(this->usesIntervalSet()){
    for(int i=1;i<=this->dim;i++)
      setencl[i]+=this->r[i];
  }

  if(useothercoord){
    coordch->coordtran(setencl,setencl);
  }
  interval rtemp[MAX_DIM+1];
  if(!this->usesIntervalSet()){
    if(useothercoord){
      // interval tB[MAX_DIM+1][MAX_DIM+1];
      intervalmatrix1 *tB;
      tB=new intervalmatrix1[this->dim+1];
      ArrArr(coordch->intv_m,this->B,tB,this->dim);
      multAV(tB,this->r,rtemp,this->dim);
      delete [] tB;
      tB=NULL;
    } else
      multAV(B,this->r,rtemp,this->dim);
    for(int i=1;i<=this->dim;i++)
      setencl[i]+=rtemp[i];
  }
  if(this->doubletonRep()){
    if(useothercoord){
     // interval tC[MAX_DIM+1][MAX_DIM+1];
      intervalmatrix1 *tC;
      tC=new intervalmatrix1[this->dim+1];
      ArrArr(coordch->intv_m,this->C,tC,this->dim);
      multAV(tC,this->r0,rtemp,this->dim);
      delete [] tC;
      tC=NULL;
    } else
      multAV(this->C,this->r0,rtemp,this->dim);
    for(int i=1;i<=this->dim;i++)
      setencl[i]+=rtemp[i];
  }

  if(this->intersect_rep){
    bool error=false;
    if(useothercoord){
      interval *rt;
      rt=new interval[this->dim+1];
      coordch->coordtran(this->enclo_intv,rt);
      error=!commonpart(setencl,rt,setencl,this->dim);
      delete [] rt;
    }else{
      error=!commonpart(setencl,this->enclo_intv,setencl,this->dim);
    }
    if(error){
      cerr << "serious error in Set::intvencl - an empty intersection " << endl;
      #if defined(WIN95) && defined(KRAK)
      waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
  }
}


void Set::intvencl(interval setencl[]) const
{
  intvencl(setencl,0,(LinCC *)NULL);
}


void Set::add(const interval y[])
  //  x=x + y;
{
  interval xsr[MAX_DIM+1];
  for(int i=1;i<=this->dim;i++){
    xc[i]+=y[i];
    xc[i].split(xsr[i]);
  }
  if(this->usesIntervalSet()){
    for(int i=1;i<=this->dim;i++)
      this->r[i]+=xsr[i];
  } else{
    intervalmatrix1 *bt;
    bt=new intervalmatrix1[this->dim+1];
    transpose(B,bt,this->dim);
    multAV(bt,xsr,xsr,this->dim);
    delete [] bt;
    bt=NULL;
    for(int i=1;i<=this->dim;i++)
      this->r[i]+=xsr[i];
  }
  if(this->intersect_rep){
    for(int i=1;i<=this->dim;i++)
      this->enclo_intv[i]+=y[i];
  }
}

bool  Set::rearrangement(interval encl[])
  //  if numLohner==LOHNER5x and switchingOn then:
     //  if size(r) > bias*size(C*r0) and switchon
     //           then we rearrange (C,r0,B,r)
     //   and we set encl=intvencl() and return <true>
     //  otherwise <encl> is unchanged , and <false> is returned
{
  bool rep_switch=false;
  if(this->switchingOn && this->doubletonRep()){
    interval cr0[MAX_DIM+1];
    multAV(this->C,this->r0,cr0,this->dim);

    #if defined(SWITCH_OLD_VER)

    if(MultiIDiam(this->r,this->dim,NORM_MAX) >
                   this->bias*MultiIDiam(cr0,this->dim,NORM_MAX)){
      rep_switch=true;
      interval newr0[MAX_DIM+1];
      interval Br[MAX_DIM+1];
      if(this->usesIntervalSet())
        copyvector(this->r,Br,this->dim);
      else
        multAV(this->B,this->r,Br,this->dim);

      for(int i=1;i<=this->dim;i++)
        newr0[i]=cr0[i]+Br[i];

      identity(this->C,this->dim);
      if(!this->usesIntervalSet())
        identity(this->B,this->dim);

      copyvector(newr0,this->r0,this->dim);
      zero(this->r,this->dim);
    }
    #else
    if(this->bias <= 1.0){
      if(MultiIDiam(this->r,this->dim,NORM_MAX) >
              this->bias*MultiIDiam(this->r0,this->dim,NORM_MAX)){  // it was:   bias*MultiIDiam(cr0,dim,NORM_MAX)
        rep_switch=true;
        interval newr0[MAX_DIM+1];
        interval Br[MAX_DIM+1];
        if(this->usesIntervalSet())
          copyvector(this->r,Br,this->dim);
        else
          multAV(this->B,this->r,Br,this->dim);

        for(int i=1;i<=this->dim;i++)
          newr0[i]=cr0[i]+Br[i];

        identity(this->C,this->dim);
        if(!this->usesIntervalSet())
          identity(this->B,this->dim);

        copyvector(newr0,this->r0,this->dim);
        zero(this->r,this->dim);
      }
    }else{
      if(MultiIDiam(this->r,this->dim,NORM_MAX) >
                 this->bias*MultiIDiam(cr0,this->dim,NORM_MAX)){
        rep_switch=true;
        intervalmatrix1 *tB=NULL;
        if(!this->usesIntervalSet()){
          tB=new intervalmatrix1[this->dim+1];
          transpose(this->B,tB,this->dim);
          ArrArr(tB,this->C,tB,this->dim);  // tB=B^T*C
          interval tBCr0[MAX_DIM+1];
          multAV(tB,this->r0,tBCr0,this->dim);
          for(int i=1;i<=this->dim;i++)
            this->r0[i]=this->r[i]+tBCr0[i];
          // now r0=r+(B^TC)r0
        }else{
          for(int i=1;i<=this->dim;i++)
            this->r0[i]=cr0[i]+this->r[i];
        }

        if(this->usesIntervalSet()){
          identity(this->C,this->dim);
        }else{
          copymatrix(this->B,this->C,this->dim); // C=B
          identity(this->B,this->dim);
        }

        zero(this->r,dim);

        delete [] tB;
      }
    }
    #endif
    this->intvencl(encl);
    // enclo_intv - stays unchanged because this is only a rearrangement

  }
  return rep_switch;
}

bool Set::test_xc_in_encl(bool write_to_err_log /*= false */)
  // checks whether this->xc is contained in this->intvencl() (the interval
  //   enclosure of this)
  // returns true , when it is contained, false otherwise
  // if <write_to_err_log>  then a report is written to  <zglerrorlogname>
{
  bool retval=true;
  interval* encl_of_this=new interval[this->dim+1];
  this->intvencl(encl_of_this);

  int must_be_in_interior=0;
  retval=MultiIn(this->xc,encl_of_this,must_be_in_interior,this->dim);

  if((!retval) &&   write_to_err_log ){
    ofstream report(zglerrorlogname,ios::app);
    entriesinerrorlog=1;
    report << "-------------------------" << endl;
    writeCompInfo(report);
    writedatetime(report);

    report << "Warning - xc is not contained in the enclosure of the set " << endl;
    report << "Bad coordinates:" << endl;
    for(int i=1;i<=this->dim;i++){
      if(!in(this->xc[i],encl_of_this[i],must_be_in_interior)){
        report << "i=" << i << "  xc[i]=" << this->xc[i] <<
                    "  encl[i]="  << encl_of_this[i]  <<
                       "   diff=" <<  (this->xc[i] -  encl_of_this[i]) << endl;
      }
    }
    report << endl;
    this->savetxt(report);
  }
  delete [] encl_of_this;
  return retval;
}



void Set::savetxt(ostream &report) const
{
  interval *setencl;
  setencl=new interval[this->dim+1];

  report << "dim=" << this->dim << endl;
  report << "num Lohner=" << this->numLohner << endl;


  if(this->switchingOn){
    report << "switching of representation: active ";
    report << "   bias=" << this->bias << endl;
  }

  report << "the representation: " ;
  if(this->numLohner==LOHNER1){
    report << "xc + r ";
  }
  if(this->QuboidRep()){
    report << "xc + B*r ";
  }
  if(this->doubletonRep() && (!this->usesIntervalSet())){
    report << "xc + C*r0 + B*r ";
  }
  if(this->doubletonRep() && this->usesIntervalSet()){
    report << "xc + C*r0 + r ";
  }
  report << endl;
  if(this->usesSmartQR())
    report << "QR-smart used" << endl;
  if(this->intersect_rep){
    report << "and an intersection with <enclo_intv>:" << endl;
    saveAllCoord(report,this->enclo_intv,this->dim,2);
    report << endl;
  }
  report << "xc : " << endl;
  saveAllCoord(report,this->xc,this->dim,1);
  if(this->doubletonRep()){
    report << "C : " << endl;
    saveArray(report,this->C,this->dim);
    report << endl;
    report << "r0 : " << endl;
    saveAllCoord(report,this->r0,this->dim,1);
    report << endl;
    multAV(this->C,this->r0,setencl,this->dim);
    report << "enclosure of C*r0 : " << endl;
    saveAllCoord(report,setencl,this->dim,2);
    report << endl;
  }
  if(!this->usesIntervalSet()){
    report << "B : " << endl;
    saveArray(report,this->B,this->dim);
    report << endl;
  }
  report << "r : " << endl;
  saveAllCoord(report,this->r,this->dim,1);
  report << endl;
  if(!this->usesIntervalSet()){
    multAV(this->B,this->r,setencl,this->dim);
    report << "enclosure of B*r : " << endl;
    saveAllCoord(report,setencl,this->dim,2);
    report << endl;
  }


  this->intvencl(setencl);

  report << "interval enclosure: " ;
  if(this->intersect_rep)
    report << " enclo_intv[] \\cap eval(Lohner set)  ";
  report << endl;

  saveAllCoord(report,setencl,this->dim,1);
  report << endl;

  delete [] setencl;
}



bool Set::intersect_with_interval(interval y[],bool seldir[])
// we intersect the set <this> with the set \{z | z_i \in y_i, for seldir[i] \}
//  we already know that y[i] is contained in enclosure(this)[i] for seldir[i]
// On output:   this(input) \cap y \subset this(output)
{

  //  independent of the current value of <intersect_rep> - this function deals only
  //  Lohner-part of the set

  // we make changes for those "i" for which seldir[i]==true
  // in the doubleton representation, we zero i-th row in the C-matrix
  //          and r0 - stays the same
  //    this means that B*r-part gives correct size in this direction

  bool ok=true;
  interval *xbar, *delta;
  xbar=new interval[this->dim+1];
  delta=new interval[this->dim+1];
  for(int i=1;i<=this->dim;i++){
    if(!(seldir[i])) xbar[i]=0;
    else xbar[i]=y[i];
  }
  DecomposeIMulti(xbar,delta,this->dim);

  if(this->usesIntervalSet()){ // x=xc +r (LOHNER1 or  (LOHNER51) x=xc+C*r0 + r
    for(int i=1;i<=this->dim;i++){
      if(seldir[i]){
        this->xc[i]=xbar[i];
        this->r[i]=delta[i];
      }
    }
  }else{
    for(int i=1;i<=this->dim;i++){
      if(seldir[i]){
        this->xc[i]=xbar[i];
      }
    }

    //  the part B*r
    intervalmatrix1 *M;
    M=new intervalmatrix1[this->dim+1];
    copymatrix(this->B,M,this->dim);
    for(int i=1;i<=this->dim;i++){
      if(seldir[i]){
        for(int j=1;j<=this->dim;j++)
          M[i][j]=0;
      }
    }

    // we now have to find   B*r(new), such that M*r(old) + delta \subset B*r(new)
    QR_rearrange(this->dim,M,this->r,delta,this->B,this->r,this->usesSmartQR());

    delete [] M;
    //// end of computation of B*r

  }

 // the Lipschitz part
  if(this->doubletonRep()){
    for(int i=1;i<=this->dim;i++){
      if(seldir[i]){
        for(int j=1;j<=this->dim;j++)
          this->C[i][j]=0;
      }
    }
  }


  delete [] delta;
  delete [] xbar;
  return ok;
}


bool Set::rear_cmpart(const interval intvset[])
//  returns true - when  the intersection of set <this> and <intvset> is non-empty
//          false - otherwise
// on output: if true is returned, then  this(input) \cap intvset \subset this(output)

{
  //  independent of the current value of <intersect_rep> - this function deals only
  //  Lohner-part of the set

  interval* xencl=new interval[this->dim+1];
  this->lointvencl(xencl);  // xencl -  is the evaluation of xc +C*r0+Br in Set
                        // or something equivalent depending on the this->snumLohner()
  interval *y;
  y=new interval[this->dim+1];
  bool *seldir;
  seldir=new bool[this->dim+1];
  int i=1;
  bool ok=true;
  bool intersect=false;
  while((i<=this->dim) && ok){
    if(!(xencl[i].subset(intvset[i])) ){
      seldir[i]=true;
      intersect=true;
      ok=commonpart(xencl[i],intvset[i],y+i);
    }else{
      seldir[i]=false; // we don't need to look at the intersection in this direction
    }
    i++;
  }
  if(intersect){
    ok=this->intersect_with_interval(y,seldir);
  }

  delete [] seldir;
  delete [] y;
  delete [] xencl;
  return ok;
}





bool Set::intersect_interval(const interval intvset[])
    //  returns true - when  the intersection of set <this> and <intvset> is non-empty
   //          false - otherwise
   // on output: if true is returned, then  this(input) \cap intvset \subset this(output)
{
  bool  ret=true;
  if(this->intersect_rep){
    ret=commonpart(this->enclo_intv,intvset,this->enclo_intv,this->dim);
    if(ret){
      int in_interior=0;
      if(!MultiIn(this->xc,this->enclo_intv,in_interior,this->dim)){ // this means that the estimates
                           // from the 'Lohner' part are quite large and xc[i]is not contained
                           //  enclo_intv[i], this happens when we integrate dissipative
                           //  ODE with relatively large  step (which violate the stability
                           //  of the explicit numerical scheme)
        ret=this->rear_cmpart(this->enclo_intv);
      }
    }
  }else{
    ret=this->rear_cmpart(intvset);
  }

  return ret;
}

void Set::get_Cmatrix(interval Ccopy[][MAX_DIM+1]) const
     //  Ccopy=this->c;
{
   copymatrix(this->C,Ccopy,this->dim);
}

void Set::get_r(interval rcopy[]) const
{
  copyvector(this->r,rcopy,this->dim);
}
void Set::get_r0(interval rcopy[]) const
{
  copyvector(this->r0,rcopy,this->dim);
}

void Set::get_xc(interval xccopy[]) const
{
  copyvector(this->xc,xccopy,this->dim);
}

void Set::get_Bmatrix(interval Bcopy[][MAX_DIM+1]) const
{
   copymatrix(this->B,Bcopy,this->dim);
}



bool Set::get_switchingOn() const
{
  return this->switchingOn;
}

//////// END SET ////////////////


//  C1Set
C1Set::C1Set(int pdim,int pnumLohner):dim(pdim),numLohner(pnumLohner)
// the initialization with identity
{
  this->currnumLo=LOHNER1;
  identity(this->Vc,this->dim);
  zero(this->Vr,this->dim);
  this->bias=0.05;  // this is rather arbitrary and should be changed later
  if(this->numLohner!=LOHNER1){
    this->currnumLo=LOHNER3;
    identity(this->Vb,this->dim);
/*    zero(this->Ve,this->dim);  // Ve and V0 are not used at the beginning
    zero(this->V0,this->dim); */
  }
}

bool C1Set::rearrangement(interval dencl[][MAX_DIM+1])
// switches the internal representation from LOHNER3 to LOHNER5
//  on output: dencl contains an enclosure of current set in case
//   is we make any change
//  returns <true> iff the switch occurred
{
  bool rep_changed=false;
  if((numLohner==LOHNER5) && (numLohner!=currnumLo)){
    intvencl(dencl);
    if(diameter(dencl,dim)>bias){
      rep_changed=true;
      currnumLo=LOHNER5;
      copymatrix(Vr,V0,dim);

   //   interval temp[MAX_DIM+1][MAX_DIM+1];
      intervalmatrix1 *temp;
      temp=new intervalmatrix1[dim+1];
      copymatrix(Vb,Ve,dim);
      split(Ve,temp,dim); // now Ve=m(Vb) and temp=Vb - Ve

      ArrArr(temp,Vr,Vr,dim);  /// Vr=(Vb-Ve)*Vr
      delete [] temp;
      temp=NULL;
      identity(Vb,dim);
      intvencl(dencl);
    }
  }
  return  rep_changed;
}


C1Set::C1Set(const C1Set &s)  // the copying constructor
{
  dim=s.dim;
  numLohner=s.numLohner;
  currnumLo=s.currnumLo;
  bias=s.bias;
  copymatrix(s.Vc,Vc,s.dim);
  copymatrix(s.Vr,Vr,s.dim);
  if(s.currnumLo!=LOHNER1){
    copymatrix(s.Vb,Vb,s.dim);
    if((s.currnumLo==LOHNER4) || (s.currnumLo==LOHNER5)){
      copymatrix(s.Ve,Ve,s.dim);
      copymatrix(s.V0,V0,s.dim);
    }
  }
}


C1Set & C1Set::operator=(const C1Set & s) // the assignment/copy operator
{
 dim=s.dim;
  numLohner=s.numLohner;
  currnumLo=s.currnumLo;
  bias=s.bias;
  copymatrix(s.Vc,Vc,s.dim);
  copymatrix(s.Vr,Vr,s.dim);
  if(s.currnumLo!=LOHNER1){
    copymatrix(s.Vb,Vb,s.dim);
    if((s.currnumLo==LOHNER4) || (s.currnumLo==LOHNER5)){
      copymatrix(s.Ve,Ve,s.dim);
      copymatrix(s.V0,V0,s.dim);
    }
  }
  return *this;
}


void C1Set::intvencl(interval setencl[][MAX_DIM+1],int useothercoord,const LinCC *coordch) const
//   computes set_encl - the interval enclosure of set
//   the result is given in coordinates (*coordch) if <useothercoord>
{
  copymatrix(Vc,setencl,dim);
  if(currnumLo==LOHNER1){
    for(int i=1;i<=dim;i++)
      for(int j=1;j<=dim;j++)
        setencl[i][j]+=Vr[i][j];
  }

  if(useothercoord){
    transLinearPart(setencl,*coordch,setencl);
  }
  if(currnumLo!=LOHNER1){
    // interval rtemp[MAX_DIM+1][MAX_DIM+1];
    intervalmatrix1 *rtemp;
    rtemp=new intervalmatrix1[dim+1];
    if(useothercoord){
     // interval tB[MAX_DIM+1][MAX_DIM+1];
      intervalmatrix1 *tB;
      tB=new intervalmatrix1[dim+1];
      ArrArr(Vb,Vr,tB,dim);
      transLinearPart(tB,*coordch,rtemp);
      delete [] tB;
      tB=NULL;
    } else
      ArrArr(Vb,Vr,rtemp,dim);

    for(int i=1;i<=dim;i++)
      for(int j=1;j<=dim;j++)
        setencl[i][j]+=rtemp[i][j];

    if((currnumLo==LOHNER4) || (currnumLo==LOHNER5)){
      if(useothercoord){
       // interval tC[MAX_DIM+1][MAX_DIM+1];
        intervalmatrix1 *tC;
        tC=new intervalmatrix1[dim+1];
        ArrArr(Ve,V0,tC,dim);
        transLinearPart(tC,*coordch,rtemp);
        delete [] tC;
        tC=NULL;
      } else
        ArrArr(Ve,V0,rtemp,dim);
      for(int i=1;i<=dim;i++)
        for(int j=1;j<=dim;j++)
        setencl[i][j]+=rtemp[i][j];
    }
    delete [] rtemp;
    rtemp=NULL;
  }
}

void C1Set::intvencl(interval setencl[][MAX_DIM+1]) const
{
  intvencl(setencl,0,(LinCC *)NULL);
}

// end C1Set


void info_err_commonpart(const interval xencl[],const interval intvset[],int dim,
           const char comment[])
{
  ofstream report(zglerrorlogname,ios::app);
  entriesinerrorlog=1;
  report << "-------------------------" << endl;
  writeCompInfo(report);
  writedatetime(report);

  report << comment << endl;
  report << "sets causing problems " << endl;
  report << "xencl:";
  saveAllCoord(report,xencl,dim,1);
  report << endl;

  report << "intvset:";
  saveAllCoord(report,intvset,dim,1);
  report << endl;


  interval *cp;
  cp=new interval[dim+1];
  if(!commonpart(xencl,intvset,cp,dim)){
    report << "an empty intersection of encl(x) and intvset in [commonpart_Set_interval]" << endl;


    report << "i   xencl[i]   intvset[i]   intersection not empty " << endl;
    for(int i=1;i<=dim;i++){
      report << i << "   " << xencl[i] <<  "   " << intvset[i] << "   ";
      interval cp;
      if(commonpart(xencl[i],intvset[i],&cp)){
        report << "not empty " << cp ;
      } else{
        report << "empty intersection " ;
      }

      report << endl;
    }


  }else{
    report << "the intersection is not empty " <<endl;
  }
  delete [] cp;
}




void test_set_intersection()
{
  int dim=3;
  intervalmatrix1 *B;
  B=new intervalmatrix1[dim+1];
  interval *r,*setencl, *xc, *intvset;
  r=new interval[dim+1];
  setencl=new interval[dim+1];
  xc=new interval[dim+1];
  intvset=new interval[dim+1];

  zero(xc,dim);

  zero(B,dim);
  int i=1;
  for(int j=1;j<=dim;j++){
    B[j][i]=1/sqrt(interval(3));
  }
  B[1][2]=1/sqrt(interval(2));
  B[2][2]=-1/sqrt(interval(2));

  B[1][3]=1/sqrt(interval(6));
  B[2][3]=1/sqrt(interval(6));
  B[3][3]=-2/sqrt(interval(6));

  int ortho_test=test_orthogonal(B,dim);

  r[1]=interval(-1,1);
  r[2]=interval(-1,1);
  r[3]=interval(-1,1);


  Set s(dim,LOHNER3,xc,B,r);
  s.intvencl(setencl);

  copyvector(setencl,intvset,dim);
  intvset[3]=setencl[3]/2 + 0.2*Right(setencl[3]);

  char fname[80];
  strcpy(fname,catwork);
  strcat(fname,"zset.txt");
  ofstream report(fname,ios::app);
  report << "============================" << endl;
  writeCompInfo(report);
  writedatetime(report);
  s.savetxt(report);

  if(!ortho_test)
    report << "BAD B - apparently not orthogonal " << endl;


  report << "intvset " << endl;
  saveAllCoord(report,intvset,dim,1);

  if(s.intersect_interval(intvset)){
    report << "After the intersecton: " << endl;
    s.savetxt(report);
  } else{
    report << "Empty intersection" << endl;
  }





  delete [] intvset;
  delete [] xc;
  delete [] r;
  delete [] setencl;
  delete [] B;

  cout << "report is in " << fname << endl;
  #if defined(KRAK) && defined(WIN95)
  waitBt();
  #endif
}


void LohnerMove(const Set &x,const interval xnxc[],intervalmatrix1* A, Set &xn)
// This function is used in the following context, we compute an image of f
// (say a timestep h map of x or the coordinate change ,
// on input:
//    the value f(x.xc),   hence it is an interval vector
//                                            with nonzero width
//   A= is the Jacobian matrix for f(x) )
//         x - the input set
//
// on output: xn = xnxc + A*(x - x.xc) (in the representation defined in <x>)
{
  xn.getsettings(x);
  xn.dim=x.dim;

  if(xn.intersect_rep){ // we compute how the interval enclosure, which is contained
                       //  in x.enclo_intv is evolving - this is independent of
                       //  the dubleton or quboid representation
    interval *er;
    er=new interval[x.dim+1];
    for(int i=1;i<=x.dim;i++){
      er[i]=x.enclo_intv[i]-x.xc[i];
    }

    multAV(A,er,er,xn.dim);  // er(output)=A*er(input)
    for(int i=1;i<=xn.dim;i++)
      xn.enclo_intv[i]=xnxc[i]+er[i];

    delete [] er;
  }

  interval z[MAX_DIM+1];

  copyvector(xnxc,xn.xc,xn.dim);
  DecomposeIMulti(xn.xc,z,xn.dim); // after this diam(xn.xc)=0 , discretization errors
                   // and round-off errors are moved to z


  intervalmatrix1 *M;
  M=new intervalmatrix1[xn.dim+1];

  interval rtemp[MAX_DIM+1];
  if (x.numLohner==LOHNER1){
    multAV(A,x.r,rtemp,xn.dim);
    for(int i=1;i<=xn.dim; i++)
      xn.r[i]=rtemp[i]+z[i];
  }


  if (x.doubletonRep()){
    copyvector(x.r0,xn.r0,xn.dim);
    ArrArr(A,x.C,xn.C,xn.dim);
  }

  if ((x.numLohner!=LOHNER4) && (x.doubletonRep())){
    for(int i=1;i<=xn.dim;i++)
      for(int j=1;j<=xn.dim;j++){
        M[i][j]=xn.C[i][j] - mid(xn.C[i][j]);
        xn.C[i][j]=mid(xn.C[i][j]);
      }
    multAV(M,x.r0,rtemp,xn.dim);
    for(int i=1;i<=xn.dim;i++)
      z[i]+=rtemp[i];
  }
  // z - contains 'error' from the computation of xn + (eventually) error from
         // from the propagation of r0 (for Lohner5- method)

  if(x.numLohner==LOHNER51){ // doubleton + interval
    multAV(A,x.r,rtemp,xn.dim);
    for(int j=1;j<=xn.dim;j++)
      xn.r[j]=z[j]+ rtemp[j];
  }

  if(!x.usesIntervalSet()){ // not (doubleton + IntervalSet or just IntervalSet)
    // QR -decomposition; computation of  xn.r, xn.B

    ArrArr(A,x.B,M,xn.dim);  // M=A*x.B

    // now our task is to find Bn*rn, Bn-orthogonal, such that:  M*x.r + z \subset Bn*rn
    bool take_shape_r_into_consideration=x.usesSmartQR();

    QR_rearrange(x.dim,M,x.r,z,xn.B,xn.r,take_shape_r_into_consideration);


  }
  delete [] M;

  // in the case of "intersection" representation
  // we intersect <enclo_intv> with the interval enclosure of xc+C*r0+B*r
  if(xn.intersect_rep){
    interval *loencl;
    loencl=new interval[xn.dim+1];
    xn.lointvencl(loencl);
    if(!commonpart(xn.enclo_intv,loencl,xn.enclo_intv,xn.dim)){

      ofstream report(zglerrorlogname,ios::app);
      entriesinerrorlog=1;
      report << "-------------------------" << endl;
      writeCompInfo(report);
      writedatetime(report);

      report << " LohnerMove - an empty intersection of enclosure(xc+C*r0+..) and enclo_intv " << endl;
      report << "Set x:" << endl;
      x.savetxt(report);
      report << endl;
      report << "Set xn:" << endl;
      xn.savetxt(report);
      report << endl;
      report.flush();


      cerr << "serious error in LohnerMove - an empty intersection " << endl;
      #if defined(WIN95) && defined(KRAK)
      waitBt();
      #endif
      exit(EXIT_FAILURE);
    }
    delete [] loencl;
  }
}




void transformSet(const Set &x,const LinCC *pcoord,int inverse,Set &xn)
// on output if inverse xn=pcoord^{-1}(x)
//           else  xn=pcoord(x)
{
  xn.getsettings(x);
  xn.setnewdim(x.sdim());
  if(pcoord){

    if(pcoord->dim!=x.sdim()){
      wrongdiminfo(pcoord->dim,x.sdim(),"transformSet");
      exit(EXIT_FAILURE);
    }

    interval *xnxc;
    xnxc=new interval[x.sdim()+1];
    if(inverse){
      pcoord->coordinvtran(x.xc,xnxc);
      LohnerMove(x,xnxc,pcoord->intv_mInv,xn);
    }else{
      pcoord->coordtran(x.xc,xnxc);
      LohnerMove(x,xnxc,pcoord->intv_m,xn);
    }
    delete [] xnxc;
  }else{
    xn=x;
  }
}


char *LohnerSetDescription( int nLohner,char *name,int lenname)
{
  char lname[80];
  strcpy(lname,"unknown set description");

  // jak jest to wypisywanie to dziala dobrze - jak nie to  w domu nie
           // wypisuje unknown set dla pozycji LOHNER33, LOHNER4, LOHNER5

 // cout << "nLohner=" << nLohner << endl;




 // similarly it works fine when I write something, like "wyb 3" below



  if(nLohner==LOHNER1 ){
    strcpy(lname,"interval set, naive interval evaluation");
  }
  if(nLohner==LOHNER3 ){
    strcpy(lname,"cuboid set, QR-decomposition");
 //   cout << "wyb 3  " << endl;
  }
  if(nLohner==LOHNER33 ){
    strcpy(lname,"cuboid set, shape-modified QR-decomposition");
  //  cout << "wyb 33  " << endl;
  }
  if(nLohner==LOHNER4 ) {
    strcpy(lname,"doubleton + cuboid set,  Lipschitz part 'thick'");
  //  cout << "wyb 4  " << endl;
  }
  if(nLohner==LOHNER5 ){
    strcpy(lname,"doubleton + cuboid set,  standard last Lohner approach");
  //  cout << "wyb 5  " << endl;
  }
  if(nLohner==LOHNER51 ){
    strcpy(lname,"doubleton + interval set");
  //  cout << "wyb 51  " << endl;
  }
  if(nLohner==LOHNER53 ){
    strcpy(lname,"doubleton + cuboid set, shape-modified QR-decomposition");
   // cout << "wyb 53  " << endl;
  }

 // cout << lname << endl;
  strncpy(name,lname,lenname-1);
  return name;
}


int whichSetTypeChose()
{
  int numLohner=LOHNER5;
  int sel;


  const int maxlen=80;
  char buffer[maxlen];

// dodalem ponizej to dodatkowe wypisywanie LOHNERxx - bo dla pozycji 3,4,5 zwracalo mi unknown set

  cout << "chose the Lohner algorithm and set representation: \n";
  cout << "1 - " << LohnerSetDescription(LOHNER1 ,buffer,maxlen) << " \n";
  cout << "2 - " << LohnerSetDescription(LOHNER3 ,buffer,maxlen) << " \n";
  cout << "3 - " << LOHNER33 << "  "<< LohnerSetDescription(LOHNER33 ,buffer,maxlen) << " \n";
  cout << "4 - " << LOHNER4 << "  " << LohnerSetDescription(LOHNER4 ,buffer,maxlen) << " \n";
  cout << "5 - "  << LOHNER5 << "  "<< LohnerSetDescription(LOHNER5 ,buffer,maxlen) << " \n";
  cout << "6 - " << LohnerSetDescription(LOHNER51 ,buffer,maxlen) << " \n";
  cout << "7 - " << LohnerSetDescription(LOHNER53 ,buffer,maxlen) << " \n";
  cin >> sel;
  #if defined(WIN95) && defined(KRAK)
    cout << endl;
  #endif
  switch(sel){
  case 1: numLohner=LOHNER1; break;
  case 2: numLohner=LOHNER3; break;
  case 3: numLohner=LOHNER33; break;
  case 4: numLohner=LOHNER4; break;
  case 5: numLohner=LOHNER5; break;
  case 6: numLohner=LOHNER51; break;
  case 7: numLohner=LOHNER53; break;
  }
  return numLohner;
}


int whichC1SetTypeChose()
{
  int numLohnerV=LOHNER5;
  int sel;
  cout << "chose Lohner algorithm for variational part : press 1,3,5 " ;
  cin >> sel;
  #if defined(WIN95) && defined(KRAK)
    cout << endl;
  #endif
  switch(sel){
  case 1: numLohnerV=LOHNER1; break;
  case 3: numLohnerV=LOHNER3; break;
  case 5: numLohnerV=LOHNER5; break;
  }
  return numLohnerV;
}


//#define TESTY

bool projectionOfTheSet(const Set &source,int dimTarget,Set & target,interval rem[] )
  // this function produces target and rem, such that
  //     source \subset target + rem[dimTarget+1,...,source.dim]
  //  It is required that: dimTarget <= source.dim
{
   bool ret=true;
   if(dimTarget > source.dim){
     ret=false;
     cerr << "projectionOfTheSet - error: dimTarget > source.dim" << endl;
      #if defined(KRAK) && defined(WIN95)
                             waitBt();
     #endif
     exit(EXIT_FAILURE);
   }
   source.intvencl(rem);  // only coordinates i > dimTarget - matter in rem


   #if defined(TESTY)
   cout << "wejscie do projectionOfTheSet: source.xc[1]=" << source.xc[1] << endl;
   #endif

   interval* txc=new interval[source.dim+1];
   interval* tr=new interval[source.dim+1];

   copyvector(source.xc,txc,dimTarget);  // xc

   #if defined(TESTY)
   cout << "TROP 3 - w projectionOfTheSet: source.xc[1]=" << source.xc[1] << endl;
   #endif



   #if defined(TESTY)
   cout << "TROP 2 - w projectionOfTheSet: source.xc[1]=" << source.xc[1] << endl;
   #endif


   //   the projection is given as follows
   //  P(xc+A*r0 + Br) \subset [P(xc) + PA((I-P)r0) + P(Br)] + P(A) * P(r0)
   //    the part [P(xc) + PA((I-P)r0) + P(Br)] - should be split into a new
   //  xc and r, with B=Id
   interval* temp=new interval[source.dim+1];
   if(source.usesIntervalSet()){  // we have reprezention with B=Id
     for(int i=1;i<=dimTarget;i++)
       txc[i]+=source.r[i];
   }else{   // we use the cuboid or doubleton with cuboid
     multAV(source.B,source.r,temp,source.dim);
     for(int i=1;i<=dimTarget;i++)
       txc[i]+=temp[i];
   }
   delete [] temp;

   if(source.doubletonRep()){
     for(int i=1;i<=dimTarget;i++){
       for(int j=dimTarget+1;j<=source.dim;j++)
         txc[i]+=source.C[i][j]*source.r0[j];
     }
   }
   DecomposeIMulti(txc,tr,dimTarget);

   #if defined(TESTY)
   cout << "TROP 1: w projectionOfTheSet: source.xc[1]=" << source.xc[1] << endl;
   #endif


   target.getsettings(source);    // copies numLohner,bias,switchingon etc , but <dim> is unchanged
   target.dim=dimTarget;
   copyvector(txc,target.xc,target.dim);
   copyvector(tr,target.r,target.dim);
   if(!target.usesIntervalSet())
     identity(target.B,target.dim);

   if(target.doubletonRep()){
     copymatrix(source.C,target.C,target.dim);
     copyvector(source.r0,target.r0,target.dim);
   }


   if(target.intersect_rep)
     copyvector(source.enclo_intv,target.enclo_intv,target.dim);



   delete [] tr;
   delete [] txc;


   #if defined(TESTY)
   cout << "przed wyjsciem z projectionOfTheSet: source.xc[1]=" << source.xc[1] << endl;
   #endif

   return ret;
}
