#ifndef _zset_h_
#define _zset_h_


#include"defo.h"
#include"coordc.h"


 // constants enumerating Lohner methods
extern const int LOHNER1;
extern const int LOHNER3;
extern const int  LOHNER4;
extern const int LOHNER5;
extern const int  LOHNER33;
extern const int  LOHNER53;
extern const int  LOHNER51;
// end of constants enumerating Lohner methods

bool doubletonRep(int NumLohnerMethod);
     // returns true when  this is doubleton representation

bool QuboidRep(int NumLohnerMethod)  ;
     // returns true when this is  a cuboid set

bool usesSmartQR(int NumLohnerMethod) ;
      // returns true, when the set uses modified QR which
      // takes the shape of r into consideration

bool usesIntervalSet(int NumLohnerMethod) ;
    // returns true when the whole or part of the set representation
    //  is an interval set (product of intervals)




class Ode;
class OneStepParamType;
class C1Set;
class InclOde;
class InclOneStepParamType;
class SetWithTail;

class Set{
//   if numLohner==LOHNER4 or LOHNER5x  set=xc+C*r0 + B*r
//             in LOHNER4 - C is an interval matrix = A[k]*A[k-1]*...*A[1]
//                      where A[i]- is the Jacobion matrix in the i-th time step
//               LOHNER5 - doubleton +QR
//               LOHNER51 - doubleton + interval
//               LOHNER53 - doubleton + modQR (QR - taking into account
//                             the shape of the set)
//   if numLohner==LOHNER3x             set=xc+ B*r
//             LOHNER3 - QR-method
//             LOHNER33 - modified QR taking into account
//                             the shape of the set)
//   if numLohner==LOHNER1             set=xc+ r

  private:


    interval xc[MAX_DIM+1];
    interval C[MAX_DIM+1][MAX_DIM+1];
    interval r0[MAX_DIM+1];
    interval B[MAX_DIM+1][MAX_DIM+1];
    interval r[MAX_DIM+1];


    //  the variables  used for switching to the doubleton representation
    bool switchingOn;
    double bias; // if size(Br) > bias*size(C*r0) and switchon
           // then we set r0=C*r0+Br, C=id; r=0, B=id


   int dim;
   int numLohner;


   void setdefaults();  // numLohner - must be allready set when using this function

   bool intersect_rep; // if true - then the set is given by (xc+C*r0 + B*r) \cap enclo_intv
   interval enclo_intv[MAX_DIM+1];

   void lointvencl(interval setencl[]) const;
  //   computes setencl - the interval enclosure of "this" : xc+B*r + C*r0
  //    this means we ignore the fact that we may have an 'intersection'
  //    representation

   bool rear_cmpart(const interval intvset[]);
   //  returns true - when  the intersection of set <x> and <intvset> is non-empty
   //          false - otherwise
   // on output: if true is returned, then  x(input) \cap intvset \subset x(output)

   bool intersect_with_interval(interval y[],bool seldir[]);

  public:
    int sdim() const; // returns dim
    int setnewdim(int pdim); //  returns previous value of dim, and sets new dim

    int snumLohner() const; // return numLohner

    bool sintersect_rep() const; // returns - <intersect_rep> if <true>, then the set is represented
                  // according to the setting of <numLohner == LOHNERxx>
                  // intersected with an interval set
                  //  otherwise is just the Lohner-type set

    bool get_switchingOn() const;

    bool doubletonRep()  const;
     // returns true when when this is doubleton representation

    bool QuboidRep()  const;
     // returns true when this is a cuboid set

    bool usesSmartQR() const;
      // returns true, when the set uses modified QR which
      // takes the shape of r into consideration

    bool usesIntervalSet() const;
    // returns true  when the whole or part of the set representation
    //  is just interval set (product of intervals)


    Set(int pdim,int pnumLohner,const interval px0[],const interval pC0[][MAX_DIM+1],
       const interval pr0[],bool pintersect_rep=true);
    Set(int pdim,const interval px0[],const interval pC0[][MAX_DIM+1],
       const interval pr0[],int pnumLohner  =LOHNER5 ,bool pintersect_rep=true );
    // if pnumLohner==LOHNER1 - xc=px0, r=pr0
    // if pnumLohner==LOHNER3xx - xc=px0, r=pr0, B=pC0
    // if  pnumLohner==LOHNER4 or LOHNER5xx - xc=px0, r0=pr0, C=pC0, B=identity, r=0


    Set(int pdim,const interval px0[],int pnumLohner =LOHNER5, bool pintersect_rep=true );
    // if pnumLohner==LOHNER1 - xc=center(px0), r=pr0-xc
    // if pnumLohner==LOHNER3xx - xc=center(px0), r=pr0-xc, B=Id
    // if  pnumLohner==LOHNER4 or LOHNER5xx - xc=center(px0), r0=pr0-xc, C=Id, B=identity, r=0




    Set();

    Set(const Set &s);
    Set& operator=(const Set & s); // the assignment/copy operator
// it is a term by term copy, but for the arrays's it
// copy only the necessary terms

    void newvalues(int pdim,int pnumLohner,const interval px0[],
            const interval pC0[][MAX_DIM+1],const interval pr0[]);
    // if pnumLohner==LOHNER1 - xc=px0, r=pr0
    // if pnumLohner==LOHNER3x - xc=px0, r=pr0, B=pC0
    // if  pnumLohner==LOHNER4 or LOHNER5xx - xc=px0, r0=pr0, C=pC0, B=identity, r=0

    void intvencl(interval setencl[],int useothercoord,const LinCC *coordch) const;
//   computes setencl - the interval enclosure of "this"
//   the result is given in coordinates (*coordch) if <useothercoord>

    void intvencl(interval setencl[]) const;
       //   computes set_encl - the interval enclosure of <this>


    void add(const interval y[]);
     //  x=x + y;


    bool intersect_interval(const interval intvset[]);
     //  returns true - when  the intersection of set <this> and <intvset> is non-empty
   //           false - otherwise
   // on output: if true is returned, then  this(input) \cap intvset \subset this(output)

    bool rearrangement(interval encl[]);
     //  if numLohner==LOHNER5x and switchingOn then:
     //  if size(r) > bias*size(C*r0) and switchon
     //           then we set r0=C*r0+Br, C=id; r=0, B=id
     //   and we set encl=intvencl() and return <true>
     //  otherwise encl is unchanged , and <false> is returned


    bool test_xc_in_encl(bool write_to_err_log = false );
  // checks whether this->xc is contained in this->intvencl() (the interval
  //   enclosure of this)
  // returns true , when it is contained, false otherwise
  // if <write_to_err_log>  then a report is written to  <zglerrorlogname>



   void set_B_identity();
   //  if numLohner!=LOHNER1 we rearange the part B*r, so that B=id
   //         r=interval enclosure of (B(old)*r(old))
   /// this makes sense if r is small when compared of other 'errors'
   //   which will be incorporated into the term B*r in the next steps
   //  of the evolution, hence having a nontrivial B introduces unnnecessary
   // wrapping



   void getsettings(const Set &s);
     // copies numLohner,bias,switchingon etc , but <dim> is unchanged

   void get_Cmatrix(interval Ccopy[][MAX_DIM+1]) const;
     //  Ccopy=this->c;

   void get_r(interval rcopy[]) const;
     //  rcopy=this->r;
   void get_r0(interval rcopy[]) const;
     //  rcopy=this->r0;
   void get_Bmatrix(interval Bcopy[][MAX_DIM+1]) const;
     //  Bcopy=this->B;
   void get_xc(interval xccopy[]) const;
     //  xccopy=this->xc


   void savetxt(ostream &report) const;


   friend void LohnerMove(const Set &x,const interval xnxc[],
                              intervalmatrix1* A, Set &xn);
// This function is used in the following context, we compute an image of f
// (say a timestep h map of x or the coordinate change ,
// on input:
//    the value f(x.xc),   hence it is an interval vector
//                                            with nonzero width
//   A= is the Jacobian matrix for f(x) )
//         x - the input set
//
// on output: xn = xnxc + A*(x - x.xc) (in the representation defined in <x>)


   friend  void Setinsertcoord(const Set &xonsec,int coord,interval cval,Set &xt);
   //  from <xonsec> we create a dim(xonsec)+1 - dimensional set,
  // by inserting the <c>-direction  and setting it to <cval>


  friend bool projectionOfTheSet(const Set &source,int dimTarget,Set & target,interval rem[] );
  // this function produces target and rem, such that
  //     source \subset target + rem[dimTarget+1,...,source.dim]
  //  It is required that: dimTarget <= source.dim



  friend int OneStepLohner(Ode &ode,const OneStepParamType &paronestep, interval h,
       Set &x,interval xd[],Set &xn, interval xdn[],interval xencl[],
       interval rem[]);

  friend int incl_OneStepLohner(InclOde &inclode,const InclOneStepParamType &onesteppar,
       interval h,Set &x,interval xd[], SetWithTail & y, Set &xn, interval xdn[],
       SetWithTail &yn,interval xencl[],SetWithTail &yencl, Set &xs,
        interval xds[],interval yc[],interval w1_encl[],interval rem[]);


  friend int C1OneStepLohner(Ode &ode,const OneStepParamType &param, interval h,
       Set &x,interval xd[],Set & xn, interval xdn[],interval xencl[],
       interval rem[], C1Set &V, interval /*Vd*/[][MAX_DIM+1],
       C1Set &Vn,interval Vdn[][MAX_DIM+1], interval Vencl[][MAX_DIM+1],
       interval Vrem[][MAX_DIM+1] );

  friend void transformSet(const Set &x,const LinCC *pcoord,int inverse,Set &xn);

  friend class LinearF; // need this for method eval(const & Set)

  friend class GeneralLinearSection;

};



class C1Set {
//  this is an "extension" of the class Set, devised to handle C1-computations
//  currnumLo==LOHNER1
//     Vc + Vr
//  currnumLo==LOHNER3 - QR-decomposition
//    Ve,Ven,V0 - are not used
//       V=Vc + Vb*Vr
//  currnumLo==LOHNER5 - inner enclosure  with QR-decomposition
//      V=Vc + Ve*V0 + Vb*Vr
   private:
     int dim;
     int numLohner;
     int currnumLo;
     interval bias; // used only when numLohner==LOHNER5
          // if  diam(dp) > bias - then we start to trace the Lipschitz part
     interval Vc[MAX_DIM+1][MAX_DIM+1];
     interval Ve[MAX_DIM+1][MAX_DIM+1];
     interval V0[MAX_DIM+1][MAX_DIM+1];
     interval Vb[MAX_DIM+1][MAX_DIM+1];
     interval Vr[MAX_DIM+1][MAX_DIM+1];

  public:

     C1Set(int pdim,int pnumLohner);
     // the initialization with identity

     C1Set(const C1Set &s);  // the copying constructor

     C1Set & operator=(const C1Set & s); // the assignment/copy operator

     void intvencl(interval setencl[][MAX_DIM+1]) const;
     void intvencl(interval setencl[][MAX_DIM+1],int useothercoord,const LinCC *coordch) const;
     //   computes set_encl - the interval enclosure of set
     //   the result is given in coordinates (*coordch) if <useothercoord>

     bool rearrangement(interval dencl[][MAX_DIM+1]);
     // switches the internal representation from LOHNER3 to LOHNER5
     //  on output: dencl contains an enclosure of current set in case
     //   is we make any change
     //  returns <true> iff the switch occurred


     friend void C1LohnerRearrangement(C1Set &V, C1Set &Vn,intervalmatrix1 *A);

     friend int C1OneStepLohner(Ode &ode,const OneStepParamType &param, interval h,
       Set &x,interval xd[],Set & xn, interval xdn[],interval xencl[],
       interval rem[], C1Set &V, interval /*Vd*/[][MAX_DIM+1],
       C1Set &Vn,interval Vdn[][MAX_DIM+1], interval Vencl[][MAX_DIM+1],
       interval Vrem[][MAX_DIM+1]);

};





void info_err_commonpart(const interval xencl[],const interval intvset[],int dim,
           const char comment[]);
// an error message to file zglerror




void transformSet(const Set &x,const LinCC *pcoord,int inverse,Set &xn);
// on output if inverse xn=pcoord^{-1}(x)
//           else  xn=pcoord(x)


int whichSetTypeChose();

char *LohnerSetDescription(int numLohner,char *name,int lenname);

int whichC1SetTypeChose();

#endif /*_zset_h */
